/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#include "ChartParserHyp.hh"
#include "UniqueCompleteHyps.hh"
//============================================================================
// CLMUtilities
//============================================================================

CWordSequence CUtilities::ComputeLMContext(const CWordSequence& lmString) {
  if (lmString.size() <= CChartParserInfo::C_HistoryLength) {
    return lmString;
  } else {
    // return the first C_HistoryLength words, followed by
    // m_lmBoundaryWI, followed by the last C_HistoryLength words
    // note that it can happen that first words and last words
    // overlap, i.e., are duplicated
    // this is intended
    CWordSequence lmContext;
    for (int i=0; i<CChartParserInfo::C_HistoryLength; i++) {
      lmContext.push_back(lmString[i]);
    }
    lmContext.push_back(CChartParserInfo::m_lmBoundaryWI);
    int beginOfLastWords = lmString.size()-CChartParserInfo::C_HistoryLength;
    for (int i=0; i<CChartParserInfo::C_HistoryLength; i++) {
      lmContext.push_back(lmString[beginOfLastWords+i]);
    }
    return lmContext;
  }
}

CWordSequence CUtilities::HarvestLMString(
    const CWordSequence& bpTargetSequence,
    const vector<CChartParserNode*>& m_backPointers) {

#ifndef NDEBUG
  bool s_debugComputeLM=false;
#endif

  // resulting LM string
  CWordSequence lmString;

  //go thru bpTargetSequence
  for (CWordSequence::const_iterator iter=bpTargetSequence.begin();
       iter!=bpTargetSequence.end();
       ++iter){
    CWordIndex nextWord = *iter; //(*(pHypRule->m_pTarget))[ti];
    //if the word index is a back pointer than process it carefully
    if (nextWord<=0){
      CChartParserNode* pEquivalenceClass = m_backPointers[-1*nextWord];
#ifndef NDEBUG
      if (s_debugComputeLM){
        cout<<"found backpointer node:";
        pEquivalenceClass->Display(CChartParserInfo::m_pVocabulary);
        cout<<endl;
      }
#endif
      // push the complete LM context (either (w1...wm)
      // where m<=C_HistoryLength or
      // (w1..w_C_HistoryLength m_lmBoundaryWI v1..v_C_HistoryLength) )
      // into lmString
      lmString.insert(lmString.end(),
                      pEquivalenceClass->m_lmContext.begin(),
                      pEquivalenceClass->m_lmContext.end());
    }else{
      // nextWord is a regular word - push it into lmString
#ifndef NDEBUG
      if (s_debugComputeLM)
        cout<<"direct target wordf: "
            <<CChartParserInfo::m_pVocabulary->GetWord(nextWord)
            <<"("<<nextWord<<")"<<endl;
#endif
      lmString.push_back(nextWord);
    }
  }//finished going thru all m_translationOutput words
  return lmString;
}

float CUtilities::ComputeLMCost(
    CLanguageModel* lm,
    const CWordSequence& lmString,
    bool estimateBoundaryCost,
    bool computeInsideCost) {

#ifndef NDEBUG
  bool s_debugProcess = false;
  if (s_debugProcess) {
    cout << "In ComputeLMCost: " <<
        (lm->m_lmType==FORWARDLM? "FORWARDLM: " : "BACKWARDLM: ");
  }
#endif

  float hypLMCost = 0; // the result, we'll keep adding costs to it
  //  CWordSequence mostRecentWords; // start with empty history
  int mostRecentWordsCount=0; // length of the current history

  // counts the total n.o. words processed
  // (i.e., the current word position, possibly relative to the
  // end of lmString in case of BACKWARDLM)
  int currentWordPos = 0;
  if (lm->m_lmType == FORWARDLM)  {
    // do everything from left to right
    for (CWordSequence::const_iterator iter=lmString.begin();
         iter!=lmString.end(); ++iter, currentWordPos++){
      CWordIndex w = *iter;
      // have we reached a star (m_lmBoundaryWI)?
      if (w==CChartParserInfo::m_lmBoundaryWI) {
	// this should only happen after at least
        // C_HistoryLength words have been
	// processed  since the last boundary
	assert(mostRecentWordsCount >= CChartParserInfo::C_HistoryLength);
	//clear the history and don't push the boundary into it
	//      mostRecentWords.clear();
	mostRecentWordsCount=0;
      } else {
	if (
	    (estimateBoundaryCost &&
             currentWordPos < CChartParserInfo::C_HistoryLength )
            // still processing hyp's first-words,
            // which are considered boundary words for FORWARDLM
	    ||
	    (computeInsideCost &&
             mostRecentWordsCount >= CChartParserInfo::C_HistoryLength)
            // beyond hyp's first-words and also
            // not right after an lmBoundaryWI
	    ) {
	  hypLMCost += lm->GetSingleWordLMCost(lmString, currentWordPos);
	}
	//push-back the word also into the history context
	//      mostRecentWords.push_back(w);
	mostRecentWordsCount++;
      }
    } // end of for loop
  } else {
    // do everything from right to left
    for (CWordSequence::const_reverse_iterator iter=lmString.rbegin();
         iter!=lmString.rend(); ++iter, currentWordPos++){
      CWordIndex w = *iter;
      // have we reached a star (m_lmBoundaryWI)?
      if (w==CChartParserInfo::m_lmBoundaryWI) {

	assert(mostRecentWordsCount >= CChartParserInfo::C_HistoryLength);
	//clear the history and don't push the boundary into it
	//      mostRecentWords.clear();
	mostRecentWordsCount=0;
      } else {
	if (
	    (estimateBoundaryCost
             && currentWordPos < CChartParserInfo::C_HistoryLength)
	    ||
	    (computeInsideCost
             && mostRecentWordsCount >= CChartParserInfo::C_HistoryLength)
	    // beyond hyp's last-words and also not
            // within C_HistoryLength words left of an lmBoundaryWI
	    ) {
	  hypLMCost +=
              lm->GetSingleWordLMCost(lmString,
                                     lmString.size()-1-currentWordPos);
	}
	mostRecentWordsCount++;
      }
    } // end of for loop
  }
  return hypLMCost;
}

CWordSequence CUtilities::HarvestLMStringWithStates(
    const CWordSequence& bpTargetSequence,
    const vector<CChartParserNode*>& m_backPointers,
    CLanguageModel* lm, // the lm for dummy lookups
    bool dodummy) {

  bool debug = false || CChartParserInfo::m_verbosity>1;
  if (debug){
    cout<<"Harvesting for: ";
    CChartParserInfo::PrintVocabSequence(bpTargetSequence);
    cout<<" BPs: "<<m_backPointers.size()<<endl;
  }
  // resulting LM string
  CWordSequence lmString;
  //go thru bpTargetSequence
  SLMHistory history;
  if (dodummy) {
    history = lm->getNullState();
  }
  for (CWordSequence::const_iterator iter = bpTargetSequence.begin();
       iter != bpTargetSequence.end();
       ++iter) {
    CWordIndex nextWord = *iter;
    //if the word index is a back pointer than process it carefully
    if (nextWord<=0){
      CChartParserNode* pEquivalenceClass = m_backPointers[-1*nextWord];
      CHECK_GT(pEquivalenceClass->m_lmContext.size(), 0);
      // immediately add on the first words
      lmString.insert(lmString.end(),
                      pEquivalenceClass->m_firstWords.begin(),
                      pEquivalenceClass->m_firstWords.end()
                      );
      if (dodummy){
        // process the first words
        for (CWordSequence::const_iterator word_iter =
                 pEquivalenceClass->m_firstWords.begin();
             word_iter != pEquivalenceClass->m_firstWords.end();
             word_iter++){
          CWordIndex w = (*word_iter);
          SLMHistory new_history;
          lm->DummyLookup(history, w, &new_history);
          history = new_history;
        }
      }
      // if we have a second state in the backpointer m_lmContext then
      // we transition directly to that state. We indicate this state
      // with a negative integer pushed into the lmString
      if (pEquivalenceClass->m_lmContext.size()>1) {

        //fw + m_lmContext.back is enough if we have seen symbols
        //but when processing this lmString we need to keep the
        //first state around!
        //lmString.push_back(-1*pEquivalenceClass->m_lmContext.front());
        lmString.push_back(-1*pEquivalenceClass->m_lmContext.back());
        if (dodummy){
          //directly shift the history to this final state
          history = pEquivalenceClass->m_lmContext.back();
        }
      }
    }else{
      // nextWord is a regular word - push it into lmString
      lmString.push_back(nextWord);
      // and predict it according to the dummy LM
      if (dodummy){
        SLMHistory new_history;
        if (debug) {
          cout<<"*** CUtilities::HarvestLMStringWithStates: ";
          cout<<"DummyP(" <<CChartParserInfo::m_pVocabulary->GetWord(nextWord);
          cout<<" | " << lm->HistoryToDebugString(history)<<" ) ";
        }
        lm->DummyLookup(history, nextWord, &new_history);
        if (debug) {
          cout<<"NewHistory: "<<lm->HistoryToDebugString(new_history);
          cout<<endl;
        }
        history = new_history;
      }
    }
  }//finished going thru all m_translationOutput words
  //LOG(INFO) << " Harvested: " << lmString;
  if (debug){
    cout<<"Harvested: lmString: ";
    CChartParserInfo::PrintVocabSequence(lmString);
    cout<<endl;
  }
  return lmString;
}

void CUtilities::ComputeLMCostAndResultContextAllLMs(
    const CWordSequence& lmString,
    vector<float>* m_negLogProbs,
    float* leftBoundaryCost,
    vector<CWordIndex>* lmContext,
    vector<CWordIndex>* firstWords){

  for (int i=0; i<CRule::C_NumLMs; i++) {
    // we are going to pass these local variables for update
    // and then look at their values to update the parameters
    // passed into this function. These local variables will be
    // set with raw (unscaled) values.
    float lmcost=0;
    float boundaryCost = 0;
    CWordSequence oneLMContext;
    CWordSequence oneFirstWords;
    CUtilities::ComputeLMCostAndResultContext(
        CLanguageModel::m_languageModels[i],
        lmString,
        &lmcost,
        &boundaryCost,
        &oneLMContext,
        &oneFirstWords);
    if (i==0){
      *lmContext = oneLMContext;
      *firstWords = oneFirstWords;
    }
    // m_negLogProbs only represent costs for a single rule
    // until n-best extraction time
    (*m_negLogProbs)[CRule::C_LMLocation+i] = lmcost;
    (*leftBoundaryCost) +=
        CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i]*boundaryCost;
  }
}

void CUtilities::ComputeLMCostAndResultContext(
    CLanguageModel* lm,
    const CWordSequence& lmString,
    float* insideCost,
    float* leftBoundaryCost,
    vector<CWordIndex>* lmContext,
    vector<CWordIndex>* firstWords) {
  // We currently only support forward state based models
  assert(lm->m_lmType==FORWARDLM);

  bool debug = false || CChartParserInfo::m_verbosity>1;
  if (debug){
    cout<<"*** BEGIN: CUtilities::ComputeLMCostAndResultContext"<<endl;
  }
  assert(lmContext->size()==0);

  if (debug){
    cout<<"lmString: ";
    CChartParserInfo::PrintVocabSequence(lmString);
    cout<<endl;
  }

  SLMHistory history = lm->getNullState();
  int history_length = 0;
  bool didBackoff = false;
  // Iterate through each word in the lmString ie
  // "a b c <STATE> g h <STATE> i"
  // - need to maintain a rolling context that might need to be shortened
  //   if the LM indicates that a back off occurred.
  // - all the costs beyond the first backoff are considered 'insideCost'
  //   since there is no way that they can be wrong in the context of a
  //   larger hyp
  // - costs for the words until the first backoff OR C_HistoryLength words
  //   are considered leftBoundaryCost since these probs could be modified
  for (CWordSequence::const_iterator iter=lmString.begin();
       iter!=lmString.end();
       ++iter) {
    CWordIndex w = *iter;
    if (w>0){
      // encountered a single word

      int historyFound;
      if (debug){
        cout<<"word: "<< CChartParserInfo::m_pVocabulary->GetWord(w)<<" ";
      }

      // Like getCostAndNewHistory but also returns (in nwords_found, if not
      // NULL) the length in words of the suffix of old_history actually
      // matched in computing the cost.
      SLMHistory new_history;
      int nwords_found = 0;
      float cost =
          lm->getCostAndNewHistoryAndNWordsFound(history,
                                             w,
                                             &new_history,
                                             &nwords_found);
      if (debug){
        cout<<"cost: "<<cost
            <<" old: "<<
            CLanguageModel::m_languageModels[0]->HistoryToDebugString(history)
            <<" new: "<<
            CLanguageModel::m_languageModels[0]->HistoryToDebugString(new_history)<<" "
            <<"nwords_found: "<<nwords_found<<" ";
      }
      // we still care about potential backoffs
      if (!didBackoff){
        // Case: We have not backed off yet
        if (debug){
          cout<<"!didBackoff ";
        }
        // If a backoff event occurs... (it can happen in 2 ways)
        // a) there was actual backoff for your history
        // b) your history has grown to be C_HistoryLength
        // c) nwords_found sends -1 which is a special signal
        //    that indicates that we should you the post backoff state
        // nwords_found returns context+word , history_length = context only
        if ((nwords_found<history_length+1) ||
            (firstWords->size() == CChartParserInfo::C_HistoryLength) ||
            nwords_found==-1 ){
          // Case: backed off now
          if (debug){
            cout<<"backoffnow ";
          }
          // This history will ONLY get used for hyp recombination. We also
          // stop recording additional words in firstWords now...
          if (nwords_found == -1 ) {
            lmContext->push_back(new_history); // set the new_history as left
            // if this backoff is happening at the beginning, then
            // add the corresponding BOS symbols to firstWords thereby
            // ensuring that firstwords is always filled
            if (firstWords->size()==0){
              for (int i=0; i<CChartParserInfo::C_HistoryLength; i++){
                firstWords->push_back(CChartParserInfo::m_bosWI);
              }
            }
          }else {
            lmContext->push_back(history); // set the left state
          }
          // setting this ensure that we dont record any more histories
          didBackoff = true;
          *insideCost += cost;
        }else{
          // Case: never backed off before, and not even now

          // no b ackoff yet, track how many words we have added and append
          // the first words vector
          history_length++;
          if (debug){
            cout<<"setting history_length: "<<history_length<<" ";
          }
          firstWords->push_back(w);

          // since we havent hit backoff yet, all these costs only
          // contribute to the leftBoundaryCost which we want to
          // keep separated from the finalized insideCost
          *leftBoundaryCost += cost;
        }
      }else{
        if (debug){
          cout<<"didBackoff ";
        }
        // Case: We have already backed off
        *insideCost += cost;
      }

      history = new_history;

    }else{

      SLMHistory nthistory =  -1*w;
      // Encountered a history - a history implies a backoff condition
      // that this hyp has inherited. Again, only deal with this case
      // if we haven't seen any backoff already
      if (!didBackoff) {

        // you should always have added firstwords from a nonterminal
        // UNLESS if that nonterminal represents an UNKNOWN word,
        // which you can see if the history that you encounter is
        // the null history. This NULL history can then become
        // part of a longer span - and can no longer be easily asserted.
        lmContext->push_back(history);
        // we no longer care about tracking the history length
        didBackoff = true;
      }

      // this is done because in lmString we store the histories
      // as negative ints. We have to make it positive again
      // before it can get used
      history = nthistory;

    }
    if (debug){
      cout<<endl;
    }
  }

  if (debug){
    cout<<"leftBoundaryCost: "<<*leftBoundaryCost<<" insideCost: "<<
        *insideCost<<endl;
  }
  // once all the words have been processed, we add the final history
  // to the context vector - this may be the only one there or it might
  // join a pre-backoff history
  lmContext->push_back(history);
  if (debug) {
    cout<<"lmContext: ";
    CLanguageModel::PrintContextSequenceAllLMs(*lmContext);
    cout<<endl;
  }
  // in a state based model only 2 conditions for lmContext size are possible
  // either no backoffs have occurred and there is only a single final state
  // that is used for equivalence classing only...
  // OR
  // there are two real states, representing an intermediate backoff
  assert(lmContext->size()<=2);
}

// Assumes that translationOutput has already had its tags replaced
// Assumes that backpointer initial size == pnodes.size
void CUtilities::HarvestAllLMStrings(
    CLanguageModel* lm,
    const CWordSequence& translationOutput,
    const vector<CProductionNode*>& pnodes,
    const int level,
    vector<CChartParserNode*>* backpointers){

  bool debug = false || (CChartParserInfo::m_verbosity>1);
  if (debug) {
    cout<<"*** BEGIN CUtilities::HarvestAllLMStrings"<<endl;
  }

  CHECK_EQ(backpointers->size(), pnodes.size());
  // base case
  if (level == pnodes.size()) {
    // true here indicates that we want to do DummyLookup
    HarvestLMStringWithStates(translationOutput, *backpointers, lm, true);
  } else {
    // recursive step - instantiating the next backpointer in the sequence
    CProductionNode* pProdNode = pnodes[level];

    // instantiate the backpointer vector will all the alternatives
    for (vector<CChartParserNode*>::iterator iter =
             pProdNode->m_lowestToHighestCostFineGrained.begin();
         iter != pProdNode->m_lowestToHighestCostFineGrained.end();
         ++iter){
      CChartParserNode* pFineNode = (*iter);
      (*backpointers)[level] = pFineNode;
      HarvestAllLMStrings(lm,
                          translationOutput,
                          pnodes,
                          level + 1,
                          backpointers);
    }
  }
  if (debug) {
    cout<<"*** END CUtilities::HarvestAllLMStrings"<<endl;
  }
}

void CUtilities::HarvestAllLMStringsAllLMs(
    const CWordSequence& translationOutput,
    const vector<CProductionNode*>& pnodes) {

  // initialize the backpointers to be the same size
  // as the pnodes vector, but will NULL fine-grained eqs
  vector<CChartParserNode*> backpointers;
  for (int i =0; i<pnodes.size(); i++){
    backpointers.push_back(NULL);
  }
  // create all the harvest strings for each LM
  for (int i=0; i<CRule::C_NumLMs; i++) {
    HarvestAllLMStrings(CLanguageModel::m_languageModels[i],
                        translationOutput,
                        pnodes,
                        0,
                        &backpointers);
  }
}


//============================================================================
// CProductionNode
//============================================================================

CProductionNode::CProductionNode(int i, int j, CWordIndex productionResult) {
  m_i = i;
  m_j = j;
  m_productionResult = productionResult;
  //lookup the pruning options
  if (CChartParserInfo::m_pruningMap.count(m_productionResult)!=0){
    m_pruningOptions = CChartParserInfo::m_pruningMap[m_productionResult];
  }else{
    assert(CChartParserInfo::m_pruningMap.size()>0);
    m_pruningOptions = CChartParserInfo::m_pruningMap[0];
  }
  m_pLowestCostHyp = NULL;
  m_cutoffLogProb = FLT_MAX/10;

  //initialized to a very large number to start with
  m_bestTotalRuleApplicationCount = FLT_MAX/10;
}

//delete all the fine grained classes within you
CProductionNode::~CProductionNode(){
  for (CChartParserNodeContainer::iterator iter = m_fineGrainedNodes.begin(); iter!= m_fineGrainedNodes.end(); ++iter){
    CChartParserNode* pNode = (*iter).second;
    delete pNode;
    pNode = NULL;
  }
  //you might be deleting BEFORE you have finalized any hyps, which means you have memory sitting in the m_unfiledHyps
  for (vector<CChartParserHypComplete*>::iterator iter = m_unfiledHyps.begin(); iter!=m_unfiledHyps.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    delete pHyp;
  }
}


bool CProductionNode::CheckBeamInsertHyp(CChartParserHypComplete* pHyp, bool addAll){
  //dont do any of the fine-grained equivalence class based work here, that is left for FinalizeProductionNode
  //for now, just check that you meet the beam constraint
  bool debug = false;

  //cout<<"Inserting into: "; Display(); cout<<"\t";
  //double hypCost = pHyp->m_totalCost + pHyp->m_scaledFutureCostEstimate;

  // the CProductionNode maintains m_unfiledHyps across ALL equivalence
  // classes, and maintains a beam of them, so its natural to consider
  // boundary costs for these words as well, so we use
  // the AllComponentCost which is equal to ScaledTMLMCost when
  // m_pruningMapUseBoundary==false
  double hypCost = pHyp->AllComponentCost();
  // when this node has FIRST been formed the m_cutoffLogProb is set to
  // FLT_MAX so the new hyp will pass
  // it's <= to make it match with the PruneUnfiledHyps which rejects
  // if the hyp cost > m_cutoffLogProb
  if (hypCost<=m_cutoffLogProb || addAll){

    //this hyp is a good hyp... it might even be the best!
    if (m_pLowestCostHyp==NULL
        || hypCost < m_pLowestCostHyp->AllComponentCost()){
      m_pLowestCostHyp = pHyp;

      m_cutoffLogProb = CChartParserInfo::GetPruningCutoff(
          hypCost,m_pruningOptions.second);
      // set this here so that you can check this property of the node
      // BEFORE you file the hyps away for real
      m_lexical = pHyp->m_lexical;
    }
    //in any case add it to the flat list of passing candidates
    m_unfiledHyps.push_back(pHyp);
    //cout<<"Added normal"<<endl;
    return true;
  }else{
    //this hyp was rejected by the beam
#ifndef NDEBUG
    if (debug){
      cout<<"Failed beam:"<<endl;
      pHyp->Display();
      cout<<endl;
      cout<<"Top Element:";
      m_pLowestCostHyp->Display();
      cout<<endl;
      cout<<" m_cutoffLogProb: "<<m_cutoffLogProb<<endl;
    }
#endif
    // delete it
    delete pHyp;
    return false;
  }
}


  //arv - current plan to deal with case where
  //      terminal -> @X, @S, and then @X->@S rules are generated
  //      clear the m_unfiledHyps calling finalize, then the p_applySelfRuleRecursive round comes in
  //      operate under a clean slate of hyps and add these additional hyps
  //      -- but retain the existing m_cutoffLogProb
void CProductionNode::FinalizeProductionNode(bool addAll){
  //first apply the beam restrictions again on the m_unfiledHyps
  PruneUnfiledHyps(addAll);
  //now file the hyps into their appropriate classes and set the internal variables of this node
  //cout<<"m_bestTotalRuleApplicationCount: "<<m_bestTotalRuleApplicationCount<<" -> ";
  for (vector<CChartParserHypComplete*>::iterator iter = m_unfiledHyps.begin(); iter!=m_unfiledHyps.end(); ++iter){
    FileHyp(*iter);
  }
  m_unfiledHyps.clear();
  //EqClassBeamPruning();

  m_numHyps = RankFineGrainedClasses();
  //cout<<m_bestTotalRuleApplicationCount<<endl;
}

//Fills the local variable m_lowestToHighestCostFineGrained - call this only after BeamPruning + p_applySelfRuleRecursive has been called
//set the m_totalRuleApplicationCount
int CProductionNode::RankFineGrainedClasses(){

  //arv: i initially thought that m_lowestToHighestCostFineGrained should always be unset coming into this function
  //     but it turns out that in the following condition, it could already be initialized
  //     terminal -> @X , @S AND @X -> @S
  //     the @S node will be created in the first round
  //     in the second round, we will file the rule app count 2 hyp into @S, thus modifying m_fineGrainedNodes
  //assert(m_lowestToHighestCostFineGrained.size()==0);
  m_lowestToHighestCostFineGrained.clear();
  //set the m_bestTotalRuleApplicationCount here because it involves check all fine-grained eq classes
  m_bestTotalRuleApplicationCount = FLT_MAX;
  int numHyps = 0;
  for (CChartParserNodeContainer::iterator iter = m_fineGrainedNodes.begin(); iter!= m_fineGrainedNodes.end(); ++iter){
    CChartParserNode* pNode = (*iter).second;

    //arv - new rule app
    m_bestTotalRuleApplicationCount = MIN(m_bestTotalRuleApplicationCount, pNode->m_totalRuleApplicationCountBestHyp);
    m_lowestToHighestCostFineGrained.push_back(pNode);
    //just verify that the overall best hyp is the m_bestTotalCost + m_scaledFutureCostEstimate for the best m_fineGrainedNodes element
    numHyps += pNode->m_hypContainer.size();
  }
  sort(m_lowestToHighestCostFineGrained.begin(), m_lowestToHighestCostFineGrained.end(), CompareChartParserNode());
  if (m_lowestToHighestCostFineGrained.size()>0){
    CChartParserNode* pLowestFine = (*(m_lowestToHighestCostFineGrained.begin()));
    //double cost1 = pLowestFine->m_bestTotalCost + pLowestFine->m_scaledFutureCostEstimate;
    double cost1 = pLowestFine->BestHypAllComponentCost();
    if (m_pLowestCostHyp!=NULL){
      //double cost2 = m_pLowestCostHyp->m_totalCost + m_pLowestCostHyp->m_scaledFutureCostEstimate;
      double cost2 = m_pLowestCostHyp->AllComponentCost();
      if (!(fabs(cost1-cost2)<0.001)){
	cout<<"cost1: "<<cost1<<" cost2: "<<cost2<<endl;
	cout<<"Lowest: "; m_pLowestCostHyp->Display();
	cout<<"pLowestFine: "; pLowestFine->Display();
      }
      assert(fabs(cost1-cost2)<0.001);
    }
  }
  assert(m_lowestToHighestCostFineGrained.size()==m_fineGrainedNodes.size());
  return numHyps;
}

void CProductionNode::PruneUnfiledHyps(bool addAll){
  bool debug = false;
  if (debug){
    cout<<"PruneUnfiledHyps: "; //Display();
    cout<<" size: "<<m_unfiledHyps.size() << endl;
    cout<<"Containing hyps: " << endl;
    for (int i=0; i<m_unfiledHyps.size(); i++) {
      m_unfiledHyps[i]->Display();
      cout << endl;
    }
    cout<<endl;
  }

  //first sort them into lowest to highest
  //CompareCompleteHyps is using the AllComponentCost
  sort(m_unfiledHyps.begin(), m_unfiledHyps.end(), CompareCompleteHyps());
  //  m_unfiledHyps.sort(CChartParserHypComplete::CompareAllComponentCosts);

  vector<CChartParserHypComplete*>::iterator lowestCostIter = m_unfiledHyps.begin();
  if (lowestCostIter==m_unfiledHyps.end()){
    return;
  }

  //if you are setting the m_pLowestCostHyp to NULL after you Rank, then this condition will hold
  //assert(m_pLowestCostHyp==(*lowestCostIter));
  //otherwise, the m_pLowestCostHyp will be mainted across iterations, while the m_unfiledHyps will be cleared
  vector<CChartParserHypComplete*> unfiledCopy;
  assert(m_pLowestCostHyp!=NULL);

  float tmpRuleAppCount = FLT_MAX;

  //go thru each m_unfiledHyps in order (after the recent sorting) and break out if the AllComponentCost was too high
  int count=0; //keep track of the number of hyps added
  float countLimit = m_pruningOptions.first; //this specifies the max number of HYPs (not equivalence classes) that we will file in this node

  if (debug){
    cout<<"countLimit: "<<countLimit<<endl;
  }

  //there is a slight problem with this scheme: in self rule mode we will enter this function with a fresh batch of m_unfiledHyps, but now all the
  //hyps from the original filings are completed already (they had to be so that we could spawn new hys). But this next wave can now fill up
  //to countLimit hyps again! so technically we allow upto m_pruningOptions*(n.o.selfrule waves) hyps

  vector<CChartParserHypComplete*>::iterator iter = m_unfiledHyps.begin();
  for (;iter!=m_unfiledHyps.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    //double currentLogProb = pHyp->m_totalCost+pHyp->m_scaledFutureCostEstimate;
    float currentLogProb = pHyp->AllComponentCost();

    if ((currentLogProb>m_cutoffLogProb|| count>=countLimit) && !addAll){
      //another edge case... you might have just inserted a whole series of elements that have the same
      //cost, and have supplanted the best cost. so now you cant arbitrarily kick out the top ones, you can
      //only kick of elements whose cost is worse
      if (currentLogProb>m_pLowestCostHyp->AllComponentCost()){
	if (debug){
	  cout<<"breaking out currentLogProb: "<<currentLogProb<<" m_cutoffLogProb: "<<m_cutoffLogProb<<" count: "<<count<<" countLimit: "<<countLimit<<endl;
	}
	break;
      }
    }
    unfiledCopy.push_back(*iter);
    tmpRuleAppCount = MIN(pHyp->m_totalRuleApplicationCount, tmpRuleAppCount);

    //if (debug){
    //cout<<"In PruneUnfiledHyps: "; (*iter)->Display();
    //}
    count++;
  }
  //delete the ones that are left - they will never be filed anywhere
  //note: its possible that you are throwing away hyps with a lower rule app count. This explains why sometimes
  //      you could increase the beam size, you could find a hyp with lower rule app count which causes more severe
  //      pruning due to rule app count diff
  for (; iter!=m_unfiledHyps.end(); ++iter){
     CChartParserHypComplete* pHyp = (*iter);
     assert(pHyp!=m_pLowestCostHyp);
     //we also track WHY the hyp was deleted (ie threshold or histogram pruning)
     //if you are being dropped for violating the threshold pruning
     if (pHyp->AllComponentCost()>m_cutoffLogProb){
       CChartParserInfo::m_hypsDeletedThreshold++;
     }
     //if the hyp violates the histogram pruning too
     if (count>countLimit){
       CChartParserInfo::m_hypsDeletedHistogram++;
     }
     //NOTE: a hyp can violate BOTH histogram pruning AND threshold pruning
     delete pHyp;
     pHyp=NULL;
     //this represents the total number of hyps deleted
     CChartParserInfo::m_hypsDeleted++;
  }

  if (debug){
    //if (count<m_unfiledHyps.size()){
      cout<<"Pruned away: "<<m_unfiledHyps.size()-count<<" against: "<<m_cutoffLogProb<<endl;
      //}
  }
  m_unfiledHyps = unfiledCopy;
}


//returns true iff m_bestTotalRuleApplicationCount has been updated
bool CProductionNode::FileHyp(CChartParserHypComplete* pHyp){
  bool updatedBestTotalRuleAppCount = false;
  assert(I() == pHyp->m_i);
  assert(J() == pHyp->m_j);
  assert(GetResult() == pHyp->m_productionResult);

  ///===== Handle the rule app count ==========
  //if we have never seen any hyps before
  if (m_fineGrainedNodes.empty()){
    m_lexical = pHyp->m_lexical;
    //arv - new rule app
    //m_bestTotalRuleApplicationCount = pHyp->m_totalRuleApplicationCount;
    updatedBestTotalRuleAppCount = true;
  }else{
    //arv - new rule app we do this work in RankFineGrainedClasses now
    //we need to see if we have a better rule application count
    //if (pHyp->m_totalRuleApplicationCount < m_bestTotalRuleApplicationCount) {
    //update best cumulative rule app counter
    //  m_bestTotalRuleApplicationCount = pHyp->m_totalRuleApplicationCount;
    //updatedBestTotalRuleAppCount = true;
    //}

  }

  HypData h(pHyp->m_lmContext, pHyp->m_productionResult);
  //size_t hashCode = FineHash();
  ///===== Handle the equivalence class work  ==========
  //1) Check if there is an existing fine grained eq that matches this hyp
  //2) If there is - add it into that node, else create a new one
  if (m_fineGrainedNodes.count(h)==0){
    //insert a new container
    CChartParserNode* fgNode = new CChartParserNode(this, CChartParserInfo::MAX_HYP_COUNT);
    fgNode->InsertHyp(pHyp);
    m_fineGrainedNodes.insert(pair<HypData, CChartParserNode*>(h, fgNode));
  }else{
    //there is already a node there, just add into it

    CChartParserNodeContainer::iterator existingNodeIter = m_fineGrainedNodes.find(h);
    CChartParserNode* fgNode = (*existingNodeIter).second;
    fgNode->InsertHyp(pHyp);
  }

    //      if ((pHyp->m_firstWords!=fgNode->m_firstWords) || (pHyp->m_lastWords!=fgNode->m_lastWords)){
    //	cout<<"ERROR: CollisionExisting: ";
    //	cout<<"hyp first:";
    //	CChartParserInfo::PrintVocabSequence(pHyp->m_firstWords);
    //	cout<<"node first:";
    //	CChartParserInfo::PrintVocabSequence(fgNode->m_firstWords);
    //	cout<<"hyp last:";
    //	CChartParserInfo::PrintVocabSequence(pHyp->m_lastWords);
    //	cout<<"node last:";
    //	CChartParserInfo::PrintVocabSequence(fgNode->m_lastWords);
    //	exit(-1);
    //      }

  //if (updatedBestTotalRuleAppCount){
  //cout<<"Updated :"; this->Display();
  //cout<<"With hyp: "; pHyp->Display();
  //}

  return updatedBestTotalRuleAppCount;
}

void CProductionNode::Display(CVocabulary* pVoc){

  cout<<"Node ";
  cout<<"I: "<<I()<<" J: "<<J();
  cout<<" prod="
      <<CChartParserInfo::m_pVocabulary->GetWord(m_productionResult)
      <<" m_bestTotalRuleApplicationCount: "
      <<m_bestTotalRuleApplicationCount
      <<" isLexical: "
      <<m_lexical
      <<" num fine: "
      <<m_fineGrainedNodes.size()
      <<endl;
}





//============================================================================
// CChartParserNode
//============================================================================
CChartParserNode::CChartParserNode(CProductionNode* parent, int maxHypCount){
  m_productionNodeParent = parent;
  m_maxHypCount = maxHypCount;
  m_totalRuleApplicationCountBestHyp = FLT_MAX;
  // m_scaledBoundaryCost = 0.0; //this will get initialized when
  // we set the first hyp
  m_pHeap = NULL; //make sure it starts of NULL
}

// insert a new hyp into this equivalence class and update the m_best variables
// only up to m_maxHypCount hyps are stored;
// when an additional hyp is supposed to inserted such that the number
// of hyps stored is higher
// than m_maxHypCount, the highest-cost hyp is removed and deleted
// (that could also be the hyp supposed to be inserted)!
bool CChartParserNode::InsertHyp(CChartParserHypComplete* pHyp) {

  assert(I() == pHyp->m_i);
  assert(J() == pHyp->m_j);
  assert(GetResult() == pHyp->m_productionResult);
  bool updatedRuleAppCount = false;

  //note: we must use the non-boundary cost version here since we are
  // *inside* the fine grained equivalence class
  //footnote: NO... in cmp=0 mode, we want to use the
  //AllComponentCost.
  // in cmp=1, all hyps must have the same firstWords anyway
  //float cCandidate = pHyp->TotalCost() + pHyp->m_scaledFutureCostEstimate;
  //float cCandidate = pHyp->ScaledTMLMCost();
  float cCandidate = pHyp->AllComponentCost();

  if (m_hypContainer.empty()) {
    //initialize the hyp dependent variables
    m_lexical = pHyp->m_lexical;
    m_bestTotalCost = pHyp->m_totalCost;
    m_scaledFutureCostEstimate = pHyp->m_scaledFutureCostEstimate;
    m_scaledBoundaryCost = pHyp->m_scaledBoundaryCost;
    // context that identifies the hyps in this fine grained eq
    m_lmContext = pHyp->m_lmContext;
    m_firstWords = pHyp->m_firstWords;

    m_totalRuleApplicationCountBestHyp = pHyp->m_totalRuleApplicationCount;
    updatedRuleAppCount = true;
  } else {
    assert(!CChartParserInfo::m_compareTargetWords ||
           m_lmContext == pHyp->m_lmContext);
    float cExisting = this->BestHypAllComponentCost();
    if (cCandidate < cExisting) {

      //update m_best
      m_bestTotalCost = pHyp->m_totalCost;
      m_scaledFutureCostEstimate = pHyp->m_scaledFutureCostEstimate;
      m_scaledBoundaryCost = pHyp->m_scaledBoundaryCost;
      // We also update the m_lmContext and m_firstWords since
      // we support the cmp0 case as well
      m_lmContext = pHyp->m_lmContext;
      m_firstWords = pHyp->m_firstWords;

      assert(m_lexical == pHyp->m_lexical);

      // only do this update base on the lowest cost hyp if we are not
      // using m_lowestRuleAppCount
      if (!CChartParserInfo::m_lowestRuleAppCount){
	m_totalRuleApplicationCountBestHyp = pHyp->m_totalRuleApplicationCount;
	updatedRuleAppCount = true;
      }
    }else{
      //cout<<"not better than cExisting"<<endl;
    }

    //if we are doing m_lowestRuleAppCount style, then we just see if the rule app count of this hyp
    //was lower than before and update it, even though the best hyp in this equivalence class might
    //have a higher rule app count
    if (CChartParserInfo::m_lowestRuleAppCount){
      if (pHyp->m_totalRuleApplicationCount<m_totalRuleApplicationCountBestHyp){
	m_totalRuleApplicationCountBestHyp = pHyp->m_totalRuleApplicationCount;
      }
      updatedRuleAppCount = true;
    }
  }
  //now set owner status
  pHyp->m_ownerNode = this;
  // and file it in the hyp container - note: this container is sorted based on
  // cCandidate which is pHyp->AllComponentCost()
  m_hypContainer.insert(std::pair<float, CChartParserHypComplete*>(cCandidate, pHyp));

  //do we have more than m_maxHypCount hyps in the container now?
  if (m_hypContainer.size() > m_maxHypCount) {
    // yepp - remove and delete the worst

    //cout<<"DELETING HYPs *************"<<endl;
    //    // exception: never delete _G hyps (important for gluing)
    //    if (GetResult()==CChartParserInfo::m_GlueNT) {
    //      if (m_hypContainer.size() == m_maxHypCount+1){
    //	//cout << "node for _G exceeded m_maxHypCount (m_maxHypCount="<<m_maxHypCount<<"). Not removing any hyps anyway."<<endl;
    //      }
    //    } else {
      CSortedHypContainer::iterator iterLastElem = m_hypContainer.end();
      iterLastElem--;
      CChartParserHypComplete* pHypToDel = (*iterLastElem).second;
      //float costToDelete = pHypToDel->m_totalCost + pHypToDel->m_scaledFutureCostEstimate;
      //float costToDelete = pHypToDel->ScaledTMLMCost();
      float costToDelete = pHypToDel->AllComponentCost();

      //evil edge case here. if you have inserted multiple elements onto the beam
      //that have the exact same cost and you try to delete the end() element because
      //the beam got too full you could delete the one that have been saved as m_pLowestCostHyp 
      //by mistake, thereby garbling your memory
      //if ((float)costToDelete > (float)(m_bestTotalCost + m_scaledFutureCostEstimate)){
      //float currentCost = m_bestTotalCost + m_scaledFutureCostEstimate;
      //float currentCost = this->BestHypScaledTMLMCost();

      float currentCost = this->BestHypAllComponentCost();
      if (costToDelete > currentCost){

	  if (pHypToDel==m_productionNodeParent->m_pLowestCostHyp){
	    cout<<"DELETING:"; pHypToDel->Display();
	    cout<<"size: "<<m_hypContainer.size()<<endl;
	    cout<<"PARENT m_pLowestCostHyp: "; m_productionNodeParent->m_pLowestCostHyp->Display();
	    cout<<"costToDelete: "<<costToDelete<<" cExisting: "<<m_bestTotalCost+m_scaledFutureCostEstimate<<endl;
	    cout<<"delete the top hyp my mistake!"<<endl;
	    exit(-1);
	  }

	  //note: while you know that you will never delete the lowest cost hyp, you may 
	  //      be deleting the hyp the provided the lowest rule app count
	  assert(pHypToDel!=m_productionNodeParent->m_pLowestCostHyp);
	  delete pHypToDel;
	  m_hypContainer.erase(iterLastElem);
      }
  }
  return updatedRuleAppCount;
}


void CChartParserNode::Display(CVocabulary* pVoc){
  if (pVoc == NULL) pVoc = CChartParserInfo::m_pVocabulary;
  cout<<"["<<pVoc->GetWord(GetResult())<<"-"<<I()<<"-"<<J()<<"-";
  CLanguageModel::m_languageModels[0]->PrintContextSequence(m_lmContext);
  cout<<"]";
}

void CChartParserNode::DisplayRecursive() {
  Display();
  cout << endl;
  cout<<" best hyp: ";
  (*(m_hypContainer.begin())).second->DisplayRecursive(CChartParserInfo::m_pVocabulary,0);
  cout << endl;
}


//NOTE: this function moves the m_scaledFutureCostEstimate into
//      the real LM spot
vector<CChartParserHypComplete*> CChartParserNode::RetrieveKBestHC(int k){

  assert(k>=1); // k==1 means retrieve 1st best
  //clear the stats for this retrieval round
  CChartParserInfo::ResetNBestStatistics();
  // k-1 since it's 0 indexed; passing true to indicate that 
  // this is the top level
  CChartParserHypComplete* pLastHyp = GetKthBest(k-1, true);

  //keep trying to ensure that we have enough unique hyps
   if (CChartParserInfo::m_extractUnique){
     cout<<
         "Top K Hyps in HC05 extracted, searching longer for unique !"
         <<endl;
    int count = 0;
    int limit = CChartParserInfo::BeamSizeFactor*k;
    int kToRetrieve = k;
    //keep searching until you have k unique hyps

    while (m_uniqueFinalNBest.Size()<k &&
           pLastHyp!=NULL && kToRetrieve<limit){
      int initialSize = m_uniqueFinalNBest.Size();
      pLastHyp = GetKthBest(kToRetrieve, true);
      int finalSize = m_uniqueFinalNBest.Size();
      assert(finalSize>=initialSize);
      //ask for the next k down
      ++kToRetrieve;
    }
  }

  CChartParserInfo::m_numUnique = m_uniqueFinalNBest.Size(); //track the number of unique translations generated
  cout<<"Log: NBest: "; CChartParserInfo::PrintNBestStatistics(); cout<<endl;

  if (CChartParserInfo::m_compareTargetWords==0){
    // we still have to sort the candidates b/c they might not be
    // in exact order due to incorrect LM estimates during the parsing process

    //create a copy! because if you use NBestSubSpans you will want to re-use these pre-calculated n-best lists
    vector<CChartParserHypComplete*> correctOrder;
    correctOrder.insert(correctOrder.end(), m_candidateVector.begin(), m_candidateVector.end());
    sort(correctOrder.begin(), correctOrder.end(), KBestCandidateMinCompare());
    reverse(correctOrder.begin(), correctOrder.end());
    return correctOrder;
  }
  return m_candidateVector;
}


CChartParserHypComplete* CChartParserNode::GetKthBest(int k, bool isTopLevel){

  bool debug = false;
  assert(k >= 0); // k==0 means get first best
  if (debug){
    cout<<"Entering GetKthBest: k="<<k<<" "; Display(); cout<<endl;
  }
  // have we already got the hyp asked for?
  if (m_candidateVector.size()>k){
    // yeah - just return it
    return m_candidateVector[k];
  }else{
    // no - gotta create it
    //if nothing is initialized yet
    if (m_pHeap==NULL){
      m_pHeap = new KBestCandidateHeap();


      for (CSortedHypContainer::iterator iter = m_hypContainer.begin();
           iter!=m_hypContainer.end(); ++iter){
	CChartParserHypComplete* pBaseHyp = iter->second;
	CChartParserHypComplete* pNewHyp =
            new CChartParserHypComplete(*pBaseHyp);
	//	//make sure the pNewHyp (the base parent hyp is null)
	//	pBaseHyp->m_pParentHyp = NULL;
        //guarantees that the refered alternatives DO exist
	bool wasInited = pNewHyp->InitializeHeapData(
            pBaseHyp, pBaseHyp->m_backPointers.size());
	if (!wasInited){
	  //delete it directly, it hasnt been stored on the tuple list!
	  cout<<" NO INIT POSSIBLE=========== "<<endl;
	  delete pNewHyp;
	  pNewHyp = NULL;
	}else{
	  //add to the heap
	  assert(m_pHeap!=NULL);
	  // this wasAdded tells you whether there was another hyp with
          // the same derivation... cant be at init time!
	  //cout<<" Adding =========== "<<endl;
	  bool wasAdded = m_pHeap->InsertBaseCombination(pNewHyp);
	  assert(wasAdded || CChartParserInfo::m_extractUnique);
	}
      }
      //if you have m_extractUnique turned on,
      // you can possible discard elements right as you try to initialize them
      assert(CChartParserInfo::m_extractUnique ||
             m_pHeap->m_candidateSet.size()==m_hypContainer.size());
      if (debug){
	cout<<"Heap initialized: ";  Display(); cout<<endl;
      }
    }

    if (isTopLevel && debug){
      cout<<"Searching till k: "<<k<<endl;
    }
    int count = 0;
    while (m_candidateVector.size()<=k){
      //track some top level stats
      if (isTopLevel){
	CChartParserInfo::m_topCallsKBest++;
      }
      //pop top of heap, add nearby neighbors if the heap wasnt empty
      assert(m_pHeap!=NULL);
      CChartParserHypComplete* pTopHyp = m_pHeap->PopHeap();
      if (pTopHyp==NULL){
	if (isTopLevel && debug){
	  cout<<"Received null at k:"<<k<<endl;
	}
	break;
      }

      //if checking at pop -
      // maintain a separate unique set here (dont want to conflict with
      // the one for the heap) besides unique-checking in InsertBaseCombination
      // right when putting a hyp on
      // the heap, we also check again later at the time of popping it from
      // the heap b/c in the meantime, a better-deserving hyp might have
      // been added
      bool checkAtPop = true;

      if (checkAtPop && CChartParserInfo::m_extractUnique){
	//see if its a duplicate dont delete it, the ptr is reference
        // in the m_uniqueTuples
	if (m_uniqueFinalNBest.CheckUnique(pTopHyp)!=NULL){
	  m_candidateVector.push_back(pTopHyp);
	}else{
	  //cout<<"Not added to m_candidateVector: ";
          //pTopHyp->Display(); cout<<endl;
	}
      }else{
	m_candidateVector.push_back(pTopHyp);
      }
      m_pHeap->AddNeighbors(pTopHyp);

      count++;
    }

    if (isTopLevel && debug){
      cout<<"Exitting GetKthBest: k="<<k<<" found: "<<
          m_candidateVector.size()<<endl; Display(); cout<<endl;
    }

    if (m_candidateVector.size()<=k){
      //cerr<<"WARNING: requested k-best that was too deep in some node";
      return NULL;
    } else {
      //cout<<"returning m_candidateVector";
      return m_candidateVector[k];
    }
  }
}


CChartParserNode::~CChartParserNode() {
  for (CSortedHypContainer::const_iterator iter2 = m_hypContainer.begin(); iter2!=m_hypContainer.end(); iter2++){
    CChartParserHypComplete* pHyp = (*iter2).second;
    delete pHyp;
  }
  //if the node was use in HC k-best retrieval
  if (m_pHeap!=NULL){
    delete m_pHeap; //this will call clear inside the heap
  }
}


//============================================================================
// CChartParserHypComplete
//============================================================================

//!Constructor completing an incomplete hyp that exact-matches a rule
/*!
  Take an incomplete hyp (the tail, containing the backpointers and
  accumulated costs) and a rule (the head, containing
  the production result, translation output, and the costs of applying
  the rule) and form a complete hyp out of these two parts
  This constructor is used in the Fundamental and in the Self rule -
  now during combo pruning in CCombinationTuple::Complete()
*/
CChartParserHypComplete::CChartParserHypComplete(
    CChartParserHypIncomplete* pIncompleteHyp,
    const vector<CChartParserNode*>& backPointers,
    CRule* pItem) :CChartParserHyp(*pIncompleteHyp) {

#ifndef NDEBUG
  bool debug = false;
#endif
  //assign the back pointer choices
  m_backPointers = backPointers;

  // the tail stuff has been copied by the inherited constructor from
  // pIncompleteHyp now we have to make the hyp complete by giving it a head
  m_lexical = false;
  m_completelyUnwound = false;
  //set below: m_totalCost
  //set below: m_scaledFutureCostEstimate
  m_productionResult = pItem->m_productionResult;
  m_rule = pItem->m_rule;
  m_negLogProbs = pItem->m_negLogProbs; //includes already a leading
					//zero for the LM
  if (CChartParserInfo::m_useSpanFeatures) {
    // add source span neg log prob to the cost vector
    double mu = CChartParserInfo::m_nonterminalOccurrenceStatsMap[m_productionResult].logSourceSpanLengthMean;
    double sigma = CChartParserInfo::m_nonterminalOccurrenceStatsMap[m_productionResult].logSourceSpanLengthVariance;
    m_negLogProbs.push_back(log(sigma));
    m_negLogProbs.push_back(pow(log(m_i)-mu,2)/2/sigma/sigma);
  }
  if (CRule::C_NumModelsTotal != m_negLogProbs.size()) {
    cerr<<"Error: Wrong number of models m_negLogProbs: "<<m_negLogProbs.size()<<" C_NumModelsTM: "<<CRule::C_NumModelsTM<<" C_NumModelsTotal: "<<CRule::C_NumModelsTotal << endl << " Underlying rule (see STDOUT): "<<endl;
    cout<<"Error: Wrong number of models m_negLogProbs: "<<m_negLogProbs.size()<<" C_NumModelsTM: "<<CRule::C_NumModelsTM<<" C_NumModelsTotal: "<<CRule::C_NumModelsTotal << endl << " Underlying rule: "<<endl;
    pItem->Display();
    Display();
    exit(2);
  }

  //=============================================================
  // add RuleAppCount feature to current count (which currently is the
  // cumulative count from the backpointed hyps)

  //if you use m_updateHypRuleAppCount then you will re-calculate the backpointer rule app counts
  //because they were based on the best fine-grained EC's RACs,
  //while now they are based on concrete fine-grained EC's RACs
  if (CChartParserInfo::m_updateHypRuleAppCount){
    float totalTABackup = m_totalRuleApplicationCount; // for sanity test purposes
    m_totalRuleApplicationCount = 0;
    //note: we are adding the m_totalRuleApplicationCount of the BestHyp in the CChartParserNode
    //not the best one findable inside this class
    for (int i=0; i<m_backPointers.size(); i++){
      assert(m_backPointers[i]->m_totalRuleApplicationCountBestHyp<=FLT_MAX/10);
      m_totalRuleApplicationCount += m_backPointers[i]->m_totalRuleApplicationCountBestHyp;
    }
    //in cmp0 there should only be one fine grained BP per class, so it should be the same as the backup value
    assert(CChartParserInfo::m_compareTargetWords || fabs(totalTABackup-m_totalRuleApplicationCount)<0.001);
  }
  // in any case add the applied rule's RAC feature
  m_totalRuleApplicationCount += m_negLogProbs[CRule::C_RuleApplicationCountLocation];
  //====================================================
  // set translation output and m_taggedTargets

  //if we have a number-tag production result AND we have no
  // back pointers, then we take the following steps
  // - replace the m_translationOutput with the production result
  // (since in UnwindHyp we directly calculate LM costs based on
  // m_translationOutput)
  // - fill the m_taggedTargets with the pItem->m_translationOutput
  //   (these are the words that will finally be subbed in)
  // support for: src_5 -> tgt_5 -> @NUM (this rule will get matched
  // directly on the source side and will enter this phrase)
  // : @NUM foreign_pt @NUM -> @NUM.@NUM -> @NUM
  // (translation output contains @NUMs, so we dont have to do anything)
  if (m_backPointers.size()==0 &&
      CChartParserInfo::m_tagsForLM.count(m_productionResult)!=0){
    //cout<<"Tagging: "; pItem->Display(); cout<<endl;
    // replaces the translation output
    m_translationOutput = CWordSequence(1, m_productionResult);
    m_taggedTargets = pItem->m_translationOutput;
    // also push on the tagged production result, so we when we
    // start aggregating we know how far to go
    m_taggedTargets.push_back(m_productionResult);

#ifndef NDEBUG
    if (debug){
      cout<<"Setting m_taggedTargets:";
      CChartParserInfo::PrintVocabSequence(m_taggedTargets); cout<<endl;
    }
#endif
  } else {
    m_translationOutput = pItem->m_translationOutput;
  }

#ifndef NDEBUG
  if (debug) {
     cout<<"Printing m_translationOutput:";
     CChartParserInfo::PrintVocabSequence(m_translationOutput);
     cout<<endl;
  }
#endif


  //========================================================================
  // LM cost calculation
  //========================================================================
  //the LM costs; these are 'estimates' in case of cmp0; in cmp1,
  // these are the actual LM costs
  //it's a new hyp, we must set this to 0...
  //arv: under this model: m_scaledFutureCostEstimate is ALWAYS 0!
  // it never gets set anywhere else!!
  //     these costs are being maintained inside m_totalCost now...
  m_scaledFutureCostEstimate = 0.0;
  //... and add the underlying hyps' estimates
  for (int i=0; i<m_backPointers.size(); i++){
    m_scaledFutureCostEstimate +=
        m_backPointers[i]->m_scaledFutureCostEstimate;
  }
  // ... and add this hyp's additional LM estimates (the
  // estimates are true costs if we're in cmp1 and C_HistoryLength
  // matches the LMs Markov order ) by traversing the target words and
  // the backpointed LM contexts of this rule

  // to do this, first obtain this samtstring to traverse
  /*
  CWordSequence lmString =
      CUtilities::HarvestLMString(m_translationOutput, m_backPointers);
  // also determine the resulting LM context
  m_lmContext = CUtilities::ComputeLMContext(lmString);
  // now, compute the different LM models' costs
  m_scaledBoundaryCost = 0.0;
  for (int i=0; i<CRule::C_NumLMs; i++) {
    // compute inside-cost of i-th LM
    m_negLogProbs[CRule::C_LMLocation+i] = CUtilities::ComputeLMCost(
        CLanguageModel::m_languageModels[i],
        lmString,
        false,  // no - don't estimate the left boundary cost
        true); // yes - compute the inside cost
    // and add this LM's boundary LM cost estimate of this context to
    // the boundary cost estimate
    // Note: instead of m_lmContext, we can also pass lmString here,
    // which will have the same result for FORWARDLM or BACKWARDLM but
    // a more accurate estimate for CENTERLM, but in the moment we're
    // not using CENTERLM during the parsing step anyway
    m_scaledBoundaryCost += CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i] *
        CUtilities::ComputeLMCost(CLanguageModel::m_languageModels[i], m_lmContext, true, false);
  }
  */
  m_scaledBoundaryCost = 0.0;
  // harvest the LM string without doing any DummyLookups
  CWordSequence lmString = CUtilities::HarvestLMStringWithStates(
      m_translationOutput, m_backPointers, NULL, false);

  CUtilities::ComputeLMCostAndResultContextAllLMs(
      lmString,
      &m_negLogProbs,
      &m_scaledBoundaryCost,
      &m_lmContext,
      &m_firstWords);
  assert(m_lmContext.size() <= 2);
  CHECK_LE(m_lmContext.size(),2);

  // total (cumulative) cost calculation
  // compute m_totalCost based on m_negLogProbs and m_scalingFactors
  // and underlying hyps' total costs
  // note: m_totalCost does now ALSO include LM costs even though
  // these might be estimates only
  m_totalCost = 0.0;
  for (int i=0; i<m_negLogProbs.size(); i++){
    m_totalCost += m_negLogProbs[i]*CChartParserInfo::m_scalingFactors[i];
  }
  //this m_bestTotalCost include the LM costs from the hyps below
  for (int i=0; i<m_backPointers.size(); i++){
    m_totalCost += m_backPointers[i]->m_bestTotalCost;
  }


#ifndef NDEBUG
  if (debug) {
    CChartParserInfo::TabStack(); cout<<"FinalCompleted: "; Display();
  }

  // also do a sanity check that the rule we used was proper in the sense of
  // having a one-to-one correspondence between source-NTs and
  // target-NT-references
  set<int> usedRefs;
  for (int i=0; i<m_translationOutput.size(); i++) {
    CWordIndex w = m_translationOutput[i];
    if (w <= 0) {
      // it's an NT reference, i.e., -w is the index into m_backPointers
      if (-w >= m_backPointers.size()) {
	cout << "Error: rule's target side has a substitution marker that is higher than the n.o. source NT sites. Rule concerned: " << m_rule << endl;
	exit(2);
      } else if (usedRefs.count(-w) > 0) {
	cout << "Error: rule's target side has at least two substitution markers with the same number. Rule concerned: " << m_rule << endl;
	exit(2);
      } else {
	usedRefs.insert(-w);
      }
    }
  }
  assert(usedRefs.size() <= m_backPointers.size());
  if (usedRefs.size() < m_backPointers.size()) {
    cout << "Error: rule's n.o. target substitution markers is less than its n.o. source NT occurrences. Rule concerned: " << m_rule << endl;
    exit(2);
  }

  // another sanity check if we have an _UNKNOWN rule, e.g., _UNKNOWN#_UNKNOWN#@NP#1 1 1
  if (m_translationOutput.size() > 0 && m_translationOutput[0] == CChartParserInfo::m_unknownWI) {
    // yes: unknown-rule
    if (m_translationOutput.size() != 1 || m_backPointers.size() != 0) {
      cout << "Error: ill-formed unknown-rule: " << m_rule << endl;
      assert(false);
    }
  }

#endif

  //we don't know the owner yet
  m_ownerNode=NULL;
}



void CChartParserHypComplete::AddParseOutputSubTreeStart(CWordSequence& parsingOutput, CChartParserHypComplete* pBackHyp){
    parsingOutput.push_back(CChartParserInfo::m_pVocabulary->GetIndex("#("));
    parsingOutput.push_back(pBackHyp->GetResult());
    //add the tags
    parsingOutput.push_back(pBackHyp->J());
    //we want to print span indices
    parsingOutput.push_back(pBackHyp->J()+pBackHyp->I()-1);
    int startZIndex = pBackHyp->J();
    for (int si=0; si<pBackHyp->I(); si++){
      //newParsingOutput.push_back((*CChartParserInfo::m_pSentence)[si+startZIndex]);
      //the parsing is indexed from 1, thus the -1
      assert(CChartParserInfo::m_pSentence.size()>si+startZIndex-1);
      parsingOutput.push_back(CChartParserInfo::m_pSentence[si+startZIndex-1]);
    }
    //add the end-of-tags symbol
    parsingOutput.push_back(CChartParserInfo::m_pVocabulary->GetIndex("#"));
  }

void CChartParserHypComplete::AddParseOutputSubTreeEnd(CWordSequence& parsingOutput){
  parsingOutput.push_back(CChartParserInfo::m_pVocabulary->GetIndex("#)"));
}


void CChartParserHypComplete::HandleUnknownAndTags(){
  int unkPos = 0; // position in pHyp->m_targetOrderedUnknownSourceWords of next unknown source word to output
  CWordSequence newTranslationOutput;
  //this call is to avoid changing the code below to update the fact that it is now a member variable
  CChartParserHypComplete* pHyp = this;
  CWordSequence::iterator tagIter = pHyp->m_taggedTargets.begin();
  for (CWordSequence::iterator iter3=pHyp->m_translationOutput.begin(); iter3!=pHyp->m_translationOutput.end(); ++iter3){
    if (*iter3 > 0) {
      // *iter3 is a terminal symbol
      if (*iter3 == CChartParserInfo::m_unknownWI) {
	// it's an unknown marker - delete it or replace it with the corresponding source word
	if (CChartParserInfo::m_noTranslationOutputOfUnknownWords) {
	  // ignore the _UNKNOWN marker
	} else {
	  // instead of the _UNKNOWN marker, output the corresponding source word
	  // you should have filled this word in the unwinding process
	  //cout<<"UNKSize: "<<pHyp->m_targetOrderedUnknownSourceWords.size()<<endl;
	  assert(pHyp->m_targetOrderedUnknownSourceWords.size()>unkPos);
	  newTranslationOutput.push_back(pHyp->m_targetOrderedUnknownSourceWords[unkPos]);
	  ++unkPos;
	}
      } else if (*iter3 == CChartParserInfo::m_bosWI || *iter3 == CChartParserInfo::m_eosWI) {
	// either <s> or </s> - skip over it
      } else if (CChartParserInfo::m_tagsForLM.count(*iter3)!=0){
	  //*iter3 IS a @NUM style marker.
	  //m_taggedTargets should now contain something like "two and a half @NUM"
	assert(tagIter!=pHyp->m_taggedTargets.end());
	  //keep writing the tags until you reach another terminal
	  while (CChartParserInfo::m_tagsForLM.count(*tagIter)==0){
	    //cout << CChartParserInfo::m_pVocabulary->GetWord(*tagIter)<<" ";
	    newTranslationOutput.push_back(*tagIter);
	    ++tagIter;
	    //there should ALWAYS be a @TAG marker after the actual surface output
	    assert(tagIter!=pHyp->m_taggedTargets.end());
	  }
	  //skip past the fake @TAG that we have put onto the m_taggedTargets as a marker
	  ++tagIter;
	  //tagIter can be at end now
      } else {
	// a regular word - preserve it
	newTranslationOutput.push_back(*iter3);
      }
    } else {
      // *iter3 is a backpointer - so the hyp is not completely unwound!
      cerr << "Error in HandleUnknownAndTags: Hyps are supposed to be completely unwound, but this hyp still has a nonterminal marker in its m_translationOutput. Hyp is: ";
      assert(pHyp!=NULL);
      cout<<"Failed on: ";
      CChartParserInfo::PrintVocabSequence(pHyp->m_translationOutput);
      cout<<" "; pHyp->PrintTranslationOutput(); cout<<endl;
      pHyp->DisplayRecursive(CChartParserInfo::m_pVocabulary,0);
      exit(1);
    }
  }
  //overwrite the m_translationOutput
  pHyp->m_translationOutput = newTranslationOutput;
}

////!Set the m_totalRuleApplicationCount based on the fine grained eq class backpointers and the current rule
//void CChartParserHypComplete::UpdateTotalRuleAppCount(){
//  }

void CChartParserHypComplete::Display(CVocabulary* pVoc){
  if (pVoc == NULL) {
    pVoc = CChartParserInfo::m_pVocabulary;
  }
  cout<<"I: "<<I()<<" J: "<<J()<<" *prod* : "<<
      pVoc->GetWord(m_productionResult);

  cout<<" m_translationOutput: ";
  for (CWordSequence::iterator iter=m_translationOutput.begin();
       iter!=m_translationOutput.end();
       ++iter){

    if ((*iter)>0){
      cout<<pVoc->GetWord(*iter)<<" ";
    }else{
      //cout<<"s"<<(*iter)<<" ";
      int prodIndex = -(*iter);
      cout<<pVoc->GetWord(m_backPointers[prodIndex]->GetResult());

      //      if (CChartParserInfo::m_compareTargetWords){
      cout<<"[";
      CLanguageModel::m_languageModels[0]->PrintContextSequence(
          m_backPointers[prodIndex]->m_lmContext);
      cout<<"]";
      //      }
      cout<<" ";

    }
  }

  cout<<" m_lmContext: ";
  CLanguageModel::m_languageModels[0]->PrintContextSequence(m_lmContext);

  cout<< " m_totalRuleApplicationCount: " << m_totalRuleApplicationCount;
  //display the cost
  cout<<",ScaledTMLM:"<<ScaledTMLMCost()<<",m_scaledBoundaryCost:" << m_scaledBoundaryCost << ",AllComp: "<<AllComponentCost()<<" m_totalCost: "<<m_totalCost<<" m_scaledFutureCostEstimate: "<<m_scaledFutureCostEstimate<<" raw [";
  for (int i=0; i<m_negLogProbs.size(); i++){
    cout<<" "<<m_negLogProbs[i];
  }
  cout<<"]";
  cout<<" m_taggedTargets: "; CChartParserInfo::PrintVocabSequence(m_taggedTargets);
  cout<<" |tgtSrc|: "<<m_targetOrderedUnknownSourceWords.size()<<endl;
  cout<<endl;
}

//tabs indicate the number of tabs to print at this level
void CChartParserHypComplete::DisplayRecursive(CVocabulary* pVoc, int tabs){
  if (CChartParserInfo::m_pVocabulary == NULL) pVoc = CChartParserInfo::m_pVocabulary;


  // display vital information about this hyp
  for (int t=0; t<tabs; t++){
    cout<<"\t";
  }
  //first output the basic display
  Display(pVoc);
  cout<<endl;

  //now it needs to work in the order of the target language,
  // like in ComputeLMData
  for (int i=0; i<m_backPointers.size(); i++){
    //m_backPointers[i]->DisplayRecursive(pVoc, tabs+1);
    CChartParserNode* pNode = m_backPointers[i];
    CSortedHypContainer* pEquivalenceClass = &pNode->m_hypContainer;
    CChartParserHypComplete* pBestBackHyp = (*(pEquivalenceClass->begin())).second;
    pBestBackHyp->DisplayRecursive(pVoc, tabs+1);
  }
  cout<<endl;

}

void CChartParserHypComplete::PrintTranslationOutput(){
  cout<<CChartParserInfo::m_pVocabulary->GetWord(GetResult())<<" -> ";
  for (CWordSequence::const_iterator iter=m_translationOutput.begin(); iter!= m_translationOutput.end(); ++iter){
    if ((*iter)<=0){
      assert(m_backPointers.size()>-1*(*iter));
      CChartParserNode* pNode = m_backPointers[-1*(*iter)];
      pNode->Display();
      //os << *iter;
    }else{
	//      cout<<" ("<<m_pVocabulary->GetWord(*iter)<<") ";
      if (CChartParserInfo::m_pVocabulary!=NULL){
	cout<<CChartParserInfo::m_pVocabulary->GetWord(*iter);
      }else{
	cout<<*iter;
      }
    }
    //if not end
    if (iter!=m_translationOutput.end()-1){
	cout<<" ";
    }
  }
}



//============================================================================
// CChartParserHypIncomplete
//============================================================================
//virtual 
void CChartParserHypIncomplete::ExpandTail(CProductionNode* pExpansionClass) {

  bool s_debugExpandHypWithComplete= false;
  if (s_debugExpandHypWithComplete){
    cout<<"ExpandTail(): ";
    cout<<"Tail-expanding incomplete hyp: ";
    Display(CChartParserInfo::m_pVocabulary);
    cout<<endl;
    cout<<"  Expanding with hypergraph node: ";
    pExpansionClass->Display(CChartParserInfo::m_pVocabulary);
    cout<<endl;
  }

  //first update m_i, the span length of this hyp; m_j (the span start) remains the same
  m_i += pExpansionClass->I();

  // if we are adding a non-lexical node, i.e., a node with nonterminal result category,
  // then add backpointer to this node
  if (!pExpansionClass->m_lexical) {
    
    if (s_debugExpandHypWithComplete){
      cout<<"Adding backpointer to pExpansionClass:";
      pExpansionClass->Display();
    }
    m_productionBackPointers.push_back(pExpansionClass);
  }

  //also expand prefix info by production result of hypergraph node pExpansionClass
  m_prefix.push_back(pExpansionClass->GetResult());

  //increase cumulative rule application count by the best rule app
  //count in the backpointed hypernode
  m_totalRuleApplicationCount += pExpansionClass->m_bestTotalRuleApplicationCount;

}

//!Add a production back pointer to this hyps backpointer vector
//void CChartParserHypIncomplete::AddProductionBP(CProductionNode* pClass){
//  m_productionBackPointers.push_back(pClass);
//}

void CChartParserHypIncomplete::Display(CVocabulary* pVoc){
  if (pVoc == NULL) pVoc = CChartParserInfo::m_pVocabulary;

  cout<<"I: "<<I()<<" J: "<<J()<<" ";

  cout<<" m_prefix: ";
  for (vector<CWordIndex>::iterator iter=m_prefix.begin(); iter!=m_prefix.end(); ++iter){
    if ((*iter)>0){
      cout<<pVoc->GetWord(*iter)<<" ";
    }else{
      cout<<"s"<<(*iter)<<" ";
    }
  }

  cout <<"backpointerNodes: ";
  for (int i=0;i<m_productionBackPointers.size();i++) {
    cout<<"I"<<m_productionBackPointers[i]->I()<<" J"<<m_productionBackPointers[i]->J()<<" *prod* : "<<pVoc->GetWord(m_productionBackPointers[i]->GetResult());
  }

  cout<< " m_totalRuleApplicationCount: " << m_totalRuleApplicationCount;

}

void CChartParserHypIncomplete::ReCalculateRuleAppCount(){
  m_totalRuleApplicationCount = 0;
  for (vector<CProductionNode*>::iterator iter = m_productionBackPointers.begin(); iter!= m_productionBackPointers.end(); ++iter){
    //make sure its actually been set
    assert((*iter)->m_bestTotalRuleApplicationCount!=FLT_MAX);
    m_totalRuleApplicationCount += (*iter)->m_bestTotalRuleApplicationCount;
  }
}

//============================================================================================
//        K-Best list utility methods for HC05
//============================================================================================

//========= To keep derivations in the n-best list separate ============================
size_t KBestCandidateHash::operator()(const CChartParserHypComplete* p) const{
   //use the address of the rule
  size_t hashCode = (long)(p->m_pParentHyp);
  //+ integer offsets for each instatiated result
  for (int i=0; i<p->m_derivationVector.size(); i++){
    int offset = p->m_derivationVector[i];
    hashCode = 31*hashCode + offset;
  }
  return hashCode;
}

bool KBestCandidateEq::operator()(const CChartParserHypComplete* p1, const CChartParserHypComplete* p2) const{
  if (p1->m_pParentHyp!=p2->m_pParentHyp){
    return false;
  }
  if (p1->m_derivationVector.size()!=p2->m_derivationVector.size()){
    return false;
  }
  for (int i=0; i<p1->m_derivationVector.size(); i++){
    if (p1->m_derivationVector[i]!=p2->m_derivationVector[i]){
      return false;
    }
  }
  return true;
}


//========= To compare derivations in the n-best list ============================
bool KBestCandidateMinCompare::operator()(const CChartParserHypComplete* p1, const CChartParserHypComplete* p2) const{
  return (p1->AllComponentCost()>p2->AllComponentCost());

  /*
  //NOTE: i put each of the three options here simply as an
  explanation for why we should always
  //      be able to use AllComponentCost

  //if using HC05 on cmp=0, the hyp are arranged by AllComponentCost in each eq class
  if (CChartParserInfo::m_compareTargetWords==0 && CChartParserInfo::m_nbestStyle==2){
    return (p1->AllComponentCost()>p2->AllComponentCost());
  }

  //all the hyps come from the sentence spanning beam, where m_scaledBoundaryCost = 0.0
  if (CChartParserInfo::m_compareTargetWords==0 && CChartParserInfo::m_nbestStyle==3){
    assert(p1->m_scaledBoundaryCost==0.0);
    assert(p2->m_scaledBoundaryCost==0.0);
    return (p1->AllComponentCost()>p2->AllComponentCost());
  }

  //if using HC05 on cmp=1, its ok, since all the hyps have the same boundary words
  if (CChartParserInfo::m_compareTargetWords==1 && CChartParserInfo::m_nbestStyle==2){
    assert(p1->m_scaledBoundaryCost==p2->m_scaledBoundaryCost);
    return (p1->AllComponentCost()>p2->AllComponentCost());
  }
  return (p1->ScaledTMLMCost()>p2->ScaledTMLMCost());
  */
}

bool CChartParserHypComplete::InitializeHeapData(
    CChartParserHypComplete* pParentHyp, int backWardStarSize){

  bool debug = false;
  //establish this candidates content
  float m_totalCostBackup = m_totalCost;
  float m_scaledFutureCostEstimateBackup = m_scaledFutureCostEstimate;

  if (debug) {
    cout<<"InitializingHeapData: ";
    this->PrintTranslationOutput();
    cout<<" unwnd: "<<m_completelyUnwound<<endl;
  }

  // assign a parent (this parent hyp is NOT one that is in the beam)
  // this m_pParentHyp is used when we want to refer back to the original
  // pristine hyp that is not participating in n-best extraction
  m_pParentHyp = pParentHyp;

  //the backwards star of this edge refers to the number of
  // backpointers that this edge has.
  for (int i=0; i<backWardStarSize; i++){
    m_derivationVector.push_back(0); //since this is the 0'th best
    //add the m_derivationLimit to the parent hyp
    //the limit is un-defined to start with
    m_pParentHyp->m_derivationLimit.push_back(INT_MAX/2);
  }
  assert(m_pParentHyp->m_derivationLimit.size()==m_derivationVector.size());

  //ComputeCostsLazyKBest makes recursive calls down the k-best tree
  //cout<<" Before Compute: "; this->PrintTranslationOutput(); cout<<" Cost: "<<ScaledTMLMCost()<<endl;
  bool foundFirstBest = ComputeCostsLazyKBest();

#ifndef NDEBUG
  //cout<<" After Compute: "; this->PrintTranslationOutput(); cout<<" Cost: "<<ScaledTMLMCost()<<endl;
  if (!foundFirstBest){
    //assert(CChartParserInfo::m_extractUnique && CChartParserInfo::m_compareTargetWords==1);
    assert(CChartParserInfo::m_extractUnique);
  } else {
    if (debug) {
      cout<<"VerifyingHeapData: "; this->PrintTranslationOutput(); cout<<" unwnd: "<<m_completelyUnwound<<endl;
    }
    //cout<<"m_totalCostBackup: "<<m_totalCostBackup<<" m_totalCost: "<<m_totalCost<<endl;
    //this condition is no longer true, since you might actually find BETTER hyps
    //assert(fabs(m_totalCostBackup + m_scaledFutureCostEstimateBackup-(m_totalCost+m_scaledFutureCostEstimate))<0.001);
  }
#endif

  return foundFirstBest;
}

void CChartParserHypComplete::DisplaySimpleKBest(){
  cout<<"KBestCandidate: ";
  CChartParserInfo::PrintVocabSequence(m_translationOutput);
  cout<<" lmContext: "; CChartParserInfo::PrintVocabSequence(m_lmContext);
  cout<<" PARSE: "; CChartParserInfo::PrintParsingOutput(m_parsingOutput);
  cout<<" AllComponentCost: "<<AllComponentCost()<<" m_totalCost: "<<m_totalCost<<" m_scaledFutureCostEstimate: "<<m_scaledFutureCostEstimate<< "BoundaryCostEstimate: " << m_scaledBoundaryCost << " m_derivationVector: ";
  copy(m_derivationVector.begin(), m_derivationVector.end(), std::ostream_iterator<int>(cout, " "));
  cout<<" m_negLogProbs: ";
  copy(m_negLogProbs.begin(), m_negLogProbs.end(), std::ostream_iterator<float>(cout, " "));
  if (m_pParentHyp==NULL){
    cout<<" m_derivationLimit: ";
    copy(m_derivationLimit.begin(), m_derivationLimit.end(), std::ostream_iterator<int>(cout, " "));
  }else{
    cout<<" (parent's)m_derivationLimit: ";
    copy(m_pParentHyp->m_derivationLimit.begin(), m_pParentHyp->m_derivationLimit.end(), std::ostream_iterator<int>(cout, " "));
  }
}

vector<CChartParserHypComplete*> CChartParserHypComplete::ExpandNeighbors(const UniqueHypsKBest& m_uniqueTuples){
  vector<CChartParserHypComplete*> spawnedTuples;

  //for each dimension in |e| ie the set of tails of E we need to increment clone a new CCombinationTuple and increment it 
  assert(m_pParentHyp!=NULL);

  //the m_pParentHyp *never* has a m_derivationVector, but it will have a m_derivationLimit
  assert(m_derivationVector.size()==m_pParentHyp->m_derivationLimit.size());
  for (int i=0; i<m_derivationVector.size(); i++){
      int baseOffset = m_derivationVector[i];

      //check if we can actually achieve a derivation from a sub-node
      if (baseOffset+1 < m_pParentHyp->m_derivationLimit[i]){
	//clone the neighbor based on this hyp, not the parent to ensure you get the m_derivationVector
	//inside ComputeCostsLazyKBest the m_totalCost etc is totally recalculated based on the parent hyp
	CChartParserHypComplete* pCloned = new CChartParserHypComplete(*this);
	assert(m_pParentHyp == pCloned->m_pParentHyp);
	assert(pCloned->m_derivationVector.size()>i);
	++pCloned->m_derivationVector[i];
	if (m_uniqueTuples.count(pCloned)==0){
	  //compute new translation output, parse output, costs, etc.
	  // and determine if we could actually instantiate a hyp this far down
	  if (pCloned->ComputeCostsLazyKBest()){
	    // success - add it
	    spawnedTuples.push_back(pCloned);
	  }else{
	    //delete it, and note very clearly that you cant ask past this level
	    m_pParentHyp->m_derivationLimit[i]= pCloned->m_derivationVector[i]; //this value is not-available
	    //this means that a) this hyp has never been seen before, and b) its not generatable, and c) we have never stored it anywhere, so delete it
	    //if you don't delete it here you will waste the memory (since you dont push it onto the spawnedTuples)
	    delete pCloned;
	    pCloned = NULL;
	  }
	}else{
	  delete pCloned;
	  pCloned=NULL;
	}//delete the duplicate
      }
  }

  //IF it is important to us that the neighbors be added to InsertBaseCombination in lowest-highest cost order
  //then do this sorting here. The reason this could be useful: ExtractUnique's criteria for uniquing will only
  //throw away LOWER cost nodes, so the order in which they are examined matters
  //for now, orderMatters is false, since this doesnt address the issue of order across set of neighbors
  bool orderMatters = false;
  //if (orderMatters && CChartParserInfo::m_extractUnique && CChartParserInfo::m_compareTargetWords==1){
  if (orderMatters && CChartParserInfo::m_extractUnique){
    sort(spawnedTuples.begin(), spawnedTuples.end(), KBestCandidateMinCompare());
    reverse(spawnedTuples.begin(), spawnedTuples.end());
  }
  return spawnedTuples;
}


vector<CChartParserHypComplete*> CChartParserHypCompleteNBest::ExpandNeighbors(const UniqueHypsKBest& m_uniqueTuples){
  vector<CChartParserHypComplete*> spawnedTuples;

  int depthDown = 1;
  //for each dimension in |e| ie the set of tails of E we need to increment clone a new CCombinationTuple and increment it 
  assert(m_pParentHyp!=NULL);

  //the m_pParentHyp *never* has a m_derivationVector, but it will have a m_derivationLimit
  assert(m_derivationVector.size()==m_pParentHyp->m_derivationLimit.size());
  for (int i=0; i<m_derivationVector.size(); i++){
      int baseOffset = m_derivationVector[i];

      //check if we can actually achieve a derivation from a sub-node
      if (baseOffset+1 < m_pParentHyp->m_derivationLimit[i]){
	//clone the neighbor based on this hyp, not the parent to ensure you get the m_derivationVector
	CChartParserHypCompleteNBest* pCloned = new CChartParserHypCompleteNBest(*this);
	assert(m_pParentHyp == pCloned->m_pParentHyp);
	assert(pCloned->m_derivationVector.size()>i);
	pCloned->m_derivationVector[i]+=1;	  
	if (m_uniqueTuples.count(pCloned)==0){
	  //if we could actually instantiate a hyp this far down
	  if (pCloned->ComputeCostsLazyKBest()){
	    spawnedTuples.push_back((CChartParserHypComplete*)pCloned);
	  }else{
	    //delete it, and note very clearly that you cant ask past this level
	    m_pParentHyp->m_derivationLimit[i]= pCloned->m_derivationVector[i]; //this value is not-available
	    //this means that a) this hyp has never been seen before, and b) its not generatable c) we have never stored it anywhere, so delete it
	    //if you dont delete it here you will waste the memory (since you dont push it onto the spawnedTuples)
	    delete pCloned;
	    pCloned = NULL;
	  }
	}else{
	  delete pCloned;
	  pCloned=NULL;
	}//delete the duplicate
      }
  }

  //IF it is important to us that the neighbors be added to InsertBaseCombination in lowest-highest cost order
  //then do this sorting here. The reason this could be useful: ExtractUnique's criteria for uniquing will only
  //throw away LOWER cost nodes, so the order in which they are examined matters
  //for now, orderMatters is false, since this doesnt address the issue of order across set of neighbors
  bool orderMatters = false;
  //if (orderMatters && CChartParserInfo::m_extractUnique && CChartParserInfo::m_compareTargetWords==1){
  if (orderMatters && CChartParserInfo::m_extractUnique){
    sort(spawnedTuples.begin(), spawnedTuples.end(), KBestCandidateMinCompare());
    reverse(spawnedTuples.begin(), spawnedTuples.end());
  }
  return spawnedTuples;
      
}



//version that performs on nbest items (which are the result of a* search)
bool CChartParserHypCompleteNBest::ComputeCostsLazyKBest(){

  bool debug = false;
  // this hyp is basically the back hyp + the additional cost,
  // models tags that were added by this hyp
  assert(m_derivationVector.size()==1);

  // if this is the first hyp that actually generates terminal words...
  // just blank out the m_scaledFutureCostEstimate and proceed
  // IFF this is m_derivationVector[0]==0
  //otherwise return false, indicating that there are no alternatives
  if (m_pBackEqClass==NULL){
    if (m_derivationVector[0]==0.0){
      return true;
    }else{
      return false;
    }
  }
  assert(m_pBackEqClass!=NULL);
  assert(m_pBackHyp!=NULL);

  // This hyp was generated by extending some hyp
  // (which is a part of the class m_pBackEqClass). Now we
  // consider some alternative from this class.
  // All language model probs have already been worked out
  // correctly since this m_pBackEqClass represents all hyps that
  // have the same last context boundary words
  // All we have to do here is shown below
  //  ------------------------------ ptr this
  //  ---------pBackHyp->m_posNT---- pBackHyp
  //  --pBackAlternative->m_posNT--- pBackAlternative
  //
  // we want upto pBackAlternative->m_posNT from pBackAlternative and then the rest from ptr this

  CChartParserHypCompleteNBest* pBackAlternative =
      m_pBackEqClass->GetKthBest(m_derivationVector[0]);
  if (pBackAlternative==NULL){
    return false;
  }

  //the m_pBackHyp is the one we want to sub out
  //cout<<"Found: "; pBackAlternative->Display(); cout<<endl;
  //cout<<"Currently: "; this->Display(); cout<<endl;

  //make sure we never make this call on one of the original hyps
  assert(m_pParentHyp!=NULL);

  //determine what has to comes from this hyp AFTER the last sub site
  assert(m_pParentHyp->m_translationOutput.begin()+m_pBackHyp->m_posNT<=m_pParentHyp->m_translationOutput.end());

  //do not include the posNT element
  CWordSequence remainingSymbols =
      CWordSequence(
          m_pParentHyp->m_translationOutput.begin() + m_pBackHyp->m_posNT,
          m_pParentHyp->m_translationOutput.end());

  // edge case, without the pop call at the end, we get a problem
  // when we do end with a non-terminal
  //Failed on hyp: @_S -> <s> <s> not so bad it should be sunny [@JJ-1-6-cloudy] </s> </s> m_posNT: 10
  //Failed on: <s> <s> not so bad it should be sunny -2
  // you cant also just pop the last element if its a
  // non-terminal as this case demonstrates
  // Failed on hyp: @_S -> [@_S-2-1-<s> <s> <*> too bad] [@MD-1-3-should] [@JJ+AUX-2-4-sunny is] [@JJ-1-6-cloudy] </s> </s> m_posNT: 0
  int posworking = pBackAlternative->m_posNT;
  CWordSequence alternativeSymbols =
      CWordSequence(pBackAlternative->m_translationOutput.begin(),
                    pBackAlternative->m_translationOutput.begin()
                    +pBackAlternative->m_posNT);

  if (debug){
    //ensure that the leading context is all terminals
    for(int i=0; i<alternativeSymbols.size(); i++){
      if (alternativeSymbols[i]<=0){
	cout<<"Failed on hyp: ";
        pBackAlternative->PrintTranslationOutput();
        cout<<" m_posNT: "<<pBackAlternative->m_posNT<<endl;
	cout<<"Failed on: ";
        CChartParserInfo::PrintVocabSequence(alternativeSymbols); cout<<endl;
	assert(alternativeSymbols[i]>0);
      }
    }
  }

  //transfer over the additional unknown words and tagged targets
  //note: we start with the tags from the pBackAlternative
  CWordSequence newTargetOrderedSourceWords =
      pBackAlternative->m_targetOrderedUnknownSourceWords;
  assert(m_pParentHyp->m_targetOrderedUnknownSourceWords.size()>=
         m_pBackHyp->m_targetOrderedUnknownSourceWords.size());
  // then we add on only those targets that m_pParentHyp had OVER
  // that which m_pBackHyp had
  for (int i=m_pBackHyp->m_targetOrderedUnknownSourceWords.size();
       i<m_pParentHyp->m_targetOrderedUnknownSourceWords.size(); i++){
    newTargetOrderedSourceWords.push_back(
        m_pParentHyp->m_targetOrderedUnknownSourceWords[i]);
  }

  CWordSequence newTaggedTargets = pBackAlternative->m_taggedTargets;
  assert(m_pParentHyp->m_taggedTargets.size()>=
         m_pBackHyp->m_taggedTargets.size());
  for (int i= m_pBackHyp->m_taggedTargets.size();
       i<m_pParentHyp->m_taggedTargets.size(); i++){
    newTaggedTargets.push_back(m_pParentHyp->m_taggedTargets[i]);
  }

  //transfer over the cost differences
  vector<float> newNegLogProbs;
  for (int m=0; m<m_pParentHyp->m_negLogProbs.size(); m++){
    newNegLogProbs.push_back(
        m_pParentHyp->m_negLogProbs[m]-
        m_pBackHyp->m_negLogProbs[m]+
        pBackAlternative->m_negLogProbs[m]);
  }

  float newTotalCost = m_pParentHyp->m_totalCost
      - m_pBackHyp->m_totalCost +
      pBackAlternative->m_totalCost;
  m_totalCost = newTotalCost;

  m_translationOutput = alternativeSymbols;
  m_translationOutput.insert(
      m_translationOutput.end(),
      remainingSymbols.begin(), remainingSymbols.end());
  //recalculate the posnt
  int newPosNT = 0;
  while (newPosNT<m_translationOutput.size()
         && m_translationOutput[newPosNT]>0){
    newPosNT++;
  }
  m_posNT = newPosNT;
  m_targetOrderedUnknownSourceWords = newTargetOrderedSourceWords;
  m_taggedTargets = newTaggedTargets;
  m_negLogProbs = newNegLogProbs;
  return true;
}

void CChartParserHypCompleteNBest::UpdatePosNT(){
  //only do this if m_posNT is not already set past the index range
  if (m_posNT<m_translationOutput.size()){
    while (m_translationOutput.size()>m_posNT && m_translationOutput[m_posNT]>0){
      m_posNT++;
    }
  }
  assert(m_posNT<=m_translationOutput.size());
}

//version that works on the basic hyp to perform hc 05 search
bool CChartParserHypComplete::ComputeCostsLazyKBest(){

  //blank out the variables that we will refill in unwound form
  m_translationOutput.clear();
  m_parsingOutput.clear();

  //first some unknown-handling
  m_targetOrderedUnknownSourceWords.clear();
  // no unknown handling should have been done during parsing
  assert(m_pParentHyp->m_targetOrderedUnknownSourceWords.size()==0);
  //if you are a @Y->_UNKNOWN#_UNKNOWN hyp, then put the corresponding source word into m_targetOrderedUnknownSourceWords
  if (m_pParentHyp->m_translationOutput.size()>0 &&
      m_pParentHyp->m_translationOutput.front()==
      CChartParserInfo::m_unknownWI) {
    assert( m_pParentHyp->m_backPointers.size()==0 && m_pParentHyp->I()==1);
    //cout<<"Adding m_targetOrderedUnknownSourceWords: "; Display(); cout<<endl;
    //cout<<"Parent: "; m_pParentHyp->Display(); cout<<endl;
    assert(CChartParserInfo::m_pSentence.size()>J()-1);
    m_targetOrderedUnknownSourceWords.push_back(
        CChartParserInfo::m_pSentence[J()-1]);
  }

  //after this method is done, we won't have any substitution sites left
  //reflect this (though that's strictly not necessary in C++) by
  //clearing out the references to now unused backpointer hyps
  m_backPointers.clear();

  // take some values from the parent hyp as this hyp might not have
  // been cloned from the parent
  // (as done in InitializeHeapData) but possibly from a sibling
  // (as done in ExpandNeighbors)
  m_taggedTargets = m_pParentHyp->m_taggedTargets;

  // if we're in cmp0, we have to recalculate the LM costs of the
  // target words of this hyp's corresponding
  // rule and of the first C_HistoryLength words of each underlying
  // (backpointed) hyp since
  // our previous LM estimates were based on underlying nodes' best
  // hyps' LM contexts.
  // If --RecomputeLMCostsDuringKBest=1, all LM costs are
  // recalculated anyway after the underlying hyps have been slurped
  // in, so in that case, we don't need to set recomputeNetLM (we
  // *will* set it in the sanity test case --RecomputeLMCostsDuringKBest=2 though)
  bool recomputeNetLM = !CChartParserInfo::m_compareTargetWords
      && !(CChartParserInfo::m_recomputeLMCostsDuringKBest==1);
  // The words whose LM-costs have to be re-calculated in
  // case of recomputeNetLM.
  // This will include the LM contexts of underlying hyps
  // and is computed in the spirit of
  // CChartParserNode::HarvestLMString with the difference that
  // BP substitution sites
  // correspond to concrete hyps rather than nodes.
  // This will also be used to determine this hyp's m_lmContext
  CWordSequence lmString;

#ifndef NDEBUG
  // do some sanity testing

  // see if ashish's old tagging moves lead to the same stuff that we have gotten from the parenthyp
  CWordSequence sanitytest_taggedTargets;
  //if this node is the @NUM hyp
  if (CChartParserInfo::m_tagsForLM.count(m_productionResult)!=0){
    assert(m_pParentHyp!=NULL);
    //cout<<"Adding: "; CChartParserInfo::PrintVocabSequence(m_pParentHyp->m_taggedTargets); cout<<endl;
    for (CWordSequence::iterator tagIter = m_pParentHyp->m_taggedTargets.begin(); tagIter!= m_pParentHyp->m_taggedTargets.end(); ++tagIter){
      sanitytest_taggedTargets.push_back(*tagIter);
    }
  }
  assert(sanitytest_taggedTargets==m_taggedTargets);

  // further, we will check whether the backpointer nodes are
  // actually being 'used' by corresponding BP
  // references in the translationOutput
  int sanitytest_NT_count = 0;
#endif

  // now if this == @NUM hyp, the m_taggedTargets are properly assigned
  // if its not the @NUM hyp, but rather some hyp above it, then the
  // m_taggedTargets will be propagated
  //similarly with _UNKNOWN hyps
  m_scaledFutureCostEstimate = 0.0; // no estimate anymore, from now on all costs will be correct
  // Start with the net neglog probs (inlcuding LM neglog probs) only
  // only; below we will add the underlying hyps' neglog probs. Note
  // that the (forward-) LM neglog probs also include the neglog probs of the
  // first-words of backpointer-substitution-sites. E.g., If this
  // hyp's m_translationOutput is 'a b @NP c' and @NP has m_lmContext
  // 'x y <*> y z', then these m_negLogProbs contain already the LM
  // neglog probs of a, b, x, y, and c (but not z).
  m_negLogProbs = m_pParentHyp->m_negLogProbs;

  CChartParserHypComplete::AddParseOutputSubTreeStart(m_parsingOutput, this);

  CWordSequence::iterator tagIter = m_pParentHyp->m_taggedTargets.begin();
  // iterate through the translation output of the hyp, slurping in for each backpointer site the next best underlying
  // hyp, creating a translation output exclusively consisting of terminals
  for (CWordSequence::iterator iter = m_pParentHyp->m_translationOutput.begin(); iter != m_pParentHyp->m_translationOutput.end(); ++iter){
    if (*iter>0){
      // a word - store it as is in the translation output
      m_translationOutput.push_back(*iter);

      if (CChartParserInfo::m_tagsForLM.count(*iter)!=0){
	//*iter IS a @NUM style marker.
	//m_taggedTargets should now contain something like "two and a half @NUM"
	assert(tagIter!=m_pParentHyp->m_taggedTargets.end());
	//keep writing the tags until you reach another terminal
	while (CChartParserInfo::m_tagsForLM.count(*tagIter)==0){
	  //cout << CChartParserInfo::m_pVocabulary->GetWord(*tagIter)<<" ";
	  m_parsingOutput.push_back(*tagIter);
	  ++tagIter;
	  //there should ALWAYS be a @TAG marker after the actual surface output
	  assert(tagIter!=m_pParentHyp->m_taggedTargets.end());
	}
	//skip past the fake @TAG that we have put onto the m_taggedTargets as a marker
	++tagIter;
	//tagIter can be at end now
      }else{
	m_parsingOutput.push_back(*iter);
      }
      if (recomputeNetLM) {
	// and also in the lmString of words to LM-recompute
	lmString.push_back(*iter);
      }
    }else{
      // a backpointer site - get the corresponding node backpointed to
      int bpIndex = -(*iter);
      assert(bpIndex<m_pParentHyp->m_backPointers.size());
      CChartParserNode* pNode = m_pParentHyp->m_backPointers[bpIndex];

#ifndef NDEBUG
      sanitytest_NT_count++;
#endif

      // here comes the basic hc05 recursion: retrieve kth best
      // (where k is stored in m_derivationVector) hyp
      // in this node
      CChartParserHypComplete* pBackCandidate =
          pNode->GetKthBest(m_derivationVector[bpIndex]);
      if (pBackCandidate==NULL){
	//we were not able to find the KthBest as requested so return false
	return false;
      }
      // pBackCandidate has itself recursively retrieved
      // instantiations for its
      // underlying backpointer hyps so that its
      // m_translationOutput now exclusively
      // consists of terminal symbols

      //cout<<"Using KthBest..."<<endl;
      //pBackCandidate->DisplaySimpleKBest();
      //cout<<endl;

      //add underlying hyp's translation output to our
      //so-far accumulated translation output
      m_translationOutput.insert(
          m_translationOutput.end(),
          pBackCandidate->m_translationOutput.begin(),
          pBackCandidate->m_translationOutput.end());

      if (recomputeNetLM) {
	// add underlying hyp's LM context to our current lmString
	// so that we can calculate the hyp's first C_HistoryLength
        // words' LM cost
	// and use the hyp's last C_HistoryLength words as history
        // context for target words to follow
	lmString.insert(lmString.end(), pBackCandidate->m_lmContext.begin(), pBackCandidate->m_lmContext.end());
      }

      // store underlying tree-form translation
      m_parsingOutput.insert(
          m_parsingOutput.end(),
          pBackCandidate->m_parsingOutput.begin(),
          pBackCandidate->m_parsingOutput.end());

      // propagate any unknown or tagged words
      // we can't have unknown-sourcewords in this hyp's underlying
      // pBackCandidate and at the same time
      // in this hyp's corresponding translation rule
      assert(pBackCandidate->m_targetOrderedUnknownSourceWords.size()==0 || m_pParentHyp->m_targetOrderedUnknownSourceWords.size()==0);
      m_targetOrderedUnknownSourceWords.insert(m_targetOrderedUnknownSourceWords.end(), pBackCandidate->m_targetOrderedUnknownSourceWords.begin(), pBackCandidate->m_targetOrderedUnknownSourceWords.end());
      assert(pBackCandidate->GetResult() != CChartParserInfo::m_unknownWI || m_targetOrderedUnknownSourceWords.size()>0);

      // similarly, we can't have taggedTargets in both this hyp and the underlying pBackCandidate
      assert(pBackCandidate->m_taggedTargets.size()==0
             || m_pParentHyp->m_taggedTargets.size()==0);
      m_taggedTargets.insert(m_taggedTargets.end(),
                             pBackCandidate->m_taggedTargets.begin(),
                             pBackCandidate->m_taggedTargets.end());

      // neg log probs from the pBackCandidate
      // (ie really accumulating the lm cost)
      for (int m=0; m<CRule::C_NumModelsTotal; m++){
	m_negLogProbs[m] += pBackCandidate->m_negLogProbs[m];
      }
      //there's no future cost estimate anymore
      assert(pBackCandidate->m_scaledFutureCostEstimate==0);
    }
  }
#ifndef NDEBUG
  assert(sanitytest_NT_count == m_pParentHyp->m_backPointers.size());
#endif

  //end the parsing output
  //m_parsingOutput.push_back(CChartParserInfo::m_pVocabulary->GetIndex("#)"));
  CChartParserHypComplete::AddParseOutputSubTreeEnd(m_parsingOutput);

  // in the  following, we'll update the LM cost and (in cmp0) the LM context and the boundary cost

  // This is for cmp0:
  if (recomputeNetLM) {
    // determine the new LM context
    m_lmContext = CUtilities::ComputeLMContext(lmString);
    // get the inside (i.e., not including first C_HistoryLength
    // words) LM costs, add it to the LM cost locations, and subtract
    // the original inside LM costs, which were only estimates.
    // Also, compute the new boundary cost estimate
    m_scaledBoundaryCost = 0.0;
    for (int i=0; i<CRule::C_NumLMs; i++) {
      // correct inside-cost of i-th LM
      m_negLogProbs[CRule::C_LMLocation+i] +=
          CUtilities::ComputeLMCost(
              CLanguageModel::m_languageModels[i], lmString, false, true)
	- m_pParentHyp->m_negLogProbs[CRule::C_LMLocation+i];
      // and add this LM's boundary LM cost estimate of this context to
      // the boundary cost estimate
      // Note: instead of m_lmContext, we can also pass lmString here,
      // which will have the same result for FORWARDLM or BACKWARDLM but
      // a more accurate estimate for CENTERLM, but in the moment we're
      // not using CENTERLM during the parsing step anyway
      m_scaledBoundaryCost +=
          CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i] *
	CUtilities::ComputeLMCost(
            CLanguageModel::m_languageModels[i], m_lmContext, true, false);
    }
  }

  // If we want to re-compute all LM costs from scratch:
  if (CChartParserInfo::m_recomputeLMCostsDuringKBest>0) {
    // m_lmContext and m_scaledBoundaryCost should not change -
    // only if we are in cmp1
    assert(m_lmContext ==
           CUtilities::ComputeLMContext(m_translationOutput)
           || !CChartParserInfo::m_compareTargetWords);
    // get the inside (i.e., not including first C_HistoryLength
    // words) LM costs, add it to the LM cost locations, and subtract
    // the original inside LM costs, which were only estimates.
    // Also, re-compute the boundary cost estimate for sanity-check purposes
    float sanitytest_scaledBoundaryCost = 0.0;
    for (int i=0; i<CRule::C_NumLMs; i++) {
      // correct the inside-cost of i-th LM
      float c = CUtilities::ComputeLMCost(CLanguageModel::m_languageModels[i], m_translationOutput, false, true);
      if ((CChartParserInfo::m_recomputeLMCostsDuringKBest==2 && fabs(c - m_negLogProbs[CRule::C_LMLocation+i]) > 0.001) && CChartParserInfo::m_compareTargetWords) {
	cerr << "Sanity check error - check STDOUT for more info." << endl;
	cout << "Sanity check error: --RecomputeLMCostsDuringKBest=2 was passed, but for the " << (i+1) <<
	  "-th LM recomputing LM costs during K-best yielded a different cost (recomputedCost=" << c << " != " <<
	  m_negLogProbs[CRule::C_LMLocation+i] << "=originalCost)! Hyp concerned:" << endl;
	DisplayRecursive();
	cout << "Cloned from m_pParentHyp:" << endl;
	m_pParentHyp->DisplayRecursive();
	exit(2);
      }
      m_negLogProbs[CRule::C_LMLocation+i] = c;
      // and add this LM's boundary LM cost estimate of this context to
      // the boundary cost estimate
      // Note: instead of m_lmContext, we can also pass m_translationOutput here,
      // which will have the same result for FORWARDLM or BACKWARDLM but
      // a more accurate estimate for CENTERLM, but in the moment we're
      // not using CENTERLM during the parsing step anyway
      sanitytest_scaledBoundaryCost +=
          CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i] *
          CUtilities::ComputeLMCost(
              CLanguageModel::m_languageModels[i], m_lmContext, true, false);
    }
    if ( (fabs(m_scaledBoundaryCost - sanitytest_scaledBoundaryCost) > 0.001) && CChartParserInfo::m_compareTargetWords) {
	cerr << "Sanity check error - check STDOUT for more info." << endl;
	cout << "Sanity check error: when recomputing LM boundary cost estimate during K-best yielded a different cost (recomputedBoundaryCost=" << sanitytest_scaledBoundaryCost << " != " <<
	  m_scaledBoundaryCost << "=originalBoundaryCost)! Hyp concerned:" << endl;
	DisplayRecursive();
	cout << "Cloned from m_pParentHyp:" << endl;
	m_pParentHyp->DisplayRecursive();
	exit(2);
    }

  } else {
    // use m_lmContext and m_scaledBoundaryCost from cloned parent/sibling
  }

  // calculate m_totalCost based on the neg log probs and feature weights m_scalingFactors
  m_totalCost = 0.0;
  for (int i=0; i<m_negLogProbs.size(); i++){
    m_totalCost += m_negLogProbs[i]*CChartParserInfo::m_scalingFactors[i];
  }
  assert(m_scaledFutureCostEstimate==0.0);
  //we were able to instantiate the k-best as proposed in the derivation vector for this hyp
  return true;
}








KBestCandidateHeap::KBestCandidateHeap(){
}
KBestCandidateHeap::~KBestCandidateHeap(){
  Clear();
}

bool KBestCandidateHeap::InsertBaseCombination(CChartParserHypComplete* toAdd){
   bool debug = false;

   //track the top level proposed additions
   CChartParserInfo::m_proposedAddedKBest++;

  if (debug){
    cout<<"InsertBaseCombination: Inserting: ";
    toAdd->Display();
  }
  //don't add duplicates from a k-best standpoint, ie same edge + configuration
  if (m_uniqueTuples.count(toAdd)==0){


    //insert into both the unique set and the candidate set
    m_uniqueTuples.insert(toAdd);
    //AS pq
    m_candidateSet.push(toAdd);
  }else{
    //the hyp even gets instantiated to save unnecessary computations
    assert(false);
    if (debug){
      cout<<"InsertBaseCombination:  *** existing match *** "<<endl;
      toAdd->Display();
    }
    delete toAdd;
    return false;
  }
  return true;
}

CChartParserHypComplete* KBestCandidateHeap::PopHeap(){
  if (m_candidateSet.size()==0){
    return NULL;
  }
  CChartParserHypComplete* topElement = m_candidateSet.top();
  assert(topElement!=NULL);
  m_candidateSet.pop();

  return topElement;
}

void KBestCandidateHeap::AddNeighbors(CChartParserHypComplete* pTopElement){
  vector<CChartParserHypComplete*> frontier = pTopElement->ExpandNeighbors(m_uniqueTuples);
  for (vector<CChartParserHypComplete*>::iterator iter = frontier.begin(); iter!= frontier.end(); ++iter){
    //m_candidateSet.insert(std::pair<double, CCombinationTuple*>((*iter)->m_cost, *iter));
    InsertBaseCombination(*iter);
  }
}

void KBestCandidateHeap::Clear(){
  // we can't delete the pointers in the hash table directly since for some weird reason iterating through
  // the hash table calls the equals function, which dereferences the pointers, causing a crash when the
  // corresponding objects don't exist anymore
  vector<CChartParserHypComplete*> del;
  for (UniqueHypsKBest::iterator tmpIter = m_uniqueTuples.begin(); tmpIter!=m_uniqueTuples.end(); ++tmpIter){
    CChartParserHypComplete* ptr = (*tmpIter);
    //cout<<"Deleting: "<<ptr<<" "; (*tmpIter)->Display();
    //cout<<endl;
    //  delete ptr;
    del.push_back(ptr);
  }
    for(vector<CChartParserHypComplete*>::iterator tmpIter = del.begin(); tmpIter != del.end(); ++tmpIter){
      CChartParserHypComplete* ptr = (*tmpIter);
      delete ptr;
      ptr=NULL;
    }
  
  m_uniqueTuples.clear();
}

void CChartParserHypCompleteEqClass::ClearRecombined(){
  for (vector<CChartParserHypCompleteNBest*>::iterator iter= m_recombinedHyps.begin(); iter!= m_recombinedHyps.end(); ++iter){
    CChartParserHypCompleteNBest* pRecHyp = (*iter);
    assert(pRecHyp!=NULL);
    delete pRecHyp;
    pRecHyp=NULL;
  }

}

//!< Add the hyp that was expanded to generate THIS edge
void CChartParserHypCompleteNBest::SetBackInfo(
    CChartParserHypCompleteEqClass* pBackEqClass,
    CChartParserHypCompleteNBest* pBackHyp){
  m_pBackEqClass=pBackEqClass;
  m_pBackHyp = pBackHyp;
}

CChartParserHypCompleteEqClass::~CChartParserHypCompleteEqClass(){
  //ClearRecombined must already have been called
  if (m_pHeap!=NULL){
    m_pHeap->Clear();
    delete m_pHeap;
    m_pHeap = NULL;
  }
}

// Just add the recombined hyps during parsing, returns true IF this
// hyp is better than the current prototype value
bool CChartParserHypCompleteEqClass::AddRecombinedHyp(
    CChartParserHypCompleteNBest* pHyp, bool& olddeactivated){

  // if are adding the very first element, or it
  // beats the cost of the existing element
  if (pHyp->ScaledTMLMCost()<this->ScaledTMLMCost() || m_recombinedHyps.size()==0){

    this->m_totalCost = pHyp->m_totalCost;
    this->m_scaledFutureCostEstimate = pHyp->m_scaledFutureCostEstimate;
    assert(fabs(this->ScaledTMLMCost()-pHyp->ScaledTMLMCost())<0.001);

    if (m_pCurrentBest==NULL){
      m_pCurrentBest = pHyp;
    }else{
      //some other hyp who IS on the agenda just got displaced, so de-activate him
      //only indicate that an old was deactivated if it was active in the first place!!
      if (m_pCurrentBest->m_active){
	olddeactivated = true;
      }
      m_pCurrentBest->Deactivate();
      m_pCurrentBest = pHyp;
    }
    //make sure you add the new hyp
    m_recombinedHyps.push_back(pHyp);

    return true;
  }else{
    //there was an existing hyp that had BETTER cost than pHyp, deactivate this one

    if (CChartParserInfo::m_saveAllAlternatives){
      m_recombinedHyps.push_back(pHyp);
    }
    
    //dont want to delete the hyp at all.. but we could de activate it    
    pHyp->Deactivate();
    return false;
  }
}


CChartParserHypCompleteNBest* CChartParserHypCompleteEqClass::GetKthBest(int k, bool isTopLevel){
  bool debug = false;

  // have we already created the hyp asked for?
  if (m_candidateVector.size()>k){
    // yes - just return it
    if (debug){
      cout<<"m_candidateVector: "<<m_candidateVector.size()<<endl;
    }
    return (CChartParserHypCompleteNBest*)(m_candidateVector[k]);
  }else{
    // no, create it
    //if nothing is initialized yet
    if (m_pHeap==NULL){
      if (debug) {
	cout<<"InitializingClass: "; this->PrintTranslationOutput(); cout<<" unwnd: "<<m_completelyUnwound<<" m_posNT: "<<m_posNT<<endl;
      }
      m_pHeap = new KBestCandidateHeap();
      
      //generate unique recombined hyps - unique in that no two hyps here point back to the same last equivalence class
      vector<CChartParserHypCompleteNBest*> uniqueRecHyps = IdentifyUniqueHyps();
      
      if (debug){
	cout<<"Hyps to initialize: "<<uniqueRecHyps.size()<<endl;
	for (vector<CChartParserHypCompleteNBest*>::iterator iter= uniqueRecHyps.begin(); iter!= uniqueRecHyps.end(); ++iter){
	  CChartParserHypCompleteNBest* pHyp = (*iter);
	  cout<<" \t Init: "; pHyp->PrintTranslationOutput(); cout<<" Cost: "<<pHyp->ScaledTMLMCost()<<endl;
	}
      }
      for (vector<CChartParserHypCompleteNBest*>::iterator iter= uniqueRecHyps.begin(); iter!= uniqueRecHyps.end(); ++iter){
	CChartParserHypCompleteNBest* pHyp = (*iter);
	CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(*pHyp);
	pHyp->m_pParentHyp = NULL;
	
	//there is only 1 back pointer when doing hc05 style search
	bool wasInited = pNewHyp->InitializeHeapData(pHyp, 1);
	if (!wasInited){
	  cout<<" NO INIT POSSIBLE=========== "<<endl;
	  delete pNewHyp;
	  pNewHyp = NULL;
	}else{
	  //add to the heap
	  assert(m_pHeap!=NULL);
	  //this wasAdded tells you whether there was another hyp with the same derivation... cant be at init time!
	  //cout<<" Adding =========== "<<endl;
	  bool wasAdded = m_pHeap->InsertBaseCombination(pNewHyp);

	  //the only reason you should get immediately rejected is if you have m_extractUnique on
	  if (wasAdded==false){
	    if (debug){
	      cout<<" Rejected on heap ===== "<<endl;
	    }
	    assert(CChartParserInfo::m_extractUnique);
	  }
	}
      }
    }
  }
  //now you have initialized
  if (debug){
    cout<<"Initialized: "; this->PrintTranslationOutput(); cout<<endl;
  }

  int count = 0;

  while (m_candidateVector.size()<=k){
    //track some top level stats
    if (isTopLevel){
      CChartParserInfo::m_topCallsKBest++;
    }
    //pop top of heap, add nearby neighbors if the heap wasnt empty
    assert(m_pHeap!=NULL);
    //CChartParserHypCompleteNBest* pTopHyp = (CChartParserHypCompleteNBest*)(m_pHeap->NextTuple());
    CChartParserHypCompleteNBest* pTopHyp = (CChartParserHypCompleteNBest*)(m_pHeap->PopHeap());

    if (pTopHyp==NULL){
      if (debug){
	cout<<"Received null at k:"<<k<<endl;
      }
      break;
    }
    
    //if checking at pop - maintain a separate unique set here (dont want to conflict with the one for the heap)
    bool checkAtPop = true;
    //if (checkAtPop && CChartParserInfo::m_extractUnique && CChartParserInfo::m_compareTargetWords==1){
    if (checkAtPop && CChartParserInfo::m_extractUnique){
      //see if its a duplicate dont delete it, the ptr is reference in the m_uniqueTuples
      if (m_uniqueFinalNBest.CheckUnique(pTopHyp)!=NULL){
	m_candidateVector.push_back(pTopHyp);
      }else{
	//cout<<"Not added to m_candidateVector: "; pTopHyp->Display(); cout<<endl;
      }
    }else{
      m_candidateVector.push_back(pTopHyp);
    }

    //add all the neighbors
    m_pHeap->AddNeighbors(pTopHyp);

    count++;
  }

  //cout<<"m_candidateVector.size(): "<<m_candidateVector.size()<<endl;
  if (m_candidateVector.size()<=k){
    //cerr<<"WARNING: requested k-best that was too deep in some node";
    return NULL;
  }
  
  assert(m_candidateVector.size()>k);
  if (debug){
    cout<<"Returning: "; this->PrintTranslationOutput(); cout<<" k: "<<k<<endl;
  }
  return (CChartParserHypCompleteNBest*)(m_candidateVector[k]);


}


void CChartParserHypCompleteEqClass::PrintNBest(){
  for (vector<CChartParserHypComplete*>::iterator iter = m_candidateVector.begin(); iter!= m_candidateVector.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    assert(pHyp->m_scaledFutureCostEstimate==0.0);
    cout<<"NBest: "; pHyp->PrintTranslationOutput(); cout<<" Cost: "<<pHyp->ScaledTMLMCost()<<endl;
  }
  
}


vector<CChartParserHypCompleteNBest*> CChartParserHypCompleteEqClass::IdentifyUniqueHyps(){
  //all the hyps in this class have the same remaining  non-terminals, so we just have to 
  //check to make sure that if you have the same parent hyp, we only propagate the lower cost one
  map<long, CChartParserHypCompleteNBest*> uniquePaths;

  //if you have only the 1 recombined hyp, then you are unique
  if (m_recombinedHyps.size()==1){
    return m_recombinedHyps;
  }
  for (vector<CChartParserHypCompleteNBest*>::iterator iter = m_recombinedHyps.begin(); iter!= m_recombinedHyps.end(); ++iter){
    CChartParserHypCompleteNBest* pHyp = (*iter);
    //if (pHyp->m_pBackHyp==NULL){
    //  cout<<"NULL Back: "; pHyp->Display();      
    //}
    //assert(pHyp->m_pBackEqClass!=NULL);
    long addr  = (long)(pHyp->m_pBackEqClass);
    if (uniquePaths.count(addr)==0){
      uniquePaths[addr] =  pHyp;
    }else{
      map<long, CChartParserHypCompleteNBest*>::iterator dupIter = uniquePaths.find(addr);
      assert(dupIter!=uniquePaths.end());
      CChartParserHypCompleteNBest* pExistingHyp = (*dupIter).second;
      if (pHyp->ScaledTMLMCost()<pExistingHyp->ScaledTMLMCost()){
	uniquePaths[addr] = pHyp;
      }
    }
  }

  // push it all onto the vector
  vector<CChartParserHypCompleteNBest*> toRet;
  for (map<long, CChartParserHypCompleteNBest*>::iterator iter = uniquePaths.begin(); iter!= uniquePaths.end(); ++iter){
    toRet.push_back((*iter).second);
  }
  return toRet;
}
