/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#include "ChartParserHypBeam.hh"
#include <math.h>

#include <set>


//============================================================================
// CHypBeam
//============================================================================

CChartParserHypCompleteEqClass* CHypBeam::LookupProto( CChartParserHypCompleteNBest* pHyp){
  TContextHyps::iterator iter = m_lmContexts.find(pHyp);
  if (iter==m_lmContexts.end()){
    return NULL;
  }else{
    CChartParserHypCompleteEqClass* pHyp =  (CChartParserHypCompleteEqClass*)(*iter);
    assert(pHyp!=NULL);
    return pHyp;
  }
}

//! Utility function to get the context words from the m_firstWords
CWordSequence CHypBeam::GetContextWords(CWordSequence& leadingTerminals){
  int count=0;
  CWordSequence lastContext;
  for (CWordSequence::reverse_iterator wordIter = leadingTerminals.rbegin(); wordIter!=leadingTerminals.rend(); ++wordIter){
    lastContext.push_back(*wordIter);
    count++;
    if (count==CChartParserInfo::C_HistoryLength){
      break;
    }
  }
  reverse(lastContext.begin(), lastContext.end());
  return lastContext;
}

float CHypBeam::GetLowestCost(){
  if (CChartParserInfo::m_breadthAStar==0){
    return ((m_store.begin()))->second->ScaledTMLMCost();
  }else{
    return m_lowestCost; //gets set in InsertHyp and reset in GetAgenda
  }
}

vector<CChartParserHypCompleteNBest*> CHypBeam::GetAgenda(double beam){

  //but it stays compatible with the old method so lets not change it yet
  float lowestCost = 9999;
  vector<CChartParserHypCompleteNBest*> agenda;
  bool exceeded = false;
  vector<CSortedHypNBestContainer::iterator> toerase;

  //sort it - now the first hyp should be the lowest cost one
  sort(m_unsortedstore.begin(), m_unsortedstore.end(), CompareCompleteHyps());

  //go thru the whole set of created elements... delete the useless elements
  //for (CSortedHypNBestContainer::iterator iter = m_store.begin(); iter!= m_store.end(); ++iter){
  
  for (vector<CChartParserHypCompleteNBest*>::iterator iter = m_unsortedstore.begin(); iter!= m_unsortedstore.end(); ++iter){
    //float cost = iter->second->ScaledTMLMCost();
    float cost = (*iter)->ScaledTMLMCost();
    //just keep the cost 
    if (iter==m_unsortedstore.begin()){
      lowestCost = cost;
    }

    //if the hyp is inactive, then it needs to be deleted as well!
    //two ways to become inactive - removal cause you were just unwound, and getting marked down due to eq classing
    if ((*iter)->m_active==false){
      //toerase.push_back(iter);
      //there is nothing to erase now...
    }else{
      //if u have already exceeded there is no reason to check!
      if (!exceeded){
	if (cost > CChartParserInfo::GetPruningCutoff(lowestCost, beam) ){		  
	  exceeded = true;
	  //toerase.push_back(iter);
	}else{
	  agenda.push_back((*iter));
	}
      }else{
	//toerase.push_back(iter);
      }
    }
  }
  //remove all the elements at the end!
  //for (vector<CSortedHypNBestContainer::iterator>::iterator iter = toerase.begin(); iter!= toerase.end(); ++iter){
  //m_store.erase(*iter);
  //}
  
  //we have cherry picked the best elements from the m_unsortedstore to survive. we can clear it now
  m_unsortedstore.clear();
  if (agenda.size()>0){
    m_lowestCost = agenda[0]->ScaledTMLMCost();
  }
  //only keep the first m_maxHypCount elements in the beam
  if (agenda.size()>m_maxHypCount){
    if (!m_capacityExceeded) {
      // first time this happened
      m_capacityExceeded = true;
      cout << "WARNING: maximum capacity of beam " << m_maxHypCount << " exceeded.\n";
    }
    //keep track of the deletion
    CChartParserInfo::m_deletedKBest++;
    agenda.resize(m_maxHypCount);
  }

  return agenda;
}

//insert a new hyp into this beam
//only up to m_maxHypCount hyps are stored; when an additional hyp is supposed to inserted such that the number of hyps stored is higher
//than m_maxHypCount, the highest-cost hyp is removed and deleted (that could also be the hyp supposed to be inserted)!
//!Store only the lowest cost m_maxHypCount objects
bool CHypBeam::InsertHyp(CChartParserHypCompleteNBest* pHyp) {

  //stats - each time you call InsertHyp you are proposing an addition, that may or may not be accepted
  //on the grounds of uniqueness or overfilling
  CChartParserInfo::m_proposedAddedKBest++;

  //if we are checking for uniqueness, see if the hyp that is returned is still valid
  //if its not, just quit
  //arv: this uniqueness check is not neccesary, the classing based check is stronger than this check
  //if (CChartParserInfo::m_extractUnique){
  //if (m_uniqueChecker.CheckUnique(pHyp)==NULL){
  ///delete it, we wont use it
  //delete pHyp;
  //pHyp=NULL;
  //return;
  //}
  //}

  //============ 
  // Hyp comes in to participate on the global beam. One of three conditions can occur
  // a) The hyp has brand new LM context. so generate a new prototype hyp, and add it to the expansion beam
  // b) The hyp has a existing LM context
  //     - and has HIGHER cost - simply file it INTO the prototype hyp that it matches
  //     - and has LOWER cost  - file it INTO the prototype hyp, but also let it onto the expansion beam
  //============
#ifndef NDEBUG
  bool debug = true;
#endif

#ifndef NDEBUG
  if (debug){
    cout<<"InsertHyp: "; pHyp->PrintTranslationOutput(); cout<<" CTX: "; CChartParserInfo::PrintVocabSequence(pHyp->m_lmContext);  cout<<" unwnd: "<<pHyp->m_completelyUnwound<<" m_posNT: "<<pHyp->m_posNT<<" cost: "<<pHyp->ScaledTMLMCost()<<endl;
  }
#endif

  //first check if this is this hyp has a unique context, it it is, then register a new equivalence class and add it the new eq class
  //to the m_lmContexts structure that manages them. (Hyps dont get stored here, they go directly to the beam if they pass these tests)
  if (m_lmContexts.count(pHyp)==0){
    CChartParserHypCompleteEqClass* pNewProto = new CChartParserHypCompleteEqClass(pHyp);
    bool olddeactivated;
    pNewProto->AddRecombinedHyp(pHyp, olddeactivated);
    assert(fabs(pNewProto->ScaledTMLMCost()-pHyp->ScaledTMLMCost())<0.001);
    //and we will proceed to add this eq class to the best. we must take care that we dont delete it somewhere else
    m_lmContexts.insert((CChartParserHypCompleteNBest*)pNewProto);
#ifndef NDEBUG
    if (debug){
      cout<<" NEW ";
    }
#endif
  }else{
    //there was an existing eq class, but you still might be better
    TContextHyps::iterator iterProto = m_lmContexts.find(pHyp);
    assert(iterProto!=m_lmContexts.end());
    CChartParserHypCompleteEqClass* pProto =  (CChartParserHypCompleteEqClass*)(*iterProto);
#ifndef NDEBUG
    if (debug){
      cout<<" EXISTING\n PROTO: "; pProto->PrintTranslationOutput();  cout<<" CTX: "; CChartParserInfo::PrintVocabSequence(pProto->m_lmContext); cout<<" unwnd: "<<pProto->m_completelyUnwound<<" ";
    }
#endif
    //make sure that you found the right hyp

    //assert(pProto->m_lmContext == pHyp->m_lmContext);
    //assert(GetContextWords(pProto->m_firstWords)==GetContextWords(pHyp->m_firstWords) || (pHyp->m_completelyUnwound==true && pProto->m_completelyUnwound==true));

    //arv : this seems wrong, you do want to keep the alternative inside the eq class for later expansion
    //if !wasBetter the hyps is deleted below. AddRecombinedHyp takes care of inactivating items that have been beaten

    bool olddeactivated = false; //find out if an element that used to be active on the beam was de-activated
    bool wasBetter = pProto->AddRecombinedHyp(pHyp, olddeactivated);

    //if olddeactivated then you must have been better
    if (olddeactivated && CChartParserInfo::m_updateBeamSize){ 
      assert(wasBetter);
      m_maxHypCount++;
#ifndef NDEBUG
      if (debug){
	cout<<"Increasing beam size to: "<<m_maxHypCount<<endl;
    }
#endif
    }

    //if you performed WORSE, then dont be added to the list of expandable items
    if (!wasBetter){
#ifndef NDEBUG
      if (debug){
	cout<<" WORSE"<<endl;
      }
#endif
      if (!CChartParserInfo::m_saveAllAlternatives){
	delete pHyp;
	pHyp = NULL;
      }

      return false;

    }
    
#ifndef NDEBUG
    if (debug){
      cout<<" BETTER";
    }
#endif

    
    if (CChartParserInfo::m_breadthAStar==0){
      //now even if you are better, you might be eq classes away. but we have to ensure
    //that the best hyp always goes back on the top
      CSortedHypNBestContainer::iterator iterFirstElement = m_store.begin();
      //dont let the beam become empty
      CChartParserHypCompleteNBest* pBestHyp = NULL;
      if (iterFirstElement!=m_store.end()){
	pBestHyp = iterFirstElement->second;
      }
      //if u are the new first best please keep going, otherwise if you have set m_dontReloadBeam then you will just
      //become a sleeper until activated by HC05onBeam
      if (CChartParserInfo::m_dontReloadBeam && !pHyp->m_completelyUnwound && iterFirstElement!=m_store.end() && pBestHyp!=NULL && pBestHyp->ScaledTMLMCost()<pHyp->ScaledTMLMCost()){
#ifndef NDEBUG
	if (debug){
	  cout<<"NO RELOAD"<<endl;
	}
#endif
	return true;
      }
    }
    
  }
#ifndef NDEBUG
  if (debug){
    cout<<endl;
  }
#endif
  
  if (CChartParserInfo::m_breadthAStar==0){
    //this beam considers all hyps to be fully sentence spanning and comparable - NOTE we only use ScaledTMLMCost here
    //m_store.insert(std::pair<float, CChartParserHypComplete*>(pHyp->TotalCost()+pHyp->m_scaledFutureCostEstimate, pHyp));
    m_store.insert(std::pair<float, CChartParserHypCompleteNBest*>(pHyp->ScaledTMLMCost(), pHyp));
    //  }
    //do we have more than m_maxHypCount hyps in the container now?
    if (m_store.size() > m_maxHypCount) {
    // yepp - remove and delete (!) the worst
      CSortedHypNBestContainer::iterator iterLastElem = m_store.end();
      iterLastElem--;
      //CChartParserHypComplete* pDelHyp = (*iterLastElem).second;
      //ARV 10/03/2006
      //delete pDelHyp;  //DONT DELETE THIS HYP - we store references to this hyp in the prototype
      //    cout << "deleted pDelHyp" << endl; //!!
      m_store.erase(iterLastElem);
    //    cout << "erased iterLastElem" << endl; //!!
    // should have exactly m_maxHypCount hyps now, otherwise something very weird is going on (such as insertion without using InsertHyp)
      assert(m_store.size() == m_maxHypCount); 
      
      if (!m_capacityExceeded) {
	// first time this happened
      m_capacityExceeded = true;
      cout << "WARNING: maximum capacity of beam " << m_maxHypCount << " exceeded.\n";
      }
    //keep track of the deletion
      CChartParserInfo::m_deletedKBest++;
    } 
  }else{
    //we dont have to deal with sorting now!
    assert(pHyp->m_scaledBoundaryCost==0);
    m_unsortedstore.push_back(pHyp);
  }
  return true;
}

//! Print the beam out 
void CHypBeam::PrintBeam(){
  bool debug = false;
  if (CChartParserInfo::m_breadthAStar==0){
    for (CSortedHypNBestContainer::iterator iter = m_store.begin(); iter!= m_store.end(); ++iter){
      CChartParserHypCompleteNBest* pHyp = iter->second;
      cout<<"BeamElt: "; pHyp->PrintTranslationOutput(); cout<<" unwnd: "<<pHyp->m_completelyUnwound; cout<<" TMCost: "<<pHyp->m_totalCost<<" Total: "<<pHyp->ScaledTMLMCost()<<" m_posNT: "<<pHyp->m_posNT<<" m_active: "<<pHyp->m_active<<endl;
      if (debug){
	pHyp->Display(); cout<<endl;
      }
    }
  }else{
    //work on the unsorted beam
    for (vector<CChartParserHypCompleteNBest*>::iterator iter = m_unsortedstore.begin(); iter!= m_unsortedstore.end(); ++iter){
      CChartParserHypCompleteNBest* pHyp = (*iter);
      cout<<"BeamElt: "; pHyp->PrintTranslationOutput(); cout<<" unwnd: "<<pHyp->m_completelyUnwound; cout<<" TMCost: "<<pHyp->m_totalCost<<" Total: "<<pHyp->ScaledTMLMCost()<<" m_posNT: "<<pHyp->m_posNT<<" m_active: "<<pHyp->m_active<<endl;
      if (debug){
	pHyp->Display(); cout<<endl;
      }
    }
  }
}

//! Print the equivalence classes in the beam
void CHypBeam::PrintClasses(){
  for (TContextHyps::iterator iter  =  m_lmContexts.begin(); iter!= m_lmContexts.end(); ++iter){
    CChartParserHypCompleteEqClass* pHyp  = ( CChartParserHypCompleteEqClass*)(*iter);
    cout<<"EQ: "; pHyp->PrintTranslationOutput(); cout<<" unwnd: "<<pHyp->m_completelyUnwound; cout<<" NumRec: "<<pHyp->m_recombinedHyps.size()<<" Cost: "<<pHyp->ScaledTMLMCost()<<endl;
    
for (vector<CChartParserHypCompleteNBest*>::iterator iterRecomb = pHyp->m_recombinedHyps.begin(); iterRecomb != pHyp->m_recombinedHyps.end(); ++iterRecomb){
      CChartParserHypCompleteNBest* pRecHyp = (*iterRecomb);
      cout<<"\t\t";  pRecHyp->PrintTranslationOutput(); cout<<" unwnd: "<<pRecHyp->m_completelyUnwound<<" Cost: "<<pRecHyp->ScaledTMLMCost(); cout<<" active: "<<pRecHyp->m_active<<" m_posNT: "<<pRecHyp->m_posNT;
      if (pRecHyp->m_pBackEqClass!=NULL){
	cout<<" hasBack ";
      }else{
	cout<<" noBack ";
      }
      cout<<endl;
    }
  }
}


void CHypBeam::Rescore() {
  /* !! temporarily disabled
  cout << "Warning: Rescore() is broken in that it is applied after tag- and unknown-handling, which messes up the LM calculations." << endl;
  cerr << "Warning: Rescore() is broken in that it is applied after tag- and unknown-handling, which messes up the LM calculations." << endl;
  exit(1);

  CSortedHypNBestContainer tmp_store;
  m_store.swap(tmp_store); // empty m_store, putting all its contents
			   // into tmp_store

  //iterate through hyps in the beam from best to worst
  //for (CSortedHypContainer::iterator iter = tmp_store.begin(); iter!= tmp_store.end(); ++iter){
  for (CSortedHypNBestContainer::iterator iter = tmp_store.begin(); iter!= tmp_store.end(); ++iter){
    //CChartParserHypComplete* pHyp = (*iter).second;
    CChartParserHypCompleteNBest* pHyp = (*iter).second;
   
    double pHypLMCost = 0; // newly calculated LM cost for pHyp

    assert(pHyp->m_completelyUnwound);
    assert(pHyp->m_firstWords.empty());
    assert(pHyp->m_lastWords.empty());
    assert(pHyp->m_scaledFutureCostEstimate == 0.0); // there are no future costs anymore, everything is settled

    //recompute the LM cost of pHyp
    CWordSequence tmpUnprocessedWords;
    CWordSequence tmpMostRecentWords; // history context is empty
    // sucessively LM-process m_translationOutput (which now only
    // contains terminal symbols), accumulating the costs into
    // pHypLMCost, thereby updating the context
    // tmpMostRecentWords and storing the first words, for which no
    // context was available, in tmpUnprocessedWords

    //CachedSriLM::s_debugLM = true;
    //    pHyp->Display(); cout << endl; 
    CChartParserInfo::ProcessLMWords(tmpUnprocessedWords,tmpMostRecentWords,pHyp->m_translationOutput,pHypLMCost);
    //cout<<"BOS EOS =========================================="<<endl;
    //calculate LM cost of first words given begin-of-sentence
    pHypLMCost += CChartParserInfo::GetBeginOfSentenceLMCost(tmpUnprocessedWords);
    //calculate LM cost of end-of-sentence given last words
    pHypLMCost += CChartParserInfo::GetEndOfSentenceLMCost(tmpMostRecentWords);
    //CachedSriLM::s_debugLM = false;
    //cout<<"END =========================================="<<endl;
    //pHyp->DisplayRecursive(CChartParser::m_pVocabulary, 0);
    
    //store the obtained LM cost for this hyp
    pHyp->m_negLogProbs[CRule::C_LMLocation] = pHypLMCost;
 
    // Update m_totalCost, recomputing it with the current value of
    // static CChartParserHyp::m_scalingFactors
    pHyp->UpdateTotalCost();
    assert(pHyp->m_backPointers.empty());
    // UpdateTotalCost processes the backpointers'
    // best-costs as well, but the backpointers of completely unwound
    // hyps are guarranteed to be empty

    // insert the hyp into the new beam storage, sorted by the new
    // m_totalCost values
    InsertHyp(pHyp);
  }
  */
}

void CHypBeam::Shrink(int K) {
}
/*
void CHypBeam::Shrink(int K) {
  //cout << "Shrinking beam..." << endl;
  CSortedHypNBestContainer tmp_store;
  m_store.swap(tmp_store); // empty m_store, putting all its contents into tmp_store
  //iterate through hyps in the beam from best to worst
  int iterationCount = 0;
  for (CSortedHypNBestContainer::const_iterator iter = tmp_store.begin(); iter!= tmp_store.end(); ++iter, ++iterationCount){
    CChartParserHypCompleteNBest* pHyp = (*iter).second;
    
    if (iterationCount < K) {
      // still haven't gotten K yet
      // copy element (cost-hyppointer pair) back into m_store
      m_store.insert(*iter);
    } else {
	// delete the hyp
	delete pHyp;
    }
  }
  //cout << "Shrinking done." << endl;
}
*/

//! Unique the hyps in the beam w.r.t. translationOutput, keeping only the best-cost hyps
int CHypBeam::Unique(bool onlyCount) {
  CSortedHypNBestContainer tmp_store;
  if (onlyCount) {
    tmp_store = m_store;
  } else {
    m_store.swap(tmp_store); // empty m_store, putting all its contents
    // into tmp_store
  }
  set<CWordSequence> dejavu; // to mark which translations we've already seen

  //iterate through hyps in the beam from best to worst
  for (CSortedHypNBestContainer::const_iterator iter = tmp_store.begin(); iter!= tmp_store.end(); ++iter){
    CChartParserHypCompleteNBest* pHyp = (*iter).second;
    
    if (!dejavu.count(pHyp->m_translationOutput)) {
      // not seen that translation yet
      if (!onlyCount) {
	// copy element (cost-hyppointer pair) back into m_store since it's the best of its kind
	m_store.insert(*iter);
      }
      // add its translationOutput to dejavu
      dejavu.insert(pHyp->m_translationOutput);
    } else {
      if (!onlyCount) {
	// seen better ones - delete the hyp
	delete pHyp;
      }
    }
  }
  return dejavu.size();
}

void CHypBeam::HandleUnknownAndTags() {
  for (CSortedHypNBestContainer::iterator iter = m_store.begin(); iter!= m_store.end(); ++iter){
    CChartParserHypCompleteNBest* pHyp = (*iter).second;
    //removes the unknown word or leaves it in as source, replaces and @NUM tags with the translated words
    pHyp->HandleUnknownAndTags();    
  }
}
  
CHypBeam::~CHypBeam() {

  //ARV 10/03 - in the past we would just delete the hyps in the store.
  //now we know that every hyp is actually safely stored in an equivalence class. the elements on 
  //the beam are just duplicate pointers. 

  //delete the actual hyps in the m_store
  //for (CSortedHypContainer::const_iterator iter2 = m_store.begin(); iter2!=m_store.end(); iter2++){
  //  CChartParserHypComplete* pHyp = (*iter2).second;
  //delete pHyp;
  //pHyp=NULL;
  //}

  vector<CChartParserHypCompleteEqClass*> toDelete;

  //delete the prototypes in the m_lmContexts. since these are prototype hyps they have recombined
  //hyps below them that can be deleted now as well. This happens in the destructor
  for (TContextHyps::iterator iter=m_lmContexts.begin(); iter!= m_lmContexts.end(); ++iter){
    CChartParserHypCompleteEqClass* pHyp = (CChartParserHypCompleteEqClass*)(*iter);
    pHyp->ClearRecombined();

    assert(!(pHyp->m_recombinedHyps.empty()));
    toDelete.push_back(pHyp);
  }
  
  //remove the relationships that are held in this set
  m_lmContexts.clear();

  //clear the memory that was held in the set
  for (vector<CChartParserHypCompleteEqClass*>::iterator iter = toDelete.begin(); iter!= toDelete.end(); ++iter){
    CChartParserHypCompleteEqClass* pHyp = (*iter);
    delete pHyp;
    pHyp=NULL;
  }
}

//void HashTargetContextHyp::LMContextAndBackpointers(const CChartParserHypCompleteNBest* pHyp, size_t& hashCode)const{
void HashTargetContextHyp::LMContextAndBackpointers(const CChartParserHypCompleteNBest* pHyp, size_t& hashCode)const{
  //need to generate a hash based on elements of the translation output PAST m_posNT and the context
  //we use the whole context here even tho the first 2 elements should always be the same in a* left to right mode
  CWordSequence sofar_context = CUtilities::ComputeLMContext(CWordSequence(pHyp->m_translationOutput.begin(), pHyp->m_translationOutput.begin()+ pHyp->m_posNT));
  for (CWordSequence::const_iterator iter = sofar_context.begin(); iter!= sofar_context.end(); ++iter){
    hashCode = 31*hashCode + (*iter);
  }
  if (!(pHyp->m_posNT<=pHyp->m_translationOutput.size())){
    cout<<"pHyp->m_posNT: "<<pHyp->m_posNT<<"pHyp->m_translationOutput.size(): "<<pHyp->m_translationOutput.size()<<endl;
    assert(pHyp->m_posNT<=pHyp->m_translationOutput.size());
  }
  
  for (CWordSequence::const_iterator iter = pHyp->m_translationOutput.begin() + pHyp->m_posNT; iter!= pHyp->m_translationOutput.end(); ++iter){
    CWordIndex nextWord = (*iter);
    if (nextWord>0){
	hashCode = 31*hashCode + nextWord;
    }else{
      hashCode = 31*31*hashCode + (long)(pHyp->m_backPointers[-1*nextWord]);      
    }
  }
  
  
  /*
  //OLD WAY
  //find the first non-terminal, in the meanwhile, just keep adding to a vector
  bool firstNonTerminalFound = false;
  CWordSequence leadingTerminals = pHyp->m_firstWords; 
  assert(!pHyp->m_completelyUnwound);

  //calculate the hash code for the last 2 terminals
  int count=0;
  for (CWordSequence::reverse_iterator wordIter = leadingTerminals.rbegin(); wordIter!=leadingTerminals.rend(); ++wordIter){
    hashCode = 31*hashCode + (*wordIter);
    count++;
    if (count==C_HistoryLength){
      break;
    }
  }

  //these are the remaining non-terminals
  for (int i=pHyp->m_firstWords.size(); i<pHyp->m_translationOutput.size(); ++i){
    CWordIndex nextWord = pHyp->m_translationOutput[i];
    if (nextWord>0){
      hashCode = 31*hashCode + pHyp->m_translationOutput[i];
    }else{
      hashCode = 31*31*hashCode + (long)(pHyp->m_backPointers[-1*nextWord]);      
    }
  }
  assert(pHyp->m_firstWords==leadingTerminals);
  */  
}


//size_t HashTargetContextHyp::operator()(const CChartParserHypCompleteNBest* pHyp)const{
size_t HashTargetContextHyp::operator()(const CChartParserHypCompleteNBest* pHyp)const{

  size_t hashCode = pHyp->m_completelyUnwound? 2 : 1; 
  if (pHyp->m_completelyUnwound){
    //cout<<"TOTALLY UNWOUND"<<endl;
    return hashCode;
  }else{
    LMContextAndBackpointers(pHyp, hashCode);    
    //cout<<"HashCode: "<<hashCode<<" "; pHyp->Display(); cout<<endl;
    return hashCode;
  }
}


//void EqTargetContextHyp::locateNT(const CChartParserHypComplete* pHyp, CWordSequence& lastWords, int& posNT, CWordSequence& remainingNonTerminals)const{

  /* !! temporarily disabled
  CWordSequence leadingTerminals = pHyp->m_firstWords;
  bool foundNonTerminals = false;

  assert(pHyp->m_firstWords.size()<=pHyp->m_translationOutput.size());
  //start out from m_firstWords and move on
  for (int i=pHyp->m_firstWords.size(); i<pHyp->m_translationOutput.size(); ++i){
    CWordIndex nextWord = pHyp->m_translationOutput[i];
    //if its a word
    if (nextWord>0){
      if (!foundNonTerminals){
	assert(false);
      }else{
	//this is a terminal symbol that is BEYOND the currently fully expanded point
	remainingNonTerminals.push_back(nextWord);
      }
    }else{
      //if this is the very first time
      if (!foundNonTerminals){
	posNT=i;
	foundNonTerminals=true;
      }
      remainingNonTerminals.push_back(nextWord);
    }
  }
  //go back thru and ID the last words
  int count=0;
  for (CWordSequence::reverse_iterator wordIter = leadingTerminals.rbegin(); wordIter!=leadingTerminals.rend(); ++wordIter){
    lastWords.push_back(*wordIter);
    count++;
    if (count==C_HistoryLength){`
      break;
    }
  }
  //reverse it, because it came in backwards
  reverse(lastWords.begin(), lastWords.end());
  */
//}

//bool EqTargetContextHyp::operator()(const CChartParserHypCompleteNBest* pHyp1, const CChartParserHypCompleteNBest* pHyp2)const{
bool EqTargetContextHyp::operator()(const CChartParserHypCompleteNBest* pHyp1, const CChartParserHypCompleteNBest* pHyp2)const{
  //find the first non-terminal, in the meanwhile, just keep adding to a vector
  bool debug = false;

  if (pHyp1->m_completelyUnwound!=pHyp2->m_completelyUnwound){
    return false;
  }

  //they are both unwound, all hyps get filed together, dont worry at ALL about what words are generated
  if (pHyp1->m_completelyUnwound && pHyp2->m_completelyUnwound){
    //cout<<"BOTH unwound"<<endl;
    return true;
  }

  //check the context words
  CWordSequence sofar_context1 = CUtilities::ComputeLMContext(CWordSequence(pHyp1->m_translationOutput.begin(), pHyp1->m_translationOutput.begin()+ pHyp1->m_posNT));
  CWordSequence sofar_context2 = CUtilities::ComputeLMContext(CWordSequence(pHyp2->m_translationOutput.begin(), pHyp2->m_translationOutput.begin()+ pHyp2->m_posNT));
  if (sofar_context1!=sofar_context2){
    return false;
  }

  //locate the remaining symbols using the post NT marker
  CWordSequence remainingSymbols1 = CWordSequence(pHyp1->m_translationOutput.begin()+pHyp1->m_posNT, pHyp1->m_translationOutput.end());
  CWordSequence remainingSymbols2 = CWordSequence(pHyp2->m_translationOutput.begin()+pHyp2->m_posNT, pHyp2->m_translationOutput.end());

  //check the remaining symbols. We cant just check for equality of the part of the translation output, we need
  //to check the pointers to see if they were the same as well

  if (remainingSymbols1.size()!=remainingSymbols2.size()){return false;}

  for (int i=0; i<remainingSymbols1.size(); i++){
    CWordIndex nextWord1 = remainingSymbols1[i];
    CWordIndex nextWord2 = remainingSymbols2[i];
    //they are xor (ie they are different)
    if ((nextWord1>0) ^ (nextWord2>0)){
      return false;
    }
    //if they are both terminals, they MUST be the same terminals
    if ((nextWord1>0) && (nextWord2>0)){
      if (nextWord1!=nextWord2){
	return false;
      }
    }else{
      if ((nextWord1<=0) && (nextWord2<=0)){
	//they need to point to the same address... 
 	long add1 =   (long)(pHyp1->m_backPointers[-1*nextWord1]);
 	long add2 =   (long)(pHyp2->m_backPointers[-1*nextWord2]);
	if (add1!=add2){ return false; }
      }else{
	//being here means that they are different, we should have exited already
	assert(false);
      }
     }
    
  }
  return true;
}
