/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#ifndef __HypBeam__
#define __HypBeam__

#include "ParserInfo.hh"
#include "ChartParserHyp.hh"
#include "UniqueCompleteHyps.hh"

/**
This class manages an ordered beam worth of CChartParserHypCompleteNBest items. This beam is available to 
outside users of this class. It also manages an internal set of equivalence classes CChartParserHypCompleteEqClass
that consider the context of the CChartParserHypCompleteNBest and the remaining terminals after the unwound marker
m_posNT in CChartParserHypCompleteNBest
 **/

//Returns a hashcode based on the last two target words unwound and the remaining un-wound non-terminals
struct HashTargetContextHyp{
  void LMContextAndBackpointers(const CChartParserHypCompleteNBest* pHyp, size_t& hashCode)const;
  size_t operator()(const CChartParserHypCompleteNBest* pHyp)const;  
};

//Checks of two hyps have the exact same last two words + remaining backpointers
struct EqTargetContextHyp{
  //void locateNT(const CChartParserHypCompleteNBest* pHyp, CWordSequence& lastWords, int& posNT, CWordSequence& remainingNonTerminals)const;
  bool operator()(const CChartParserHypCompleteNBest* pHyp1, const CChartParserHypCompleteNBest* pHyp2)const;
};

//! Bin hyps by context
typedef hash_set<CChartParserHypCompleteNBest*, HashTargetContextHyp, EqTargetContextHyp > TContextHyps;
typedef multimap<float, CChartParserHypCompleteNBest*> CSortedHypNBestContainer;
//! Class that holds CChartParserHypComplete pointers during the n-best list extraction
/*!
  Contains a single multimap that associates costs and complete hyps, and add functionality to only store the top (lowest cost)
  K objects. Uses the CSortedHypContainer as the backing store
  contains completed hyps, sorted by costs (lowest cost first)
  use InsertHyp to insert new hyp
  only up to m_maxHypCount hyps are stored; when an additional hyp is supposed to inserted such that the number of hyps stored is higher
  than m_maxHypCount, the highest-cost hyp in the beam is removed and deleted (that could also be the hyp supposed to be inserted)!
*/
class CHypBeam{
public:

  int m_maxHypCount;    //!< The maximum number of hyps stored in the beam - this can be manipulated from outside
  bool m_capacityExceeded; //!< just for info purposes
  CSortedHypNBestContainer m_store; //!< the sorted list of elements  - this function is used in the old astar search
  vector<CChartParserHypCompleteNBest*> m_unsortedstore; //!Use this when you dont need a constantly sorted list
  float m_lowestCost;  //!track the cost of the best hyp

  CUniqueCompleteHyps m_uniqueChecker; //!keep track of the unique hyps only
  TContextHyps m_lmContexts; //! lm contexts that maintain prototype - this class manages the hyps in this class

  //! arv: this variable doesnt seem to be used anywhere
  vector<CChartParserHypCompleteNBest*> m_fullSentenceHyps; //!Sentences that are spanning the whole sentence

  //! Initialize with the a limit of ojects
  CHypBeam(int K) {
    m_maxHypCount = K;
    m_capacityExceeded = false;
    m_lowestCost = 9999;
  }

  //check if this hyp is just going to generate more of the same translations
  // if it is, then delete it and return NULL
  //CChartParserHypComplete* CheckUnique(CChartParserHypComplete* pHyp);


  //!Insert complete hyps, only the top m_maxHypCount are stored
/*! only up to m_maxHypCount hyps are stored; when an additional hyp is supposed to inserted such that the number of hyps stored is higher
 than m_maxHypCount, the highest-cost hyp in the beam is removed and deleted (that could also be the hyp supposed to be inserted)!
 return false if the hyp was recombined and was worse - could be used to single the main loop
*/
  bool InsertHyp(CChartParserHypCompleteNBest* pHyp);

  //!Returns the agenda best elements according to this beam and erases the rest from the store
  vector<CChartParserHypCompleteNBest*> GetAgenda(double beam);
  float GetLowestCost();


  //! Return the prototypical hyp from the m_lmContexts
  //CChartParserHypCompleteNBest* LookupProto( CChartParserHypCompleteNBest* pHyp);
  CChartParserHypCompleteEqClass* LookupProto( CChartParserHypCompleteNBest* pHyp);
 
  //! Get the last C_HistoryLength words from a set of leadingTerminals
  CWordSequence GetContextWords(CWordSequence& leadingTerminals);
 
  //!simple access to the data store
  CSortedHypNBestContainer* GetStorePtr(){
    return &(m_store);
  }

  CChartParserHypCompleteNBest* GetTopHyp(){
    assert(m_store.size()>0);
    return (*(m_store.begin())).second;
  }

  vector<CChartParserHypComplete*> AsVector(){
    vector<CChartParserHypComplete*> toRet;
    for (CSortedHypNBestContainer::iterator iter=m_store.begin(); iter!= m_store.end(); ++iter){
      toRet.push_back((CChartParserHypComplete*)(iter->second));
    }
    return toRet;
  }

  //! Print out the elements on the beam
  void PrintBeam();

  //! Print out the equivalence classes
  void PrintClasses();

  //! Rescore the hyps in the beam.
  /*! 
    Rescore the completely unwound hyps by calculating the exact
    language model costs for the translations and by recalculating the
    m_totalCost variable using the current value of
    CChartParserHyp::m_scalingFactors. Thereby, the hyps are re-sorted
    according to the new m_totalCost value
  */
  void Rescore();

  //! Unique the hyps in the beam w.r.t. translationOutput, keeping only the best-cost hyps; return n.o. unique translations in beam
  int Unique(bool onlyDoCounting = false);

  //!Return number of elements in beam
  int Size(){ return m_store.size(); }

  //! Shrink the beam down to the K best hypotheses
  void Shrink(int K);

  //! Removes or replaces _UNKNOWN, according to whether m_noTranslationOutputOfUnknownWords is true
  void HandleUnknownAndTags();

  //! destructor - frees the memory for all hyps that have been inserted
  ~CHypBeam();

};

#endif
