/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#include "CombinationBeam.hh"

size_t CombinationTuplePrtHash::operator()(const CCombinationTuple* p) const{
  //use the address of the rule
  size_t hashCode = (long)(p->m_pRule) * 31 + long(p->m_pIncomplete);
  for (int i=0; i<p->m_derivationVector.size(); i++){
    hashCode = 31*hashCode + long(*(p->m_derivationVector[i]));
  }
  // we must NOT consider m_pComplete as this pointer is deleted whenever hyps are added to the chart
  return hashCode;
}

bool CombinationTuplePrtEq::operator()(const CCombinationTuple* p1, const CCombinationTuple* p2) const{

  if (p1->m_pRule!=p2->m_pRule){
    return false;
  }

  if (p1->m_pIncomplete!=p2->m_pIncomplete){
    return false;
  }

  assert(p1->m_derivationVector.size()==p2->m_derivationVector.size()); // since the incomplete hyps are the same, the dimensions must match, too
  if (p1->m_derivationVector != p2->m_derivationVector) {
      return false;
  }

  // we must NOT consider m_pComplete as this pointer is deleted whenever hyps are added to the chart


  return true;
}

bool CombinationTupleMinCompare::operator()(const CCombinationTuple* p1, const CCombinationTuple* p2) const{
  return (p1->Cost() > p2->Cost());
}

CCombinationTuple::CCombinationTuple(CRule* pRule,  CChartParserHypIncomplete* pIncomplete, const vector< std::vector<CChartParserNode*>::iterator >& derivationVector ) :
  m_pRule(pRule),
  m_pIncomplete(pIncomplete),
  m_pComplete(NULL),
  m_derivationVector(derivationVector) {
}


void CCombinationTuple::Complete() {
  // first determine the fine-grained EC backpointer vector expressed by m_derivationVector
  vector<CChartParserNode*> m_fineGrainedBackPointers;
  for (int i=0; i < m_derivationVector.size(); i++) {
    m_fineGrainedBackPointers.push_back(*(m_derivationVector[i]));
  }
  // now create a complete hyp and calculate its LM cost
  m_pComplete = new CChartParserHypComplete(m_pIncomplete, m_fineGrainedBackPointers, m_pRule);
}

CCombinationTuple::~CCombinationTuple(){
  // might have to remove m_pComplete, unless it has actually been filed somewhere (in which case the pointer will have been set to NULL)
  if (m_pComplete != NULL) {
             delete m_pComplete;
  }
}


void CCombinationTuple::Display(){
  cout<<"Tuple: ";
  m_pRule->Display();
  cout<<"COST: "<<Cost()<<" ";
  //production nodes

  for (int i=0; i<m_derivationVector.size(); i++){
    //display its size
    cout<<CChartParserInfo::m_pVocabulary->GetWord(m_pIncomplete->m_productionBackPointers[i]->GetResult())<<"-"<<m_pIncomplete->m_productionBackPointers[i]<<"("<<m_pIncomplete->m_productionBackPointers[i]->m_fineGrainedNodes.size()<<")-";
    //and which number down this tuple represents
    cout<<"[ ";
    cout<<m_derivationVector[i]-m_pIncomplete->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.begin()<<" ("<<*m_derivationVector[i]<<") = ";
    CChartParserNode* pPNode = *m_derivationVector[i];

    CLanguageModel::PrintContextSequenceAllLMs(pPNode->m_lmContext);
    cout<<" => bestRuleCost:"<<pPNode->m_bestTotalCost<<" + scaledFutureCost:"<<pPNode->m_scaledFutureCostEstimate
      <<" + boundaryLM:"<<pPNode->m_scaledBoundaryCost;
    cout<<" ] ";
  }
  cout<<endl;
}

CCombinationTupleBeam::CCombinationTupleBeam(){
  //m_k = k;
  //m_percDiff = percDiff;
  m_numTuplesProposed = 0; //no tuples initially
}


//returns true if this element was added to the beam (always)
bool CCombinationTupleBeam::InsertAndCompleteBaseCombination(
    CCombinationTuple* toAdd){

#ifndef NDEBUG
  bool debug = false;
  if (debug){
    cout<<"InsertBaseCombination: Inserting: ";
    toAdd->Display();
  }
#endif

  // track each time we call InsertBaseCombination since this tuple
  // must have been well formed
  m_numTuplesProposed++;
  // create complete hyp from the tuple and do LM cost calculation etc.
  toAdd->Complete();
  //insert into both the unique set and the candidate set
  m_uniqeTuples.insert(toAdd);
  //AS pq
  m_candidateSet.push(toAdd);
  //AS multimap
  //m_candidateSet.insert(std::pair<double, CCombinationTuple*>(toAdd->m_cost, toAdd));
  assert(m_uniqeTuples.count(toAdd)==1);
  return true;
}

//returns the next tuple from the m_finalNBest list - leave all the cutoff logic to the level that uses this CCombinationTupleBeam
CCombinationTuple* CCombinationTupleBeam::NextTuple(bool expandNeighbors){
  if (m_candidateSet.size()==0){
    return NULL;
  }
  CCombinationTuple* topElement = m_candidateSet.top();
  m_candidateSet.pop();

  if (expandNeighbors) {
    //we WILL return topElement, but we also put his neighbors on the m_candidateSet
    ExpandNeighbors(topElement);

    //  for (vector<CCombinationTuple*>::iterator iter = frontier.begin(); iter!= frontier.end(); ++iter){
    //    //m_candidateSet.insert(std::pair<double, CCombinationTuple*>((*iter)->m_cost, *iter));
    //    InsertBaseCombination(*iter);
    //  }
  }

  return topElement;
}

bool CCombinationTupleBeam::ExpandNeighbors(CCombinationTuple* pTuple){

  bool didBetter = false;

  //az: disables as it seems more effective to increase ComboPruningBeamsize instead
  //  int depthDown = 1;
  //  if (CChartParserInfo::m_comboPruningUseLM){
  //    depthDown = 10; //make this a parameter later
  //  }

  //for each dimension in |e| ie the set of tails of E we need to clone a new CCombinationTuple and increment it forward in one dimension
  for (int i=0; i< pTuple->m_derivationVector.size(); i++){ // iteration through the dimensions
    assert(pTuple->m_derivationVector[i] != pTuple->m_pIncomplete->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.end());

    //    for (int d=1; d<=depthDown; d++){
      //if there are more options in this dimension
      //this should always hold:      if (m_derivationVector[i] != m_pIncomplete->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.end()){

    if (pTuple->m_derivationVector[i]+1 != pTuple->m_pIncomplete->m_productionBackPointers[i]->m_lowestToHighestCostFineGrained.end()) {
      // there's still an alternative backpointed fine grained EC available

      //if you are using cmp=0 then there shouldn't be any neighbors to expand
      assert(CChartParserInfo::m_compareTargetWords==1);

      //now basically clone pTuple, but modify its derivationVector in dimension i to point to the next-best element there
      vector< std::vector<CChartParserNode*>::iterator > newDerivationVector = pTuple->m_derivationVector;
      newDerivationVector[i]++;

      CCombinationTuple* cCloned = new CCombinationTuple(pTuple->m_pRule, pTuple->m_pIncomplete, newDerivationVector);

      //we dont want to do the cost calculations especially the LM calculations if this node is just going
      //to be a duplicate. so we only complete and add the tuple if it doesn't exist yet
      if (m_uniqeTuples.count(cCloned)==0){
	InsertAndCompleteBaseCombination(cCloned);

	// just for diagnostics purposes
	if (pTuple->Cost() - cCloned->Cost() > 0.00001){
	  //cout<<"Bettered at depth: "<<d<<" m_cost: "<<m_cost<<" cCloned->m_cost: "<<cCloned->m_cost<<endl;
	  didBetter = true;
	}
      }else{
	//it's a duplicate... it was never added to any data structure
	delete cCloned;
      }
    }
  }

  return didBetter;
}


void CCombinationTupleBeam::Display(){
  cout<<"Candidate Set PQ ================: "<<m_candidateSet.size()<<endl;

  /*
  for (SortedCombinations::iterator tmpIter = m_candidateSet.begin(); tmpIter!=m_candidateSet.end(); ++tmpIter){
    (*tmpIter).second->Display();
    cout<<endl;
  }
  */
  //  cout<<"NBest Set ================"<<endl;
  //  for (SortedCombinations::iterator tmpIter = m_finalNBest.begin(); tmpIter!=m_finalNBest.end(); ++tmpIter){
  //    (*tmpIter).second->Display();
  //    cout<<endl;
  ///  }

  cout<<"Unique Set ================"<<endl;
  for (UniqueCombinations::iterator tmpIter = m_uniqeTuples.begin(); tmpIter!=m_uniqeTuples.end(); ++tmpIter){

    CCombinationTuple* ptr = (*tmpIter);
    cout<<"Printing: "<<ptr<<" "; (*tmpIter)->Display();
    cout<<endl;
  }

}



//void CCombinationTupleBeam::Clear(){
CCombinationTupleBeam::~CCombinationTupleBeam(){
  // remove all the m_uniqeTuples
  // this will also call the destructors of the tuples, deleting the tuples' complete hyps which didn't get added to the chart

  //cout<<"Unique Set ================"<<endl;

  // we can't delete the pointers in the hash table directly since for some weird reason iterating through
  // the hash table calls the equals function, which dereferences the pointers, causing a crash when the
  // corresponding objects don't exist anymore
  vector<CCombinationTuple*> del;
  for (UniqueCombinations::iterator tmpIter = m_uniqeTuples.begin(); tmpIter!=m_uniqeTuples.end(); ++tmpIter){

    CCombinationTuple* ptr = (*tmpIter);
    //cout<<"Deleting: "<<ptr<<" "; (*tmpIter)->Display();
    //cout<<endl;
    del.push_back(ptr);
    //delete ptr;
  }
    for(vector<CCombinationTuple*>::iterator tmpIter = del.begin(); tmpIter != del.end(); ++tmpIter){
       CCombinationTuple* ptr = (*tmpIter);
       delete ptr;
    }

  //  m_uniqeTuples.clear();
  //m_finalNBest.clear();
  //m_candidateSet.clear();
}
