/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#ifndef __CombinationBeam__
#define __CombinationBeam__

#include <vector>
#include <map>
#include <set>
#include <queue>

#include "ParserInfo.hh"
#include "ChartParserHyp.hh"

/**
 * Structure of N-Best list retrieval
 *  - CCombinationTuple represents a hyper-edge pointer and backpointers to the backward star of this edge
 *  - CCombinationTuple can be compared against each other and maintained in a priority queue
 *
 *  - CCombinationTupleBeam is a priority queue wrapper that also maintains uniqueness checking for its elements
 *
 *  Usage: In the CChartParserNode class (fine-grained) there already exists a sorted list of hyps
 *         But each hyp represents structured decisions (ie using back-pointers)
 *         We want to build a non-backpointer based (ie samtstring based + scores) n-best list in each node
 *         So we augment the CChartParserNode with a CCombinationTupleBeam to manage this search process
 *
 **/

class CCombinationTuple;

struct CombinationTuplePrtHash{
  size_t operator()(const CCombinationTuple* p) const;
};
struct CombinationTuplePrtEq{
  bool operator()(const CCombinationTuple* p1, const CCombinationTuple* p2) const;
};

struct CombinationTupleMinCompare{
  bool operator()(const CCombinationTuple* p1, const CCombinationTuple* p2) const;
};

typedef hash_set<CCombinationTuple*, CombinationTuplePrtHash, CombinationTuplePrtEq> UniqueCombinations;

//keep the matches rules separated by their production result, to avoid pruning out all rules of a particular production result
//typedef multimap<CWordIndex, pair<CChartParserHypIncomplete*, vector<CRule> > > RulesByProductionResult;
typedef multimap<CWordIndex, pair<CChartParserHypIncomplete*, CRule> > RulesByProductionResult;

//Tuple - representing a derivation during the KBest derivations when combining rules + CProductionNode (s) to produce => rule + CChartParserNode (s)
class CCombinationTuple{
  
public:
  
  //! only initialize the variables, nothing else
  CCombinationTuple(CRule* pRule,  CChartParserHypIncomplete* pIncomplete, const vector< std::vector<CChartParserNode*>::iterator >& derivationVector);
  //  CCombinationTuple(CRule* pRule,  CChartParserHypIncomplete* pIncomplete, const std::vector<CProductionNode*>& productionBackPointers);

  CCombinationTuple(CCombinationTuple& t) {
    cerr << "Copy constructor not allowed b/c then we'll delete m_pComplete twice!" << endl;
    assert(false);
    exit(2);
  }

  CCombinationTuple& operator=(const CCombinationTuple& t) {
    cerr << "Assignment operator not allowed b/c then we'll delete some m_pComplete twice and the other not at all!" << endl;
    assert(false);
    exit(2);    
    // copy the stuff
    //    return *this;
  }

  //! instantiate the complete hyp, created from m_pRule and m_pIncomplete - this will also compute the LM costs of that hyp
  void Complete();

  //  CCombinationTuple(const CCombinationTuple& c);
  ~CCombinationTuple();

  //do the cost calculation - optionally do the LM calculations too and set the individual values in the specified fields
  //void GetTupleCost(double& cost, double& scaledTM, double& scaledFutureCostEstimate );
  //  void UpdateTupleCost();

  float Cost() const{
    assert(m_pComplete!=NULL);
    return m_pComplete->AllComponentCost();
  }

  //print out the node
  void Display();

  //  CCombinationTuple& CCombinationTuple::operator = (const CCombinationTuple& r); //!!az: what for?

  CRule* m_pRule; //!< rule that is being applied

  CChartParserHypIncomplete* m_pIncomplete; //!< the incomplete hyp (with references to its backpointers that just got completed

  vector< std::vector<CChartParserNode*>::iterator > m_derivationVector; //!< for each dimension (i.e. NT site), an iterator to the currently  considered fine grained EC within the coarse grained EC represented by the NT site

  CChartParserHypComplete* m_pComplete; //!< here, a complete hyp will be created from m_pRule, m_pIncomplete, and the state of m_derivationVector represented by this hyp

  //both of these will always have the same length in the real k-best implementation
  //  std::vector<CProductionNode*> m_productionBackPointers; //!< CProductionNode backpointers havent been instantiated yet
  //  std::vector<CChartParserNode*> m_fineGrainedBackPointers; //! CChartParserNode backpointers HAVE been instantiated

  //  float m_cost; //!< Cost of real m_fineGrainedBackPointers instantiated + best costs of m_productionBackPointers 

  //data to avoid recomputation when the hyp is finally created
  //  float m_scaledTM; //!Scaled cost of the TM component
  //  float m_scaledFutureCostEstimate; //!< Scaled cost of the LM component (including the m_scaledFutureCostEstimate from below)
  //  CWordSequence m_firstWords;
  //  CWordSequence m_lastWords;
};

//keep pointers to CombinationTuple* objects sorted by a double value
//typedef std::multimap<float, CCombinationTuple*> SortedCombinations;

//! a priority queue over tuples, where fast access to the lowest cost element is provided
typedef std::priority_queue<CCombinationTuple*, vector<CCombinationTuple*>, CombinationTupleMinCompare> CombinationPQ;

class CCombinationTupleBeam{

public: 
  
  CCombinationTupleBeam();
  
  //!Remove the last (highest cost) element in the beam - return a ptr to the element that is going to be removed
  //so that you can manually delete it for yourself.
  //CANT do this with a PQ
  //CCombinationTuple* RemoveLast();

  //!Add this tuple to the beam - it may or may not be added
  //if actually added, we will also complete the tuple, i.e., create a complete hyp for it and do LM cost calculation etc.
  //returns true if this element actually made it onto the beam
  bool InsertAndCompleteBaseCombination(CCombinationTuple* toAdd);

  //!Returns the next tuple on the list; if expandNeighbors then add all neighbors of the tuple to the candidate heap
  CCombinationTuple* NextTuple(bool expandNeighbors);

  //! returns true if in CChartParserInfo::m_comboPruningUseLM mode
  //a  hyp lower down in the derivation produced a better score
  bool ExpandNeighbors(CCombinationTuple* pTuple);

  int Size(){
    return m_candidateSet.size();
  }
  //!Return the number of tuples proposed in this usage of the combination beam
  int NumTuplesProposed(){
    return m_numTuplesProposed;
  }

  //Fills and returns the NBest list
  //SortedCombinations* FillNBestReturnPtr();

  void Display();

  //! destruct all ever created CCombinationTuple objects
  ~CCombinationTupleBeam();

  //  void Clear();

  //  vector<CCombinationTuple*> m_tuplesToDelete;

  CombinationPQ m_candidateSet; //!< maintains candidate set
  //SortedCombinations m_candidateSet; //!< maintains candidate set
  //  SortedCombinations m_finalNBest; //!<where the final nbest list is stored
  UniqueCombinations m_uniqeTuples; //! keep track of ptrs to unique tuples

  //Utility function that gives you a UniqueCombinations structure for easy lookup from a SortedCombinations
  //structure which gives you easy ordering
  //  static void ConvertSortedToUnique(SortedCombinations* pFrom, UniqueCombinations* pTo);

protected:
  
  //int m_k; //!< beam size
  //double m_percDiff; //!< prob difference
  bool m_capacityExceeded;
  int m_numTuplesProposed;
};

#endif
