/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#ifndef __COMPACTTSEQ__
#define __COMPACTTSEQ__


//#include <config.h>


#include <map>
#include <sstream>
#include "Vocabulary.hh"
//#include <NumericalArray.hh>
#include "Types.hh"
#include <iostream>

template<class T1, class T2> class CPair;


template<class T1, class T2> 
class CPair{

public:
  T1 m_first;
  T2 m_second;

  CPair(){}

  CPair(T1 first, T2 second){
    m_first = first;
    m_second = second;
  }
  

  CPair<T1, T2> operator + ( const CPair<T1,T2>& Pair) { 
    CPair<T1, T2> toReturn;
    toReturn.m_first = m_first + Pair.m_first;
    toReturn.m_second = m_second + Pair.m_second;
    return toReturn; 
  } 

  friend std::ostream& operator <<(std::ostream &OutStream, const CPair<T1, T2> Pair ){
    OutStream<<Pair.m_first<<" "<<Pair.m_second;
    return OutStream;
  }
  
  CPair<T1,T2> & operator = ( const CPair<T1,T2> &Pair ){
    if ( this == &Pair ) 
      //-- Kopie auf sich selber kann ignoriert werden
      return *this;
    m_first = Pair.m_first;
    m_second = Pair.m_second;
  }
  
};


typedef int CTransducerState;
typedef int KeyType;

class CTSStateWord
{
public:
  CTSStateWord( CTransducerState State, CWordIndex WordIndex ){
    m_State = State;
    m_WordIndex = WordIndex;
  }
  CTransducerState m_State;
  CWordIndex   m_WordIndex;
};

struct CompareTStateWord
{
  bool operator()(const CTSStateWord & StateWord1, const CTSStateWord & StateWord2 ) const
  {
    return (StateWord1.m_State < StateWord2.m_State) 
	  || ((StateWord1.m_State == StateWord2.m_State) 
		   && (StateWord1.m_WordIndex < StateWord2.m_WordIndex) );
  }
};


template <class T> class CCompactTSequence;


//the item at the end of the trie, which holds the scaling factor
template <class T> 
class CCompactTSequenceItem{
  
public:
  
  CCompactTSequenceItem(std::vector<CWordIndex> sequence){
    m_sequence = sequence;
    m_key = -1;
  }

  ~CCompactTSequenceItem(){
    m_sequence.clear();
  }
  void Display(){
    cout<<"Sequence:";
    for (int i=0; i<m_sequence.size(); i++){
      cout<<m_sequence[i]<<" ";
    }
    cout<<" Key:"<<m_key<<" Value:"<<m_value<<endl;
  }

  void Display(CVocabulary* pVoc){
    cout<<"Sequence:";
    for (int i=0; i<m_sequence.size(); i++){
      //const CString* strPtr = pVoc->GetWordPtr(m_sequence[i]);
      const samtstring* strPtr =  pVoc->GetWordPtr(m_sequence[i]);
      if (strPtr!=NULL){
	//cout<<(*strPtr)<<"("<<m_sequence[i]<<") ";
      }else{
	cerr<<"Unable to get word for: "<<m_sequence[i]<<" might have been added to the vocab "<<endl;
      }
    }
    cout<<" Key:"<<m_key<<" Value:"<<m_value<<endl;
  }
  
  //manipulates the value field
  //void IncrementValue(double value);  

  void SetValue(T value){
    m_value = value;
  }

  T GetValue(){
    return m_value;
  }
  T* GetValueAddress(){
    return &m_value;
  }

  //manipulates the identifier key for this sequence item
  KeyType GetKey(){
    return m_key;
  }
  void SetKey(KeyType key){
    m_key = key;
  }

  std::vector<CWordIndex> GetSequence(){
    return m_sequence;
  }
  
  void PrintSequence(CVocabulary* pVoc){
    for (std::vector<CWordIndex>::iterator iter=m_sequence.begin(); iter!=m_sequence.end(); ++iter){
      if (pVoc!=NULL){
	//cout<<"Looking for voc!!!"<<endl;
	const samtstring* s = pVoc->GetWordPtr(*iter);
	if (s!=NULL){
	  cout<<*s;
	}else{
	  cerr<<"Unable to display word index:"<<(*iter)<<" might have been added to dictionary"<<endl;
	  cout<<"("<<(*iter)<<")";
	}
      }else{
	cout<<(*iter);
      }
      if (iter!=m_sequence.end()-1){
	cout<<" ";
      }
    }
  }
  
private:
  T m_value;
  KeyType m_key;
  
  //the sequence used to get to this node
  std::vector<CWordIndex> m_sequence;

};


template <class T> 
class CCompactTSequence{


  typedef hash_map< CTransducerState , CCompactTSequenceItem<T>* > EmissionTable;
  typedef std::multimap< CTransducerState , CTransducerState > ExtensionMap;


public:
  
  CCompactTSequence(){
    s_NotFound = -1;  
    m_NumberOfStatesSource=0;
    m_pVoc = NULL;
  }
  
  ~CCompactTSequence(){
    for(typename EmissionTable::iterator iter= m_EmissionTableSource.begin(); iter!=m_EmissionTableSource.end(); ++iter){
    CCompactTSequenceItem<T>* item = (*iter).second;
    delete item;
  }
    m_EmissionTableSource.clear();
  }

  void Init(CVocabulary* pInputVocabulary){
    m_pVoc = pInputVocabulary;
  }

  CCompactTSequenceItem<T>* InsertPhrasePair(std::vector<CWordIndex> sequence, T value){
    //sets up the trie to accept new sequences
    m_BuildMode = true;
    
    CTransducerState State = 0;
    CTransducerState NextState = 0;
    
    for(std::vector<CWordIndex>::iterator iter = sequence.begin(); iter!=sequence.end(); ++iter){
      if (NextSource( State, (*iter), NextState ) == 0) {
	std::cout << "CTransducer::BuildSourceTables: Warning: NextSource() returned 0 (whatever this means)" << std::endl;
      }
      State = NextState;
    }
    
    //state now holds the transducer state for the final word
    //cout<<"EmissionTableSize("<<NextState<<")"<<m_EmissionTableSource.count(NextState)<<endl;
    // cout<<"Before adding"<<endl;
    //cout<<"Table Size:"<<m_EmissionTableSource.size()<<endl;
    //DisplayEmmissionTable();

    if (m_EmissionTableSource.count(NextState)==0){
      CCompactTSequenceItem<T>* oneItem = new CCompactTSequenceItem<T>(sequence);
      oneItem->SetValue(value);
      //m_TransducerItems.Append(oneItem);
      m_TransducerItems.push_back(oneItem);
      m_EmissionTableSource.insert(std::pair<CTransducerState, CCompactTSequenceItem<T>* >(NextState, oneItem));

      //cout<<"End of new item========="<<endl;
      //cout<<"Table Size:"<<m_EmissionTableSource.size()<<endl;
      //DisplayEmmissionTable();
      return oneItem;
    }else{
      typename EmissionTable::iterator iter = m_EmissionTableSource.find(NextState);
      CCompactTSequenceItem<T>* oneItem = NULL;
      if (iter==m_EmissionTableSource.end()){
	cerr<<"ERROR: Tried to add to an element that wasn't acutally there"<<endl;
	return NULL;
      }else{
	oneItem = (*iter).second;
      }
      //cout<<"OneItemAtIncrement======="<<oneItem<<endl;
      oneItem->SetValue(oneItem->GetValue() + value);
      //cout<<"End of used item========="<<endl;
      //cout<<"Table Size:"<<m_EmissionTableSource.size()<<endl;
      //DisplayEmmissionTable();
      return oneItem;
    }
  }
  CCompactTSequenceItem<T>* GetCompactTSequence(std::vector<CWordIndex> sequence){
    m_BuildMode = false;
    

    CTransducerState State = 0;
    CTransducerState NextState = 0;
    
    for(std::vector<CWordIndex>::iterator iter = sequence.begin(); iter!=sequence.end(); ++iter){
      if (NextSource( State, (*iter), NextState ) == 0) {
	//std::cout << "CTransducer::BuildSourceTables: Warning: NextSource() returned 0 (whatever this means)" << std::endl;
	return NULL;
      }
      State = NextState;
    }
    
    //cout<<"End of Get:=========="<<endl;
    //cout<<"Table Size:"<<m_EmissionTableSource.size()<<endl;
    //DisplayEmmissionTable();
    CCompactTSequenceItem<T>* toRet = NULL;
    typename EmissionTable::iterator iter = m_EmissionTableSource.find(NextState);
    
    if (iter==m_EmissionTableSource.end()){
      toRet=NULL;
    }else{
      toRet = (*iter).second;
    }
    //cout<<"Table Size:"<<m_EmissionTableSource.size()<<endl;
    //cout<<"End of GetBefore Ret:=========="<<endl;
    //DisplayEmmissionTable();
    return toRet;
  }

  CCompactTSequenceItem<T>* GetCompactTSequenceLexical(CWordIndex e, CWordIndex f){
    std::vector<CWordIndex> sequence;
    sequence.push_back(e);
    sequence.push_back(f);
    CCompactTSequenceItem<T>* item = GetCompactTSequence(sequence);
    return item;
  }

  //the leaf nodes
  //CNumericalArray< CCompactTSequenceItem<T> * > m_TransducerItems;
  vector< CCompactTSequenceItem<T> * > m_TransducerItems;

  int NumItems(){
    //return m_TransducerItems.Size();
    return m_TransducerItems.size();
  }

  void DisplayAllItems(const char* label="SeqKey"){
    //for (int i=m_TransducerItems.MinIndex(); i<=m_TransducerItems.MaxIndex(); i++){
    for (int i=0; i<m_TransducerItems.size(); i++){
      CCompactTSequenceItem<T>* item = m_TransducerItems[i];
      cout<<label<<": "<<item->GetKey()<<" ";
      item->PrintSequence(m_pVoc);
      cout<<" Value: "<<item->GetValue()<<endl;
    }
  }

  void DisplayEmmissionTable(){
    for(typename EmissionTable::iterator iter = m_EmissionTableSource.begin(); iter!= m_EmissionTableSource.end(); ++iter){
      std::pair<CTransducerState , CCompactTSequenceItem<T>* > data = (*iter);
      cout<<"State:"<<data.first<<" ";
      data.second->Display();
    }

    for(typename ExtensionMap::iterator iter = m_childExtensions.begin(); iter!= m_childExtensions.end(); ++iter){
      std::pair<CTransducerState , CTransducerState > data = (*iter);
      cout<<"StateRoot:"<<data.first<<" StateChild:"<<data.second<<endl;
    }
    
  }
  
  void ClearAll(){
    for (int i=0; i<m_TransducerItems.size(); i++){
      CCompactTSequenceItem<T>* ptr = m_TransducerItems[i];
      delete ptr;
      ptr = NULL;
    }
    //m_TransducerItems.Clear();
    m_TransducerItems.clear();
    m_EmissionTableSource.clear();
    m_TransitionTableSource.clear();
  }

  // If the element already exists, add to the value, otherwise add it
  // We dont carry the keys over since these are tied into the
  // instance of the perceptron learning method
  void Add(CCompactTSequence<T>* toAdd) { 
    //for (int i=toAdd->m_TransducerItems.MinIndex(); i<=toAdd->m_TransducerItems.MaxIndex(); i++){
    for (int i=0; i<toAdd->m_TransducerItems.size(); i++){
      CCompactTSequenceItem<T>* ptr = toAdd->m_TransducerItems[i];
      CCompactTSequenceItem<T>* existing = GetCompactTSequence(ptr->GetSequence());
      if (existing!=NULL){
	existing->SetValue(existing->GetValue() + ptr->GetValue());
      }else{
	InsertPhrasePair(ptr->GetSequence(), ptr->GetValue());
      }
    }
  }

  double L1(){
    double sum =0.0;
    for (int i=0; i<m_TransducerItems.size(); i++){
      CCompactTSequenceItem<T>* ptr = m_TransducerItems[i];
      sum += ptr->GetValue();
    }
    return sum;
  }
  
  /*
  double L2(){
    double sum =0.0;
    for (int i=0; i<m_TransducerItems.size(); i++){
      CCompactTSequenceItem<T>* ptr = m_TransducerItems[i];
      sum += pow(ptr->GetValue(), 2);
    }
    return pow(sum, 0.5);
  }
  */
  //divide each item's value by a fixed constant
  void DivideBy ( const double value) { 
    for (int i=0; i<m_TransducerItems.size(); i++){
      CCompactTSequenceItem<T>* ptr = m_TransducerItems[i];
      ptr->SetValue(ptr->GetValue()/value);
    }
  }

  void WriteOut(std::ostream& out){
    for (int i=0; i<m_TransducerItems.size(); i++){
      CCompactTSequenceItem<T>* ptr = m_TransducerItems[i];
      for (int s=0; s<ptr->GetSequence(); s++){
	out<<s<<" ";
      }
      out<<"V "<<ptr->GetValue()<<endl;
    }
  }

  

  

private:

  CTransducerState s_NotFound;

  bool m_BuildMode;
  CVocabulary *m_pVoc;
  
  // trie operations
  // transition from internal node to internal node
  std::map< CTSStateWord, CTransducerState, CompareTStateWord > m_TransitionTableSource;


  // emission from node to leaf item
  //hash_map< CTransducerState , CCompactTSequenceItem<T>* > m_EmissionTableSource;
  EmissionTable m_EmissionTableSource;

  //keeps track of extensions that have been made to a transducer state
  //to allow submatching
  ExtensionMap m_childExtensions;

  // track num states
  int m_NumberOfStatesSource;
  //move the state forward
  int NextSource( CTransducerState State, CWordIndex Word, CTransducerState &NextState){
    //Trace( 1, "CTransducer::NextSource " << State << " " << Word );
    
    // if in BuildMode we make new entry into transition table
    // otherwise: error
    
    CTSStateWord StateWord( State, Word );
    //cout<<"\tSTWord State:"<<State<<" Word:"<<Word<<endl;
    std::map< CTSStateWord, CTransducerState, CompareTStateWord >::const_iterator IterState;
    IterState = m_TransitionTableSource.find( StateWord );
    
    if ( IterState == m_TransitionTableSource.end() ) {
      if ( m_BuildMode ) {
	m_NumberOfStatesSource++;
	NextState = m_NumberOfStatesSource;
	m_TransitionTableSource[ StateWord ] = NextState;
	m_childExtensions.insert(std::pair<CTransducerState, CTransducerState>(State, NextState));
	
	//if (m_pVoc!=NULL){
	  //cout<<"\tInsert " << State 
	  //    << " " << m_pVoc->GetWord( Word ) 
	  //    << " " << NextState<<endl;
	//}else{
	  //cout<<"\tInsert " << State 
	  //    << " " <<  Word 
	  //    << " " << NextState<<endl;
	//}
      } else {
	//std::cout << "ERROR: in CTransducer::Next" << std::endl
	//<< "ERROR: State " << State << " is not in the transition table" << std::endl;
	NextState = s_NotFound;
	return 0;
      }
    }else {
      if (m_pVoc!=NULL){
	//cout<<"\tFound " << m_pVoc->GetWord( IterState->first.m_WordIndex ) 
	//<< " leading from state " << IterState->first.m_State 
	//  << " to state " << " " << IterState->second<<endl;
      }else{
	//cout<<"\tFound " <<IterState->first.m_WordIndex
	//<< " leading from state " << IterState->first.m_State 
	//<< " to state " << " " << IterState->second<<endl;
      }
      NextState = IterState->second;
    }
    //cout<<"Returning with NextState:"<<NextState<<endl;
    return 1;
  }
};

#endif
