/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#include "FastLatticeChartParser.hh"
#include "HypFiller.h"


void CChartParser::DisplayCompletedHypsInCell(int i, int j){

  CCompleteHypChartCell& chartCell = m_completeHypsChart[i][j];
  for (CCoarseGrainedEq::const_iterator iter =
           chartCell.m_productionNodes.begin();
       iter!= chartCell.m_productionNodes.end();
       ++iter
       ) {
    CProductionNode* node = (*iter).second;
    int initialSize = node->m_fineGrainedNodes.size();
    cout<<"Production Result === "<<
        CChartParserInfo::m_pVocabulary->GetWord(node->GetResult())<<endl;
    //now the equivalence classees
    for (CChartParserNodeContainer::iterator fiter =
             node->m_fineGrainedNodes.begin();
         fiter!= node->m_fineGrainedNodes.end();
         ++fiter){
      cout<< " Nodes: "<<endl;
      CChartParserNode* pChartNode = (*fiter).second;
      cout<<"\t";
      pChartNode->Display();
      cout<<endl;

      //now display all the hyps
      cout<< " Hyps: "<<endl;
      for (CSortedHypContainer::iterator iterHyp =
               pChartNode->m_hypContainer.begin();
           iterHyp!=pChartNode->m_hypContainer.end();
           ++iterHyp){
	cout<<"\t\t";
	CChartParserHypComplete* pTopHyp = (*iterHyp).second;
	pTopHyp->Display();
      }
      cout<<endl;
    }
    cout<<"Production Result Completed === "<<
        CChartParserInfo::m_pVocabulary->GetWord(node->GetResult())<<endl;
  }
}

void CChartParser::DisplayIncompletedHypsInCell(int sourceCellRow, int sourceCellColumn){

  // the incomplete hyps in the source cell
  CSortedIncompleteHypContainer& incompleteHyps = m_incompleteHypsChart[sourceCellRow][sourceCellColumn];
  int totalIncompleteHypCount = incompleteHyps.size();

  cout <<" totalIncompleteHypCount: "<<totalIncompleteHypCount<<endl;
  //iterate through the incompleted hyps of current IC chart cell
  //(sourceCellRow,sourceCellColumn)
  bool continueIncompleteHypIteration = true;
  int iterationCountIH = 0;
  for (CSortedIncompleteHypContainer::const_iterator iterHyps = incompleteHyps.begin(); continueIncompleteHypIteration != 0 && iterHyps != incompleteHyps.end(); ++iterHyps, ++iterationCountIH){

    float ruleAppCountIH = (*iterHyps).first;
    CChartParserHypIncomplete* pIncompleteHyp = (*iterHyps).second; // the current incomplete hyp
    pIncompleteHyp->Display();
    //      cout << "        (rule-app-count according to CSortedIncompleteHypContainer: " << ruleAppCountIH << endl;
    cout << endl;
    assert(ruleAppCountIH == pIncompleteHyp->m_totalRuleApplicationCount);
  }
}


void CChartParser::DisplayCell(int i, int j){

  cout<<" Cell i : "<<i<<" j : "<<j<<" ================"<<endl;
  cout<<"   Completed hyps: ========="<<endl;
  DisplayCompletedHypsInCell(i, j);
  cout<<"   Incompleted hyps: ======="<<endl;
  DisplayIncompletedHypsInCell(i, j);
  cout<<" === End of Cell: "<<i<<" "<<j<<endl;

}

//go thru the complete and incomplete hyps and give a full row wise dump of the chart at this point
void CChartParser::Display(){
  for (int i=1; i<=CChartParserInfo::m_sourceLength; i++){
    for (int j=1; j<=CChartParserInfo::m_sourceLength-i+1; j++){
      DisplayCell(i,j);
    }
  }
}


void CChartParser::InitSearch(const CWordSequence& pSentence){

  // reset the combination counter
  m_combinationCount = 0;

  assert(m_newlyCreatedNodes.empty());


  //step 1
  //fill in the completed hyps table with all the words in the
  //sentence, this will involve creating new completed CChartParserHyp
  // elements for each word in the sentence

  CChartParserInfo::m_pSentence = pSentence;
  CChartParserInfo::m_sourceLength = pSentence.size();

  //allocate space for the equivalence class containers
  SetupChartCells();

  for (int w = 0; w<pSentence.size(); w++){
    cout<<"Log: Checking word: "<<w+1<<" = "<<" index: "<<
        pSentence[w]<<" word: "<<
        CChartParserInfo::m_pVocabulary->GetWord(
            CChartParserInfo::m_pSentence[w])<<endl;

    //reset the stack level - which is ONLY used for
    // pretty printing during recursive calls
    CChartParserInfo::m_stackLevel = 0;

    // check whether this cell should get a special status
    if (pSentence[w] == CChartParserInfo::m_bosWI) {
      //+1 since the parser works on 1 indexing
      GetCompleteCell(1,w+1).m_cellType = 1; // BOS
    } else if (pSentence[w] == CChartParserInfo::m_eosWI) {
      //+1 since the parser works on 1 indexing
      GetCompleteCell(1,w+1).m_cellType = 2; // EOS
    }

    // create Terminal complete hyp
    //+1 since the parser works on 1 indexing
    CChartParserHypComplete* terminalHyp =
        new CChartParserHypComplete(pSentence[w], 1, w+1);

    // and add it to the CH chart
    bool wasAdded = HypFiller::AddCompletedHyp(terminalHyp,
                                               true,
                                               &(m_completeHypsChart)[1][w+1],
                                               &m_newlyCreatedNodes);
    assert(wasAdded==true);
    if (!wasAdded){
      cout<<"ERROR:terminalHyp was rejected by the online beam criteria"<<endl;
      exit(-1);
    }
    //Reset the expansion statistics
    CChartParserInfo::ResetCellStatistics();
    //+1 since the parser works on 1 indexing
    IndexNodesAccordingToRuleAppCount(1,w+1);
  }
  cout<<"Initialization completed"<<endl;
}

void CChartParser::SetupChartCells(){

  //All parsing indices start at 1 and then go uptil <= m_sourceLength
  m_completeHypsChart.reserve(CChartParserInfo::m_sourceLength+1);
  m_incompleteHypsChart.reserve(CChartParserInfo::m_sourceLength+1);
  for (int i=0; i<=CChartParserInfo::m_sourceLength; i++){
    vector<CCompleteHypChartCell> vComplete(CChartParserInfo::m_sourceLength+1);
    m_completeHypsChart.push_back(vComplete);
    vector<CSortedIncompleteHypContainer> vIncomplete(CChartParserInfo::m_sourceLength+1);
    m_incompleteHypsChart.push_back(vIncomplete);
  }

}


CChartParser::CChartParser(CVocabulary *pVocabulary,
                           CWordIndex unknownWI,
                           bool noTranslationOutputOfUnknownWords,
                           float maxRuleAppCountDifference,
                           RuleSet* ruleSet,
                           int maxHypCount,
                           int MaxCombinationCount,
                           double aFutureCostImpact,
                           int aBeamSizeFactor,
                           bool extractUnique) :
  m_maxRuleAppCountDifference(maxRuleAppCountDifference),
  m_bTree(ruleSet)
{
  CChartParserInfo::m_pVocabulary = pVocabulary;
  CChartParserInfo::m_noTranslationOutputOfUnknownWords=noTranslationOutputOfUnknownWords;
  CChartParserInfo::FutureCostImpact = aFutureCostImpact;
  CChartParserInfo::BeamSizeFactor = aBeamSizeFactor;
  CChartParserInfo::m_extractUnique = extractUnique;
  CChartParserInfo::m_unknownWI = unknownWI;
  CChartParserInfo::MAX_HYP_COUNT = maxHypCount;
  CChartParserInfo::m_maxCombinationCount = MaxCombinationCount;

  // note: it is important that m_bosWI refer to the samtstring "<s>" because of the LM call to the SRI, which treats
  // the samtstring <s> as beginning of sentence marker, so don't change this!
  CChartParserInfo::m_bosWI = CChartParserInfo::m_pVocabulary->GetIndex("<s>");
  CChartParserInfo::m_eosWI = CChartParserInfo::m_pVocabulary->GetIndex("</s>");
  CChartParserInfo::m_lmBoundaryWI = CChartParserInfo::m_pVocabulary->GetIndex("<*>");

  m_pBeamToDelete = NULL;
}

void CChartParser::ClearCHCCell(int i, int j){

  CCoarseGrainedEq& prodNodes = m_completeHypsChart[i][j].m_productionNodes;

  //CProductionNode(s) are newed, and inside them ChartParserNode(s) are newed and within them the hyps are stored
  for(CCoarseGrainedEq::iterator iter = prodNodes.begin(); iter!=prodNodes.end(); ++iter){
    CProductionNode* pProdNode = (*iter).second;
    delete pProdNode;
    pProdNode = NULL;
  }
}

void CChartParser::ClearIHCCell(int i, int j){
  for (CSortedIncompleteHypContainer::iterator iter =  m_incompleteHypsChart[i][j].begin(); iter!= m_incompleteHypsChart[i][j].end(); ++iter){
    CChartParserHypIncomplete* pIncompleteHyp = (*iter).second;
    delete pIncompleteHyp;
  }
}

void CChartParser::ClearSearch(){
  //first clear the completed hyps chart. when clearing a container, first clear all its elements
  for (int i=1; i<=CChartParserInfo::m_sourceLength; i++){
    for (int j=1; j<=CChartParserInfo::m_sourceLength; j++){
      ClearCHCCell(i, j);
    }
    m_completeHypsChart[i].clear();
  }
  m_completeHypsChart.clear();

  //  delete m_pCompleteTranslationsNode;

  assert(m_newlyCreatedNodes.empty());

  //clear the cache(s) AFTER rescoring
  if (!CChartParserInfo::m_globalLMCache){
    for (int i=0; i<CLanguageModel::m_languageModels.size(); i++)
      CLanguageModel::m_languageModels[i]->ClearCache();
  }

  //clear any hyps that exist as a result of n-best retrieval
  for (vector<CChartParserHypCompleteNBest*>::iterator iter=m_poppedHyps.begin(); iter!= m_poppedHyps.end(); ++iter){
    CChartParserHypCompleteNBest* pHyp = (*iter);
    delete pHyp;
    pHyp=NULL;
  }

  if (m_pBeamToDelete!=NULL){
    delete m_pBeamToDelete;
    m_pBeamToDelete = NULL;
  }
}





void CChartParser::AddIncompleteHyp(CChartParserHypIncomplete* pIncompleteHyp) {
  bool debug = false;
  if (debug) {
    for (int i=0; i<CChartParserInfo::m_stackLevel; i++) cout<<"\t";
    cout<<  "AddIncompleteHyp:";
    pIncompleteHyp->Display();
    cout<<endl;
  }
  //CSortedIncompleteHypContainer& cell = m_incompleteHypsChart.GetRow(pIncompleteHyp->I())[pIncompleteHyp->J()];
  CSortedIncompleteHypContainer& cell = m_incompleteHypsChart[pIncompleteHyp->I()][pIncompleteHyp->J()];
  cell.insert(std::pair<float, CChartParserHypIncomplete*>(pIncompleteHyp->m_totalRuleApplicationCount, pIncompleteHyp));
}






void CChartParser::UnwindHyp_HandleUnknownAndTags(CWordIndex word, CChartParserHypComplete* pNewHyp, CChartParserHypComplete* pBackHyp){
  if (word == CChartParserInfo::m_unknownWI) { // is it an unknown marker?
    // yes - save the corresponding source word
    // since this must come from an unknown-rule of the form _UNKNOWN#_UNKNOWN#@X#weights, we expect this rule to have
    // no backpointers (because no nonterminals are involved) and a span of one
    if ((pBackHyp->m_backPointers.size() != 0) || (pBackHyp->I() != 1)) {
      cerr << "In UnwindHyp: wrong format of unknown-rule employed. Unknown-rule's hyp is: ";
      pBackHyp->DisplayRecursive(CChartParserInfo::m_pVocabulary, 0);
    }
    // push the corresponding source word into m_targetOrderedUnknownSourceWords
    //pNewHyp->m_targetOrderedUnknownSourceWords.push_back((*CChartParserInfo::m_pSentence)[pBackHyp->J()]);
    //parsing is indexed from 1
    pNewHyp->m_targetOrderedUnknownSourceWords.push_back(CChartParserInfo::m_pSentence[pBackHyp->J()-1]);
  }else{

    //if you are not an m_unknownWI, then you are one of two things
    // a) a regular terminal, and you have been added onto the newTranslationOutput and the newParsingOutput
    // b) a non-terminal symbol that occurs in translation output. This is different from regular rules, which get converted into "I gave -1 a -2" form
    //    in this case we see "I gave him @NUM dollars"

    //if we have hit a tagged word backpointer (like @NUM) then we ONLY want to convert it into a m_taggedTargets element if it has no back pointers
    //in order to support multiple levels of tagged hierarchy. If its not a tag then process as usual
    if (CChartParserInfo::m_tagsForLM.count(word)!=0){
      //implies pBackHyp was empty AND the its production result was a tag
      //insert into the pNewHyp at the end of its m_taggedTargets since we are going left to right
      //we add ALL the m_taggedTargets since @NUM could represent  to "two and a half"
      pNewHyp->m_taggedTargets.insert(pNewHyp->m_taggedTargets.end(), pBackHyp->m_taggedTargets.begin(), pBackHyp->m_taggedTargets.end());
    }
  }
}

//CChartParserHypCompleteNBest* CChartParser::UnwindHyp(CSortedHypNBestContainer::iterator& iterPHyp, CHypBeam& beam, float lowestCostHyp){

CChartParserHypCompleteNBest* CChartParser::UnwindHyp(CChartParserHypCompleteNBest* pHyp, CHypBeam& beam, float lowestCostHyp){

#ifndef NDEBUG
  bool debug = false;
#endif
  //pre-conditions:
  //pHyp has a m_posNT (0 based index) that points to the current location of the first NT
  //if m_posNT goes past the size of the lists, then we should not be coming back to this point
  //pHyp->m_scaledFutureCostEstimate is the language model probability of the pHyp->m_translationOutput sequence
  //accounting for first best approximations when using non-terminals
  //CChartParserHypCompleteNBest* pHyp = iterPHyp->second;
  //eraase the beam from the hyp using its iterator
  //beam.m_store.erase(iterPHyp); //this is done before calling UnwindHyp now


  //th next line is not compatible with g++ 4.1
  //iterPHyp = 0; // just to make sure it's not being used again b/c it's now invalid

  //you should NOT be completely unwound - since there is nothing to expand further
  //but you MAY be ready for the m_completelyUnwound designation. The reason for this two part process where you first get filed
  //in the m_completelyUnwound=0 state and then get transfered to m_completelyUnwound=true is to make sure these unwound hyps
  //actually stay on the beam (and then we dont pop them again in RetrieveKBestParses)
  assert(pHyp->m_completelyUnwound!=true);


   //ensure that this is a hyp that we actually want to do calculations for
  if (!pHyp->m_active && !CChartParserInfo::m_dontReloadBeam){
    //should make this a counted variable
#ifndef NDEBUG
    if (debug){
      cout<<"Rejecting inactive: "; pHyp->PrintTranslationOutput(); cout<<" cost: "<<pHyp->m_totalCost<<" unwnd: "<<pHyp->m_completelyUnwound<<endl;
    }
#endif
    //we have removed it, remove the extra buffer that we had provided
    if (CChartParserInfo::m_updateBeamSize){
      beam.m_maxHypCount--;
#ifndef NDEBUG
    if (debug){
      cout<<"Decreasing beam size to: "<<beam.m_maxHypCount<<endl;
    }
#endif
    }
    return NULL;
  }

#ifndef NDEBUG
  if (debug){
    cout<<"\n\nUnwind Hyp: ";
    pHyp->PrintTranslationOutput();
    cout<<" m_posNT: "<<pHyp->m_posNT<<" orig totalcost: "
        <<pHyp->m_totalCost<<" neg: ";
    copy(pHyp->m_negLogProbs.begin(),
         pHyp->m_negLogProbs.end(),
         std::ostream_iterator<float>(cout, " "));
    cout<<" m_taggedTargets: ";
    CChartParserInfo::PrintVocabSequence(pHyp->m_taggedTargets);
    cout<<endl;
  }
#endif

  //if you meet the qualifications for unwound ness
  if (pHyp->m_posNT==pHyp->m_translationOutput.size()){
    CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(*pHyp);
    pNewHyp->m_completelyUnwound = true;
    pNewHyp->m_backPointers.clear();
#ifndef NDEBUG
    if (debug){
      cout<<"Setting completely unwound!!!"<<endl;
      pNewHyp->PrintTranslationOutput(); cout<<" m_taggedTargets: "; CChartParserInfo::PrintVocabSequence(pNewHyp->m_taggedTargets); cout<<endl;
    }
#endif
    pNewHyp->SetBackInfo(beam.LookupProto(pHyp), pHyp);
    beam.InsertHyp(pNewHyp);

    //this is the only time that u return any value
    return pNewHyp;
  }


  assert(pHyp->m_posNT<=pHyp->m_translationOutput.size()-1);
  //do the same with the m_parsingOutput - unexpanded non-terminals point into the same locations as in the m_translationOutput case
  //but there are additional symbols for [, NT, etc in the m_parsingOutput which are real symbols in the vocab
  //calculate the posNTParse
  int posNTParse = 0;
  for (CWordSequence::iterator itP = pHyp->m_parsingOutput.begin(); itP!= pHyp->m_parsingOutput.end() && *itP>0; ++itP){
    ++posNTParse;
  }

  CWordSequence sofar = CWordSequence(pHyp->m_translationOutput.begin(), pHyp->m_translationOutput.begin()+ pHyp->m_posNT);
  //sofar has no backpointers (its all terminals), while remaining could have several back pointers
  CWordSequence remaining = CWordSequence(pHyp->m_translationOutput.begin()+pHyp->m_posNT+1, pHyp->m_translationOutput.end());
  //first remove future costs associated with this non-terminal
  CWordIndex currentNT = pHyp->m_translationOutput[pHyp->m_posNT];
  //if (debug){
  //cout<<"sofar: "; CChartParserInfo::PrintVocabSequence(sofar); cout<<endl;
  //cout<<"rem: "; CChartParserInfo::PrintVocabSequence(remaining); cout<<endl;
  //cout<<"currentNT: "<<currentNT<<endl;
  //cout<<endl;
  //}

  assert(currentNT<=0);
  // get the non-terminal object
  CChartParserNode* pEquivalenceClass = pHyp->m_backPointers[-1 * (currentNT)];

  //COSTS: pHyp->m_totalCost has the accumulated costs for pHyp, including all LM decisions
  //       pHyp->m_negLogProbs have ONLY the surface LM costs as well as the surface (single rule only) translation costs
  //             -- these surface costs include real LM costs of a hyp as well as estimates where nt boundaries occured
  //             -- at then end of this, the pNewHyp needs to have accumulated m_negLogProbs costs
  //
  //       new hyps will be put on the beam based on m_totalCost - we need to remove the LM cost due to the replacement operation

  float totalCostBase = pHyp->m_totalCost;

  //now remove language model probs for the left and right boundary around this non-terminal
  CWordSequence sofar_context = CUtilities::ComputeLMContext(sofar);
  //compute the remaining context back on potential non-terminals that exist in the remaining symbols
  CWordSequence remaining_context =
      CUtilities::ComputeLMContext(
          CUtilities::HarvestLMString(remaining, pHyp->m_backPointers));

  //if (debug){
  //cout<<"sofar_context: "; CChartParserInfo::PrintVocabSequence(sofar_context); cout<<endl;
  //cout<<"remaining_context: "; CChartParserInfo::PrintVocabSequence(remaining_context); cout<<endl;
  //}

  //subtract out these costs
  //left context
  CWordSequence forLeftContextInitial = sofar_context;
  forLeftContextInitial.insert(forLeftContextInitial.end(), pEquivalenceClass->m_lmContext.begin(), pEquivalenceClass->m_lmContext.end());
  //we care bout saving the actual probs for the lm parts that we are removing, since these surface costs need to be deleteHC05d in each alternative
  vector<float> forLeftContextRemoval;
  for (int i=0; i<CRule::C_NumLMs; i++) {

    float remval = CUtilities::ComputeLMCost(CLanguageModel::m_languageModels[i], forLeftContextInitial, false, true);
    //if (debug){
    //cout<<"Removing left: "; CChartParserInfo::PrintVocabSequence(forLeftContextInitial); cout<<" val: "<<remval<<endl;
    //}
    forLeftContextRemoval.push_back(remval);
    totalCostBase -= forLeftContextRemoval[i]*CChartParserInfo::m_scalingFactors[i];
  }
  //right context
  CWordSequence forRightContextInitial = CUtilities::ComputeLMContext(forLeftContextInitial);
  forRightContextInitial.insert(forRightContextInitial.end(), remaining_context.begin(),  remaining_context.end());

  //we care bout saving the actual probs for the lm parts that we are removing, since these surface costs need to be deleted in each alternative
  vector<float> forRightContextRemoval;
  for (int i=0; i<CRule::C_NumLMs; i++) {
    float remval = CUtilities::ComputeLMCost(CLanguageModel::m_languageModels[i], forRightContextInitial, false, true);
    //if (debug){
    //cout<<"Removing right: "; CChartParserInfo::PrintVocabSequence(forRightContextInitial); cout<<" val: "<<remval<<endl;
    //}
    forRightContextRemoval.push_back(remval);
    totalCostBase -= forRightContextRemoval[i]*CChartParserInfo::m_scalingFactors[i];
  }
  //pick up the alternatives and spawn of new hyps
  CSortedHypContainer* sortedHyps = &(pEquivalenceClass->m_hypContainer);
  //CChartParserHypComplete* pOldFirstBest =  (sortedHyps->begin())->second;

  //======== Go thru each alternative ==========
  for (CSortedHypContainer::const_iterator iter2 = sortedHyps->begin(); iter2 != sortedHyps->end(); iter2++){


    //pNewHyp is the hyp that represents subbing in pBackHyp for the NT as position m_posNT
    //pNewHyp needs to be full configured to participate on the beam. This involves...
    // - setting its new m_totalCost and accumulating m_negLogProbs
    // - moving the m_posNT to the appopriate position
    // - changing its m_translationOutput to include the subbed in symbols
    // - updates its backpointers to reflect any new non-terminals that were seen
    CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(*pHyp);
    //pNewHyp has m_negLogProbs that are ONLY the surface level costs - ie incorrect LM surface costs that need to be deleted out

    //pBackHyp represents the alternative translation that we are considering
    CChartParserHypComplete* pBackHyp = (*iter2).second;
    //if (!(pBackHyp->m_totalCost>=pOldFirstBest->m_totalCost)){
    //cout<<"curr: "<<pBackHyp->m_totalCost<<" best: "<<pOldFirstBest->m_totalCost<<endl;
    //assert(pBackHyp->m_totalCost>=pOldFirstBest->m_totalCost);

    //}
    //if (debug){
    //cout<<"Back Hyp: "; pBackHyp->PrintTranslationOutput(); cout<<" m_taggedTargets: "; CChartParserInfo::PrintVocabSequence(pBackHyp->m_taggedTargets); cout<<endl;
    //}

    //this cost has no language boundary costs for the old hyp because we cut them out above - so its the accumulated total cost
    //based on first best decisions for TM and LM, minus the old boundary costs
    float totalCost = totalCostBase;
    //left context
    CWordSequence forLeftContext = sofar_context;
    forLeftContext.insert(forLeftContext.end(), pBackHyp->m_lmContext.begin(), pBackHyp->m_lmContext.end());
    for (int i=0; i<CRule::C_NumLMs; i++) {
      //update the surface level costs for the left context
      float updatedLM = CUtilities::ComputeLMCost(CLanguageModel::m_languageModels[i], forLeftContext, false, true);
      //if (debug){
      //cout<<"Addback left: "; CChartParserInfo::PrintVocabSequence(forLeftContext); cout<<" val: "<<updatedLM<<endl;
      //}
      pNewHyp->m_negLogProbs[CRule::C_LMLocation+i] += updatedLM - forLeftContextRemoval[i]; //remove the old boundary costs, add in a new one
      totalCost += updatedLM * CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i];
    }
    //right context
    CWordSequence forRightContext = CUtilities::ComputeLMContext(forLeftContext);
    forRightContext.insert(forRightContext.end(), remaining_context.begin(), remaining_context.end());
    for (int i=0; i<CRule::C_NumLMs; i++) {
      //update the surface level costs for the right context
      float updatedLM = CUtilities::ComputeLMCost(CLanguageModel::m_languageModels[i], forRightContext, false, true);
      //if (debug){
      //	cout<<"Addback right: "; CChartParserInfo::PrintVocabSequence(forRightContext); cout<<" val: "<<updatedLM<<endl;
      //}
      pNewHyp->m_negLogProbs[CRule::C_LMLocation+i] += updatedLM - forRightContextRemoval[i]; //remove the old boundary costs, add in a new one
      totalCost += updatedLM * CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i];
    }

    //if (debug){
    //cout<<"pHyp->m_totalCost:"<<pHyp->m_totalCost<<endl;
    //}
    //the simple move of switching out the old vs new for all the internal costs
    totalCost += pBackHyp->m_totalCost - pEquivalenceClass->m_bestTotalCost;


    //sanity check assersions for the first best hyp. we know that the estimates were calculated based on first best choices
    //so we should have deleted (before the loop) and added back costs to result in the same value as pHyp->m_totalCost
    if (iter2 == sortedHyps->begin()){
      assert(fabs(totalCost-pHyp->m_totalCost)<0.01);
      //assert(pNewHyp->m_negLogProbs == pHyp->m_negLogProbs); //need to check this hear before the accumulation starts
      assert(pNewHyp->m_negLogProbs.size()==pHyp->m_negLogProbs.size());
      for (int i=0; i<pNewHyp->m_negLogProbs.size(); i++){
	assert(fabs(pNewHyp->m_negLogProbs[i]-pHyp->m_negLogProbs[i])<0.01);
      }
    }

    //now the totalCost and pNewHyp->m_negLogProbs looks as they would if this alternative pBackHyp was taken during parsing
    //we still need to perform accumulation on the pNewHyp->m_negLogProbs.
    //do the final assignment of costs for the new hyp before it gets considered for addition onto the beam
    //add pBackHyp's neg log probs to pNewHyp - NOTE during parsing these m_negLogProbs were never accumulated
    //now they finally get summed to represent decisions made over multiple rules
    for (int m=0; m<CRule::C_NumModelsTotal; m++){
      pNewHyp->m_negLogProbs[m] += pBackHyp->m_negLogProbs[m]; //no need to delete the old hyps costs, since m_negLogProbs are only accumulated now
    }


    //adjust m_totalCost variable by the difference between pBackHyp and best hyp in pEquivalenceClass
    pNewHyp->m_totalCost = totalCost;
    //at this point the cost is set... you can check to see if its ok

    //set up the parsing output
    CWordSequence newParsingOutput(pNewHyp->m_parsingOutput.begin(), pNewHyp->m_parsingOutput.begin()+posNTParse);
    CChartParserHypComplete::AddParseOutputSubTreeStart(newParsingOutput, pBackHyp);

    //create a new translation, replacing the not terminal negative indicies with their appropriate values
    CWordSequence newTranslationOutput = sofar; //needs to have the corrected back pointer indices here
    for (CWordSequence::iterator iter3=pBackHyp->m_translationOutput.begin(); iter3!=pBackHyp->m_translationOutput.end(); ++iter3){
      if (*iter3 > 0) {
	// *iter3 is a terminal symbol
	// copy it over into the new translation output
	newTranslationOutput.push_back(*iter3);
	newParsingOutput.push_back(*iter3);
	//add the unknown src word and the tagged output to appropriate background vectors
	UnwindHyp_HandleUnknownAndTags(*iter3, pNewHyp, pBackHyp);
	//pNewHyp->NoteUnknownAndTags(*iter3, pBackHyp);
      } else {
	// *iter3 is a backpointer - gotta adjust it to point to its back node's new location
	newTranslationOutput.push_back(*iter3 - pHyp->m_backPointers.size());
	// also store the backpointer in the translation info tree
	newParsingOutput.push_back(*iter3 - pHyp->m_backPointers.size());
      }
    }
    newTranslationOutput.insert(newTranslationOutput.end(), remaining.begin(), remaining.end());
    CChartParserHypComplete::AddParseOutputSubTreeEnd(newParsingOutput);
    //and copy over the rest of the original hyp's translation info tree
    newParsingOutput.insert(newParsingOutput.end(), pNewHyp->m_parsingOutput.begin()+posNTParse+1, pNewHyp->m_parsingOutput.end());

    pNewHyp->m_translationOutput = newTranslationOutput;
    pNewHyp->m_parsingOutput = newParsingOutput;
    pNewHyp->m_backPointers.insert(pNewHyp->m_backPointers.end(),pBackHyp->m_backPointers.begin(),pBackHyp->m_backPointers.end());
    //itendify the location of m_posNT

    //identify the number of terminal words in pBackHyp's m_translationOutput (it could be all of them)
    int newPosNT = pHyp->m_posNT;
    //terminals are greater than 0
    while (newPosNT < newTranslationOutput.size() && newTranslationOutput[newPosNT]>0){
      newPosNT++;
    }
    pNewHyp->m_posNT = newPosNT;
    //this may indicate that you are at the end
    //cout<<"final posnt: "<<pNewHyp->m_posNT<<endl;

    if (pNewHyp->ScaledTMLMCost() > CChartParserInfo::GetPruningCutoff(lowestCostHyp, CChartParserInfo::m_aStarBeam) ){
      //this hyp will NEVER be expanded (since if you pop it you will terminate search). Dont add it to the beam!
#ifndef NDEBUG
      if (debug){
	cout<<"Deleting hyp: "; pNewHyp->PrintTranslationOutput(); cout<<" m_totalCost: "<<pNewHyp->m_totalCost<<" lowest: "<<lowestCostHyp<<endl;
      }
#endif
      delete pNewHyp;
      //continue; //dont do any more processing of this hyp
      break; //actually stop processing this root hyp
    }else{
#ifndef NDEBUG
      if (debug){
	cout<<" Adding hyp: "; pNewHyp->PrintTranslationOutput(); cout<<" m_totalCost: "<<pNewHyp->m_totalCost<<" lowest: "<<lowestCostHyp<<endl;
      }
#endif
      //set the new lowestCostHyp if its lower...
      if (pNewHyp->ScaledTMLMCost()<lowestCostHyp){
	lowestCostHyp = pNewHyp->ScaledTMLMCost();
      }
    }


    //now we are ready to add to the beam. This newly created hyp could get deleted so dont use pNewHyp anymore
    //set the back trace info to the equivalence class of the original pHyp, and pHyp itself
    pNewHyp->SetBackInfo(beam.LookupProto(pHyp), pHyp);
    beam.InsertHyp(pNewHyp);

  }
  return NULL;

}

//! Standard K-best retrieval based on the full-sentence node
vector<CChartParserHypComplete*> CChartParser::RetrieveKBestParses(int K){
  // make sure we have an _S hyp spanning the full sentence
  if (GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes.count(CChartParserInfo::m_FullSentenceNT)==0) {
    cout << "Error: Sentence has no translation (i.e., _S hyp). Sentence spanning chart cell contents:" << endl;
    DisplayCell(CChartParserInfo::m_sourceLength,1);
    exit(1);
  }

  // there should be only the full-sentence _S production node...
  assert(!CChartParserInfo::m_GlueAndEOSAssumptions || GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes.size() == 1);
  // ... , which only should contain the [<s> <s> <*> </s> </s>] fine-grained EC
  if (GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes[CChartParserInfo::m_FullSentenceNT]->m_fineGrainedNodes.size() != 1){
    cout<<"Full Sentence Completed hyps: ========="<<endl;
    DisplayCompletedHypsInCell(CChartParserInfo::m_sourceLength, 1);
  }
  assert(!CChartParserInfo::m_GlueAndEOSAssumptions || GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes[CChartParserInfo::m_FullSentenceNT]->m_fineGrainedNodes.size() == 1);

  return RetrieveKBestParsesSpan(K, GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes[CChartParserInfo::m_FullSentenceNT]);
}

//!Returns the k best hyps under a particular cell - on all production nodes separately
/*
 * This method should not normally be used.
 *  There are two ways to do this
 *   -- call k/#num prods for each of prod result then add boundary word cost and put them all together in one beam
 *   -- calculate boundary word cost for each hyp on the beam and update the model cost before putting it on the the beam
 *      and then operate as if we have a full spanning hyp during k-best - this is more involved...
 */
vector<CChartParserHypComplete*> CChartParser::RetrieveKBestParsesCell(int K, int i, int j){

  //pCompleteProductionNode is NOT in the chartCell
  if (i==CChartParserInfo::m_sourceLength && j==1){
    cerr << "Warning: Retrieving not only full-sentence span _S hyps, but also other full-sentence hyps. Hope you know what you're doing..." << endl;
    //    return RetrieveKBestParsesSpan(K, m_pCompleteTranslationsNode);
  }
  CCompleteHypChartCell& chartCell = m_completeHypsChart[i][j];
  vector<CChartParserHypComplete*> cellNBestList;

  //go thru each production node
  int fineGrainedK = MAX((int)((double)K/(double)chartCell.m_productionNodes.size()), 1);
  for (CCoarseGrainedEq::const_iterator iter = chartCell.m_productionNodes.begin();
       iter!= chartCell.m_productionNodes.end();
       ++iter
       ){
    CProductionNode* pNode = (*iter).second;
    vector<CChartParserHypComplete*> prodNBestList = RetrieveKBestParsesSpan(fineGrainedK, pNode);
    cellNBestList.insert(cellNBestList.end(), prodNBestList.begin(), prodNBestList.end());
  }
  return cellNBestList;
}

//!Returns the k best hyps under a particular production node
/*
 * There are two ways to do this
 *   -- call k/#num classes for each of the equivalence classes then add boundary word cost and put them all together in one beam
 *   -- calculate boundary word cost for each hyp on the beam and update the model cost before putting it on the the beam
 *      and then operate as if we have a full spanning hyp during k-best - this is more involved...
 */
//CHypBeam* CChartParser::RetrieveKBestParsesSpan(int K, CProductionNode* pTopNode){
vector<CChartParserHypComplete*> CChartParser::RetrieveKBestParsesSpan(int K, CProductionNode* pTopNode){

  //CHypBeam* pBeam = new CHypBeam(K * CChartParserInfo::BeamSizeFactor);
  //the top level beam is K best and will handle the introduction of these hyps appropriately calling ScaledTMLMCost(), which is why
  //we need to update the costs directly before filing them into the top beam. But its very expensive to compute upto K best for
  //each fine-grained class...

  vector<CChartParserHypComplete*> prodNBestList;

  //ask for upto fineGrainedK per fine-grained equivalence class
  //note that this is only relevant if there are more than one fine grained EC (which is not the
  //case for the full-sentence _S node, which only has the [<s> <s> <*> </s> </s>] EC
  int fineGrainedK = MAX((int)((double)K/(double)pTopNode->m_fineGrainedNodes.size()), 1);
  for (CChartParserNodeContainer::iterator fineIter = pTopNode->m_fineGrainedNodes.begin(); fineIter!= pTopNode->m_fineGrainedNodes.end(); ++fineIter){

    CChartParserNode* pFine = fineIter->second;
    if (CChartParserInfo::m_verbosity>0){
      cout<<"Retrieving upto: "<<fineGrainedK<<" for: "; pFine->Display(); cout<<endl;
    }

    // Boundary word cost
    // When you are calculating sub-sentence spans, you need to account for the first CChartParserInfo::C_HistoryLength words
    // This cost will be zero if this is the full-sentence node or any _S subspan with j==1 since [<s> <s>] has
    // no costs, so we don't have to worry about it.
    // If you want to use other kinds of subspans, you perhaps should add the boundary cost estimate for the first
    // CChartParserInfo::C_HistoryLength words to the LM neglob probs (don't do it as in the next line though as this assumes
    // all hyps to have the same first words, which is not true in cmp0)
    //    float boundaryCostUnscaled = CUtilities::ComputeLMCost(pFine->m_lmContext, true);
    assert(pFine->m_scaledBoundaryCost==0.0);

    vector<CChartParserHypComplete*> fineNBestList; //will contain fresh hyps that need to be deleted
    cout<<"Using m_nbestStyle: "<<CChartParserInfo::m_nbestStyle<<endl;

    if (CChartParserInfo::m_nbestStyle==3){

      time_t time_start, time_end;
      clock_t time_start_CPU, time_end_CPU;
      time(&time_start);
      time_start_CPU = clock();

      //CHypBeam* pFineBeam = RetrieveKBestParses(fineGrainedK, pFine);
      CHypBeam* pFineBeam;
      if (CChartParserInfo::m_breadthAStar){
	cout<<"Using A* Recombine Breadth ";
	pFineBeam = RetrieveByAgenda(fineGrainedK, pFine);
      }else{
	cout<<"Using A* 1st best ";
	pFineBeam = RetrieveKBestParses(fineGrainedK, pFine);
      }
      time(&time_end);
      time_end_CPU = clock();

      double NumSecsAStar = difftime(time_start, time_end);
      double NumSecsAStar_CPU = ((double) (-time_start_CPU + time_end_CPU)) / CLOCKS_PER_SEC;


      //dont handle the unknown words just yet
      time_start_CPU = clock();
      fineNBestList = HC05onBeam(fineGrainedK, *pFineBeam);
      time_end_CPU = clock();
      double NumSecsHC05_CPU = ((double) (-time_start_CPU + time_end_CPU)) / CLOCKS_PER_SEC;

      //print this calculation to calculate which component was spent on the A* segment
      cout<<"TIME-Sent: NumSecsAStar: "<<NumSecsAStar<<" NumSecsAStar_CPU: "<<NumSecsAStar_CPU<<" NumSecsHC05_CPU: "<<NumSecsHC05_CPU<<endl;

      //this will perform the following operations
      // - enter the beam and delete each of the equivalence classes stored within
      // - before deleting the equivalence class, they call ClearRecombined to remove all the recombined intermediate hyps
      // - delete the eq class ALSO delete all hyps generated during the HC05 process
      //
      // arv : we dont want to do this in here.. the hc05 method gets to do all its clearing OUTSIDE
      // delete pFineBeam;
      m_pBeamToDelete = pFineBeam;


    }else if (CChartParserInfo::m_nbestStyle==2){
      cout<<"Using Huang/Chiang(2005) K-best"<<endl;
      vector<CChartParserHypComplete*> hcNBestList = pFine->RetrieveKBestHC(fineGrainedK);
      //the memory associated with these hyps will be deleted when the node is deleted, clone it here
      for (vector<CChartParserHypComplete*>::iterator iter = hcNBestList.begin(); iter != hcNBestList.end(); ++iter){
	CChartParserHypComplete* pHyp = (*iter);
	fineNBestList.push_back(new CChartParserHypComplete(*pHyp));
      }
      //memory for HC05 will get released when you delete the nodes in the parse

    }else if (CChartParserInfo::m_nbestStyle==1){
      cout<<"USING A* normal: (not currently supported"<<CChartParserInfo::m_extractUnique<<endl;
      exit(-1);
      //      CHypBeam* pFineBeam = RetrieveKBestParses(fineGrainedK, pFine);
      //      for (CSortedHypNBestContainer::iterator iter = pFineBeam->m_store.begin(); iter != pFineBeam->m_store.end(); ++iter){
      //	//just cast it down for now, we are only cloning a non-nest item
      //	CChartParserHypComplete* pHyp = (CChartParserHypComplete*)(*iter).second;
      //	fineNBestList.push_back(new CChartParserHypComplete(*pHyp));
      //      }
      //      delete pFineBeam; //clearing all memory associated with the n-best list process
    }else{
      cout<<"ERROR: Unknown n-best extration style, exitting"<<endl;
      cerr<<"ERROR: Unknown n-best extration style, exitting"<<endl;
      exit(-1);
    }

    //Now do all the post processing for hyps after n-best extraction


    //if we haven't done any LM calculations or we want to verify that the piecewise LM calculations
    //were correct - do it now before calling HandleUnknownAndTags messes up the translationOutput
    if (CChartParserInfo::m_rescoreLM){
      RescoreLM(fineNBestList);
    }

    // - Call HandleUnknownAndTags which modifies the final m_translationOutput
    for (vector<CChartParserHypComplete*>::iterator iter = fineNBestList.begin(); iter != fineNBestList.end(); ++iter){
      CChartParserHypComplete* pHyp = (*iter);
      //DO IT HERE, its safe! since we know these hyps will never be referred to again in the search
      pHyp->HandleUnknownAndTags();
    }

    //now add all hyps to this production result n-best list
    prodNBestList.insert(prodNBestList.end(), fineNBestList.begin(), fineNBestList.end());

  }
  //now this beam should contain hyps across all the equivalence classes with the correct boundary costs
  return prodNBestList;

}


vector<CChartParserHypComplete*> CChartParser::HC05onBeam(int K, CHypBeam& beam){

  //SANITY check to make sure all hyps got filed into the same class
  //int iterationCount = 0;
  //for (; iter!= beam.m_store.end() && iter->second->m_completelyUnwound && iterationCount<K; ++iter, ++iterationCount){
  //}
  //at this point you should only have 1 element, the top most one
  //assert(iterationCount==1);
  //END SANITY check

  CSortedHypNBestContainer::iterator iter = beam.m_store.begin();
  iter = beam.m_store.begin();


  CChartParserHypCompleteNBest* pTopHyp;
  //the beam might actually be empty in the m_dontReloadBeam case and the m_breadthAStar case
  if (CChartParserInfo::m_dontReloadBeam || CChartParserInfo::m_breadthAStar){
    pTopHyp = m_pFullSentenceHyp;
    assert(pTopHyp!=NULL);
  }else{
    assert(iter!=beam.m_store.end());
    pTopHyp= iter->second;
  }
  //the equivalence class for all the hyps
  CChartParserHypCompleteEqClass* pAllEq = beam.LookupProto(pTopHyp);
  assert(pAllEq!=NULL);
  CChartParserHypCompleteNBest* pTopHC05 = pAllEq->GetKthBest(K-1,true);
  //pAllEq->PrintNBest();

  vector<CChartParserHypComplete*> toRet;
  float lastElementCost = 9999;
  for (vector<CChartParserHypComplete*>::iterator iter= pAllEq->m_candidateVector.begin(); iter!= pAllEq->m_candidateVector.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    assert(pHyp->m_scaledFutureCostEstimate==0.0);
    lastElementCost = pHyp->ScaledTMLMCost();
    //give cloned memory
    toRet.push_back(new CChartParserHypComplete(*pHyp));
  }
  //track the cost of the very last element coming out of HC05
  cout<<"Log: A* Rec + HC05 Kth Cost: "<<lastElementCost<<endl;
  return toRet;
}


// Note: RescoreLM has to do DummyLookup(s) in order to ensure that
// all LM costs are available. From initial inspection this does not
// look like its neccesary, since in cmp1 mode, all the fine grained
// classes are separated by terms that have already been combined
// with the consequent node and therefor dummy and real lookups are done.
// In fact: this is only approximately true... Consider the following case
// Hyp1 : [is C four three]
//      : we process this hyp, and we successfully predict "is C four"
//      : and then hit backoff at "three". We would then store
//      : firstWords="is C four" and state [three]
// Consequent: [number <Hyp1>] can be correctly scored during decoding
//           : but during n-best retrieval we see...
//           : "number is C four three"
//           : if P(is|number) is in the model, we will keep extending
//           : our history, finally requesting
//           : P(three | number is C four) which has never been queued!!!
void CChartParser::RescoreLM(vector<CChartParserHypComplete*>& nBestList){

  bool debug = false;

  if (CChartParserInfo::m_rescoreLM==2) {
    cout << "Sanity-test ";
  }
  cout << "LM-Rescoring" << endl;

  // Queue up all the dummy lookups
  for (vector<CChartParserHypComplete*>::iterator iter =
           nBestList.begin(); iter!= nBestList.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    CLanguageModel::DummyLookupSequenceAllLMs(pHyp->m_translationOutput);
    // m_translationOutput already has appropraite <s> and </s> tags
  }
  CLanguageModel::WaitForLookupsAllLMs();

  for (vector<CChartParserHypComplete*>::iterator iter =
           nBestList.begin(); iter!= nBestList.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    // there are no future costs anymore, everything is settled
    assert(pHyp->m_scaledFutureCostEstimate == 0.0);

    if (debug) {
      cout<<"BeforeRescoring: ";
      pHyp->Display();
      cout<<endl;
      cout<<endl<<"PARSE: ";
      CChartParserInfo::PrintParsingOutput(pHyp->m_parsingOutput);
      cout<<endl;
    }

    //compute LM costs, including (in case this is not a full sentence
    // hyp starting with <s> <s>), an estimate for
    // the first CChartParserInfo::C_HistoryLength words
    for (int i=0; i<CRule::C_NumLMs; i++) {

      //float c = CUtilities::ComputeLMCost(CLanguageModel::m_languageModels[i], pHyp->m_translationOutput, true, true);

      vector<CWordIndex> lmContext;
      vector<CWordIndex> firstWords;
      float insideCost = 0;
      float leftBoundaryCost = 0;
      CUtilities::ComputeLMCostAndResultContext(
          CLanguageModel::m_languageModels[i],
          pHyp->m_translationOutput,
          &insideCost,
          &leftBoundaryCost,
          &lmContext,
          &firstWords);
      float c = insideCost + leftBoundaryCost;

      if (CChartParserInfo::m_rescoreLM==2 &&
          fabs(c - pHyp->m_negLogProbs[CRule::C_LMLocation+i]) > 0.001) {
	cerr << "Sanity check error - check STDOUT for more info." << endl;
	cout << "Sanity check error: --RescoreLM=2 was passed, but for the "
             << (i+1)
             <<"-th LM recomputing LM costs during Rescoring yielded a different cost (recomputedCost=" << c << " != "
             <<pHyp->m_negLogProbs[CRule::C_LMLocation+i]
             <<"=originalCost)! Hyp concerned:" << endl;
	pHyp->DisplayRecursive();
        exit(2);
      }
      pHyp->m_negLogProbs[CRule::C_LMLocation+i] = c;
    }

    // recompute total cost
    pHyp->m_totalCost = 0.0;
    for (int i=0; i<pHyp->m_negLogProbs.size(); i++){
      pHyp->m_totalCost +=
          pHyp->m_negLogProbs[i]*CChartParserInfo::m_scalingFactors[i];
    }
    if (debug){
      cout<<"AfterRecoring: "; pHyp->Display(); cout<<endl;
    }
  }
  //re-sort after rescoring
  sort(nBestList.begin(), nBestList.end(), KBestCandidateMinCompare());
  reverse(nBestList.begin(), nBestList.end());
}

void CChartParser::InitializeBeam(CHypBeam& beam, CSortedHypContainer* sortedHyps){
  int counter = 0;
  cout<<"\nLog: A* Rec: TopHyps: "<<sortedHyps->size()<<endl;
  for (CSortedHypContainer::const_iterator iter = sortedHyps->begin(); iter!= sortedHyps->end(); ++iter){
    counter++;

    CChartParserHypComplete* pOldHyp = (*iter).second;
    CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(*((*iter).second));

    //this is a very special case of a full-span full-sentence category complete hyp that came into existence via the
    //unknown-rule - this rule must thus have span-length 1 and thus our source sentence must also have only one word
    //(the unknown word)
    if (pNewHyp->m_translationOutput.size() == 1 && pNewHyp->m_translationOutput.front() == CChartParserInfo::m_unknownWI) {
      // save the corresponding source word
      // since this must come from an unknown-rule of the form _UNKNOWN#_UNKNOWN#@X#weights, we expect this rule to have
      // no backpointers and a span of one
      if ((pNewHyp->m_backPointers.size() != 0) || (pNewHyp->I() != 1)) {
	cerr << "In UnwindHyp: wrong format of unknown-rule employed. Unknown-rule's hyp is: ";
	pNewHyp->DisplayRecursive(CChartParserInfo::m_pVocabulary, 0);
      }

      //parsing is indexed from 1 while the m_pSentence is indexed from 0
      assert(CChartParserInfo::m_pSentence.size()>pNewHyp->J()-1);
      pNewHyp->m_targetOrderedUnknownSourceWords.push_back(CChartParserInfo::m_pSentence[pNewHyp->J()-1]);
      assert(pNewHyp->J() == 1);
    }

    //initialize the parse output - THIS IS BEFORE ANY UNWINDING IS DONE
    CWordSequence newParsingOutput;
    CChartParserHypComplete::AddParseOutputSubTreeStart(newParsingOutput, pNewHyp);

    newParsingOutput.insert(newParsingOutput.end(), pNewHyp->m_translationOutput.begin(), pNewHyp->m_translationOutput.end());
    CChartParserHypComplete::AddParseOutputSubTreeEnd(newParsingOutput);
    pNewHyp->m_parsingOutput = newParsingOutput;

    //make it very clear that these are the purely unwound hyps (even tho they might be all terminals)
    pNewHyp->SetBackInfo(NULL, NULL);

    //the minute these hyps are inserted they might create equivalance classes, remember these are classes based on
    //the contet before m_posNT (which is at 0 right now), followed by remaining content
    //cout<<" Before insertion: m_active"<<pNewHyp->m_active<<endl;
    beam.InsertHyp((pNewHyp));
  }
}

CHypBeam* CChartParser::RetrieveByAgenda(int K, CChartParserNode* pTopFineNode){
#ifndef NDEBUG
  bool debug = true;
#endif
   //clear the stats for this retrieval round
  CChartParserInfo::ResetNBestStatistics();
  CSortedHypContainer* sortedHyps = &(pTopFineNode->m_hypContainer);
  CHypBeam* pBeam = new CHypBeam(K * CChartParserInfo::BeamSizeFactor);
  CHypBeam& beam = *pBeam;
  InitializeBeam(beam, sortedHyps); //beam now has all the starter elements in it
  m_pFullSentenceHyp = NULL; //we will convey the first best thru this
#ifndef NDEBUG
  if (debug){
    cout<<"Initialized beam size:"<<beam.m_store.size()<<endl;
  }
#endif
  bool done = false;
  while (!done){
    vector<CChartParserHypCompleteNBest*> agenda = beam.GetAgenda(CChartParserInfo::m_aStarBeam);
    vector<CChartParserHypCompleteNBest*>::iterator iter = agenda.begin();
    //skip past all the elements that are commpletely unwound!
    int iterationCount = 0;
#ifndef NDEBUG
    if (debug){
      cout<<"Agenda size:"<<agenda.size()<<endl;
    }
#endif
    if (agenda.size()==0){
      assert(m_pFullSentenceHyp!=NULL);
      return pBeam;
    }
    CChartParserHypCompleteNBest* pBestHyp = (*iter);
    float lowestCost = pBestHyp->ScaledTMLMCost(); //this is the lowest cost on the agenda
#ifndef NDEBUG
    if (debug){
      cout<<"LowestCost: "<<lowestCost<<endl;
    }
#endif

    //move thru the completely unwound elements
    for (; iter!= agenda.end() && (*iter)->m_completelyUnwound; ++iter, ++iterationCount){
    }

    //this means all the elements on the agenda are totally unwound
    if (iter==agenda.end()){
      //the agenda is stocked with a set of completely unwound items. Files they guys onto the real
      //beam, they will all go into the same class
      //for (vector<CChartParserHypCompleteNBest*>::iterator it = agenda.begin(); it!=agenda.end(); ++it){
      //assert((*it)->m_completelyUnwound);
      //beam.InsertHyp();
      //}
      assert(agenda.size()>0);
      m_pFullSentenceHyp = agenda[0]; //setting m_pFullSentenceHyp allows the HC05onBeam to start
      done = true;
    }else{
    //now the iter is at the first hyp that we should unwind..lets unwind until the end!
      for (; iter!= agenda.end(); ++iter){
	//children within the lowestCost + m_agendaBeam get a chance on the beam
	//children above the lowestCost + m_astarbeam terminates the unwinding process
	CChartParserHypCompleteNBest* pCurrentHyp = (*iter);
	if (!pCurrentHyp->m_completelyUnwound){
	  CChartParserHypCompleteNBest* pFullTmp = UnwindHyp(pCurrentHyp, beam, lowestCost);
	  if (pFullTmp!=NULL){
	    m_pFullSentenceHyp = pFullTmp;
	  }
	  pCurrentHyp->m_active = false; // this ensures that this hyp wont get back on the agenda
	}
	//if the hyp is unwound and  on the agenda, it would not be bumped off
	//to get online updates, you have to look at the real beam now, not the agenda
	//lowestCost = ((beam.m_store.begin()))->second->ScaledTMLMCost();
	//arv: we no longer store anything sorted... so we need a variable instead
	lowestCost = beam.GetLowestCost();
      }
    }
#ifndef NDEBUG
    if (debug) {
      cout<<"\n\nBeam after expansion ==================== "<<endl;
      beam.PrintBeam();
    }
#endif
  }


  return pBeam;
}

//! returns a pointer to a newly created hyp beam of the K best
//  completely unwound translation hypotheses.
/*!
  An A* style lazy unwinding algorithm
  Don't forget to delete the returned beam! The beam contains only
  freshly created hyps. These will all be deleted by the beam's destructor.
*/
CHypBeam* CChartParser::RetrieveKBestParses(int K, CChartParserNode* pTopFineNode) {
#ifndef NDEBUG
  bool debug = true;
#endif
  //clear the stats for this retrieval round
  CChartParserInfo::ResetNBestStatistics();
  CSortedHypContainer* sortedHyps = &(pTopFineNode->m_hypContainer);

  //our beam: a multimap of complete hyps sorted by their totalcosts with an overridden insert function that only keeps upto K elements in the beam
  // if we correctly separate equivalence classes by first words and last words, we only need to keep track of the best
  // K hyps on the beam at any time.
  // If we are in CompareTargetWords==false though, the 'best cost' value for each EC is only an estimate, and so it is
  // better to keep track of the best K * BeamSizeFactor hyps
  // the same applies when we do uniquing while K-best extracting (ExtractUnique=1) because then we're constantly deleting
  // undesired hyps from the beam, thus making it smaller than K
  CHypBeam* pBeam = new CHypBeam(K * CChartParserInfo::BeamSizeFactor);
  CHypBeam& beam = *pBeam;

  //the full sentence spanning hyp
  CChartParserHypCompleteNBest* pFullHyp = NULL;

  int counter = 0;
  //the set of hyps that are sentence spanning
  //we used to restrict to only pull up the first K elements, but this is not the point of the astar search, we want to keep
  //pulling until we are happy with the result
  //for (CSortedHypContainer::const_iterator iter = sortedHyps->begin(); iter!= sortedHyps->end() && counter < K; ++iter){

  cout<<"\nLog: A* Rec: TopHyps: "<<sortedHyps->size()<<endl;
  for (CSortedHypContainer::const_iterator iter = sortedHyps->begin(); iter!= sortedHyps->end(); ++iter){
    counter++;

    CChartParserHypComplete* pOldHyp = (*iter).second;

#ifndef NDEBUG
    if (debug){
      cout<<"Copying from old: "; pOldHyp->PrintTranslationOutput(); cout<<" m_taggedTargets: "; CChartParserInfo::PrintVocabSequence(pOldHyp->m_taggedTargets); cout<<endl;
    }
#endif
    //clone retrieved hyp (b/c we will delete all the elements on the beam and don't want to try delete things twice)
    //CChartParserHypComplete* pNewHyp = new CChartParserHypComplete((*iter).second);

    CChartParserHypCompleteNBest* pNewHyp = new CChartParserHypCompleteNBest(*((*iter).second));

    //pNewHyp has its m_posNT correctly set, but its the job of unwind hyp to clone this hyp and make it completely unwound
    //if (pNewHyp->m_posNT==pNewHyp->m_translationOutput.size()){
    //pNewHyp->m_completelyUnwound=true;
    //}


    //arv: this parent cost tracking doesnt quite work
    //make sure to do this BEFORE multiplying in the FutureCostImpact
    //this is just for tracking the hyp as it unrolls
    //pNewHyp->AddParentHypCost(pNewHyp->ScaledTMLMCost());

    //arv: this was the old style. now the cost is kept purely in the m_totalCost
    // the current RetrieveKBestParses and UnwindHyp will re-calculate the LM from scratch and keep adding
    // the so-far-LM-processed words' LM costs to m_negLogProbs[CRule::C_LMLocation]
    // m_scaledFutureCostEstimate will contain the optimistic estimate of LMCost*FutureCostImpact for the not-yet
    // LM-recalculated words
    //pNewHyp->m_scaledFutureCostEstimate *= CChartParserInfo::FutureCostImpact;
    //assert(pNewHyp->m_negLogProbs[CRule::C_LMLocation] == 0.0);


    //arv: no such thing as first words
    // initialize m_firstWords for use in UnwindHyp
    // m_firstWords will now be abused to contain all words of the translationOutput that have been processed so far
    //pNewHyp->m_firstWords.clear();

    //this is a very special case of a full-span full-sentence category complete hyp that came into existence via the
    //unknown-rule - this rule must thus have span-length 1 and thus our source sentence must also have only one word
    //(the unknown word)
    if (pNewHyp->m_translationOutput.size() == 1 && pNewHyp->m_translationOutput.front() == CChartParserInfo::m_unknownWI) {
      // save the corresponding source word
      // since this must come from an unknown-rule of the form _UNKNOWN#_UNKNOWN#@X#weights, we expect this rule to have
      // no backpointers and a span of one
      if ((pNewHyp->m_backPointers.size() != 0) || (pNewHyp->I() != 1)) {
	cerr << "In UnwindHyp: wrong format of unknown-rule employed. Unknown-rule's hyp is: ";
	pNewHyp->DisplayRecursive(CChartParserInfo::m_pVocabulary, 0);
      }

      //parsing is indexed from 1 while the m_pSentence is indexed from 0
      assert(CChartParserInfo::m_pSentence.size()>pNewHyp->J()-1);
      pNewHyp->m_targetOrderedUnknownSourceWords.push_back(CChartParserInfo::m_pSentence[pNewHyp->J()-1]);
      assert(pNewHyp->J() == 1);
    }

    //add the hyp to the beam
#ifndef NDEBUG
    if (debug) {
      cout << "Adding to beam the full-sentence spanning hyp: \n";
      pNewHyp->DisplayRecursive();
      cout << endl;
    }
#endif

    //initialize the parse output - THIS IS BEFORE ANY UNWINDING IS DONE
    CWordSequence newParsingOutput;
    CChartParserHypComplete::AddParseOutputSubTreeStart(newParsingOutput, pNewHyp);

    newParsingOutput.insert(newParsingOutput.end(), pNewHyp->m_translationOutput.begin(), pNewHyp->m_translationOutput.end());
    CChartParserHypComplete::AddParseOutputSubTreeEnd(newParsingOutput);
    pNewHyp->m_parsingOutput = newParsingOutput;

    //adding to beam
    //cout<<"Adding to beam: "; pNewHyp->Display(); cout<<endl;
    //CChartParserInfo::PrintVocabSequence(pNewHyp->m_parsingOutput); cout<<endl;

    //make it very clear that these are the purely unwound hyps (even tho they might be all terminals)
    pNewHyp->SetBackInfo(NULL, NULL);
    //the minute these hyps are inserted they might create equivalance classes, remember these are classes based on
    //the contet before m_posNT (which is at 0 right now), followed by remaining content
    //cout<<" Before insertion: m_active"<<pNewHyp->m_active<<endl;
    beam.InsertHyp((pNewHyp));
    //cout<<" After insertion: m_active"<<pNewHyp->m_active<<endl;
  }

  //cout<<"Unwinding hyps..."<<endl;
  bool done = false;
  cout<<"Log: A* Rec: Init: "<<beam.m_store.size()<<endl;
#ifndef NDEBUG
    if (debug) {
      cout<<"\n\nBeam INIT ==================== "<<endl;
      beam.PrintBeam();
    }
#endif

  vector<float> lowestUnwoundCosts;
  vector<float> lowestAllCosts;
  while (!done) {

    //iterate through the beam through increasingly costly hyps and find the best hyp that is still not completely unwound
    //if all hyps are completely unwound, we are done
    //CSortedHypContainer::iterator iter = beam.m_store.begin();

    CSortedHypNBestContainer::iterator iter = beam.m_store.begin();
    if (iter!=beam.m_store.end()){
      CChartParserHypCompleteNBest* pTopHyp = iter->second;
      lowestAllCosts.push_back(pTopHyp->ScaledTMLMCost());
    }


    int iterationCount = 0;
    //look for the lowest cost unwound hyp
    float worstCompleteHyp = 9999;
    float lowestCostHyp = 9999;
    //for (; iter!= beam.m_store.end() && iter->second->m_completelyUnwound && iterationCount<K; ++iter, ++iterationCount){

    //remove the restriction the we only go past the first K elements, we want to end up at the next
    //unwound hyp regardlesss if the first K have been fully un-wound
    for (; iter!= beam.m_store.end() && iter->second->m_completelyUnwound; ++iter, ++iterationCount){
      if (iterationCount==0){
	lowestUnwoundCosts.push_back(iter->second->ScaledTMLMCost());
	lowestCostHyp = iter->second->ScaledTMLMCost();
      }
      assert(iter->second->m_scaledFutureCostEstimate==0.0);
      worstCompleteHyp = iter->second->ScaledTMLMCost();
    }
    //post-condition: worstCompleteHyp holds the cost

#ifndef NDEBUG
    if (debug){
      cout<<"Popping beam iterationCount: "<<iterationCount<<" size: "<<beam.m_store.size()<<endl;
    }
#endif


    //if the beam is empty, lets terminate.
    if (iter==beam.m_store.end()) {
      //yes, so we're done
      done = true;

      //we should indicate the termination reason
      cout<<"Log: A* Rec: ";
      if (iter==beam.m_store.end()){
	cout<<" beam empty : iterationCount:"<<iterationCount;
      }
      if (iterationCount == K){
	cout<<" first k unwound ";
      }
      cout<<endl;

      //remove the hyps beyond the first K from the beam
      //only reason to do this is to save memory before running hc05
      //beam.Shrink(K);
    } else {

      //pop the top element of
      CChartParserInfo::m_topCallsKBest++;
      //no, so unwind hypothesis iter->second
      CChartParserHypCompleteNBest* pTopHyp = iter->second;
      //cout<<"Unwinding: "; pTopHyp->PrintTranslationOutput(); cout<<" Cost: "<<pTopHyp->ScaledTMLMCost()<<endl;

      //quit if the next hyps cost is much higher
      if (pTopHyp->ScaledTMLMCost() > CChartParserInfo::GetPruningCutoff(lowestCostHyp, CChartParserInfo::m_aStarBeam) ){

	//display the cost of the rejected item to compare it to HC05's n-best extraction
	//we want the K'th best
	cout<<"Log: A* Rec: Rejected Item Cost: "<<pTopHyp->ScaledTMLMCost()<<endl;
	//so we're done
	done = true;
      }else{

	//the beam could be made smaller here, because you could have found a not complete hyp that has been deactivated and delete him
	//we use the MIN because lowestCostHyp is potentially undefined (since its defined by the existence of m_completelyUnwound hyps)
	//at the top of the beam.
	CChartParserHypCompleteNBest* toUnwind = iter->second;
	beam.m_store.erase(iter);
	CChartParserHypCompleteNBest* pFullHypTmp = UnwindHyp(toUnwind, beam, MIN(pTopHyp->ScaledTMLMCost(), lowestCostHyp));
	if (pFullHypTmp!=NULL){
#ifndef NDEBUG
    if (debug) {
      cout<<"Setting COMPLETE: "; pFullHypTmp->PrintTranslationOutput(); cout<<endl;
    }
#endif
	  pFullHyp = pFullHypTmp;
	}

      }

    }

#ifndef NDEBUG
    //if (debug) {
    //cout<<"\n\nBeam after ==================== "<<endl;
    //beam.PrintBeam();

    //cout<<"\n\nEQ after ==================== "<<endl;
    //beam.PrintClasses();
    //}
#endif
  }
#ifndef NDEBUG
  if (debug) {
    cout<<"\n\nBeam astar end ==================== "<<endl;
    beam.PrintBeam();

    cout<<"\n\nEQ astart end  ==================== "<<endl;
    beam.PrintClasses();
  }
#endif

  CChartParserInfo::m_numUnique = pBeam->m_uniqueChecker.Size();
  cout<<"Log: NBest: "; CChartParserInfo::PrintNBestStatistics(); cout<<endl;
  //cout<<"Log: NBestLowestUnwound: "; copy(lowestUnwoundCosts.begin(), lowestUnwoundCosts.end(), ostream_iterator<float>(cout, " ")); cout<<endl;
  //cout<<"Log: NBestLowestAll: "; copy(lowestAllCosts.begin(), lowestAllCosts.end(), ostream_iterator<float>(cout, " ")); cout<<endl;


  //set the hyp that was sentences spanning
  m_pFullSentenceHyp=NULL;
  m_pFullSentenceHyp = pFullHyp;
  assert(m_pFullSentenceHyp!=NULL);
  //return the created upto-K-best beam of completely-unwound hyps
  return pBeam;
}


//outputs the Huang and Chiang N-Best list
/*
  void CChartParser::OutputHC(vector<CChartParserHypComplete*> candidateSet, int translationCount){

  int count = 0;
  for (vector<CChartParserHypComplete*>::iterator iter=pCompleteProductionNodeFine->m_candidateVector.begin(); count<translationCount && iter!=pCompleteProductionNodeFine->m_candidateVector.end(); ++count, ++iter){
  CChartParserHypComplete* pTopHyp = (*iter);
  //cout<<"RETRIEVED: "; pTopHyp->DisplaySimpleKBest(); cout<<endl;

  }
  }
*/

//! iterate through the beam from best hyp to worst (assuming all hyps
//are completely unwound and unknown-handled) and output the hyp's translations
//void CChartParser::OutputBeam(CHypBeam* pBeam, int translationCount) {
void CChartParser::OutputNBest(vector<CChartParserHypComplete*>::iterator first, vector<CChartParserHypComplete*>::iterator last, int translationCount) {

  for (vector<CChartParserHypComplete*>::iterator iter = first; translationCount-- != 0 && iter!= last; ++iter){
    //CChartParserHypComplete* pHyp = (*iter).second;
    CChartParserHypComplete* pHyp = (*iter);
    cout<<endl;
    cout<<"ScaledTMLMCost: "<<pHyp->ScaledTMLMCost()<<" m_totalCost: "<<pHyp->m_totalCost<<" raw:";
    for (int i=0; i<pHyp->m_negLogProbs.size(); i++){
      cout<<" "<<pHyp->m_negLogProbs[i];
    }
    cout << " UNKNOWN-TAGGED-handled translation: ";
    //output the parse tree too
    CChartParserInfo::PrintVocabSequence(pHyp->m_translationOutput);
    cout<<" PARSE: "; CChartParserInfo::PrintParsingOutput(pHyp->m_parsingOutput);

  }
  cout<<endl;

}


void CChartParser::PruneCell(const int i, const int j){

  bool debug = false;


  //CCompleteHypChartCell& chartCell = m_completeHypsChart.GetRow(i)[j];
  CCompleteHypChartCell& chartCell = m_completeHypsChart[i][j];
  if (debug){
    cout<<"====== Before Pruning ==== "<<endl;
    DisplayCell(i, j);
  }
  vector<CProductionNode*> allProductionNodes;
  float lowestNTNodeCost = FLT_MAX;
  //fill the allProductionNodes and also determine lowest All-Component-Cost across all noterminal production nodes
  for (CCoarseGrainedEq::const_iterator iter = chartCell.m_productionNodes.begin();
       iter!= chartCell.m_productionNodes.end();
       ++iter
       ){
    CProductionNode* node = (*iter).second;
    allProductionNodes.push_back(node);
    if (!node->m_lexical) {
      lowestNTNodeCost = MIN(lowestNTNodeCost,node->m_pLowestCostHyp->AllComponentCost());
    }
  }
  //az: sorting unnecessary, we'll iterate through all of them anyway
  //  sort(allProductionNodes.begin(), allProductionNodes.end(), CompareProductionNode());

  if (allProductionNodes.size()==0){
    return;
  }


  double safeCutoffLogProb = CChartParserInfo::GetPruningCutoff(lowestNTNodeCost, CChartParserInfo::m_maxCostDifferencePerCell);


  //clear out what was left this these nodes
  chartCell.m_productionNodes.clear();

  assert(safeCutoffLogProb!=0.0);
  pair<int, int> prodKeptAll(0, 0);
  //  vector<CProductionNode*> toDelete;
  for (vector<CProductionNode*>::iterator iter= allProductionNodes.begin(); iter!= allProductionNodes.end(); ++iter){
    CProductionNode* pNode = (*iter);

    if (pNode->m_pLowestCostHyp->AllComponentCost() > safeCutoffLogProb ){
      //as long as you are not a glue or full sentnence node
      if (!pNode->m_lexical && pNode->GetResult()!=CChartParserInfo::m_GlueNT &&  pNode->GetResult()!=CChartParserInfo::m_FullSentenceNT){
	if (CChartParserInfo::m_verbosity>1){
	  cout<<"Deleting: "; pNode->Display(); cout<<" lowest: ";pNode->m_pLowestCostHyp->AllComponentCost(); cout<<endl;
	}
	//	toDelete.push_back(pNode);
	//delete the entry from the m_newlyCreatedNodes
	CCoarseGrainedEq::iterator toEraseIter = m_newlyCreatedNodes.find(pNode->GetResult());
	//if its also a newly created node, then erase it from that map
	//az: if it's not newly created, we shouldn't delete it b/c the Self rule might already have created hyps
	//that depend on it
	if (toEraseIter!=m_newlyCreatedNodes.end()){
	  m_newlyCreatedNodes.erase(toEraseIter);
	  delete pNode;
	} else {
	  chartCell.m_productionNodes[pNode->GetResult()] = pNode;
	  prodKeptAll.first++;
	}
      }else{
	//this means you have lowest cost but are a @_G or @_S
	//cout<<"Tried to delete glue or s, but we need to keep it: "; pNode->Display(); cout<<endl;
	chartCell.m_productionNodes[pNode->GetResult()] = pNode;
	prodKeptAll.first++;
      }
    }else{
      chartCell.m_productionNodes[pNode->GetResult()] = pNode;
      prodKeptAll.first++;
    }
    prodKeptAll.second++;
  }

  cout<<"Log: PruneCell production nodes kept: "<<prodKeptAll.first<<" / "<<prodKeptAll.second<<" safeCutoffLogProb: "<<safeCutoffLogProb<<endl;
}


int CChartParser::FinalizeAllNodes(const CCoarseGrainedEq& productionNodes, bool addAll){
  int numEqClasses = 0;
  for (CCoarseGrainedEq::const_iterator iter = productionNodes.begin(); iter!= productionNodes.end(); ++iter){
    CProductionNode* node = (*iter).second;
    node->FinalizeProductionNode(addAll);
    //cout<<"Finalized: "; node->Display();cout<<endl;
    numEqClasses += node->m_fineGrainedNodes.size();
  }

  return numEqClasses;
}




void CChartParser::ParseChart(){
  // main CKY+ loop
  //int i,j; //i is the length of the produced rule, and j is the starting word index of that rule
  for (int i=2; i<=CChartParserInfo::m_sourceLength; i++){
    cout << "Log: ParseChart(): Building completed hyps of size i=" << i << "... " << endl;
    int skippingCount = 0;
    int total_count_EC_IH_pairs = 0;

    //track the span - limit rules based on this span
    m_combinationCount = i;

    //bool maxCombinationCountExceeded = m_combinationCount > CChartParserInfo::m_maxCombinationCount;
    //now do the actual combinations
    if (m_combinationCount == CChartParserInfo::m_maxCombinationCount+1){
      cout << endl << endl << "MaxCombinationCount " << CChartParserInfo::m_maxCombinationCount << " exceeded." << endl << endl;
      //CChartParserInfo::s_maxCombinationCountExceeded = true;
    }
    //left side index
    //az: we used to have an optimization here that creates only cells
    //    with j=1 once the max combi count has been exceeded
    //this opti ruled out combining lexial rules spanning more than
    //max-combi-count words, so I commented it out again
    //but made sure that in fundamental hyp only cells with j=1 are
    // explored for which complete hyp and incomplete hyp
    // are purely lexical, so basically the opti has been moved into
    // ApplyFundamentalRule
    for (int j=1; j<=CChartParserInfo::m_sourceLength-i+1; j++){
      //    for (int j=1; j<=jIndex; j++){

      //try to apply the fundamental rule to fill up cell i,j with new
      //complete hyps


      //but first pre-compute the best rule app count we will achieve over
      //all our new complete hyps combinations if we just glue them together
      // We base our computation on adjacent complete hyps, as we don't
      // know yet which incomplete hyps will combine with which complete
      // hyps.
      // bestCombinedRuleAppCount will always be achieved, since for
      // CHs X and Y, gluing to an _S hyp can take place
      // costing only one additional rule app count

      /* here's how rule app count works now for an example:

      Example:

      <s> the ass </s>
      the#DT, ass#NN, the ass#NP, DT NN#NP

      i\j         1        2           3    4
      1          <s>      the          ass   </s>
      _S(1)   DT(1)        NN(1)
      _S.(1)  NP.(0) (from 'the')
      NP.(1) (from 'DT')

      2:RAC_min= 1+1+1    1+1+1        1+0+1
      _S(3)    NP(1)
      _S.(3)   NP(3)

      3:RAC_min= 1+1+1
      _S(3) (combining _S.[1,1] with NP[2,2])
      _S.(3)

      4:RAC_min= 3+0+1
      _S(4)

      */

      float bestCombinedRuleAppCount = FLT_MAX/2; // divide by 2 to
      // avoid overflow
      // when adding
      // stuff
      int k_max=-1;
      for (int k=1; k<i; k++) { // k: length of source (left) cell
	// the ECs (hypergraph nodes) in the source cell
	CRuleAppCountSortedNodeIndex& eqClasses1 = m_completeHypsChart[k][j].ruleAppCountSortedNodeIndex;
	//	int totalEC1Count = eqClasses1.size();

	// the ECs (hypergraph nodes) in the target cell
	CRuleAppCountSortedNodeIndex& eqClasses2 = m_completeHypsChart[i-k][j+k].ruleAppCountSortedNodeIndex;
	//	int totalEC2Count = eqClasses2.size();

	// is there anything to combine at all?
	//	if (totalEC1Count>0 && totalEC2Count>0) {
	//yes

	//az: we used to check only for _S as the left-side cell's hyps. Now, we don't have _S hyps anymore
	//except in cells with j=1, so we have to check all hyps and add 1 b/c of the transition to glueing,
	//which costs 1 rule-app-count

	// determine best value of m_bestTotalRuleApplicationCount amongst all
	// hyps in the left cell
	float bestRuleAppCountEC1Iteration;

	//	  if (GetCompleteCell(k,j).m_cellType == 1) {
	//	    // commented out as for that case we have a complete _S hyp on top of the j==1 BOS hyp
	//	    // this means we have a BOS cell k=1, j=1
	//	    assert(k==1); assert(j==1);
	//	    // the rule app count for this cell will be 0 as it contains the terminal hyp <s> which will be subsumed by
	//	    // the glue rule
	//	    bestRuleAppCountEC1Iteration = 0;
	//	  } else
	{
	  //cout<<"Checking cell: "<<k<<","<<j<<" with "<<i-k<<","<<j+k<<endl;
	  // in case of j==1, check for complete _S hyp
	  CProductionNode* pClass1;
	  if (j==1 && CChartParserInfo::m_GlueAndEOSAssumptions) {
	    CRuleAppCountSortedNodeIndex::iterator itClasses1;
	    for (	itClasses1 = eqClasses1.begin(); itClasses1!=eqClasses1.end(); ++itClasses1){
	      pClass1 = itClasses1->second;
	      //cout<<"c1 option: "; pClass1->Display(); cout<<endl;
	      if (pClass1->GetResult()==CChartParserInfo::m_FullSentenceNT) {
		break;
	      }
	    }
	    if (itClasses1 == eqClasses1.end()) {
	      cerr << "Grammar's Glue rule is not working. Since j==1, the _S hyp should have been in the cell" << endl;
	      cout << "Grammar's Glue rule is not working. Since j==1, the _S hyp should have been in the cell" << endl;
	      assert(false);
	    } else {
	      bestRuleAppCountEC1Iteration = pClass1->m_bestTotalRuleApplicationCount;
	    }
	  } else {
	    // our grammar has direct glue rules: @_S @N # @1 @2 # @_S
	    // thus we'll use the very first (i.e., minimum-rule-app-count)
            // complete hyp
	    // that is not a terminal symbol hyp
	    CRuleAppCountSortedNodeIndex::iterator itClasses1;
	    for (	itClasses1 = eqClasses1.begin(); itClasses1!=eqClasses1.end(); ++itClasses1){
	      pClass1 = itClasses1->second;
	      //cout<<"c1 option: "; pClass1->Display(); cout<<endl;
	      if (!pClass1->m_lexical) {
		break;
	      }
	    }
	    if (itClasses1 == eqClasses1.end()) {
	      // found no non-lexical hyp
	      assert(eqClasses1.begin()==eqClasses1.end()); // then there should be no hyps at all
	      // set best possible RAC to the best RAC that we can get by gluing subcells together to get
	      // a translation spanning this cell k,j
	      bestRuleAppCountEC1Iteration = GetCompleteCell(k,j).m_bestGluedRuleAppCount;
	    } else {
	      // MIN of:
	      // best value of m_bestTotalRuleApplicationCount
              // amongst all hyps in the left cell
	      // and best RAC that we can get by gluing subcells
              // together to get
	      // a translation spanning this cell k,j
	      bestRuleAppCountEC1Iteration = MIN(GetCompleteCell(k,j).m_bestGluedRuleAppCount,
						 pClass1->m_bestTotalRuleApplicationCount);
	    }
	  }
	}

	// now determine best value of m_bestTotalRuleApplicationCount amongst all
	// hyps in the right cell
	float bestRuleAppCountEC2Iteration;

	if (GetCompleteCell(i-k,j+k).m_cellType == 2) {
	  // this means we have a EOS cell
	  assert(i-k==1); assert(j+k==CChartParserInfo::m_sourceLength);
	  // the rule app count for this cell will be 0 as it contains the terminal hyp </s> which will be subsumed by
	  // the glue rule
	  bestRuleAppCountEC2Iteration = 0;
	} else {
	  // our grammar has direct glue rules <s> @_S @_N # <s> <s> @1 @2 # @_S
	  // thus we'll use the very first (i.e., minimum-rule-app-count) complete hyp
	  // that is not a terminal symbol hyp
	  CProductionNode* pClass2;
	  CRuleAppCountSortedNodeIndex::iterator itClasses2;
	  for (	itClasses2 = eqClasses2.begin(); itClasses2!=eqClasses2.end(); ++itClasses2){
	    pClass2 = itClasses2->second;
	    //cout<<"c2 option: "; pClass2->Display(); cout<<endl;
	    if (!pClass2->m_lexical) {
	      break;
	    }
	  }
	  float minRAC2 = (itClasses2 == eqClasses2.end()) ?
	    // found no non-lexical hyp
	    FLT_MAX/10
	    :
	    pClass2->m_bestTotalRuleApplicationCount;
	  bestRuleAppCountEC2Iteration = (j==1)?
	    // left cell used for the RAC calculation is a glue _S node
	    // so for this right cell always use the actual bestRuleAppCount since
	    // m_bestGluedRuleAppCount (if better) cannot be achieved due to pruning
	    minRAC2
	    :
	    // MIN of:
	    // best value of m_bestTotalRuleApplicationCount amongst all hyps in the right cell
	    // and best RAC that we can get by gluing subcells together to get
	    // a translation spanning this cell i-k,j+k
	    bestRuleAppCountEC2Iteration = MIN(minRAC2,GetCompleteCell(i-k,j+k).m_bestGluedRuleAppCount);
	}
	//cout<<"checking *** "<<endl;
	//cout<<"long cell: "; pClass1->Display();
	//cout<<"short cell: "; pClass2->Display();
	if (bestRuleAppCountEC1Iteration + bestRuleAppCountEC2Iteration + 1 < bestCombinedRuleAppCount) {
	  // gluing left-cell EC1 and right-cell EC2 together (+1 RAC b/c of the use of the glue rule)
	  // sets a new record RAC for this combined cell
	  bestCombinedRuleAppCount = bestRuleAppCountEC1Iteration + bestRuleAppCountEC2Iteration + 1;
	  //cout<<" *** re-setting? **** ";
	  //cout<<endl;
	  k_max=k;
	}

      }//go thru each k
      if (k_max==-1){
	//if we are completely unable to find any matching hyps
	cout<<"ERROR: Unable to find any glue rule style matching for cell: "<<i<<","<<j<<endl;
	exit(-1);
      }
      // set this cell's best RAC achievable by gluing subcells
      GetCompleteCell(i,j).m_bestGluedRuleAppCount = bestCombinedRuleAppCount;


      //now that we know what the bestCombinedRuleAppCount is, do the actual ApplyFundamentalRule application

      //if (CChartParserInfo::m_verbosity>0){
      cout<<"Building Cell: i="<<i<<" j="<<j<<" bestCombinedRuleAppCount: "<<bestCombinedRuleAppCount<<" at: ("<<k_max<<","<<j<<") ("<<i-k_max<<","<<j+k_max<<") "<<endl;
      //}
      //Reset the expansion statistics
      CChartParserInfo::ResetCellStatistics();

      //combo pruning across values of k
      RulesByProductionResult hypRuleCombinations;
      vector<CChartParserHypIncomplete*> incompleteHypsToDelete;
      assert(CChartParserInfo::m_hypsAdded==0);
      for (int k=1; k<i; k++){
        // k: length of source (left) cell; i-k: length of target (right) cell
	if (CChartParserInfo::m_verbosity>0){
	  cout<<"Calling ApplyFundamentalRule src: ["<<k<<","<<j<<"] ["<<i-k<<","<<j+k<<"] bestCombinedRuleAppCount: "<<bestCombinedRuleAppCount<<endl;
	}

	pair<int,int> tmppair =
	  ApplyFundamentalRule(k, j, (i-k), (j+k),
                               bestCombinedRuleAppCount,
                               hypRuleCombinations,
                               incompleteHypsToDelete);
	// just some statistics
	skippingCount += tmppair.first;
	total_count_EC_IH_pairs += tmppair.second;
	if (CChartParserInfo::m_verbosity>0){
	  cout<<"Skipped: "<<tmppair.first<<" out of "
              <<tmppair.second<<" due to too-high rule-app count"<<endl;
	}
      }

      //spawn out any complete hyps in global mode
      if (CChartParserInfo::m_comboPruningGlobal){
	if (hypRuleCombinations.size()>0){
	  //now we have built up a series of CChartParserHypIncomplete - rule vector pairs, that we can process from the top down
	  pair<int, int> expandedAndPossible =
              HypFiller::SpawnKBestCompleteHyps(hypRuleCombinations,
                                                false,
                                                &(m_completeHypsChart[i][j]),
                                                &m_newlyCreatedNodes
                                                );
	  //if (CChartParserInfo::m_verbosity>0){
	  //cout<<"Log: ExpPossGlobal: "<<expandedAndPossible.first<<" / "<<expandedAndPossible.second<<endl;
	  //}
	  hypRuleCombinations.clear();
	}
      } else {
	// not in global mode - this means we have done the Spawning already in ApplyFundamentalRule
      }
      //and delete the incomplete hyps that had no more matches
      for (vector<CChartParserHypIncomplete*>::iterator dIter = incompleteHypsToDelete.begin(); dIter!= incompleteHypsToDelete.end(); ++dIter){
	CChartParserHypIncomplete* pNewIncompleteHypPrototype = (*dIter);
	delete pNewIncompleteHypPrototype;
	(*dIter)=NULL;
      }
      incompleteHypsToDelete.clear();

      // now sort all created nodes in cell i,j according to best
      // rule-app-count and more importantly, call SelfRule
      IndexNodesAccordingToRuleAppCount(i,j);
    } // end j loop

    cout << "Log: " << skippingCount << " out of " <<
        total_count_EC_IH_pairs <<
        " possible combinations of incomplete hyps and hypergraph nodes were skipped due to too high combined Rule-Application-Counts."
	 << endl;
  } // end i loop

  //we don't need the incomplete hyps anymore - clear them
  //when clearing a container, first clear all its elements
  for (int i=1; i<=CChartParserInfo::m_sourceLength; i++){
    for (int j=1; j<=CChartParserInfo::m_sourceLength; j++){
      ClearIHCCell(i, j);
    }
    m_incompleteHypsChart[i].clear();
  }
  m_incompleteHypsChart.clear();
}




//! returns: first: n.o. source IH/target node pairs rule-app-diff skipped;
//second: total n.o. source/target pairs
pair<int,int> CChartParser::ApplyFundamentalRule(int sourceCellRow, int sourceCellColumn, int targetCellRow, int targetCellColumn, float bestResultingRuleAppCountEstimate, RulesByProductionResult& hypRuleCombinations, vector<CChartParserHypIncomplete*>& incompleteHypsToDelete){

  // the bestResultingRuleAppCountEstimate estimate of the minimum-rule-app-count rule created by this ApplyFundamentalRule method
  // will change if we create a lower-rule-app-count rule

  // correspondingly, maxResultingRuleAppCountAllowed will also adapt
  float maxResultingRuleAppCountAllowed
    = CChartParserInfo::m_multiplicativeRA ? (int)(ceil(bestResultingRuleAppCountEstimate*m_maxRuleAppCountDifference)) : (int)(bestResultingRuleAppCountEstimate + m_maxRuleAppCountDifference);


  int skippingOccurred = 0;

  //iterate through the completed items in the target side, and for each production, check to
  //see if any are available on the incomplete side
  bool s_debugApplyFund = false;
  //allow all lexical only elements to pass thru the m_maxCombinationCount filter which only allows S, G combination thru
  bool allowLexicalsThru = true;

  // the (coarse-grained) ECs (hypergraph nodes) in the target cell
  CRuleAppCountSortedNodeIndex& eqClasses = m_completeHypsChart[targetCellRow][targetCellColumn].ruleAppCountSortedNodeIndex;
  int totalECCount = eqClasses.size();
  int pairsExplored = 0; //count the actual number of pair explored (less than totalECCount*totalIncompleteHypCount) due to MCC

  // the incomplete hyps in the source cell
  CSortedIncompleteHypContainer& incompleteHyps = m_incompleteHypsChart[sourceCellRow][sourceCellColumn];
  int totalIncompleteHypCount = incompleteHyps.size();

  if (s_debugApplyFund){cout<<"totalECCount: "<<totalECCount<<" totalIncompleteHypCount: "<<totalIncompleteHypCount<<endl;}
  // is there anything to combine at all?
  if ((totalECCount>0 && totalIncompleteHypCount>0) &&
      // optimization: if we are supposed to create cells spanning over the last source word (the </s> word),
      // that can only be possible if we are creating full-sentence-cells since all rules in the grammar
      // involving </s> are assumed to be either the glue rule (which is left-most) or a rule with source side <s> ... </s>
      (!CChartParserInfo::m_GlueAndEOSAssumptions || // don't do the opti
       targetCellRow+targetCellColumn-1 < CChartParserInfo::m_sourceLength || // not reaching over the </s> word
       sourceCellColumn == 1 // starting with <s> word (which, when reaching over the </s> word, means, we are spanning the whole sentence
       )      ) {
    //yes

    // best value of m_bestTotalRuleApplicationCount amongst all nodes
    // hyps in the target cell
    float bestRuleAppCountECIteration = (*eqClasses.begin()).first;

    // best value of m_totalRuleApplicationCount amongst all incomplete
    // hyps in the source cell
    float bestRuleAppCountIncompleteHypIteration = (*incompleteHyps.begin()).first;

    if (s_debugApplyFund){
      //  cout<<"Num of eq classes: "<<eqClasses.size()<<endl;
    }

    //iterate through the incompleted hyps of current IC chart cell
    //(sourceCellRow,sourceCellColumn)
    bool continueIncompleteHypIteration = true;
    int iterationCountIH = 0;
    for (CSortedIncompleteHypContainer::const_iterator iterHyps =  incompleteHyps.begin(); continueIncompleteHypIteration != 0 && iterHyps != incompleteHyps.end(); ++iterHyps, ++iterationCountIH){

      float ruleAppCountIH = (*iterHyps).first;

      if (ruleAppCountIH + bestRuleAppCountECIteration + 1 > maxResultingRuleAppCountAllowed) {
        // no chance of getting an allowed combined rule app count anymore
	// don't consider this and any subsequent IHs of
	// this cell anymore
	continueIncompleteHypIteration = false;
	skippingOccurred += (totalIncompleteHypCount - iterationCountIH)*totalECCount;

      } else {


	CChartParserHypIncomplete* pIncompleteHyp = (*iterHyps).second; // the current incomplete hyp: our candidate for a match with pNode
	bool incompleteIsLexical = pIncompleteHyp->m_productionBackPointers.size() == 0;

	// we will only do the combination if the source span hasn't exceeded the panic threshold
	// or if the inc. hyp is lexical (since we don't want to miss out on long phrase-pair rules)
	// or if this incomplete hyp starts at source word 1, i.e., the corresponding rule's source side starts
	// with <s> (e.g., glue hypotheses)
	if (m_combinationCount <= CChartParserInfo::m_maxCombinationCount ||
	    (sourceCellColumn==1) ||
	    (incompleteIsLexical&& allowLexicalsThru)) {

	  //do the expansion for each equivalence class, until the EC's rule app
	  //counts get too bad
	  bool continueECIteration = true;
	  int iterationCountEC = 0;
	  for (CRuleAppCountSortedNodeIndex::const_iterator iter = eqClasses.begin();  continueECIteration && iter!= eqClasses.end(); ++iter, ++iterationCountEC){

	    float ruleAppCountEC = (*iter).first;

	    if (ruleAppCountIH + ruleAppCountEC + 1 > maxResultingRuleAppCountAllowed) {
	      // don't consider this and any subsequent incomplete hyps of
	      // this cell anymore; their rule app count is too high
	      continueECIteration = false;
	      skippingOccurred += totalECCount - iterationCountEC;
	    } else {

	      // our current node that we want to combine with pIncompleteHyp
	      CProductionNode* pNode = (*iter).second;
	      bool completeIsLexical = pNode->m_lexical;


	      // ...and check that we're still below the max source span
	      if (m_combinationCount <= CChartParserInfo::m_maxCombinationCount ||
		  (sourceCellColumn==1) ||
		  //		    (sourceCellColumn==1 && pIncompleteHyp->m_prefix[0] == CChartParserInfo::m_FullSentenceNT && !completeIsLexical) ||
		  (incompleteIsLexical&&completeIsLexical&&allowLexicalsThru)) {
		//def going to check out this pair
		pairsExplored++;


		vector<CWordIndex> newPrefix = pIncompleteHyp->m_prefix; // copy the current prefix from the selected incomplete hyp
		newPrefix.push_back(pNode->GetResult()); // append the production result of the candidate-matching completed node to the prefix

		//if (incompleteIsLexical && completeIsLexical && m_combinationCount > CChartParserInfo::m_maxCombinationCount){

		if (CChartParserInfo::m_verbosity>5){
		  cout<<"      Try FundRule-Combining: "; pIncompleteHyp->Display(); cout<<endl;
		  cout<<"      with: "; pNode->Display(); cout<<endl;
		}

		RuleQueryResult bTreeQueryResult = m_bTree->QueryRuleSet(
                    newPrefix); // look up newPrefix in the rule set

		if (bTreeQueryResult.first != 0) { // are there any matches at all?
		  // good, so it's worth expanding pIncompleteHyp to a prototypical incomplete hyp that already encompasses the
		  // backpointer to and the cost info of the hypergraph node pNode to be subsumed

		  if (s_debugApplyFund || CChartParserInfo::m_verbosity>5){
		    cout<<"        matched: ";
		    CChartParserInfo::PrintVocabSequence(newPrefix);
		    cout<<endl;
		  }

		  //clone the current incomplete hyp
		  CChartParserHypIncomplete* pNewIncompleteHypPrototype = new CChartParserHypIncomplete(*pIncompleteHyp);

		  // expand the incomplete hyp's tail with pNode, meaning: extending the span length m_i, extending
		  // prefix and backpointer info, updating cost variables and LM stuff
		  if (s_debugApplyFund){
		    cout<<"\t Creating new incomplete hyp PROTOTYPE by tail-expanding incompl. hyp:\t";
		    pNewIncompleteHypPrototype->Display(CChartParserInfo::m_pVocabulary);
		    cout<<"\t with hypergraph node:\t";
		    pNode->Display(CChartParserInfo::m_pVocabulary);
		  }
		  pNewIncompleteHypPrototype->ExpandTail(pNode); //adds another backpointer + increments the m_totalRuleApplicationCount for pNewIncompleteHypPrototype
		  assert(pNewIncompleteHypPrototype->I() == sourceCellRow+targetCellRow);
		  assert(pNewIncompleteHypPrototype->J() == sourceCellColumn);
		  // now go through the exact matches and the possible prefix match and file one complete hyp to the chart for each exact match
		  // further, in case of prefix match, file pNewIncompleteHypPrototype, otherwise, delete pNewIncompleteHypPrototype

		  if ((bTreeQueryResult.first & C_ExactMatch) != 0  // at least one exact match
		      && (ruleAppCountIH + ruleAppCountEC + 1 < bestResultingRuleAppCountEstimate)
		      ) {
		    // created a new hyp achieving a new rule-app-count record low
		    bestResultingRuleAppCountEstimate = ruleAppCountIH + ruleAppCountEC + 1;
		    // correspondingly, maxResultingRuleAppCountAllowed will also adapt
		    maxResultingRuleAppCountAllowed = CChartParserInfo::m_multiplicativeRA ?
		      (ceil(bestResultingRuleAppCountEstimate*m_maxRuleAppCountDifference)) :
		      (bestResultingRuleAppCountEstimate + m_maxRuleAppCountDifference);
		  }


		  // call ProcessBTreeResult
		  //ProcessBTreeResult is liable to call AddCompleteHyp, reset the stack level
		  CChartParserInfo::m_stackLevel = 0;

		  //pNewIncompleteHypPrototype = ProcessBTreeResult(bTreeQueryResult,pNewIncompleteHypPrototype, false);
		  if (!ProcessBTreeResult(hypRuleCombinations, bTreeQueryResult,pNewIncompleteHypPrototype)){
		    incompleteHypsToDelete.push_back(pNewIncompleteHypPrototype);
		  }else{
		    if (s_debugApplyFund || CChartParserInfo::m_verbosity>5){
		      cout << "    prefix-match, so calling AddIncompleteHyp." << endl;
		    }
		    //we can just file the pNewIncompleteHypPrototype with its current m_totalRuleApplicationCount right here
		    //because its back pointers are to nodes that arent going to have changing elements
		    AddIncompleteHyp(pNewIncompleteHypPrototype);
		  }
		  //if we found something in the btree
		}else{
		  //cout<<" no match "<<endl;
		}
		//as long as we abide by the MaxCombinationCount conditions
	      }else{
		//cout<<" prefix[0]!=@_S "<<endl;
	      }
	    }//rule app count difference has not been exceeded
	  }//go thru each incomplete hyp
	}//MCC count check
      }//rule app count check
    }//gone thru each equivalence class
  }

  // deleting all the incompleted hyps is done outside ApplyFundamentalRule
  if (CChartParserInfo::m_verbosity>0){
    cout<<"Log: ApplyFundamentalRule skipped: "<<skippingOccurred<<" explored: "<<pairsExplored<<" possible: "<<totalIncompleteHypCount*totalECCount<<endl;
  }
  return pair<int,int>(skippingOccurred,totalIncompleteHypCount*totalECCount);
} // end of ApplyFundamentalRule





void CChartParser::IndexNodesAccordingToRuleAppCount(int i, int j) {

  bool debug = false;
  cout<<"Entering IndexNodesAccordingToRuleAppCount i: "<<i<<" j: "<<j<<endl;

  // here we prune the m_productionNodes (i.e., coarse-grained pruning),
  // removing elements from the m_newlyCreatedNodes
  if (CChartParserInfo::m_maxCostDifferencePerCell<FLT_MAX/10){
    PruneCell(i, j);
  }

  // current chart cell in the complete hyp chart
  CCompleteHypChartCell& chartCell = m_completeHypsChart[i][j];
  int numEqClassesBeforeSelf = 0;
  numEqClassesBeforeSelf += FinalizeAllNodes(chartCell.m_productionNodes);

  vector<CChartParserHypIncomplete*> selfRuleIncompleteHyps;

  int waveCount = 0; // counts the waves of self-rule applications
  bool unknownHandling = false; // just for info purposes
  while (!m_newlyCreatedNodes.empty()) {
    // next wave of self-rule applications
    waveCount++;
    //copy the set of nodes newly created in last wave
    //vector<CProductionNode*> newlyCreatedNodesCopy = m_newlyCreatedNodes;
    CCoarseGrainedEq newlyCreatedNodesCopy = m_newlyCreatedNodes;
    //empty m_newlyCreatedNodes
    m_newlyCreatedNodes.clear();
    if (debug){
      cout<<"*** Beginning self rule round: "<<endl;
      for (CCoarseGrainedEq::iterator iter=newlyCreatedNodesCopy.begin(); iter!= newlyCreatedNodesCopy.end(); ++iter){
	(*iter).second->Display();
	cout<<endl;
      }
      cout<<" Starting now..."<<endl;
    }
    // now apply self-rule to all newly created nodes
    //for (vector<CProductionNode*>::const_iterator it = newlyCreatedNodesCopy.begin(); it != newlyCreatedNodesCopy.end(); ++it) {
    for (CCoarseGrainedEq::const_iterator it = newlyCreatedNodesCopy.begin(); it != newlyCreatedNodesCopy.end(); ++it) {

      //OLD style, now is a single finalization step
      //compute the sorted structure - we have to do this here because a) PruneCell must have been called
      //(*it)->RankFineGrainedClasses();
      //(*it)->FinalizeProductionNode();

      // according to the assumptions we make in section "Self Rule" at the comments in the
      // beginning of FastTranslateChart.cc, we never have a 'shortcut' case such as NN->NP->_S and NN->_S in the
      // rule file. There is, however, one problematic case: _S _G -> _S (which will fire an _S complete hyp
      // within Fundamental Rule execution), and with IndexNodesAccordingToRuleAppCount again as self-rule for _S:
      // _S _G -> _S (which will trigger an IH for _S based on _S's current hyp's rule-app-count), but also within
      // IndexNodesAccordingToRuleAppCount the self-rule for e.g. NN:
      // NN -> _S
      // which will add hyps to the _S EC possibly after _S's self rule has already been fired.
      // As a remedy, call Self Rule on _S at the very last.
      // Note: Strictly this is not necessary anymore now, since we re-calculate IH's rule-app-counts
      // based on the latest backpointer's ECs within ReCalculateRuleAppCount
      if ((*it).first == CChartParserInfo::m_FullSentenceNT && newlyCreatedNodesCopy.size() != 1) {
       	// make sure that m_FullSentenceNT is expanded at the very last
       	//it ensures that glue rule node gets correct m_bestTotalRuleApplicationCount
	//(this shouldn't be necessary anymore)
       	//don't treat it in this loop yet but later
       	//m_newlyCreatedNodes.push_back(*it);
	m_newlyCreatedNodes.insert(std::pair<CWordIndex, CProductionNode*>((*it).first, (*it).second));
      } else {
	//if ((*it)->GetResult() == CChartParserInfo::m_FullSentenceNT && newlyCreatedNodesCopy.size() == 1) {
	//cout<<"ALONE ---- "<<endl;
	//}
	CChartParserInfo::m_stackLevel++;
	//output settings
	if (CChartParserInfo::m_verbosity>0){
	  CChartParserInfo::TabStack(); cout<<"SelfFor: "; cout<<CChartParserInfo::m_pVocabulary->GetWord((*it).second->GetResult()); cout<<" ";
	}
	p_applySelfRuleRecursive((*it).second, selfRuleIncompleteHyps);
	CChartParserInfo::m_stackLevel--;
	//cout<<"Updating done......"<<endl;
      }
    }

    if (i==1) {
      // we're on the lexical level
      if (waveCount == 1) {
	// first self-rule wave is over: check if the lexical hyp
        // triggered any self-rules filing completed hyps in the cell
	if (m_newlyCreatedNodes.empty() &&
	    // exception: BOS or EOS b/c for these no unknown rules are used
	    // but rather we rely on at least one complete-sentence
            // rule making use of BOS and EOS
	    GetCompleteCell(i,j).m_cellType == 0) {

	  // add _UNKNOWN lexical hyp edge as the actual word was
	  // unmatched by any rules all we now do is add a second
	  // terminal hyp, this time with result m_unknownWI ("_UNKNOWN")
	  // this hyp can then be expanded by rules in the rule file
          // such as: _UNKNOWN#_UNKNOWN#@NP#1 1 1
	  cout << "Inserting _UNKNOWN at i,j=" << i << "," << j << endl;
	  CChartParserHypComplete* pUNKHyp =
              new CChartParserHypComplete(CChartParserInfo::m_unknownWI, i, j);
	  bool wasAdded = HypFiller::AddCompletedHyp(
              pUNKHyp,
              true,
              &m_completeHypsChart[i][j],
              &m_newlyCreatedNodes);

	  assert(wasAdded==true);
	  // 	  if (!wasAdded){
	  // 	    cout<<"ERROR: pUNKHyp was rejected by the online beam criteria";
	  // 	    exit(-1);
	  // 	  }
	  unknownHandling = true;
	  assert(!m_newlyCreatedNodes.empty()); // should contain the new node
	}
      } else if (waveCount == 2 && unknownHandling && m_newlyCreatedNodes.empty()) {
	// even the _UNKNOWN hyp didn't trigger any new completed hyps to be filed - seems there
	// are no unknown-rules in the DB
	cerr << "Warning: Unknown-hyp for word " << j <<
	  " didn't trigger any new hyps. Seems you don't have any _UNKNOWN rules in the rule database."
	     << endl;
	cout << "Warning: Unknown-hyp for word " << j <<
	  " didn't trigger any new hyps. Seems you don't have any _UNKNOWN rules in the rule database."
	     << endl;
      }
    }


    //re-prune the m_productionNodes (i.e., coarse-grained pruning), removing elements from the m_newlyCreatedNodes
    if (CChartParserInfo::m_maxCostDifferencePerCell<FLT_MAX/10){
      PruneCell(i, j);
    }

    //re-finalize at the end of the round - with the addAll param set to true
    //no - dont set this addAll to true, we want all hyps to be compared on equal ground
    //FinalizeAllNodes(chartCell.m_productionNodes, true);
    FinalizeAllNodes(chartCell.m_productionNodes, false);
  }


  int numEqClasses = 0;
  int numHyps = 0;
  // index the nodes according to rule app count
  if (CChartParserInfo::m_verbosity>0){
    cout<<"\nLog: IndexCell: ";
  }
  for (CCoarseGrainedEq::const_iterator iter = chartCell.m_productionNodes.begin();
       iter!= chartCell.m_productionNodes.end();
       ++iter
       ){
    CProductionNode* node = (*iter).second;
    int initialSize = node->m_fineGrainedNodes.size();
    if (CChartParserInfo::m_verbosity>0){
      cout<<CChartParserInfo::m_pVocabulary->GetWord(node->GetResult())<<" ("<<initialSize<<" eq "<<node->NumHyps()<<" hyps ) ";
    }
    numEqClasses +=initialSize;
    numHyps += node->NumHyps();
    // add node to the rule-app-count sorted index
    chartCell.ruleAppCountSortedNodeIndex.insert(pair<float,CProductionNode*>(node->m_bestTotalRuleApplicationCount, node));
  }
  if (CChartParserInfo::m_verbosity>0){
    cout<<endl;
  }

  //index all of the self rule incomplete hyps after asking them to recalculate their backpointers
  for (vector<CChartParserHypIncomplete*>::iterator iter = selfRuleIncompleteHyps.begin(); iter!= selfRuleIncompleteHyps.end(); ++iter){
    CChartParserHypIncomplete* pIncompleteHyp = (*iter);
    pIncompleteHyp->ReCalculateRuleAppCount();
    AddIncompleteHyp(pIncompleteHyp);
  }

  cout<<"Log: Production Nodes: "<<chartCell.m_productionNodes.size()<<" Eq Classes: before self "<<numEqClassesBeforeSelf<<" after: "<<numEqClasses<<endl;
  //numHyps is the actual number of hyps left in this cell after FinalizeAllNodes has been called
  //m_hypsDeleted counts those that have been added due to cube pruning, but then deleted cause they didnt fit in the histogram or threshold beam
  //m_hypsAdded is the total number of hyps added due to the cube pruning expansion step. As you increase the fuzz, this value should increase too
  //m_totalCalls is the number of calls to request additional elements from the cube pruning process
  cout<<"Log: numHyps: "<<numHyps<<" m_hypsDeleted: "<<CChartParserInfo::m_hypsDeleted<<" m_hypsAdded: "<<CChartParserInfo::m_hypsAdded<<" m_numTopCalls: "<<CChartParserInfo::m_numTopCalls<<" m_allPossibleExpansions: "<<CChartParserInfo::m_allPossibleExpansions<<" numAddedPostFailure: "<<CChartParserInfo::m_numAddedPostFailure<<" m_numTuplesProposed: "<<CChartParserInfo::m_numTuplesProposed<<" m_hypsDeletedHistogram: "<<CChartParserInfo::m_hypsDeletedHistogram<<" m_hypsDeletedThreshold: "<<CChartParserInfo::m_hypsDeletedThreshold<<endl;

  cout <<"Log(end-of IndexNodesAccordingToRuleAppCount): " << chartCell.m_productionNodes.size() << " nodes, " << m_incompleteHypsChart[i][j].size() << " incomplete hyps in cell "<< i << "," << j << ". " << endl << endl;

  if (!CChartParserInfo::m_GlueAndEOSAssumptions && CChartParserInfo::m_lowestRuleAppCount && j==1) {
    // test whether the estimated best-possible RAC achievable by gluing ended up being achieved by a complete _S hyp filed
    if (chartCell.m_bestGluedRuleAppCount < chartCell.m_productionNodes[CChartParserInfo::m_FullSentenceNT]->m_bestTotalRuleApplicationCount) {
      cout << "Minor Warning: For _S, best-possible RAC achievable by gluing m_bestGluedRuleAppCount="
	   << chartCell.m_bestGluedRuleAppCount << " < "
	   <<  chartCell.m_productionNodes[CChartParserInfo::m_FullSentenceNT]->m_bestTotalRuleApplicationCount
	   << "=m_bestTotalRuleApplicationCount (best RAC actually achieved). You might want to use less severe pruning."
	   << endl;
    }
  }

  //in very verbose mode
  if (debug || CChartParserInfo::m_verbosity>1){
    cout<<"====== Cell "<<i<<","<<j<<" ==== "<<endl;
    DisplayCell(i, j);
    cout<<"================================="<<endl;
  }

}




/*!
  After a chart cell has been filled with all lexical hyps or the hyps
  resulting from the Fundamental rule, p_applySelfRuleRecursive is called
  on each coarse-grained EC in the cell, i.e., on each result nonterminal,
  in order to look up rules whose source side starts with that NT.
  In case of exact matches, new complete hyps might be fired in the same cell.
  In case new result NTs arise this way, p_applySelfRuleRecursive
  has to be called for these new nonterminals again.
  p_applySelfRuleRecursive is called by IndexNodesAccordingToRuleAppCount
  The self rule queries the B-tree for rules that begin with that production
  result pNode. In case of a strict prefix match, it adds these rules as
  incomplete hyps into the IH chart. In case of an exact match, it files
  them as complete hyps to the CH chart.
*/
void CChartParser::p_applySelfRuleRecursive(
    CProductionNode* pNode,
    vector<CChartParserHypIncomplete*>& scheduleForAdding){

  bool s_debugApplySelf = false;
  if (s_debugApplySelf){
    //only display information if successful
    cout<<"APPLYING SELF RULE***********"<<endl;
    cout<<"Recursing for: "<<CChartParserInfo::m_pVocabulary->GetWord(pNode->GetResult())<<endl;
    cout<<"Recursing node: ";
    pNode->Display(CChartParserInfo::m_pVocabulary);
    cout<<endl;
  }

  // create empty vector of source symbols ...
  vector<CWordIndex> newPrefix;
  // ... and append the production result of the node
  newPrefix.push_back(pNode->GetResult());
  if (s_debugApplySelf){
    cout<<"querying: ";
    CChartParserInfo::PrintVocabSequence(newPrefix); cout<<endl;
  }
  // look up newPrefix in the rule set
  RuleQueryResult bTreeQueryResult = m_bTree->QueryRuleSet(newPrefix);
  if (s_debugApplySelf){
    cout<<"returned bTreeQueryResult"<<endl;
  }

  // are there any matches at all?
  if (bTreeQueryResult.first != 0) {
    // good, so it's worth expanding pIncompleteHyp to a
    // prototypical incomplete hyp that already encompasses the
    // backpointer to and the cost info of the hypergraphf node
    // pNode to be subsumed

    if (s_debugApplySelf){
      cout<<"***APPLY_SELF_RULE(";
      pNode->Display();
      cout<<"Found rules exact- or prefix-matching with: "
          <<CChartParserInfo::m_pVocabulary->GetWord(pNode->GetResult())<<endl;
      cout<<"Creating new incomplete hyp prototype by tail-expanding a new empty incompl. hyp with hypergraph node:\t";
      pNode->Display(CChartParserInfo::m_pVocabulary);
      cout<<endl;
    }

    // Ereate a new incomplete hyp to be used as prototype for
    // all the B-tree matches
    CChartParserHypIncomplete* pNewIncompleteHypPrototype =
        new CChartParserHypIncomplete(pNode->J());

    // expand the incomplete hyp's (still empty) tail with pNode,
    // meaning: extending the span length m_i, extending
    // prefix and backpointer info
    pNewIncompleteHypPrototype->ExpandTail(pNode);
    // now go through the exact matches and the possible prefix match
    // and file one complete hyp to the chart for each exact match
    // further, in case of prefix match, file
    // pNewIncompleteHypPrototype, otherwise, delete pNewIncompleteHypPrototype
    // get back the value of the pNewIncompleteHypPrototype to see
    // if it has been deleted

    RulesByProductionResult hypRuleCombinations;
    bool incompleteUsed = ProcessBTreeResult(hypRuleCombinations,
                                             bTreeQueryResult,
                                             pNewIncompleteHypPrototype);
    if (hypRuleCombinations.size()>0){
      if (CChartParserInfo::m_verbosity>0){
        cout<<"Self-Hyps found: ..."<<endl;
      }

      //az: changed to false to also prune in SelfRule mode
      bool addAll = false; //make sure everything is added in addAll==true mode
      HypFiller::SpawnKBestCompleteHyps(hypRuleCombinations,
                                        addAll,
                                        &m_completeHypsChart[pNode->I()]
                                        [pNode->J()],
                                        &m_newlyCreatedNodes);
    }else{
      if (CChartParserInfo::m_verbosity>0){cout<<"NoCompleteFound "<<endl;}
    }
    // delete the pNewIncompleteHypPrototype if it turns out that we arent
    // filing it anywhere
    if (!incompleteUsed){
      delete pNewIncompleteHypPrototype;
      pNewIncompleteHypPrototype = NULL;
    }else{
      // need to call AddIncompleteHyp on pNewIncompleteHypPrototype
      // but it should be with the finalized backpointer nodes
      if (CChartParserInfo::m_verbosity>0){
        cout << "Scheduling for adding: ";
        pNewIncompleteHypPrototype->Display(); cout << endl;}
      scheduleForAdding.push_back(pNewIncompleteHypPrototype);
    }
  }else{
    if (CChartParserInfo::m_verbosity>0){cout<<"NothingFound: "<<endl;}
    if (s_debugApplySelf){
      cout<<"no matches at all"<<endl;
    }
  }
}


// called by ApplyFundamentalRule and p_ApplySelfRule to create and
// file new complete hyps from pNewIncompleteHypPrototype
// and the matching transducer rule for each exact match returned
// and possibly file pNewIncompleteHypPrototype in the IH chart if
// exists(strict prefix match)
// if there was no strict prefix match, pNewIncompleteHypPrototype is
// deleted at the end of this method
// arv: return the ptr to the prototypical incomplete hyp, so that
// ProcessBTreeResult can indicate whether it delete the incomplete hyp or not
//if it returns false you should delete the pNewIncompleteHypPrototype
bool CChartParser::ProcessBTreeResult(
    RulesByProductionResult& hypRuleCombinations,
    RuleQueryResult& bTreeQueryResult,
    CChartParserHypIncomplete* pNewIncompleteHypPrototype) {
  bool s_debugApplyFund = false;

  // go through the exact matches and the possible prefix match
  if ((bTreeQueryResult.first & C_ExactMatch) != 0) {
    // at least one exact match
    // get the production result from the rule - and register the
    // incomplete hyps (for the backpointers) and the generating rule
    //go thru each rule that was matched
    for ( vector<CRule>::iterator iter = bTreeQueryResult.second.begin(); iter!=bTreeQueryResult.second.end(); ++iter){
      CRule& oneItem = (*iter); // the rule found
      CWordIndex productionResult = oneItem.m_productionResult;

      pair<CChartParserHypIncomplete*, CRule> payLoad(
          pNewIncompleteHypPrototype, oneItem);
      hypRuleCombinations.insert(
          pair<CWordIndex, pair<CChartParserHypIncomplete*, CRule> >(
              productionResult, payLoad));

    }

  }

  // Checking the second least significant bits...
  // we ONLY delete the incomplete hyp
  //if there are NO strict prefix matches. If there ARE strict prefix matches,
  //we add the incomplete hyp to the IC chart to account for this
  if (
      (bTreeQueryResult.first & C_StrictPrefixMatch) != 0 &&
      // optimization: don't create incomplete hyps if we've reached the complete-sentence span
      !(pNewIncompleteHypPrototype->I()==CChartParserInfo::m_sourceLength)
      ) {
    // strict prefix match
    // we merely need to file pNewIncompleteHypPrototype in the IH chart
    if (s_debugApplyFund){
      cout<<"Found at least one strict prefix-match. ";
      cout	  <<"Therefore *scheduling* filing the new incomplete hyp prototype in the IH chart."<<endl;
    }
    return true;
  } else {
    //return false, to make it clear that you should delete the pNewIncompleteHypPrototype but ONLY after you have done any spawning that hypRuleCombinations has listed
    return false;
  }
}

//! ScalingFactors is the return, not the init value; the last entry
//of trackScalingFactorScores contains the initial ScalingFactors used
void CChartParser::OptimizeParametersInternal( vector<double>& ScalingFactors,
                                               vector<NBestList*>& nBestLists,
                                               vector< vector<NBestList*> >& suppNBestLists,
                                               COptimize* pOptimizer,
                                               vector< vector<double> >& rangeAndInit,
                                               ScoreScalingFactorMap& trackScalingFactorScores){


  //harvest the feats data

  vector<LogProbList*> allFeats;
  vector<RawCounts*> allScores;
  vector<int> allCands;

  //from the base n-best list
  for (vector<NBestList*>::iterator iter = nBestLists.begin(); iter!= nBestLists.end(); ++iter){
    NBestList* oneSent = (*iter);
    oneSent->LoadOptimizationData(allFeats, allScores, allCands );
  }

  int fullSentCands = allCands.size();
  cout<<"Log: OptimizeParametersInternal initial num sent: "<<fullSentCands;
  //from the supplemental n-best lists where each sentence has a vector of n-best lists
  for (int i=0; i<suppNBestLists.size(); i++){
    for (int j=0; j<suppNBestLists[i].size(); j++){
      suppNBestLists[i][j]->LoadOptimizationData(allFeats, allScores, allCands );
    }
  }
  cout<<" after supp: "<<allCands.size()<<endl;

  pair<double, vector<double> > pr = (*(trackScalingFactorScores.rbegin()));
  if (rangeAndInit.size()==2){
    rangeAndInit.push_back(pr.second);
  }

  //we are actively working with parameters
  if (rangeAndInit.size()==3){
    assert(rangeAndInit.size()>0);
    assert(pr.second.size()==rangeAndInit[0].size());
    rangeAndInit[2]= pr.second;
  }

  assert(rangeAndInit.size()==3);

  cout<<"Optimization beginning [rangeAndInit="<<rangeAndInit.size()<<" allFeats="<<allFeats.size()<<" allCands="<<allCands.size()<<"]"<<endl;
  cout<<"Left: "; copy(rangeAndInit[0].begin(), rangeAndInit[0].end(), std::ostream_iterator<double>(cout, "\t")); cout<<endl;
  cout<<"Right: "; copy(rangeAndInit[1].begin(), rangeAndInit[1].end(), std::ostream_iterator<double>(cout, "\t")); cout<<endl;
  cout<<"Init: "; copy(rangeAndInit[2].begin(), rangeAndInit[2].end(), std::ostream_iterator<double>(cout, "\t")); cout<<endl;



  cout<<"Running opti"<<endl;
  vector<double> returnLambdas = pOptimizer->minModelCostMaxMetricScore(
      allFeats, allScores, allCands, rangeAndInit[2], fullSentCands);
  cout<<"Got Params: "; copy(returnLambdas.begin(), returnLambdas.end(), std::ostream_iterator<double>(cout, "_"));   cout<<endl;
  ScalingFactors = returnLambdas;
}

RawCounts CChartParser::BuildNBest(redi::pstream& pbuf,
                     NBestList* oneNBestList,
                     MetricSearch* pMetric,
                     vector<CChartParserHypComplete*>& nBestVector,
                     int translationCount,
                     int sentenceNumber,
                     bool normalizeTranslations,
                     int& TotalNewTranslations,
                     double artificialRefLength) {

  int numFeatures;
  RawCounts topTranslationRaw;
  //for (CSortedHypContainer::iterator iter = pBeam->m_store.begin(); translationCount-- != 0 && iter!= pBeam->m_store.end(); ++iter){
  //CChartParserHypComplete* pHyp = (*iter).second;
  for (vector<CChartParserHypComplete*>::iterator iter = nBestVector.begin(); translationCount-- != 0 && iter!= nBestVector.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    numFeatures = pHyp->m_negLogProbs.size();

    stringstream ss(ios::in | ios::out);
    CChartParserInfo::PrintVocabSequence(pHyp->m_translationOutput, ss);
    samtstring translation = ss.str();
    CVocabulary::string_trim(translation);

    if (normalizeTranslations){
      stringstream ss(stringstream::in | stringstream::out);
      //cout<<"Printing to ss"<<endl;
      //cout<<"translation before norm: "<<translation<<endl;
      pbuf<<translation<<"\n"; //az: used to be: <<"\n\r"
      pbuf.flush();
      samtstring str;
      std::getline(pbuf.out(), str);
      translation = str;

      //string returnedTranslation = NormalizeStringPBUF(translation);
      //cout<<"after norm: "<<returnedTranslation<<endl;
      pHyp->m_translationOutput = CChartParserInfo::makeSequence< CWordSequence >(translation);
      //cout<<"after norm list: "<<endl;
      //PrintVocabSequence(pHyp->m_translationOutput);
      //cout<<endl;
      //translation = returnedTranslation;
    }

    RawCounts raw;
    pMetric->ScoreSingleHyp(sentenceNumber, translation, raw);

    //make sure we are using the BLEU / NIST scores
    if (artificialRefLength!=-1){
      raw.referenceLengths = artificialRefLength;
    }

    if (iter==nBestVector.begin()){
      topTranslationRaw = raw;
    }
    //file it

    TotalNewTranslations+= oneNBestList->InsertCandidate(pHyp->m_negLogProbs, CChartParserInfo::makeSequence<CWordSequenceVector>(translation), raw);

  }
  return topTranslationRaw;

}


void CChartParser::LoadText(const char* fileName,
                                      vector< CWordSequence >& corpus,
                                      CVocabulary* pVoc,
                                      int AddBOSEOSInternally){

  ifstream fp(fileName);
  if (!fp.good()){
    cerr<<"Unable to open:"<<fileName<<endl;
  }

  char buf[MAXBUF];
  buf[0] = '\0';
  for (; !fp.eof(); ){
    fp.getline(buf, MAXBUF);
    samtstring sourceString(buf);
    // remove leading or trailing whitespace
    CVocabulary::string_trim(sourceString);
    if (sourceString.size()>0){
      if (AddBOSEOSInternally) {
	sourceString = "<s> " + sourceString + " </s>";
      }
      corpus.push_back(CChartParserInfo::makeSequence<CWordSequence>(sourceString));
    } else {
      fp.getline(buf, MAXBUF);
      if (!fp.eof()) {
	cerr << "Error loading test corpus: empty (or whitespace-only) line in the corpus!" << endl;
	exit(1);
      }
    }
  }
}

void CChartParser::DeleteNBestLists(
    vector<CChartParserHypComplete*>& fullSentenceNBestList,
    vector<pair< vector<CChartParserHypComplete*>, double> >& subSentenceNBestListAndAvg){

  for (vector<CChartParserHypComplete*>::iterator iter = fullSentenceNBestList.begin(); iter!= fullSentenceNBestList.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    delete pHyp;
    pHyp= NULL;
  }
  fullSentenceNBestList.clear();

  for (vector< pair<vector<CChartParserHypComplete*>, double> >::iterator subIter = subSentenceNBestListAndAvg.begin(); subIter!= subSentenceNBestListAndAvg.end(); ++subIter){
    //pair<CHypBeam*, double> beamAndAvg = (*subIter);
    //if the supplemenary n best list is the same size as the element you need to add, then create a new n-best list
    //on the next pass around you dont need to do this
    pair< vector<CChartParserHypComplete*>, double>& beamAndAvg = (*subIter);
    for (vector<CChartParserHypComplete*>::iterator iter = beamAndAvg.first.begin(); iter!=  beamAndAvg.first.end(); ++iter){
      CChartParserHypComplete* pHyp = (*iter);
      delete pHyp;
      pHyp= NULL;
    }
    beamAndAvg.first.clear();
  }
  subSentenceNBestListAndAvg.clear();


}
