/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#ifndef LatticeChart_hh
#define LatticeChart_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <cfloat>
#include <climits>
#include <map>
#include <vector>
#include <iostream>
#include <sstream>
#include <set>
#include <iterator>

// Interface to the rules
#include "RuleSet.hh"
//includes all the hyp classes
#include "ParserInfo.hh"
#include "ChartParserHyp.hh"
#include "ChartParserHypBeam.hh"
#include "CombinationBeam.hh"

// Interface to the scoring class
#include "MetricRaw.hh"
#include "MetricSearch.hh"
#include "OptimizeMER.hh"
#include "NBestList.hh"
//!Keep track (sorted) of the score achieved for each parameter setting across iterations of MER
typedef map<double, vector<double> > ScoreScalingFactorMap;
//! Define an epsilon difference to see when certain processes can terminate
#define EPSILON 0.001

//! Class that drives the search procedure based on smart CKY parsing
/*
  Contains several methods that are used while combining and processing hyps, like LM calculation
*/
class CChartParser
{
public:

  //!Sets static variables for CChartParser
  CChartParser(CVocabulary *pVocabulary,
               CWordIndex unknownWI,
               bool noTranslationOutputOfUnknownWords,
               float maxRuleAppCountDifference,
               RuleSet* ruleSet,
               int maxHypCount,
               int MaxCombinationCount,
               double aFutureCostImpact,
               int aBeamSizeFactor,
               bool extractUnique);

  //! accessor to cells of the complete hyp chart
  CCompleteHypChartCell& GetCompleteCell(int i, int j) {
    return m_completeHypsChart[i][j];
  }

  //! accessor to cells of the incomplete hyp chart
  CSortedIncompleteHypContainer& GetIncompleteCell(int i, int j) {
    return m_incompleteHypsChart[i][j];
  }

  //scaling factors for ranking in decoding
  void SetScalingFactors(vector<double> scalingFactors){
    if (scalingFactors.size()!= CRule::C_NumModelsTotal){
      cerr<<"SetScalingFactors: scalingFactors.size(): "<<scalingFactors.size()<<" BUT C_NumModelsTotal: "<<CRule::C_NumModelsTotal<<" they must match"<<endl;
      exit(2);
    }
    m_scalingFactors = scalingFactors;
    CChartParserInfo::m_scalingFactors = scalingFactors; // az: also set for the static variable in CChartParserHyp
  }

  //display a rule with backpointers as given by the pIncompleteHyp
  void DisplayInstantiatedRule(CRule* pRule, CChartParserHypIncomplete* pIncompleteHyp);


  //!go thru the complete and incomplete hyps and give a full row wise dump of the chart at this point
  void Display();

  //!display the completed cells, in equivalence class batches
  void DisplayCompletedHypsInCell(int i, int j);

  //! Display the incomplete hyps of the cell i,j
  void DisplayIncompletedHypsInCell(int i, int j);

  //! display the complete and incomplete hyps of a single cell in the chart
  void DisplayCell(int i, int j);

  static void SetVocabulary(  CVocabulary *pVocabulary ){
    CChartParserInfo::m_pVocabulary = pVocabulary;
  }


  static void SetMaxHypCount(int max){
    CChartParserInfo::MAX_HYP_COUNT = max;
  }


  // initialize the search process with the following steps
  // apply all the source words as completed edges in the m_completeHypsChart
  // we don't need any special handling for fully lexicalized rules since terminals have completed
  //void InitSearch(CSentence* pSentence);
  void InitSearch(const CWordSequence& pSentence);

  //! Reserve space for the chart
  void SetupChartCells();

  // true if we want to keep hyps separated by first and last target words
  void SetCompareTargetWords(bool b){
    CChartParserInfo::m_compareTargetWords = b;
  }

  // true if we want to keep hyps separated by first and last target words
  void SetUseLMInCoarse(bool b){
    CChartParserInfo::m_lmInCoarse = b;
  }

  // true if we want to keep hyps separated by first and last target words
  bool GetCompareTargetWords() const{
    return CChartParserInfo::m_compareTargetWords;
  }

  // let the parser know what it takes to be considered a completed sentence
  void SetFullSentenceNonTerminal(CWordIndex w){
    CChartParserInfo::m_FullSentenceNT = w;
  }
  // not needed anymore
  void SetGlueNonTerminal(CWordIndex w){
    CChartParserInfo::m_GlueNT = w;
  }

  //!call ClearSearch() after each sentence to clear out the memory used for the previous sentence
  void ClearSearch();
  //! clear cell in the complete hyps chart
  void ClearCHCCell(int i, int j);
  //! clear each cell in the incomplete hyps chart
  void ClearIHCCell(int i, int j);

  //! prune away coarse grained classes - only prunes within newlyAddedProductionNodes b/c other nodes could already by backpointed to by SelfRule-induced hyps
  void  PruneCell(const int i, const int j);

  //!Call finalize method on all nodes in productionNodes
  int FinalizeAllNodes(const CCoarseGrainedEq& productionNodes,  bool addAll=false);

  //! Called after all lexical or fundamental rules have been applied to create new nodes in cell i,j;
  /*!
    Should be renamed b/c it does much more than indexing
    This function is the finalizer for a cell, we call this function when we know that no more hyps
    will ever be added into this cell.
    adds special lexical _UNKNOWN hyp for unknown words
    the strategy for unk words is as follows. when the word is initialized we
    will know if it has any possible translations by looking at the complete edges at the cell (1,w)
    if there are no translations, use the word info initialized here to create
    an edge that will participate with
    other edges in normal search either by glue rule or by special _UNKNOWN rules in the rule file.
    this unknown-edge is a complete lexical (i.e., initial) hyp with  production result unknownWI
    implicit assumption here: if no word-rule (i.e., rule whose source side is the
    source word itself) for a source word can be found,
    then also no phrase-rule (i.e., longer rule containing the source word amongst
    other word and/or NTs) will be found
    this assumption is satisfied by our rule extraction method
  */
  void IndexNodesAccordingToRuleAppCount(int i, int j);

  //returns the first-best parse by setting the chart by applying the CYK+ algorithm using the
  void ParseChart();

  // Optimize the parameters of the translation process using nBestLists over the development
  // corpus.
  static void OptimizeParametersInternal(
      vector<double>& ScalingFactors,
      vector<NBestList*>& nBestLists,
      vector< vector<NBestList*> >& suppNBestLists,
      COptimize* pOptimizer,
      vector< vector<double> >& rangeAndInit,
      ScoreScalingFactorMap& trackScalingFactorScores);

  // Build an nbest list endowed with metric evaluations suitable
  // for optimization
  static RawCounts BuildNBest(redi::pstream& pbuf,
                              NBestList* oneNBestList,
                              MetricSearch* pMetric,
                              vector<CChartParserHypComplete*>& nBestVector,
                              int translationCount,
                              int sentenceNumber,
                              bool normalizeTranslations,
                              int& TotalNewTranslations,
                              double artificialRefLength=-1);

  //delete the n-best list as output by the parser
  static void DeleteNBestLists(vector<CChartParserHypComplete*>& fullSentenceNBestList,
                        vector<pair< vector<CChartParserHypComplete*>,
                        double> >& subSentenceNBestListAndAvg);

  // Load a corpus from a text file into the format used during translation
  static void LoadText(const char* fileName,
                       vector< CWordSequence >& corpus,
                       CVocabulary* pVoc,
                       int AddBOSEOSInternally);



  // simply file the incomplete hyp into the IH chart
  void AddIncompleteHyp(CChartParserHypIncomplete* pIncompleteHyp);


  /*!
    Returns true iff skipping of hypergraph nodes or of incomplete
    hyps due to too high rule application counts occurred
  */
  std::pair<int,int> ApplyFundamentalRule(int sourceCellRow,
                                          int sourceCellTarget,
                                          int targetCellRow,
                                          int targetCellColumn,
                                          float bestResultingRuleRuleAppCountEstimate,
                                          RulesByProductionResult& hypRuleCombination,
                                          vector<CChartParserHypIncomplete*>& incompleteHypsToDelete);

  //!Check where the production results generated in the hypRuleCombinations are valid when considered for self rule application on node pNode
  /*
   * Production results are invalid if they are generating hyps to be filed into already finalized CProductionNode (s)
   */
  //bool SelfRuleCompliance(CProductionNode* pNode, const RulesByProductionResult& hypRuleCombinations);

  //!expand the equivalence class node and adds more complete and incomplete hyps to chart
  /*!
    Calls the btree lookup on the production result of this node
    Calls ProcessBTreeResult to handle incomplete hyps (file or delete)
    Adds new pNewIncompleteHypPrototype to the scheduleForAdding list so that they can be filed with AddIncompleteHyp after ReCalculateRuleAppCount is called on them
    at the end of IndexNodesAccordingToRuleAppCount
  */
  void p_applySelfRuleRecursive(CProductionNode* pHyp, vector<CChartParserHypIncomplete*>& scheduleForAdding);

  //initialize the beam with the RulesByProductionResult elements before calling FillNBestReturnPtr on the beam
  void InitializeComboBeam(CCombinationTupleBeam* pBeam, RulesByProductionResult::iterator begin, RulesByProductionResult::iterator end, int& allPossibleExpansions, bool addAll);

  // Fills the hypRuleCombinations with incomplete hyps that have been
  // completed (along with completing rule), sorted by their production
  // result.
  bool ProcessBTreeResult(
      RulesByProductionResult& hypRuleCombinations,
      RuleQueryResult& bTreeQueryResult,
      CChartParserHypIncomplete* pNewIncompleteHypPrototype);

  //!Top level n-best list functions
  /*!
   * Will perform n-best extraction either using HC05 or Heuristic Search
   */
  vector<CChartParserHypComplete*> RetrieveKBestParses(int K);
  vector<CChartParserHypComplete*> RetrieveKBestParsesCell(int K, int i, int j);
  vector<CChartParserHypComplete*> RetrieveKBestParsesSpan(int K, CProductionNode* pTopNode);

  //! Uses the A* search method
  void InitializeBeam(CHypBeam& beam, CSortedHypContainer* sortedHyps);
  CHypBeam* RetrieveByAgenda(int K, CChartParserNode* pTopFineNode); //do it by expanding all items on the agenda
  CHypBeam* RetrieveKBestParses(int K, CChartParserNode* pTopFineNode);
  CChartParserHypCompleteNBest* m_pFullSentenceHyp;

  //! Call Huang,Chiang(2005) K-best on a beam from the A* method
  vector<CChartParserHypComplete*> HC05onBeam(int K, CHypBeam& beam);

  //! Rescore the n-best list with the LM
  void RescoreLM(vector<CChartParserHypComplete*>& nBestList);

  CProductionNode* CompleteTranslationNode(){
    if (GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes.count(CChartParserInfo::m_FullSentenceNT)==0) {
      return NULL;
    } else {
      return GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes[CChartParserInfo::m_FullSentenceNT];
    }
  }

  //! Output the unwound hypotheses referenced by the iterators
  /*!
   * Standardizes the way in which n-best output is written
  */
  //void OutputBeam(CHypBeam* pBeam, int translationCount = -1);
  void OutputNBest(vector<CChartParserHypComplete*>::iterator first, vector<CChartParserHypComplete*>::iterator last, int translationCount);


  //ARV 09/24/06 this function used to make appropriate annotations in the pNewHyp to handle UNKs and tagged words that the pBackHyp generated
  // without modifying the m_translationOutput of the pNewHyp so that the LM operated over these tags
  void UnwindHyp_HandleUnknownAndTags(CWordIndex word, CChartParserHypComplete* pNewHyp, CChartParserHypComplete* pBackHyp);

  //!sub used by RetrieveKBestParses
  /*! unwinds pHyp

    pHyp has to be on the beam and will be removed from the beam before unwinding unless pHyp turns out to be completely unwound

    auto-delete hyps worse than k best when inserting into CHypContainer
    if you do identify the first sentence spanning hyp eq class, then return it
  */
  CChartParserHypCompleteNBest* UnwindHyp(CChartParserHypCompleteNBest* pHyp, CHypBeam& beam, float lowestCostHyp);

  //! total n.o. combination trials between nodes and incomplete hyps performed;
  /*! once this number exceeds MaxCombinationCount, only _G rules are allowed to combine
   */
  int m_combinationCount;




  /*! maximum allowed difference in ApplyFundamentalRule between m_totalRuleApplicationCount (in case of
    incomplete hyps) respectively m_bestTotalRuleApplicationCount (in
    case of hypergraph nodes) and the respective lowest values of the
    incomplete hyps resp. hypergraph nodes considered. Nodes
    resp. incompl. hyps with higher differences are ignored.
  */
  float m_maxRuleAppCountDifference;


  vector< vector<CSortedIncompleteHypContainer> > m_incompleteHypsChart; //!< Incomplete hyps corresponding to the type-2 list in the CYK+ paper

  vector< vector<CCompleteHypChartCell> > m_completeHypsChart;  //!< Complete hyps in course equivalence class buckets

 private:

  //az06/11/01: now stored in GetCompleteCell(sourcelength,1), see also accessor CompleteTranslationNode()
  //  CProductionNode* m_pCompleteTranslationsNode;   //!< Node containing the complete translations

  vector<CChartParserHypCompleteNBest*> m_poppedHyps; //! Hyps that have been popped in n-best retrieval (clear after sentence)

  //! the nodes that have been newly created within the current hyp-adding wave for the current CH chart cell
  //vector<CProductionNode*> m_newlyCreatedNodes;
  //its a hash map now so that we can easily kill elements from within it
  CCoarseGrainedEq m_newlyCreatedNodes;

  // Access to the translation model which is applied during parsing
  // This variable is set during initialization of the CChartParser
  RuleSet* m_bTree;

  //scaling factors for the underlying models
  vector<double> m_scalingFactors;

  //arv: in the future we should change the timing operation, not the place of the deletion...
  CHypBeam* m_pBeamToDelete; //beam that needs to be deleted after the sentences is handled after retrievekbest

};


#endif
