/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

/* TECHNICAL OVERVIEW OF PARSING
   read this carefully before coding or asking questions in the forum*/
/*!

This code base implements the parser introduced in "
!!add info that backpointers are only stored for non-lexical hyps, i.e., for nonterminal symbols

Consider the rule set

ich#I#PN
schlief#slept#V
schlief#slept#VP
sehr#very#ADV
gut#good#ADJ
gut#well#ADV
gut#nicely#ADV
@1 @2#@1 @2#PN VP _S
@1 @2#@1 @2#ADV ADV ADVP
@1 @2 @3 !#@1 @2 @3 !#PN V ADVP _S

and assume we want to parse

ich schlief sehr gut !

Then we build up as the complete hypothesis (CH) and incompl. hyp (IH) tables in CKY fashion (see also notes below):

CH:j 1                          2                        3                        4                   5
i

1    undef/undef/ich (1)        undef/undef/schlief (3)  undef/undef/sehr (6)     und./und./gut (8)   und./und./! (11)
1/I/PN (2)                 3/slept/V (4)            6/very/ADV (7)           8/good/ADJ (9)
3/slept/VP (5)                                    8/well/ADV (10)
8/nicely/ADV (10)

2    2 5/@1 @2/_S (12)                                   7 10/@1 @2/ADVP (13)

3

4

5    2 4 13 11/@1 @2 @3 !/_S


IH:j 1                          2                        3                        4                   5
i

1    2/PN                                                7/ADV                   10/ADV

2    2 4/PN V

3

4    2 4 13/PN V ADVP

5

Notes:
CHs are quadruples of the form a/b/c (d), where:
a: back pointer sequence, pointing back to the chart equivalence classes matching the source-side of the rule triggering the CH
b: target side of the rule triggering the CH
c: production result (from the rule triggering the CH)
d: chart equivalence class ID. Chart equivalence classes of hyps are merely partioned by i,j and the production result, NOT by language model features! Such sub-partioning can take place internally within on chart equivalence class in order to restrict pruning. Note for example that 8/well/ADV and 8/nicely/ADV in (1,4) both belong to chart equivalence class 10.

DHs are pairs a/b, where:
a: back pointer sequence, pointing back to the chart equivalence classes matching a strict prefix of the source-side of some rule in the rule set. Strict prefix means: not the complete source side is matched.
b: prefix that matched


Let's see how the tables above were created: First, CH(1,1) was filled up with undef/undef/ich by the source-word rule, creating our first chart equivalence class (1). Then application of the self-rule to this newly created eq. class matched the grammar rule ich#I#PN, triggering the CH 1/I/PN (2). Recursive application of the self-rule to EC 2 matches another grammar rule, but only prefix-wise, as we get to know from the B-tree query. We don't pursue this match further just yet, but merely create the incomplete hypothesis 1/PN in IH(1,1).

Similarly, the entries (i,j)=(1,2),(1,3),(1,4), and (1,5) of CH and IH are filled up, and we arrive at entry (i,j)=(2,1). For k=1, the fundamental rule finds a candidate for a match pair in 2/PN [IH(1,1)] and EC 5 [CH(1,2)]. Indeed, querying the B-tree for PN VP yields a match with @1 @2#@1 @2#PN VP _S, triggering complete hyp. 2 5/@1 @2/_S (12). Further, for candidate pair 2/PN and EC 4, querying for PN V returns a prefix match, thus we store 2 4/PN V in IH(2,1).

The next cell (2,2) yields no new hypotheses, but (i,j)=(2,3), k=1, allows for application of the fundamental rule to 7/ADV [IH(1,3)] and EC 10 [CH(1,4)]; querying the B-tree for ADV ADV yields a match, triggering compl. hyp. 7 10/@1 @2/ADVP (13).

The next new hypothesis will be created in IH(4,1), resulting as application of the fundamental rule to 2 4/PN V [IH(2,1)] and EC 13 [CH(2,3)]. Finally, cell CH(5,1) is filled with a complete hyp. spanning the whole sentence, resulting from application of the fundamental rule to 2 4 13/PN V ADVP and EC 11.






Construction of complete and incomplete hyps

There only two ways to construct complete hyps:

1)  CChartParserHypComplete(CWordIndex wi, int i, int j);
//initializes a complete chart item with a lexical entry, most importantly, it does
//not set the transducer item feature, only the m_productionResult is set
//This constructor is used by InitSearch and by HandleUntranslatedWords

2) CChartParserHypComplete( CChartParserHypIncomplete* pIncompleteHyp, CTransducerRule* pItem) :CChartParserHyp((CChartParserHyp*) pIncompleteHyp);
//take an incomplete hyp (the tail, containing the backpointers and accumulated costs) and a transducer rule (the head, containing
//the production result and the costs of applying the rule) and form a complete hyp out of these two parts
//this constructor is used in the Fundamental and in the Self rule

Once complete hyps are constructed, they are to remain constant! No change of costs or translated words or span positions is allowed anymore. After all, they are COMPLETED. That means a new completed hyp that spans (sits on top of) N other completed hyp equivalence classes is always created the following way: First the left-most two spanned completed hyp ECs are assembled together to a new incomplete hyp in the IC chart. This continues with the other spanned complete hyp ECs, until the left-most N-1 spanned EC are assembled together to an incompleted hyp sitting in the IC chart. Then, applying the fundamental rule the N-1st time, the incomplete hyp is expanded one last time to also span the Nth completed hyp EC. These expansion processes are the tail-expansions realized by the Fundamental Rule by first cloning the incomplete hyp with constructor

CChartParserHypIncomplete(CChartParserHypIncomplete* pHyp):CChartParserHyp((CChartParserHyp*) pHyp);

and then doing the tail-expansion

void CChartParserHypIncomplete::ExpandTail(CChartParserNode* pExpansionClass);

Once the tail is created, the incomplete hyp can be completed by assigning its transducer rule, i.e., using the CChartParserHypComplete construction method 2).

The Self Rule works very similar: First an empty incomplete hyp is created via the constructor

// creates an empty incomplete hyp at source word position j
CChartParserHypIncomplete(int j);

Then the transducer rule is applied to this empty incomplete hyp again via CChartParserHypIncomplete construction method 2).



Pruning
There are several possibilities of pruning in our Chart parser. First, note that
reduction of the number of hypergraph nodes is more desirable than reduction
in the total number of hypotheses: During the parsing process, the number of
hypotheses in a hypergraph node is merely a space issue; it has no effect
on computation time, since each hypergraph nodes represents all its contained
hypotheses simultaneously whenever the self rule or fundamental rule is applied.
Also during K-best retrieval, where individual hypothesis are retrieved from
the hypergraph nodes, this A*-type unwinding-only-when-needed is hardly slowed down when the
number of hypotheses in a hypergraph node is high.

1) Assuming you want the K best hypotheses, limit the maximum number of hypotheses
m_maxHypCount in the CChartParserNode objects to something less than K. That is a
very crude pruning mechanism that also doesn't avoid the blowup in the number of
hypergraph nodes due to the partitioning according to first two and last two words.

2) Keep the hypergraph nodes fine-grained, but prune across nodes that have the same
production symbol based on m_bestTotalCost+m_scaledFutureCostEstimate.
Smarter than 1), but also doesn't avoid the blowup in the number of
hypergraph nodes due to the partitioning according to first two and last two words.
To implement this, the CChartParserNodeContainer would have to be replaced by
a CChartParserCoarseECContainer that is a multimap storing the nodes that have
the same production result, sorted by m_bestTotalCost+m_scaledFutureCostEstimate

3) Make the hypergraph nodes coarse-grained
Do do that, set m_compareTargetWords to false. Now equivalence class
partitioning is only done according to production result.

-- 3a) Back-offs to unigram and bigram are used to determine the LM costs of
first and second target word, respectively.
Problem: The back-off estimates might be too inexact such that we might keep hyps that combine badly with other hyps at the cost of
losing hyps that would have combined well with other hyps.

-- 3b) Like 3a), but when calculating the total LM costs of the
hypotheses/hyperarcs e for a vertice/node during the parsing process, we use for each underlying tail node in T(e) the first-words and last-words of that tail node's best hypothesis/hyperarc. This involves assigning each vertice a 'best' first-two-words and a 'best' last-two-words vector. Example: When filing the hyp

S -> yesterday the @NN @VP .

we look up the underlying vertice @NN's assigned best first-two-words (say it's only one word: 'car') and best last-two-words (in this case, that must also be 'car') as well as the @VP vertice's best first-two-words (say, 'has unfortunately') and last-two-words ('right tire'). Thus we store for this hyp the best-first-words 'yesterday the' and the best-last-words 'tire .' and calculate the best LM prob as

p(car|yesterday the) * "underlying @NN's best hyp's LM prob (must be 1 in this case)" * p(has|the car) * p(unfortunately|car has) * "underlying @VP's best hyp's LM prob" * p('.'|right tire)

-- 3c) Make the hypergraph nodes coarse-grained and set LM cost to zero during parsing. Only after K-best retrieval, the true LM costs
are determined for each translation and added to the total costs. The assumption here is that the syntactic rules are good enough to
select the K best translations, such that the LM is only used as a rescoring feature.


Self rule

Self rules can lead to new nodes to be placed into the current chart cell, whose production results again might have to be looked up  in the rule base. The easiest-seeming way of handling this is to apply the self-rule recursively whenever a new node is filed in the chart cell. This, however, may cause false best-cost values of the nodes:

Consider rules

1: ich#I#PN
2: schlief#napped#V
3: schlief#slept#V
4: V # @1 # VP
5: PN V # @1 @2 # _S
6: VP # @1 # _S
7: schlief # slept # VP

and the source sentence "ich schlief".

Assume for simplicity that we don't separate EC's according to first words / last words and assume that rule 2 has higher total costs than rule 3. After filing a lexical hyp for ich in CH(1,1) and by self-rule filing a hyp for rule 1 in CH(1,1), we get to CH(1,2), file a lexical hyp for schlief and via self-rule a hyp for rule 2. This triggers the self-rule to file a new hyp for rule 4 in CH(1,2) with backpointer to the V-node in CH(1,2). When calculating the costs for this hyp, the current best-cost info for the V-node is used. However, later on, rule 3 will be added to the V-node, giving it a lower best-cost. This best-code is not propagated anymore to the new S-node in CH(1,2)! A similar problem exists with the incomplete hyp for rule 5, which is also filed too early in IH(2,1).

In order to apply the self-rule properly, the triggering node needs to have all hyps in it already. But since the self-rule can lead to the filing of new nodes, as it happened above, after each 'wave' of self-rule applications, we have to start a new wave by applying self-rules to the nodes newly created during the last wave.

The assumption here is that one wave does not alter the best-cost values of nodes from previous waves. This assumption is violated when a given sequence of rule-applications leading to a certain node results in worse costs than a longer sequence leading to the same node. Assume for example that rule 7 is of higher cost than the combined costs of rules 3 and 4. Note that the hyp for rule 7 is filed in the same 'wave' as the hyp for rule 3, namely after the lexical hyp for 'schlief' was filed. In the next wave, rule 6 is filed to create a new _S node backpointing to rule 7's VP node and accumulating rule 7's costs, which are the current best-costs for the VP node. In the same wave, but perhaps after rule 7 was applied, rule 4 is applied to rule 3's V-node, creating a new VP-hyp that changes the best-cost of the VP node. Since rule 6 has been applied already, however, the _S node will have higher costs than it should have.

*/

#define NO_TRACE 1
#include <iostream>
#include <fstream>
//#include <Trace.hh>

//! Allows you to make a call to the shell
#include "pstream.h"

//! Allows you to make a call to the shell
#include "Parameter.h"

//! Generic Interface for scoring MT output
#include "MetricSearch.hh"

//! Bleu and NIST scoring of MT output
#include "BleuSearch.hh"

//! TER score
#include "TERSearch.hh"


//!Manage an NBest list (feature data + translation output)
#include "NBestList.hh"

//!MER optimizer with expanded features
#include "OptimizeMER.hh"

//!Interface to access rules
#include "RuleSet.hh"
//!A BTree implementation to access rules
#include "BTree.hh"

//! The main parsing engine
#include "FastLatticeChartParser.hh"

#include "LanguageModelSAMT.hh"
//language model that uses a database as a backend
#include "DatabaseSAMTLM.hh"
//!language model thats uses the sri toolkit as a backend
#include "CachedSriLM.hh"

// flags that prodlm wants declared
DECLARE_double(prodlm_stupid_factor);
DECLARE_bool(prodlm_service_load_locally);
DECLARE_bool(sstable_service_load_locally);
DECLARE_bool(logtostderr);

//!classes to take in command line options
//#include <param.hh>
//#include <paramdef.hh>
#include <time.h>
#include <sys/times.h>

//!for file locking
#include <fcntl.h> // for open()
#include <cerrno> // for errno
#include <cstdio> // for perror()
//! for file-exists checking
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

//! Define an epsilon difference to see when certain processes can terminate
#define EPSILON 0.001

// ====================
// prototypes
// ====================

//!Keep track (sorted) of the score achieved for each parameter setting across iterations of MER training
typedef map<double, vector<double> > ScoreScalingFactorMap;

//!MER optimization setting - numPermutes seems to have minimal effect on score
int OptiNumPermutes=1;
//! MER optimization setting - numStarts seems to have a more significant effect
int OptiNumStarts=1;
//! MER optimization setting - just experimental see details in the code or ask us on the mailing list
int OptiNumModify=4;
//! MER optimization setting - Normalize the scaling factors after each optimization of the scaling factors
int OptiDoNorm=1;
//! MER optimization setting - after optimizing, blank out each feature (set it to 0) and re-optimize, see if you get improvements
int OptiTwoStage=1;
//! MER optimization setting - determine convergence of optimization process
double OptiEpsilon=0.0;
//! MER optimization setting - scale up convergence threshold for two stage optimization
double OptiEpsilonScale=1.0;


//! extract n-best from a non sentence completing cell
//vector< pair<CHypBeam*, double> > GenerateSubsentenceNBest(CChartParser& parser, MetricSearch* pMetric, int sentenceNumber, int sourceLength, const int& NBest, int NBestSubSpans);
vector< pair< vector<CChartParserHypComplete*>, double> > GenerateSubsentenceNBest(CChartParser& parser, MetricSearch* pMetric, int sentenceNumber, int sourceLength, const int& NBest, int NBestSubSpans);

//delete the n-best list as output by the parser
void DeleteNBestLists(vector<CChartParserHypComplete*>& fullSentenceNBestList,  vector<pair< vector<CChartParserHypComplete*>, double> >& subSentenceNBestListAndAvg);

//! Augments the NBestList data structure with count data provided by the scoring function (MetricSearch)
//RawCounts BuildNBest(redi::pstream& pbuf, NBestList* oneNBestList,  MetricSearch* pMetric, CHypBeam* pBeam, int translationCount, int sentenceNumber, bool normalizeTranslations, int& TotalNewTranslations, double artificialRefLength=-1);
RawCounts BuildNBest(redi::pstream& pbuf, NBestList* oneNBestList,  MetricSearch* pMetric, vector<CChartParserHypComplete*>& nBestVector, int translationCount, int sentenceNumber, bool normalizeTranslations, int& TotalNewTranslations, double artificialRefLength=-1);

void BuildNBestSubsentence(redi::pstream& pbuf, vector< vector<NBestList*> > suppNBestLists, MetricSearch* p_bSearch, vector<pair< vector<CChartParserHypComplete*>, double> > subSentenceNBestListAndAvg, int subSentenceWorkingNBest, int sentenceNumber, bool useNormalizing, int& AdditionalTranslations );

//!Runs MER optimization and sets the ScalingFactors
void OptimizeParametersInternal( vector<double>& ScalingFactors, vector<NBestList*>& nBestLists, vector< vector<NBestList*> >& suppNBestLists, MetricSearch* pMetric,  vector< vector<double> >& rangeAndInit, ScoreScalingFactorMap& trackScalingFactorScores, samtstring timeAdded);

//!Loads a corpus of text into a vector of CWordSequence (s)
void LoadText(const char* fileName, vector< CWordSequence >& corpus, CVocabulary* pVoc, int AddBOSEOSInternally);

//!Parses a parameter line that is _ separated into a vector of doubles
void ParseDoubleSequence(const char* floatSequenceStr, vector<double>& doubleVector);
void ParseDoubleSequence(samtstring floatSequenceStr, vector<double>& doubleVector);

//! check the return code of open() as follows:
int getlock(const char* filename) // returns the lock's descriptor on success
{
  int fd;
  fd=open(filename, O_WRONLY | O_CREAT | O_EXCL);
  if (fd<0 && errno==EEXIST)
    {
      // the file already exist; another process is
      // holding the lock
      // cout<<"the file is currently locked; try again later";
      return -1;
    }
  else if (fd < 0)
    {
      // perror() appends a verbal description of the current
      // errno value after the user-supplied string
      perror("Error: Locking failed for the following reason: ");
      cerr << "Exit due to locking problem." << endl;
      exit(2);
      // return -1;
    }
  // if we got here, we own the lock
  return fd;
}


//! checks if the file exists and returns the last-modified time
time_t FileExists(const char* strFilename) {
  struct stat stFileInfo;
  int intStat;
  // Attempt to get the file attributes
  intStat = stat(strFilename,&stFileInfo);
  if(intStat == 0) {
    // We were able to get the file attributes
    // so the file obviously exists.
    time_t tmpt = time(NULL);
    cout << "File " << strFilename << " exists and was last modified at " << ctime(&stFileInfo.st_mtime);
    cout << " (current time: "
	 << ctime(&tmpt) << ", difftime: " << difftime(tmpt,stFileInfo.st_mtime) << ")" << endl;
    return stFileInfo.st_mtime;
  } else {
    // We were not able to get the file attributes.
    // This may mean that we don't have permission to
    // access the folder which contains this file. If you
    // need to do that level of checking, lookup the
    // return values of stat which will give you
    // more details on why stat failed.
    cout << "File " << strFilename << " doesn't exist" << endl;
    return 0;
  }
}

int main(int argc, char *argv[])
{

  //====================================================================================
  // Parameters that are passed in from the command line, see the lines where
  // they are added below Param.AddParamDef(...) for details as to what they do
  //====================================================================================

  //int testme = 1;
  //getGlobalParSet().insert(new Parameter<int>("testme",ParameterChangedFlag,"output path",testme,1,0));
  //parseArguments(argc, argv);
  //cout<<"testme: "<<testme<<endl;
  //exit(-1);

  double FutureCostImpact = 1.0;
  double NBest=1000; //its a double cause it COULD be a multiplicative factor
  float MaxRuleAppCountDifference = 2;

  //! If non-NULL, we're in Worker-bee mode, and write per-sentence n-best list in the directory specified
  samtstring WorkerBeeDir = "";

  int Verbosity=0;
  int DisplayNBestData=40;
  int MaxSentenceProcessingTime = 480;

  samtstring WordLM = "";
  samtstring SAMTLM_MinHistoryFrequency = "";
  samtstring SAMTLM_MinBackoffWeight = "";
  samtstring SAMTLM_HistoryFrequencyConvergenceSpeed = "";
  samtstring SAMTLM_BreakEvenHistoryFrequency = "";
  samtstring SAMTLM_UniqueEventCountImpact = "";
  int SAMTLM_UseBackwardLM = 0;

  //pass in the rule  file here
  samtstring TransMemParams = "";
  samtstring SentenceList = "";
  samtstring Vocabulary = "";

  int AddBOSEOSInternally = 1;
  samtstring FullSentNonTerminal = "@_S";
  samtstring UnknownWord = "_UNKNOWN";
  samtstring TagSetList = ""; //!< Tags in LM mode
  //  samtstring GlueNonTerminal;

  samtstring FeatureWeightsParsing = "";
  samtstring FeatureWeightsRescoring = "";
  samtstring MinFeatureWeightsParsing = "";
  samtstring MaxFeatureWeightsParsing = "";
  samtstring MinFeatureWeightsRescoring = "";
  samtstring MaxFeatureWeightsRescoring = "";
  samtstring MaxCostDifferencePerCell = "inf"; //!< Pruning across NTs
  samtstring AStarBeam = "inf";
  samtstring PruningMap = "0-100-5-@_S-200-5"; //!< Pruning per non-terminal

  //!< Use boundary word LM costs for pruning costs
  int PruningMapUseBoundary=1;
  samtstring ComboPruningBeamSize = "10000";
  samtstring ComboPruningMinBeamSize = "30";
  samtstring ComboPruningFuzzCostDifference = "5";
  int ComboPruningGlobal = 1;
  int ComboPruningUseLM = 1; //!< Use the LM during combo pruning
  //  int ComboPruningUseLexical; //!< Use the Lexical hyps in the same beam
  int CompareTargetWords = 1;
  int UseLMInCoarse = 1;

  //default values for optional parameters, scoring MER params
  int MaxCombinationCount = 10;
  int GenOracleScores=1;
  int MEROptimize=0;
  //currently not in use:  int UniqueBeam=1;
  int RemoveUnt=0;

  int NBestMult=0; //treat NBest as base to calculate working NBest per sentence (1.8 seems to be a nice value)
  int NBestMin=100; //min value of the working nbest if the exp generates very few translations for the short sentences

  int BeamSizeFactor=10;
  int NBestSubSpans=0; //!< use sub spans in mer training

  int ExtractUnique = 1;

  //!The max. number of hyps in a fine grained equivalence class
  int HypsPerNode=1000;

  int AlwaysRescore = 0;
  int NoRescoring = 1;
  int GlueAndEOSAssumptions = 1;
  int IterationLimit=1;

  samtstring ScoringMetric = "IBMBLEU";
  samtstring NormalizingScript= "";
  samtstring GainFile= "";
  samtstring ReferenceList= "";
  int NumReferences=1;

  //==================================
  // PARAMS
  //==================================

  getGlobalParSet().insert(new Parameter<int>("DisplayNBestData",ParameterChangedFlag, "Write out translation model scores and path for DisplayNBestData UNIQUE elements in the n-best list",DisplayNBestData,0));

  getGlobalParSet().insert(new Parameter<int>("Verbosity",ParameterChangedFlag, "Verbosity (0) for general experiments (1) to see the flow of the program (2) to debug problem areas",Verbosity,0));

  getGlobalParSet().insert(new Parameter<int>("AddBOSEOSInternally",ParameterChangedFlag, "(0): Assume input sentences of the form <s> ... </s>; (1) Add <s> and </s> internally to the input sentences",AddBOSEOSInternally,0));

  getGlobalParSet().insert(new Parameter<int>("GenOracleScores",ParameterChangedFlag, "saves nbest lists across iterations of MER (more memory) and outputs an Oracle score after each iteration",GenOracleScores,0));

  getGlobalParSet().insert(new Parameter<int>("MEROptimize",ParameterChangedFlag, "runs iterations of Minimum Error rate training on the nbest lists generated in each iteration",MEROptimize,0));

  getGlobalParSet().insert(new Parameter<samtstring>("ScoringMetric",ParameterChangedFlag, "Scoring Metric to run MER or evaluate your translations against the reference",ScoringMetric,0));

  getGlobalParSet().insert(new Parameter<int>("IterationLimit",ParameterChangedFlag, "Limit to the number of iterations that MER will be run for",IterationLimit,0));

  getGlobalParSet().insert(new Parameter<int>("NumReferences",ParameterChangedFlag,"num references per source sentence ",NumReferences,0));


  getGlobalParSet().insert(new Parameter<double>("NBest",ParameterChangedFlag, "number of translations to put in the N-best list for parameter optimization",NBest,0));




  getGlobalParSet().insert(new Parameter<samtstring>("GainFile",ParameterChangedFlag, "information Gain file for NIST scoring  ",GainFile,0));



  getGlobalParSet().insert(new Parameter<samtstring>("ReferenceList",ParameterChangedFlag,"reference sentences for scoring ",ReferenceList,0));


  getGlobalParSet().insert(new Parameter<int>("NBest_Mult",ParameterChangedFlag,"Treat the NBest value as a the base of an nbest^srclength if NBestMult==1 : nbest*srclength if NBestMult=2",NBestMult,0));


  getGlobalParSet().insert(new Parameter<int>("NBest_Min",ParameterChangedFlag, "Base value of working NBest for short sentences",NBestMin,0));

  getGlobalParSet().insert(new Parameter<int>("HypsPerNode",ParameterChangedFlag, "number of complete hypotheses a hypergraph node can maximally hold (exception: the _S node will keep all its hyps)",HypsPerNode,0));

getGlobalParSet().insert(new Parameter<int>("BeamSizeFactor",ParameterChangedFlag, "during N-best retrieval the temporary size of the beam will be N*BeamSizeFactor - set this to a value higher than 1 if you are forcing unique extraction (ExtractUnique=1) or if you don't do LM intersection (CompareTargetWords=0)",BeamSizeFactor,0));

getGlobalParSet().insert(new Parameter<int>("ExtractUnique",ParameterChangedFlag,"Force the nbest list extraction to extract NBest unique translations - must set BeamSizeFactor>1 will slow down retrieval   ",ExtractUnique,0));

getGlobalParSet().insert(new Parameter<int>("ExtractRecombine",ParameterChangedFlag, "In the A* style search form equivalence classes on the fly and drop hyps that are re-combined",CChartParserInfo::m_extractRecombine,0));

getGlobalParSet().insert(new Parameter<int>("NBestStyle",ParameterChangedFlag, "Style of n-best list retrieval (1==A* style) (2==HC05)",CChartParserInfo::m_nbestStyle,0));

getGlobalParSet().insert(new Parameter<int>("NBestSubSpans",ParameterChangedFlag, "Use NBest extraction from the sub-spans 0 (dont do sub span addition) 1 (do it with just cell m_maxCombinationCount 1) 2( use the whole m_maxCombinationCount row)   ",NBestSubSpans,0));

getGlobalParSet().insert(new Parameter<samtstring>("AStarBeam",ParameterChangedFlag,"Per word beam for A* search step",AStarBeam,0));

getGlobalParSet().insert(new Parameter<samtstring>("FeatureWeightsParsing",ParameterChangedFlag, "Feature weights (when doing MER training these are the initial values) separated by commas",FeatureWeightsParsing,0));

getGlobalParSet().insert(new Parameter<samtstring>("FeatureWeightsRescoring",ParameterChangedFlag, "",FeatureWeightsRescoring,0));

getGlobalParSet().insert(new Parameter<samtstring>("MinFeatureWeightsParsing",ParameterChangedFlag, "Minimum-values allowed for each feature weight separated by commas. By default we assign -100 to all",MinFeatureWeightsParsing,0));

getGlobalParSet().insert(new Parameter<samtstring>("MaxFeatureWeightsParsing",ParameterChangedFlag, "Maximum-values allowed for each feature weight separated by commas. By default we assign 100 to all",MaxFeatureWeightsParsing,0));

getGlobalParSet().insert(new Parameter<samtstring>("MinFeatureWeightsRescoring",ParameterChangedFlag,"",MinFeatureWeightsRescoring,0));

getGlobalParSet().insert(new Parameter<samtstring>("MaxFeatureWeightsRescoring",ParameterChangedFlag,"",MaxFeatureWeightsRescoring,0));

getGlobalParSet().insert(new Parameter<int>("UseSpanFeatures",ParameterChangedFlag, "Whether to append feature vector by feature computing (up to a constant) the neg-log probability of the created complete hypothesis given only its source span length",CChartParserInfo::m_useSpanFeatures,0));

getGlobalParSet().insert(new Parameter<samtstring>("PruningMap",ParameterChangedFlag, "Pruning options for each non-terminal (i.e. coarse-grained EC) specifying the max. n.o. hyps per result NT in each cell and the maximally allowed cost difference from the lowest-cost hyp across all fine-grained ECs of that result NT (all worse-cost hyps will be pruned away). (@X-10-2.5) means keep at most top 10 hyps in coarse-grained EC with LHS NT @X that are all at most 2.5 plus the cost of the top hyp. Use 0 instead of @X to define behavior for all NTs not otherwise specified. You can also specify the value inf (for infinity) as a cost-difference.",PruningMap,0));

getGlobalParSet().insert(new Parameter<samtstring>("MaxCostDifferencePerCell",ParameterChangedFlag, "Max. allowed cost that a hyp can deviate from the best hyp in its cell. This enables pruning across coarse-grained ECs. Used in PruneCell",MaxCostDifferencePerCell,0));

getGlobalParSet().insert(new Parameter<int>("PruningMapUseBoundary",ParameterChangedFlag, "Use the boundary words to pruning equivalence classes",PruningMapUseBoundary,0));

getGlobalParSet().insert(new Parameter<samtstring>("ComboPruningBeamSize",ParameterChangedFlag, "When combing the incomplete hyp and a rule for a particular cell we can do some lossy pruning here to avoid calculating LM scores for all the completed hyps most of which will be discarded. This is the max. n.o. hyps considered per coarse-grained EC (you can also use inf denoting infinity here)",ComboPruningBeamSize,0));

getGlobalParSet().insert(new Parameter<samtstring>("ComboPruningMinBeamSize",ParameterChangedFlag, "When combing the incomplete hyp and a rule for a particular cell we can do some lossy pruning here to avoid calculating LM scores for all the completed hyps most of which will be discarded. This is the min. n.o. hyps considered per coarse-grained EC",ComboPruningMinBeamSize,0));

getGlobalParSet().insert(new Parameter<samtstring>("ComboPruningFuzzCostDifference",ParameterChangedFlag, "Allowed cost (neg-log prob) a hyp during ComboPruning is allowed to differ from last officially allowed cost (i.e from bestHypCost+PruningMapCostDifference) to still be included into the combo pruning beam. However such an element is only put on the beam if ComboPruningBeamSize has not been exceeder yet. Further elements are added to the beam *disregarding* ComboPruningFuzzCostDifference as long as ComboPruningMinBeamSize has not been exceeded yet. You can also use inf denoting infinity here",ComboPruningFuzzCostDifference,0));

getGlobalParSet().insert(new Parameter<int>("ComboPruningGlobal",ParameterChangedFlag, "Combo pruning can be done within a single incomplete rule or across all the ones that match on a cell (global) ",ComboPruningGlobal,0));

getGlobalParSet().insert(new Parameter<int>("ComboPruningUseLM",ParameterChangedFlag, "Use the LM during ComboPruning should trigger the k-best seearch to add a few additional neighbors down in case they end up filter upto the top ",ComboPruningUseLM,0));

getGlobalParSet().insert(new Parameter<int>("BreadthAStar",ParameterChangedFlag, "Do AStar by breadth ",CChartParserInfo::m_breadthAStar,0));

getGlobalParSet().insert(new Parameter<int>("DontReloadBeam",ParameterChangedFlag, "Dont reload the beam in astar ",CChartParserInfo::m_dontReloadBeam,0));

getGlobalParSet().insert(new Parameter<int>("SaveAllAlternatives",ParameterChangedFlag, "Keep all alternatives in the EqClass ",CChartParserInfo::m_saveAllAlternatives,0));

getGlobalParSet().insert(new Parameter<int>("UpdateBeamSize",ParameterChangedFlag, "Try to ensure that we have BeamSizeFactor active hyps not just all hyps ",CChartParserInfo::m_updateBeamSize,0));

getGlobalParSet().insert(new Parameter<int>("MaxCombinationCount",ParameterChangedFlag, "Limit on span of non glue rules - effectively a time out for the expensive parsing step",MaxCombinationCount,0));

getGlobalParSet().insert(new Parameter<int>("HistoryLength",ParameterChangedFlag, "History length assumed for the language models: this will determine the length of the first-words and last-words of a hyp's m_lmContext and thus the granularity of the equivalence classes. This value needn't necessarily correspond to the true Markov-order of the LMs.",CChartParserInfo::C_HistoryLength,0));

getGlobalParSet().insert(new Parameter<int>("CompareTargetWords",ParameterChangedFlag, "Each cell in the parsing contains several equivalence classes - hyp classes that cannot be compared and pruned away because hyps from that class my give best performance when combined later. Setting CompareTargetWords==1 use LM context length as a separator if you use this you MUST use the PruningMap ",CompareTargetWords,0));

getGlobalParSet().insert(new Parameter<int>("RecomputeLMCostsDuringNBest",ParameterChangedFlag, "1: During N-best retrieval each time an underlying hypothesis is slurped in recompute all LM costs from scratch as we have more history context available now. This slows down K-best retrieval and is only useful when the Markov order of one of the LMs used is higher than --HistoryLength. You can also set this option to 2 which will recompute the LM cost and sanity-check that the results are the same as compared to only recomputing the LM costs necessary (use this as a sanity check if your LMs' Markov orders are all less or equal --HistoryLength).",CChartParserInfo::m_recomputeLMCostsDuringKBest,0));

getGlobalParSet().insert(new Parameter<int>("RescoreLM",ParameterChangedFlag, "After the N best translations have been retrieved recompute their LM costs from scratch and re-sort the translations. The CPU time to do this is insignificant. You should only set this parameter to zero if the Markov order of your LMs is not higher than --HistoryLength. If --RecomputeLMCostsDuringKBest > 0 then --RescoreLM=1 won't have any effect. You can also set this option to 2 which will recompute the LM cost and sanity-check that the results are the same as they were after K-best retrieval (use this as a sanity check if your LMs' Markov orders are all less or equal --HistoryLength or you had --RecomputeLMCostsDuringKBest > 0).",CChartParserInfo::m_rescoreLM,0));

getGlobalParSet().insert(new Parameter<double>("FutureCostImpact",ParameterChangedFlag, "Undocumented and not important. Always use 1.0 for CompareTargetWords==1 possibly use lower values for CompareTargetWords=0",FutureCostImpact,0));

getGlobalParSet().insert(new Parameter<int>("NoRescoring",ParameterChangedFlag, "Don't optimize parsing features and rescoring features separately base everything on FeatureWeightsParsing and optimize only those if in MER-mode",NoRescoring,0));

getGlobalParSet().insert(new Parameter<float>("MaxRuleAppCountDifference",ParameterChangedFlag, "Largest deviation of the rule-app-count feature from the lowest-rule-app-count hyp that a hyp can have in order to still be filed in the chart. Crucial pruning parameter. It relies on our type of glue rules though so if you are using different types of glue rules or not using the rule-app-count feature then set this value to a very high number.",MaxRuleAppCountDifference,0));

getGlobalParSet().insert(new Parameter<int>("RA_UpdateHyp",ParameterChangedFlag, "Update the hyp's rule app count with the fine-grained eq class counts",CChartParserInfo::m_updateHypRuleAppCount,0));

getGlobalParSet().insert(new Parameter<int>("RA_LowestCount",ParameterChangedFlag,"Use the lowest rule app count instead of the count of the lowest cost hyp",CChartParserInfo::m_lowestRuleAppCount,0));

getGlobalParSet().insert(new Parameter<int>("RA_Mult",ParameterChangedFlag, "Use MaxRuleAppCountDifference as a multiplicative factor instead of a difference",CChartParserInfo::m_multiplicativeRA,0));

getGlobalParSet().insert(new Parameter<samtstring>("RuleDB",ParameterChangedFlag, "File name of the grammar rule database",TransMemParams,0));

getGlobalParSet().insert(new Parameter<samtstring>("SentenceList",ParameterChangedFlag,  "name of file that contains a list of sentences in plain text format",SentenceList,0));

getGlobalParSet().insert(new Parameter<samtstring>("Vocabulary",ParameterChangedFlag,"Optional. Name of file that contains the vocabulary will be augmented in memory based on vocab in rule database and test set. Becomes mandatory if an SAMT language model is used.",Vocabulary,0));

getGlobalParSet().insert(new Parameter<samtstring>("LMFiles",ParameterChangedFlag, "Filename(s) for the language models used. If more than one separate them by comma. If the file has the extension '.samtlm' we will assume it to be an SAMT language model; if the file extension is '.srilmbackward' we will assume it to be an SRI LM trained on reversed (last word first) sentences; if the extension is '.srilm' we assume it to be a regular SRI language model.",WordLM,0));

getGlobalParSet().insert(new Parameter<int>("SAMTLM_UseBackwardLM",ParameterChangedFlag, "If set to 1 then use backward (right-to-left) models for the SAMT LMs in addition to the standard forward (left-to-right) models.",SAMTLM_UseBackwardLM,0));

getGlobalParSet().insert(new Parameter<int>("SRIHistoryLength",ParameterChangedFlag, "Actual Markov order of the SRI language models used (e.g. 2 for trigram). Note that this is NOT used for determining the length of m_lmContext and thus the granularity of the equivalence classes (use parameter HistoryLength for that).",CChartParserInfo::C_SRI_HistoryLength,0));

getGlobalParSet().insert(new Parameter<samtstring>("SAMTLM_MinHistoryFrequency",ParameterChangedFlag, "Value(s) for MinHistoryFrequency parameter for the SAMT LMs. If more than one separate them by comma. In case of multiple values for either of the SAMTLM parameters we will fork off creating a different SAMT LM model for each combination of SAMT LM database specified SAMTLM_MinHistoryFrequency value SAMTLM_MinBackoffWeight value and SAMTLM_HistoryFrequencyConvergenceSpeed and SAMTLM_UniqueEventCountImpact value",SAMTLM_MinHistoryFrequency,0));

getGlobalParSet().insert(new Parameter<samtstring>("SAMTLM_MinBackoffWeight",ParameterChangedFlag, "Value(s) for MinBackoffWeight parameter for the SAMT LMs. If more than one separate them by comma. In case of multiple values for either of the SAMTLM parameters we will fork off creating a different SAMT LM model for each combination of SAMT LM database specified SAMTLM_MinHistoryFrequency value SAMTLM_MinBackoffWeight value and SAMTLM_HistoryFrequencyConvergenceSpeed value and SAMTLM_UniqueEventCountImpact value",SAMTLM_MinBackoffWeight,0));

getGlobalParSet().insert(new Parameter<samtstring>("SAMTLM_HistoryFrequencyConvergenceSpeed",ParameterChangedFlag, "Value(s) for SAMTLM_HistoryFrequencyConvergenceSpeed parameter for the SAMT LMs. If more than one separate them by comma. In case of multiple values for either of the SAMTLM parameters we will fork off creating a different SAMT LM model for each combination of SAMT LM database specified SAMTLM_MinHistoryFrequency value SAMTLM_MinBackoffWeight value and SAMTLM_HistoryFrequencyConvergenceSpeed value and SAMTLM_UniqueEventCountImpact value",SAMTLM_HistoryFrequencyConvergenceSpeed,0));

getGlobalParSet().insert(new Parameter<samtstring>("SAMTLM_BreakEvenHistoryFrequency",ParameterChangedFlag,"Value(s) for SAMTLM_BreakEvenHistoryFrequency parameter for the SAMT LMs. If more than one separate them by comma. In case of multiple values for either of the SAMTLM parameters we will fork off creating a different SAMT LM model for each combination of SAMT LM database specified SAMTLM_MinHistoryFrequency value SAMTLM_MinBackoffWeight value and SAMTLM_HistoryFrequencyConvergenceSpeed value and SAMTLM_UniqueEventCountImpact value",SAMTLM_BreakEvenHistoryFrequency,0));

getGlobalParSet().insert(new Parameter<samtstring>("SAMTLM_UniqueEventCountImpact",ParameterChangedFlag, "Value(s) for SAMTLM_UniqueEventCountImpact parameter for the SAMT LMs. If more than one separate them by comma. In case of multiple values for either of the SAMTLM parameters we will fork off creating a different SAMT LM model for each combination of SAMT LM database specified SAMTLM_MinHistoryFrequency value SAMTLM_MinBackoffWeight value and SAMTLM_HistoryFrequencyConvergenceSpeed value and SAMTLM_UniqueEventCountImpact value",SAMTLM_UniqueEventCountImpact,0));

getGlobalParSet().insert(new Parameter<int>("SAMTLM_UseBERFormat",ParameterChangedFlag, "If 1 then use Perl BER compressed integers for the Berkeley DB word indices and frequency counts (corresponding to perl pack / unpack 'w' type)",CChartParserInfo::SAMTLM_UseBERFormat,0));

getGlobalParSet().insert(new Parameter<int>("RemoveUnt",ParameterChangedFlag, "Remove untranslated words from translation output (used during MER too)",RemoveUnt,0));

getGlobalParSet().insert(new Parameter<samtstring>("TagSetList",ParameterChangedFlag,"Comma-separated list of tags: nonterminals that when occurring as LHS of an applied rule will be used directly (instead of the rule's translation output) when querying the LM model. Example: @CD @NUM @DATE",TagSetList,0));

getGlobalParSet().insert(new Parameter<samtstring>("FullSentNonTerminal",ParameterChangedFlag, "The nonterminal marker for a complete sentence signals also when EOS BOS can be used",FullSentNonTerminal,0));

getGlobalParSet().insert(new Parameter<samtstring>("UnknownWord",ParameterChangedFlag, "All occurrences of the UnknownWord in the source side of rules in the rule database serve as templates for any word that is unknown (i.e. is not equal to any rule's source side).",UnknownWord,0));

getGlobalParSet().insert(new Parameter<int>("GlueAndEOSAssumptions",ParameterChangedFlag, "If set to nonzero will assume our kind of glue rules and our kind of rules involving </s>. This will cause optimizations to make parsing faster and also will cause assertions checking that the glue rule is firing. If you set this to zero you might want to also set MaxRuleAppCountDifference to a very high number.",GlueAndEOSAssumptions,0));

getGlobalParSet().insert(new Parameter<samtstring>("NormalizingScript",ParameterChangedFlag,"Normalizing script used on references and translation output - do things like lowercasing/truecasing/detokenization; each input line is a translation each output line is the normalized translation; script has to be executable and flush input and output buffer line-by-line; see example punctuation-postprocess-eval4.pl for required structure of perl script for Normalizing",NormalizingScript,0));

getGlobalParSet().insert(new Parameter<int>("NormalizeReferences",ParameterChangedFlag, "Normalize the references with the NormalizingScript (this is dangerous)",CChartParserInfo::m_normalizeReferences,0));

getGlobalParSet().insert(new Parameter<int>("Opti_NumStarts",ParameterChangedFlag, "Number of starts of f the optimizer 1==just use the initial values ",OptiNumStarts,0));

getGlobalParSet().insert(new Parameter<int>("Opti_NumPermutes",ParameterChangedFlag, "Number of times you change a few dimension and try to reconverg 1==just use what the initial values ",OptiNumPermutes,0));

getGlobalParSet().insert(new Parameter<int>("Opti_NumModify",ParameterChangedFlag,"Number of features that you try zeroing out per MER iteration (this starts with the first Opti_NumModify features and continues with the next Opti_NumModify in the next iteration step and so on modulo the n.o. features)",OptiNumModify,0));

getGlobalParSet().insert(new Parameter<int>("Opti_DoNorm",ParameterChangedFlag,"Normalize scaling factors after each iteration ",OptiDoNorm,0));

getGlobalParSet().insert(new Parameter<int>("Opti_TwoStage",ParameterChangedFlag,"Run 2 stage optimization - nulling out certain fields and re-optimizing. Use this together with Opti_NumModify. Don't use this option AND Opti_NumStarts / Opti_NumPermutes > 1",OptiTwoStage,0));

getGlobalParSet().insert(new Parameter<double>("Opti_Epsilon",ParameterChangedFlag, "Optimize until the last dimension's error gain is within Opti_Epsilon of the previous value",OptiEpsilon,0));

getGlobalParSet().insert(new Parameter<double>("Opti_EpsilonScale",ParameterChangedFlag, "Scale up the threshold when using two stage optimization to save time",OptiEpsilonScale,0));

getGlobalParSet().insert(new Parameter<int>("GlobalLMCache",ParameterChangedFlag, "If non-zero we will never clear the LM cache. If zero we'll clear it after each sentence.",CChartParserInfo::m_globalLMCache,0));

getGlobalParSet().insert(new Parameter<samtstring>("WorkerBeeDir",ParameterChangedFlag, "If specified we use the worker-bee mode. Then this is the directory in which the Worker Bee files like locks and per-sentence n-best lists and MER params are stored.",WorkerBeeDir,0));

getGlobalParSet().insert(new Parameter<int>("MaxSentenceProcessingTime",ParameterChangedFlag,"Maximum time in seconds that processing one sentence takes. Relevant only for Worker-bee mode: The worker bee will assume that a process got killed if nothing happened on a sentence for more than MaxSentenceProcessingTime seconds.",MaxSentenceProcessingTime,0));



  //interpret the parameters
  parseArguments(argc, argv);

  /*
  if (! Param.Interpret( argv, argc) ) {
    cerr << "\nParameters could not be interpreted correctly\n ... exiting!\n";
    exit(-1);
  }
  */

  //===========================================================================
  // Check for glaring inconsistencies with the parameter settings
  //===========================================================================
  if ((ScoringMetric == "NIST") && (GainFile=="")){
    cerr << "\n--ScoringMetric=NIST requires a GainFile... exiting!\n";
    exit(-1);
  }

  if (SentenceList == ""){
    cout<<"SentenceList variable not set"<<endl;
    cerr<<"SentenceList variable not set"<<endl;
    exit(2);
  }else{
    cout<<"Loading sentence from: "<<SentenceList<<endl;
  }

  if (FeatureWeightsParsing==""){
    cout<<"ERROR: You must set a FeatureWeightsParsing vector to begin translation. It must have the |RuleWts|+|LmFiles| wts";
    exit(-1);
  }

  if (TransMemParams==""){
    cout<<"ERROR: You must set a RuleDB vector to perform translation";
    exit(-1);
  }

  if (CChartParserInfo::m_multiplicativeRA && (MaxRuleAppCountDifference<=1)){
    cout<<"ERROR: if you are using RA_Mult you need to set a MaxRuleAppCountDifference > 1.0";
    exit(-1);
  }

  if (((CChartParserInfo::m_recomputeLMCostsDuringKBest>0) && (CChartParserInfo::m_rescoreLM>0)) && (NBest<=1.0)){
    cout<<"ERROR: if you are using m_recomputeLMCostsDuringKBest or m_rescoreLM that means you need an NBest to operate on: You have set NBest = 1, set it >1";
    exit(-1);
  }

  if (CChartParserInfo::m_pruningMapUseBoundary && !CChartParserInfo::m_comboPruningUseLM){
    cout<<"ERROR: if you are want to use the boundary words in beam pruning, then you need to have m_comboPruningUseLM to make sure you have right LM estimates";
    exit(-1);
  }


  if (CompareTargetWords==1 && FutureCostImpact!=1.0){
    cout<<"ERROR: If you are using CompareTargetWords there is no notion of FutureCostImpact, all the LM costs are real"<<endl;
    exit(-1);
  }
  //cout<<"***** Parameter settings *****"<<endl;
  //Param.Display(cout);
  //cout<<endl;
  //===========================================================================
  // Operational settings
  //===========================================================================
  CChartParserInfo::m_verbosity = Verbosity;

  CChartParserInfo::m_GlueAndEOSAssumptions = GlueAndEOSAssumptions > 0;

  //==========================================================================
  // Load up data
  //==========================================================================
  CVocabulary *pVoc = new CVocabulary;

  if (Vocabulary!=""){
    //Trace( 0, "Loading vocabulary from " << VocFile );
    if ( !pVoc->Load( Vocabulary.c_str() ) ) {
      cerr << "main(): Could not load " << Vocabulary;
      //const char *source = Param.getSourceFromKey("Vocabulary");
      const char* source = Vocabulary.c_str();
      if (source != NULL) {
	// param file associated to param
	cerr << ", referred to by param file " << source;
      } // endif: param file associated to parameter
      cerr<< ". Exiting." << endl << flush;
      exit(1);
    } // endif: error loading voc file
  }else{
    //    cout<<"No vocab file provided, loading from phrase table and test set"<<endl;
    //    cerr << "main(): no Vocabulary file specified: loading from phrase table and test set " <<endl;
  }


  //=======================================
  // Load normalizing script
  //=======================================
  bool useNormalizing = false;
  redi::pstream pbuf;
  if (NormalizingScript!= ""){

    //tmp, try just initializing it here
    cout<<"Initializing normalizing process ======================== "<<endl;
    samtstring perlNormalizer = //"perl " +
     samtstring(NormalizingScript);
    pbuf.open(perlNormalizer, std::ios_base::out | std::ios_base::in);
    cout<<"done...\n";
    //InitializePBUF(NormalizingScript);
    //pbuf_global_inited = true;
    useNormalizing = true;
  }else{
    //    cout<<"This version of FastTranslateChart requires a NormalizingScript\n";
    //    exit(-1);
    useNormalizing = false;
  }
  CWordIndex unkWI = pVoc->AddWord(samtstring(UnknownWord));

  //now the rules are loaded, we can try parsing!
  //override the source corpus variable, see if we can just get a sentence in
  cout<<"SentenceList: "<<SentenceList<<endl;
  bool noTranslationOutputOfUnknownWords = RemoveUnt;

  //initialize the parser
  CRule::pVoc = pVoc;

  RuleSet* rules = new CBTree(TransMemParams);
  CChartParser parser(pVoc,
                      unkWI,
                      noTranslationOutputOfUnknownWords,
                      MaxRuleAppCountDifference,
                      rules,
                      HypsPerNode,
                      MaxCombinationCount,
                      FutureCostImpact,
                      BeamSizeFactor,
                      ExtractUnique);

  cout<<"MaxRuleAppCountDifference: "<<MaxRuleAppCountDifference<<endl;
  cout<<"MaxCombinationCount: "<<MaxCombinationCount<<endl;
  cout << "CompareTargetWords " << CompareTargetWords << endl;
  if (CompareTargetWords==1){
    parser.SetCompareTargetWords(true);
  }else{
    parser.SetCompareTargetWords(false);
  }
  cout << "UseLMInCoarse " << UseLMInCoarse << endl;
  if (UseLMInCoarse){
    parser.SetUseLMInCoarse(true);
  }else{
    parser.SetUseLMInCoarse(false);
  }

  cout << "HypsPerNode" << HypsPerNode << " NBest" << NBest  << " BeamSizeFactor" << BeamSizeFactor <<" ExtractUnique: "<<ExtractUnique<<endl;

  //scaling factors for the parser
  vector<double> ScalingFactors;
  cout << "Scaling factors (feature weights) for parsing:";
  ParseDoubleSequence(FeatureWeightsParsing, ScalingFactors);
  cout << endl;
  //set total Number of models
  CRule::C_NumModelsTotal= ScalingFactors.size();


  vector<double> tmpfloatvector;
  //for opti of parsing
  vector< vector<double> > featureRange;

  //only care about this if we are optimizing
  if (MEROptimize==1){
    if (MinFeatureWeightsParsing == "") {
      // initialize with default value -100
      tmpfloatvector.clear();
      for (int i=0;i<CRule::C_NumModelsTotal;i++) {
	tmpfloatvector.push_back(-100);
      }
    } else {
      cout << "Min Scaling factors (feature weights) for parsing:";
      ParseDoubleSequence(MinFeatureWeightsParsing, tmpfloatvector);
      cout << endl;
    }
    featureRange.push_back(tmpfloatvector);
    if (tmpfloatvector.size() != CRule::C_NumModelsTotal) {
      cerr << "N.o. models in MinFeatureWeightsParsing not consistent with FeatureWeightsParsing" << endl;
      exit(1);
    }
    if (MaxFeatureWeightsParsing == "") {
      // initialize with default value 100
      tmpfloatvector.clear();
      for (int i=0;i<CRule::C_NumModelsTotal;i++) {
	tmpfloatvector.push_back(100);
      }
    } else {
      cout << "Max Scaling factors (feature weights) for parsing:";
      ParseDoubleSequence(MaxFeatureWeightsParsing, tmpfloatvector);
      cout << endl;
    }
    featureRange.push_back(tmpfloatvector);
    if (tmpfloatvector.size() != CRule::C_NumModelsTotal) {
      cerr << "N.o. models in MaxFeatureWeightsParsing not consistent with FeatureWeightsParsing" << endl;
      exit(1);
    }
  }

  //if you are not optimizing dont set the NBestSubSpans
  if (MEROptimize!=1){
    if (NBestSubSpans!=0){
      cout<<"WARNING: MEROptimize!=1, but you still asked for NBestSubSpans, overriding this setting\n";
      NBestSubSpans=0;
    }
  }
  //scaling factors for rescoring
  vector<double> ScalingFactorsRescoring;
  //for opti of rescoring
  vector< vector<double> > featureRangeRescoring;

  if (NoRescoring) {
    cout << "No separate rescoring-feature-optimization.\n";
  } else {
    cout << "Scaling factors (feature weights) for rescoring:";
    //initialize all with 1
    //scf.insert(scf.end(),C_NumModelsTotal,1.0);
    ParseDoubleSequence(FeatureWeightsRescoring, ScalingFactorsRescoring);
    cout << endl;
    if (ScalingFactorsRescoring.size() != CRule::C_NumModelsTotal) {
      cerr << "N.o. models in FeatureWeightsRescoring not consistent with FeatureWeightsParsing" << endl;
      exit(1);
    }

    cout << "Min Scaling factors (feature weights) for rescoring:";
    ParseDoubleSequence(MinFeatureWeightsRescoring, tmpfloatvector);
    featureRangeRescoring.push_back(tmpfloatvector);
    cout << endl;
    if (tmpfloatvector.size() != CRule::C_NumModelsTotal) {
      cerr << "N.o. models in MinFeatureWeightsRescoring not consistent with FeatureWeightsParsing" << endl;
      exit(1);
    }
    cout << "Max Scaling factors (feature weights) for rescoring:";
    ParseDoubleSequence(MaxFeatureWeightsRescoring, tmpfloatvector);
    featureRangeRescoring.push_back(tmpfloatvector);
    cout << endl;
    if (tmpfloatvector.size() != CRule::C_NumModelsTotal) {
      cerr << "N.o. models in MaxFeatureWeightsRescoring not consistent with FeatureWeightsParsing" << endl;
      exit(1);
    }
  }

  cout<<"Number of features/models: "<<ScalingFactors.size()<<endl;
  cout << endl;

  //==================================================================
  // Set the LM(s)
  // you can run the parser without a LM if you REALLY want to,
  // it will just report 1.0 every time its asked for a LM cost
  // this is horrible for translation, but useful if you are
  // just playing around with settings and you dont want to wait for
  //the LM to load
  if (WordLM!= ""){
    // can be several LMs, separated by comma; so process all
    vector<samtstring> tmpscf;
    CRule::Tokenize(WordLM,tmpscf,",");
    // Go thru each file provided
    for (int i = 0; i<tmpscf.size(); i++) {

      if (tmpscf[i].substr(tmpscf[i].size()-7) == ".samtlm" || tmpscf[i].substr(tmpscf[i].size()-12) == ".samtlmbtree" ) {
	cout<<"Loading SAMT language model data base: "<<tmpscf[i]<<endl;
	// we have an SAMT language model database -- create both a
	// forward and possibly a backward LM for each
	// combination of the values in the SAMTLM_... parameter lists
	if (SAMTLM_HistoryFrequencyConvergenceSpeed == "" || SAMTLM_MinBackoffWeight == "" || SAMTLM_MinHistoryFrequency == "" || SAMTLM_BreakEvenHistoryFrequency == "" || SAMTLM_UniqueEventCountImpact == "") {
	  cerr << "Error: When using SAMT LMs, you have to specify values for the parameters --SAMTLM_MinHistoryFrequency, --SAMTLM_MinBackoffWeight, --SAMTLM_HistoryFrequencyConvergenceSpeed, and --SAMTLM_BreakEvenHistoryFrequency and --SAMTLM_UniqueEventCountImpact." << endl;
	  exit(1);
	}
	if (Vocabulary== "") {
	  cerr << "Error: When using SAMT LMs, you have to specify the Vocab file created by the trained LMs in parameter --Vocabulary" << endl;
	  exit(1);
	}
	// first create the database object linked to the datbase file
	// tmpscf[i] and register it
	CDatabaseSAMTLM* samtDB = new CDatabaseSAMTLM(tmpscf[i], tmpscf[i].substr(tmpscf[i].size()-12) == ".samtlmbtree");
	CDatabaseSAMTLM::m_samtlmDatabases.push_back(samtDB);
	// now iterate through the different SAMTLM parameter values
	vector<samtstring> v1;
	CRule::Tokenize(SAMTLM_MinHistoryFrequency, v1, ",");
	vector<samtstring> v2;
	CRule::Tokenize(SAMTLM_MinBackoffWeight, v2, ",");
	vector<samtstring> v3;
	CRule::Tokenize(SAMTLM_HistoryFrequencyConvergenceSpeed, v3, ",");
	vector<samtstring> v4;
	CRule::Tokenize(SAMTLM_BreakEvenHistoryFrequency, v4, ",");
	vector<samtstring> v5;
	CRule::Tokenize(SAMTLM_UniqueEventCountImpact, v5, ",");
	for (int i1 = 0; i1<v1.size(); i1++) {
	  for (int i2=0; i2<v2.size(); i2++) {
	    for (int i3=0; i3<v3.size(); i3++) {
	      for (int i4=0; i4<v4.size(); i4++) {
		for (int i5=0; i5<v5.size(); i5++) {
		  // first the forward version
		  CLanguageModelSAMT::m_languageModels.push_back(new CLanguageModelSAMT(samtDB->m_db, FORWARDLM, CChartParserInfo::from_string<CFrequency>(v1[i1]), CChartParserInfo::from_string<double>(v2[i2]), CChartParserInfo::from_string<double>(v3[i3]),CChartParserInfo::from_string<double>(v4[i4]), CChartParserInfo::from_string<double>(v5[i5])));

		  if (SAMTLM_UseBackwardLM) {
		    // now the backward version
		    CLanguageModelSAMT::m_languageModels.push_back(new CLanguageModelSAMT(samtDB->m_db, BACKWARDLM, CChartParserInfo::from_string<CFrequency>(v1[i1]), CChartParserInfo::from_string<double>(v2[i2]), CChartParserInfo::from_string<double>(v3[i3]),CChartParserInfo::from_string<double>(v4[i4]), CChartParserInfo::from_string<double>(v5[i5])));
		  }
		}
	      }
	    }
	  }
	}
      } else if (tmpscf[i].substr(tmpscf[i].size()-6) == ".srilm") {
	// we have a regular SRI language model
	cout<<"Loading SRI language model: "<<tmpscf[i]<<endl;
	CLanguageModelSAMT::m_languageModels.push_back(new CachedSriLM(tmpscf[i], FORWARDLM));
      } else  if (tmpscf[i].substr(tmpscf[i].size()-14) == ".srilmbackward") {
	// we have an SRI language model trained on reversed sentences
	cout<<"Loading backward (reverse) SRI language model(s): "<<tmpscf[i]<<endl;
	CLanguageModelSAMT::m_languageModels.push_back(new CachedSriLM(tmpscf[i], BACKWARDLM));
      } else {
        cerr<<"Unable to determine the format of the LM"<<endl;
        exit(-1);
      }
    }
  }else{
    cout<<"No LM model loaded!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"<<endl;
  }
  // set n.o. LMs
  CRule::C_NumLMs = CLanguageModel::m_languageModels.size();
  // also shift the index of Glue and RuleAppCount models
  CRule::C_GlueLocation = CRule::C_NumLMs;
  CRule::C_RuleApplicationCountLocation = CRule::C_NumLMs+1;
  // set n.o. translation models (i.e., cost values per rule in the rule database)
  CRule::C_NumModelsTM = CRule::C_NumModelsTotal - CRule::C_NumLMs - (CChartParserInfo::m_useSpanFeatures? 2:0);
  // also set up the location of the source span penalty feature (this location is only valid if m_useSpanFeatures is true)
  CRule::C_SourceSpanPenaltyLocation = CRule::C_NumModelsTotal - 2;

  //===========================================================================
  //more customization settings
  cout<<"FullSentNonTerminalString: "<<FullSentNonTerminal<<endl;
  CWordIndex fullWI =  pVoc->AddWord(samtstring(FullSentNonTerminal));
  cout<<"fullWI: "<<fullWI<<endl;
  parser.SetFullSentenceNonTerminal(fullWI);

  //  cout<<"GlueNonTerminalString: "<<GlueNonTerminal<<endl;
  //  CWordIndex glueWI = pVoc->AddWord(string(GlueNonTerminal));
  //  cout<<"glueWI: "<<glueWI<<endl;
  //  parser.SetGlueNonTerminal(glueWI);

  cout<<"About to begin parsing"<<endl;
  // BEGIN PARSING OPERATION


  vector<CWordSequence> sourceCorpus;
  LoadText(SentenceList.c_str(), sourceCorpus, pVoc, AddBOSEOSInternally);

  int NGramsConsidered = 1;
  //if there is a reference list
  MetricSearch* p_bSearch=NULL;
  if (ReferenceList!= ""){
    //if (ScoringMetric==NULL){
    if (ScoringMetric == ""){
      cerr<<"You must provide a ScoringMetric if you are providing a ReferenceList"<<endl;
      exit(2);
    }
    cout<<"Metric Initialized for:"<<ScoringMetric<<endl;

    if ((ScoringMetric =="NIST") || (ScoringMetric == "NISTBLEU") || (ScoringMetric == "IBMBLEU")){
      NGramsConsidered = 4;
      if (NSZ<4){
	NGramsConsidered=NSZ;
	cerr<<"WARNING: Dropping NGramsConsidered for ScoringMetric to: "<<NGramsConsidered<<" since NSZ = "<<NSZ<<endl;
      }
      bool normalizeReference = (CChartParserInfo::m_normalizeReferences>0) ? 1 : 0;
      p_bSearch = new BleuSearch(ReferenceList.c_str(), pbuf, normalizeReference, GainFile.c_str(), ScoringMetric, NumReferences, NGramsConsidered, pVoc);
    }else if ((ScoringMetric=="TER")){
      bool normalizeReference = (CChartParserInfo::m_normalizeReferences>0) ? 1 : 0;
      p_bSearch = new TERSearch(ReferenceList.c_str(), pbuf, normalizeReference, NumReferences, pVoc);
    }else{
      cerr<<"Unkown metric="<<ScoringMetric<<endl;
      cerr<<endl;
      exit(2);
    }
  }

  //verify number of references is correct
  if (p_bSearch!=NULL){
    //if (pTextF->NumberOfSentences()*NumReferences!=p_bSearch->TotalNumReferences()){
    if (sourceCorpus.size()*NumReferences!=p_bSearch->TotalNumReferences()){
      cout<<"Mismatch between number of required references, and the number found in ReferenceList"<<endl;
      cerr<<"Mismatch between number of required references, and the number found in ReferenceList"<<endl;
      exit(2);
    }
  }

  //load the pruning map for CompareTargetWords==1 AND CompareTargetWords=0
  //in CompareTargetWords==0 mode, it places a restriction on the number of rules that that can be used under a production result
  //this primarily affects nbest extraction times

  //if (CChartParserInfo::m_compareTargetWords){
  CChartParserInfo::m_maxCostDifferencePerCell = CChartParserInfo::StringToFloat(MaxCostDifferencePerCell);
  CChartParserInfo::m_aStarBeam = CChartParserInfo::StringToFloat(AStarBeam);
  CChartParserInfo::LoadPruningMap(PruningMap.c_str());
  if (PruningMapUseBoundary==1){
    CChartParserInfo::m_pruningMapUseBoundary = true;
  }else{
    CChartParserInfo::m_pruningMapUseBoundary = false;
  }
  cout<<"PruningMapUseBoundary: "<<CChartParserInfo::m_pruningMapUseBoundary<<endl;
  //}

  //if we have a set of tags
  if (TagSetList!=""){
    CChartParserInfo::LoadTagsForLM(TagSetList.c_str());
  }

  //set combo pruning options
  if (ComboPruningGlobal==1){
    CChartParserInfo::m_comboPruningGlobal = true;
  }else{
    CChartParserInfo::m_comboPruningGlobal = false;
  }

  //use the LM during combo pruning
  if (ComboPruningUseLM==1){
    CChartParserInfo::m_comboPruningUseLM = true;
  }else{
    CChartParserInfo::m_comboPruningUseLM = false;
  }

  //use the Lexical entries in the same beam as the hierarchical entries
  //  if (ComboPruningUseLexical==1){
  //    CChartParserInfo::m_comboPruningUseLexical = true;
  //  }else{
  //    CChartParserInfo::m_comboPruningUseLexical = false;
  //  }

  CChartParserInfo::m_comboPruningBeamSize = CChartParserInfo::StringToFloat(ComboPruningBeamSize);
  CChartParserInfo::m_comboPruningMinBeamSize = CChartParserInfo::StringToFloat(ComboPruningMinBeamSize);
  CChartParserInfo::m_comboPruningFuzzCostDifference = CChartParserInfo::StringToFloat(ComboPruningFuzzCostDifference);


  cout<<"ComboPruningBeamSize: "<<CChartParserInfo::m_comboPruningBeamSize<<
    " ComboPruningMinBeamSize: "<<CChartParserInfo::m_comboPruningMinBeamSize<<
    " ComboPruningFuzzCostDifference: "<<CChartParserInfo::m_comboPruningFuzzCostDifference<<" Global: "<<CChartParserInfo::m_comboPruningGlobal<<" UseLM: "<<CChartParserInfo::m_comboPruningUseLM<<endl;

  //tracking for the top level n-best lists
  vector<NBestList*> nBestLists; //track them all
  //for (int i=0; i<pTextF->NumberOfSentences(); i++){
  for (int i=0; i<sourceCorpus.size(); i++){
    //CNBest* pNB = new CNBest(ScalingFactors.size());
    //pNB->Init(pVoc);
    NBestList* pNB = new NBestList;
    //dont keep any translations
    pNB->Init(false, true);
    nBestLists.push_back(pNB);
  }
  //tracking for all the sub n-best lists - create the new elements as you need them
  vector< vector<NBestList*> > suppNBestLists;
  for (int i=0; i<sourceCorpus.size(); i++){
    suppNBestLists.push_back(vector<NBestList*>());
  }


  //keep track of scaling factors
  ScoreScalingFactorMap trackScalingFactorScores;

  //set to true if TotalNewTranslations = 0 twice
  //only used in MER Training
  bool MERComplete = false;
  int iteration = 0;

  //save sentence IDs for which no translation was found
  vector<int> noTranslationSentences;

  //save the start time of the whole process
  samtstring timeAdded;
  stringstream sst(stringstream::in | stringstream::out);
  sst<<time(NULL);
  timeAdded = sst.str();

  //start the main MER loop
  bool exitTranslationLoop = false;
  //when translation starts
  time_t run_time;
  time ( &run_time ); //fill the time object
  printf ( "TIMER start: %s", ctime (&run_time) );

  // relevant in worker-bee mode: true if we've already gone through all sentences once
  bool workerBeeCompletedFirstPass = false;
  // relevant in worker-bee mode: true if we have to try to process all sentences now, even recently locked ones
  bool workerBeeGetThemAllNow = false;

  do {
    //DO FOLLOWING INSIDE ITERATION

    cout<<"Initializing modelBestCounts to handle upto: "<<NGramsConsidered<<"-grams"<<endl;
    RawCounts modelBestCounts;  //this object is properly initialized in the constuctor

    int sentenceNumber = 0;
    int TotalNewTranslations = 0;
    //track timing figures across iterations in wall clock seconds

    double NumSecsDecoding = 0;
    double NumSecsNBest = 0;
    double NumSecsSuppNBest = 0;

    //track timing figures across iterations in CPU seconds
    double NumSecsDecoding_CPU = 0;
    double NumSecsNBest_CPU = 0;
    double NumSecsSuppNBest_CPU = 0;

    // relevant in worker-bee mode: true if at least one sentence is still unprocessed
    //this is moved to the beginning of each full pass to allow the bee to see if there are any more sentences left
    bool workerBeeAtLeastOneSentenceUnprocessed = false;
    // true if at least one sentence was processed in current loop
    bool workerBeeAtLeastOneSentenceProcessed = false;

    for (vector<CWordSequence>::iterator corpusIter = sourceCorpus.begin(); corpusIter != sourceCorpus.end(); ++corpusIter){
      bool skipSentence = false;
      samtstring lockfilename(""); // only relevant if in worker-bee mode
      // only relevant if in worker-bee mode - we'll write the n-best list into these files
      samtstring nbestcandsfilename("");
      samtstring nbestfeatsfilename("");
      samtstring nbestsentencefilename("");
      if (WorkerBeeDir != "") {
	ostringstream oss1;
	oss1 << WorkerBeeDir << "/" << sentenceNumber << ".lock";
	lockfilename = oss1.str();
	ostringstream oss2;
	oss2 << WorkerBeeDir << "/" << sentenceNumber << ".feats";
	nbestfeatsfilename = oss2.str();
	ostringstream oss3;
	oss3 << WorkerBeeDir << "/" << sentenceNumber << ".cands";
	nbestcandsfilename = oss3.str();
	ostringstream oss4;
	oss4 << WorkerBeeDir << "/" << sentenceNumber << ".translations";
	nbestsentencefilename = oss4.str();

	time_t lockFileModTime = FileExists(lockfilename.c_str());

	//you can only pass thru if you see both the feats and cands there AND the lock file has been properly removed
	//this picks up cases where the process was terminated during writing the files
	if (FileExists(nbestfeatsfilename.c_str()) && FileExists(nbestcandsfilename.c_str()) && FileExists(nbestsentencefilename.c_str()) && !lockFileModTime) {
	  cout<<"Sentence was proper, skipping: "<<sentenceNumber<<endl;
	  skipSentence = true;
	} else {
	  int lockHandle = getlock(lockfilename.c_str());
	  if (lockHandle < 0) {
	    cout<<"Sentence has lock: "<<sentenceNumber<<endl;
	    // handle is locked already
	    if (!workerBeeCompletedFirstPass) {
	      // we're still in the first pass - just skip the sentence
	      skipSentence = true;
	      workerBeeAtLeastOneSentenceUnprocessed = true;
	      cout<<"Haven't finished one pass, skipping: "<<sentenceNumber<<endl;
	    } else {
	      // get the difference in time between lock file timestamp and current time
	      // and sleep until that difference is greater than MaxSentenceProcessingTime

	      // if the lock file didn't exist (lockFileModTime == 0), then we should never have had trouble getting the lock
	      if (!lockFileModTime) {
		cerr << "Error: couldn't get lock even though the lock file doesn't exist." << endl;
		exit(2);
	      }
	      time_t timenow;
	      time(&timenow);
	      double time_to_sleep = MAX(0.0, (double) MaxSentenceProcessingTime - difftime(timenow, lockFileModTime));
	      cout << "Has been locked for " << (double) difftime(timenow,lockFileModTime) << " seconds; MaxSentenceProcessingTime="
		   << MaxSentenceProcessingTime << "; MaxSentenceProcessingTime-difftime=" <<
		(double) MaxSentenceProcessingTime - difftime(timenow,lockFileModTime) << "; time_to_sleep="
		   << time_to_sleep << endl;

	      if (time_to_sleep > 0 && !workerBeeGetThemAllNow) {
		cout << "Sentence got locked only recently, skipping: " << sentenceNumber << endl;
		skipSentence = true;
		workerBeeAtLeastOneSentenceUnprocessed = true;
	      } else {
		cout<<"Going to sleep " << (time_to_sleep) << " seconds waiting for the lock of sentence: "<<sentenceNumber<<endl;
		usleep((long) (time_to_sleep * 1000000.0));
		// check if the lock file still exists and its modification time hasn't changed
		time_t lockFileModTimeNew = FileExists(lockfilename.c_str());
		if (!lockFileModTimeNew) {
		  // file doesn't exist anymore -- that means it got done after all
		  // so we can safely skip the sentence
		  skipSentence = true;
		} else {
		  if (lockFileModTimeNew == lockFileModTime) {
		    // lock file still exists and time stamp hasn't changed
		    // assume process has died and remove the lock
		    cout<<"Removing lock of dead process: "<<sentenceNumber<<endl;
		    remove(lockfilename.c_str());
		    // and create a new lock
		    lockHandle = getlock(lockfilename.c_str());
		    if (lockHandle < 0) {
		      cerr << "Error: Deleted lock file but it's still locked!" << endl;
		      exit(2);
		    }
		  } else {
		    // lock file still exists but some work has been done on it (or someone else killed it and re-grabbed it)
		    // skip sentence for now but make sure you check on it later
		    skipSentence = true;
		    workerBeeAtLeastOneSentenceUnprocessed = true;
		  }
		}
	      }
	    }
	  }
	}
      }
      if (!skipSentence) {
	workerBeeAtLeastOneSentenceProcessed = true;
	cout<<endl<<endl<<endl<<"============================================================================";
	cout<<"Sentence " << sentenceNumber << " (starting with 0): ";
	CWordSequence& pSourceSentence = (*corpusIter);
	CChartParserInfo::PrintVocabSequence(pSourceSentence);
	cout<<endl;

	// set the Scaling factors back into parsing mode
	vector<double> ScalingFactorsBeforeRescoring = ScalingFactors;
	cout << "Using parsing feature weights:";
	for (int i=0; i<ScalingFactorsBeforeRescoring.size(); i++) {
	  cout << " " << i<<"="<<ScalingFactorsBeforeRescoring[i];
	}
	cout << endl;
	parser.SetScalingFactors(ScalingFactorsBeforeRescoring); //!runs static initialization of
	//CChartParserHyp as well

	time_t time_parseStart;
	clock_t time_parseStartCPU;
	time(&time_parseStart);
	time_parseStartCPU = clock();
	//      cout<<" \n beginning search..."<<endl;
	parser.InitSearch(pSourceSentence);
	//      cout<<"InitSearch completed"<<endl;

	parser.ParseChart();
	time_t time_parseEnd;
	clock_t time_parseEndCPU;
	time(&time_parseEnd);
	time_parseEndCPU = clock();

	NumSecsDecoding += difftime(time_parseEnd, time_parseStart);
	double Sentence_NumSecsDecoding_CPU  = ((double) (-time_parseStartCPU + time_parseEndCPU)) / CLOCKS_PER_SEC;
	NumSecsDecoding_CPU += Sentence_NumSecsDecoding_CPU;

	cout<<"========================="<<endl;
	int workingNBest;
	//if cmp=0, NBest determines the search strategy, so lets be more careful
	//but we still want to use the NBestMult strategy for retrieval in cmp=1 too
	if (NBestMult==1){
	  workingNBest = (int) MAX(pow(NBest, (double)(pSourceSentence.size())), NBestMin); //ensure we always get at least 100
	}else{
	  workingNBest = (int)NBest;
	}
	cout<<"Retrieving N-best (" << workingNBest << ") parses... "<<endl;

	//code to test retrieving a sub n-best element
	// right now testing with node at m_maxCombinationCount - PASSES after check on pHyp span to apply EOS
	// need to test at m_maxCombinationCount, j>1 (should fail due to automatic BOS being applied) -  PASSES after checking pHyp->m_BOSDone
	// need to test at i<m_maxCombinationCount, j>1, j==1 - PASSES after previous check passed
	//vector<pair<CHypBeam*, double> > beamsAndRefLengths;

	//pairing an n-best list with the average reference length
	vector<pair< vector<CChartParserHypComplete*>, double> > subSentenceNBestListAndAvg;

	int subSentenceWorkingNBest = NBestMin;
	if (pSourceSentence.size()>CChartParserInfo::m_maxCombinationCount && NBestSubSpans>0){

	  //beamsAndRefLengths = ExamineSubsentenceNBest(parser, pMetric, NBest, DisplayNBestData);
	  if (NBestMult==1){
	    subSentenceWorkingNBest = (int)MAX(pow(NBest, (double)CChartParserInfo::m_maxCombinationCount), NBestMin); //ensure we always get at least 100
	  }else{
	    subSentenceWorkingNBest = (int)NBest;
	  }
	  cout<<"Looking into max combo count condition: extracting "<<subSentenceWorkingNBest<<" NBest list\n";
	  time(&time_parseStart);
	  time_parseStartCPU = clock();

	  subSentenceNBestListAndAvg =  GenerateSubsentenceNBest(parser, p_bSearch, sentenceNumber, CChartParserInfo::m_sourceLength, subSentenceWorkingNBest, NBestSubSpans);
	  time(&time_parseEnd);
	  time_parseEndCPU = clock();
	  NumSecsSuppNBest += difftime(time_parseEnd, time_parseStart);
	  NumSecsSuppNBest_CPU += ((double) (-time_parseStartCPU + time_parseEndCPU)) / CLOCKS_PER_SEC;
	}


	//CProductionNode* pCompleteProductionNode = parser.CompleteTranslationNode();
	//assert(pCompleteProductionNode->m_fineGrainedNodes.size()==1);
	//CChartParserNode* pCompleteProductionNodeFine = (pCompleteProductionNode->m_fineGrainedNodes.begin())->second;

	//time(&time_parseStart);
	//CChartParserHypComplete* pHypLastHC = pCompleteProductionNodeFine->GetKthBest(workingNBest-1); //its workingNBest-1 since its 0 indexed
	//vector<CChartParserHypComplete*> hcNBestList = pCompleteProductionNodeFine->RetrieveKBestHC(workingNBest);
	//time(&time_parseEnd);
	//cout<<" ======== HC NBest BEGIN ============= "<<endl;
	//parser.OutputNBest(hcNBestList.begin(), hcNBestList.end(), DisplayNBestData);
	//cout<<" ======== HC NBest END ============= "<<endl;
	//double NumSecNBestHC = difftime(time_parseEnd, time_parseStart);

	time(&time_parseStart);
	time_parseStartCPU = clock();
	//CHypBeam* pBeam = parser.RetrieveKBestParses(workingNBest, pCompleteProductionNodeFine);
	//vector<CChartParserHypComplete*> beamNBestList = pBeam->AsVector();
	vector<CChartParserHypComplete*> fullSentenceNBestList = parser.RetrieveKBestParses(workingNBest);


	time(&time_parseEnd);
	time_parseEndCPU = clock();

	//cout<<"NumSecNBestBeam: "<<NumSecNBestBeam<<" NumSecNBestHC: "<<NumSecNBestHC<<endl;
	NumSecsNBest += difftime(time_parseEnd, time_parseStart);
	double Sentence_NumSecsNBest_CPU = ((double) (-time_parseStartCPU + time_parseEndCPU)) / CLOCKS_PER_SEC;
	NumSecsNBest_CPU += Sentence_NumSecsNBest_CPU;
	if (fullSentenceNBestList.empty()){
	  cerr << "ERROR: No translation found for sentence "<<sentenceNumber<<endl;
	  cout << "ERROR: No translation found for sentence "<<sentenceNumber<<endl;
	  exit(-1);
	}


	//TESTING KBEST BEGIN
	/*
	  int count =0;
	  if (CChartParserInfo::m_compareTargetWords==1 && CChartParserInfo::m_extractUnique==false){
	  CSortedHypContainer::iterator iterBeam = pBeam->m_store.begin();
	  vector<CChartParserHypComplete*>::iterator iterHC=pCompleteProductionNodeFine->m_candidateVector.begin();
	  int count = 0;
	  for (; iterBeam!= pBeam->m_store.end(), iterHC!=pCompleteProductionNodeFine->m_candidateVector.end(); ++iterBeam, ++iterHC, ++count){
	  CChartParserHypComplete* pHypBeam = (*iterBeam).second;
	  CChartParserHypComplete* pHypHC = (*iterHC);
	  assert(fabs(pHypBeam->ScaledTMLMCost()-pHypHC->ScaledTMLMCost())<0.01);
	  //check all the model costs
	  assert(pHypHC->m_negLogProbs.size() == pHypBeam->m_negLogProbs.size());
	  for (int m=0; m<pHypBeam->m_negLogProbs.size(); m++){
	  if(fabs(pHypHC->m_negLogProbs[m]-pHypBeam->m_negLogProbs[m])<0.01){

	  }else{
	  cout<<"Failed on nbest: "<<count<<endl;
	  cout<<"HC: "; copy(pHypHC->m_negLogProbs.begin(), pHypHC->m_negLogProbs.end(), ostream_iterator<float>(cout, " ")); cout<<endl;
	  cout<<"Beam: "; copy(pHypBeam->m_negLogProbs.begin(), pHypBeam->m_negLogProbs.end(), ostream_iterator<float>(cout, " "));cout<<endl;
	  assert(fabs(pHypHC->m_negLogProbs[m]-pHypBeam->m_negLogProbs[m])<0.01);
	  }
	  }
	  }
	  }
	*/
	//TESTING KBEST END


	/*
	  ARV 09/24 Implement Rescoring for general n-best list
	  if (!NoRescoring) {
	  if (AlwaysRescore || !MEROptimize) {
	  // set different Scaling factors for the rescoring process
	  vector<double> ScalingFactorsForRescoring =
	  MEROptimize ? ScalingFactors : ScalingFactorsRescoring;
	  cout << "Rescoring the beam using rescoring feature weights:";
	  for (int i=0; i<ScalingFactorsForRescoring.size(); i++) {
	  cout << " " << i <<"="<<ScalingFactorsForRescoring[i];
	  }
	  cout << endl;
	  parser.SetScalingFactors(ScalingFactorsForRescoring); //!runs static initialization ofCChartParserHyp as well
	  } else {
	  cout << "Rescoring the beam using same (parsing) feature weights." << endl;
	  }
	  pBeam->Rescore();
	  }
	*/

	//cout<<"NumElements in pBeam="<<pBeam->GetStorePtr()->size()<<endl;
	/*
	//ARV 09/24 Move this call INTO the RetrieveCalls
	if (UniqueBeam) {
	cout<<"Uniquing the beam... "<<endl;
	pBeam->Unique();
	cout<<"NumElements in pBeam="<<pBeam->GetStorePtr()->size()<<endl;
	}
	*/

	//check if search metric was defined, if so, we can augment the NBestList (s) with some scoring data
	if (p_bSearch!=NULL){

	  //augment the NBestList (s) and return the scoring data (num n-grams matched, reference length of closest etc)
	  //RawCounts topTranslationRaw = BuildNBest(pbuf, nBestLists[sentenceNumber], p_bSearch, pBeam, workingNBest, sentenceNumber, useNormalizing, TotalNewTranslations);
	  //RawCounts topTranslationRaw = BuildNBest(pbuf, nBestLists[sentenceNumber], p_bSearch, beamNBestList, workingNBest, sentenceNumber, useNormalizing, TotalNewTranslations);
	  RawCounts topTranslationRaw = BuildNBest(pbuf, nBestLists[sentenceNumber], p_bSearch, fullSentenceNBestList, workingNBest, sentenceNumber, useNormalizing, TotalNewTranslations);
	  int NBestSizeBeforeSub = nBestLists[sentenceNumber]->Size();

	  int AdditionalTranslations = 0;
	  BuildNBestSubsentence(pbuf, suppNBestLists, p_bSearch, subSentenceNBestListAndAvg, subSentenceWorkingNBest, sentenceNumber, useNormalizing, AdditionalTranslations );

	  /*
	    vector<CWordIndex> tmpTranslation;
	    if (pBeam->m_store.empty()) {
	    cerr << "WARNING: No translation found for sentence "<<sentenceNumber<<endl;
	    cout << "WARNING: No translation found for sentence "<<sentenceNumber<<endl;
	    noTranslationSentences.push_back(sentenceNumber);
	    //!! doesn't work
	    tmpTranslation.push_back(unkWI);
	    } else {
	    tmpTranslation.insert(tmpTranslation.end(),pBeam->GetTopHyp()->m_translationOutput.begin(), pBeam->GetTopHyp()->m_translationOutput.end() );
	    }
	  */

	  modelBestCounts = modelBestCounts + topTranslationRaw;
	}

	// before deleting the beam, let's get some stats about unique
	// translations
	//if (UniqueBeam)int uniqueTranslationsCount = pBeam->Unique();
	//cout<<"NumElements in full sentence pBeam="<<pBeam->GetStorePtr()->size()<<endl;
	//cout <<"Log: main pBeam "<<pBeam->GetStorePtr()->size() << " total translations. Top :" <<DisplayNBestData<< endl;

	//this is for STTK compatibility:
	cout << "TRANSLATION_RESULT_SENTENCE " << sentenceNumber << endl;
	cout <<"Log: fullSentenceNBestList "<<fullSentenceNBestList.size() << " total translations. Top :" <<DisplayNBestData<< endl;
	if (DisplayNBestData>0){
	  //parser.OutputNBest(beamNBestList.begin(), beamNBestList.end(), DisplayNBestData);
	  parser.OutputNBest(fullSentenceNBestList.begin(), fullSentenceNBestList.end(), (WorkerBeeDir != "" ? 1 : DisplayNBestData));
	  //ARV make sure you output the subsentence n-best lists if you want to here

	  cout <<"\nLog: subSentenceNBestListAndAvg: "<<subSentenceNBestListAndAvg.size() << " total sub-cells"<<endl;
	  int count = 1;
	  for (vector< pair<vector<CChartParserHypComplete*>, double> >::iterator subIter = subSentenceNBestListAndAvg.begin(); subIter!= subSentenceNBestListAndAvg.end(); ++subIter){
	    pair< vector<CChartParserHypComplete*>, double>& beamAndAvg = (*subIter);
	    cout<<"Log: subSentenceNBestListAndAvg: cell: "<<CChartParserInfo::m_maxCombinationCount<<", "<<count<<endl;
	    parser.OutputNBest(beamAndAvg.first.begin(), beamAndAvg.first.end(), 2);
	    cout<<endl;
	    count++;
	  }
	}
	cout<<endl;
	//before deleting just count how many unique were seen
	/*
	  if (!UniqueBeam){
	  pBeam->Unique();
	  cout <<"Log: main pBeam "<<pBeam->GetStorePtr()->size() << " unique translations"<<endl;
	  }
	  delete pBeam;
	*/

	if (WorkerBeeDir != "") {
	  // write out the n-best list
	  samtstring featsFileName = nbestfeatsfilename;
	  samtstring candsFileName = nbestcandsfilename;
	  //	samtstring initFileName = string("init.opt.tmp.")+timeAdded;

	  ofstream outF(featsFileName.c_str());
	  ofstream outC(candsFileName.c_str());
	  //	ofstream outI(initFileName.c_str());

	  nBestLists[sentenceNumber]->PrintOptimizationData(sentenceNumber, NSZ, outF, outC);


	  // output DisplayNBestData top translations into nbestsentencefilename
	  ofstream outS(nbestsentencefilename.c_str());
	  int translationCount = DisplayNBestData;
	  for (vector<CChartParserHypComplete*>::iterator iter = fullSentenceNBestList.begin(); translationCount-- != 0 && iter!= fullSentenceNBestList.end(); ++iter){
	      //CChartParserHypComplete* pHyp = (*iter).second;
	      CChartParserHypComplete* pHyp = (*iter);

	      outS<<"ScaledTMLMCost: "<<pHyp->ScaledTMLMCost()<<" m_totalCost: "<<pHyp->m_totalCost<<" raw:";
	      for (int i=0; i<pHyp->m_negLogProbs.size(); i++){
		outS<<" "<<pHyp->m_negLogProbs[i];
	      }
	      outS << " UNKNOWN-TAGGED-handled translation: ";
	      CChartParserInfo::PrintVocabSequence(pHyp->m_translationOutput, outS);
	      outS<<" PARSE: ";
	      CChartParserInfo::PrintParsingOutput(pHyp->m_parsingOutput, outS);
	      outS << endl;
	  }

	  // close the file streams
	  outF.close();
	  outC.close();
	  outS.close();

	  //and remove the lock file
	  remove(lockfilename.c_str());
	}
	// if there is any need for global modelling of nbest
	// lists then we cant delete them immediately
	if ((WorkerBeeDir== "") && (GenOracleScores || MEROptimize)){
	  //leave the nbest list alone!
	}else{
	  //BUT we dont clear the sub-sentence n-best lists...
	  nBestLists[sentenceNumber]->ClearAll();
	}

	//delete the other n-best lists

	DeleteNBestLists(fullSentenceNBestList, subSentenceNBestListAndAvg);

	//must be called after each sentence is translated
	time_t time_delStart = clock();
	parser.ClearSearch();
	time_t time_delEnd = clock();
	double del_CPU = ((double) (-time_delStart + time_delEnd)) / CLOCKS_PER_SEC;

	cout<<"Time-Sentence DecodingSecondsCPU: " << Sentence_NumSecsDecoding_CPU <<
	  " NBestSecondsCPU: " << Sentence_NumSecsNBest_CPU << " ClearMemorySecondsCPU: "<<del_CPU<<" len: "<<pSourceSentence.size()<<endl;
      }
      sentenceNumber++;
    }
    //done with current pass thru data
    if (WorkerBeeDir == "") {
      time ( &run_time ); //fill the time object
      printf ( "TIMER one translation loop: %s", ctime (&run_time) );
      cout<<"TIME one translation loop:NumSecsDecoding: "<<NumSecsDecoding<<" NumSecsNBest: "<<NumSecsNBest<<" NumSecsSuppNBest: "<<NumSecsSuppNBest<<endl;
      cout<<"TIME_CPU one translation loop:NumSecsDecoding_CPU: "<<NumSecsDecoding_CPU<<" NumSecsNBest_CPU: "<<NumSecsNBest_CPU<<" NumSecsSuppNBest_CPU: "<<NumSecsSuppNBest_CPU<<" Spawn: "<<CChartParserInfo::m_spawntimecpu<<" NumLMCalls: "<<CChartParserInfo::m_lmCall<<" NumLMCacheHit: "<<CChartParserInfo::m_lmCacheHit<<endl;
      //reset counters for next round
      NumSecsDecoding=0;
      NumSecsNBest=0;
      CChartParserInfo::m_lmCall = 0;
      CChartParserInfo::m_lmCacheHit = 0;
      CChartParserInfo::m_spawntimecpu = 0;

      cout<<"All sentences processed ======================="<<endl;
      if (p_bSearch!=NULL){
	cout<<"Model best: ";
	for (int i=0; i<NGramsConsidered; i++){
	  cout<<modelBestCounts.numCorrect[i]<<" / "<<modelBestCounts.numSuggested[i]<<" ";
	}
	cout<<" ref: "<<modelBestCounts.referenceLengths<<endl;
	cout<<endl;


	AbstractScore detailedScore = p_bSearch->getMetricScoreDetailed(modelBestCounts);
	assert(detailedScore.size()>0);
	double tscore = detailedScore[0];
	trackScalingFactorScores.insert(pair<double, vector<double> >(tscore, ScalingFactors));
	cout<<"TrainTranslationScore: ";
	copy(detailedScore.begin(), detailedScore.end(), std::ostream_iterator<double>(cout, " "));
	cout<<" for n.o. ScalingFactors: "<<ScalingFactors.size()<<endl;
	cout<<"TotalNewTranslations: "<<TotalNewTranslations<<endl;

	//this criteria is probably a bit too strict on its own. so we also check to make sure the metric score hasnt stagnated. we can
	//check this by looking at the score tracker
	if (TotalNewTranslations==0){
	  MERComplete = true;
	  cout<<"Potential early termination triggered since TotalNewTranslations==0"<<endl;
	}

	if (GenOracleScores){
	  AllSentRawCountsVector accumRawCounts;
	  for( vector< NBestList* >::iterator iter=nBestLists.begin(); iter!=nBestLists.end(); ++iter ) {
	    NBestList* pNB = (*iter);
	    accumRawCounts.push_back(pNB->GetRawCounts());
	  }

	  Configuration config = p_bSearch->GetOracleConfiguration(accumRawCounts);
	  cout<<"OracleTranslationScore: "<<p_bSearch->ScoreConfiguration(accumRawCounts, config)[0]<<endl;
	  delete [] config;
	}


      }else{
	cout<<"No scoring possible since metric search object was not initialized\n, probably because no references were provided"<<endl;
      }

      cout<<"Done with my first pass!!!"<<endl;
      iteration++;
      // if there is optimization going on
      if (MEROptimize){
	//OptimizeParameters( ScalingFactors, nBestLists, ScoringMetric, NGramsConsidered, ParameterRestrictionFile,trackScalingFactorScores,timeAdded);
	//OptimizeParametersInternal( ScalingFactors, nBestLists, p_bSearch,  featureRange, trackScalingFactorScores, timeAdded);
	OptimizeParametersInternal( ScalingFactors, nBestLists, suppNBestLists, p_bSearch,  featureRange, trackScalingFactorScores, timeAdded);

	time ( &run_time ); //fill the time object
	printf ( "TIMER one translation + mer loop: %s", ctime (&run_time) );
	//(now done at beginning of loop)      parser.SetScalingFactors(ScalingFactors);
      }
      cout<<"Loop completed\n";
      if (noTranslationSentences.size() != 0) {
	cerr << "WARNING: No translations were found for the following sentences (indices start with 0), this could mess up the scoring metric:";
	cout << "WARNING: No translations were found for the following sentences (indices start with 0), this could mess up the scoring metric:";
	for (int i=0; i<noTranslationSentences.size(); i++) {
	  cerr << " " << noTranslationSentences[i];
	  cout << " " << noTranslationSentences[i];
	}
	cerr << endl;
	cout << endl;
	noTranslationSentences.clear();
      }
      // modify the exit condition (e.g., no new translations occurred)
      // in the line below
      exitTranslationLoop = iteration >= IterationLimit || !MEROptimize || MERComplete;
      if (exitTranslationLoop && MEROptimize) {

	time ( &run_time ); //fill the time object
	printf ( "TIMER parsing optimized: %s", ctime (&run_time) );

	cout << "Final feature weights for parsing: ";
	for (int i=0; i<ScalingFactors.size(); i++) {
	  cout << (i == 0 ? "" : "_") << ScalingFactors[i];
	}
	cout <<endl;
	if (!NoRescoring) {
	  cout << "Do rescoring optimization starting with latest parsing feature vector" << endl;
	  ScoreScalingFactorMap trackScalingFactorScoresRescoring1;
	  trackScalingFactorScoresRescoring1.insert(pair<double, vector<double> >(0, ScalingFactors));
	  //	OptimizeParametersInternal( ScalingFactors, nBestLists, p_bSearch,  featureRangeRescoring, trackScalingFactorScoresRescoring1, timeAdded);
	  OptimizeParametersInternal( ScalingFactors, nBestLists, suppNBestLists, p_bSearch,  featureRangeRescoring, trackScalingFactorScoresRescoring1, timeAdded);
	  cout << "Final feature weights for rescoring: ";
	  for (int i=0; i<ScalingFactors.size(); i++) {
	    cout << (i == 0 ? "" : "_") << ScalingFactors[i];
	  }
	  cout <<endl;
	  cout << "Sanity test: Do rescoring optimization starting with value from --FeatureWeightsRescoring" << endl;
	  ScoreScalingFactorMap trackScalingFactorScoresRescoring2;
	  trackScalingFactorScoresRescoring2.insert(pair<double, vector<double> >(0, ScalingFactorsRescoring));
	  //OptimizeParametersInternal( ScalingFactors, nBestLists, p_bSearch,  featureRangeRescoring, trackScalingFactorScoresRescoring2, timeAdded);
	  OptimizeParametersInternal( ScalingFactors, nBestLists, suppNBestLists, p_bSearch,  featureRangeRescoring, trackScalingFactorScoresRescoring2, timeAdded);
	  ScalingFactors.clear(); // to avoid misunderstandings (this isn't the final feature weight vector)

	  time ( &run_time ); //fill the time object
	  printf ( "TIMER rescoring optimized: %s", ctime (&run_time) );
	}
      } // end: if (exitTranslationLoop && MEROptimize)
    } else { // else: if (WorkerBeeDir == NULL)
      // if this was already the second (or even higher) pass and we didn't process any sentences, then we also have
      // to tackle recently locked sentences in the next pass (if there's still any unprocessed sentences remaining)
      workerBeeGetThemAllNow = workerBeeCompletedFirstPass && !workerBeeAtLeastOneSentenceProcessed;

      workerBeeCompletedFirstPass = true; // we're not in first pass anymore from now on
      cout<<"Status after pass: unprocessed: "<<workerBeeAtLeastOneSentenceUnprocessed << ", processed: " << workerBeeAtLeastOneSentenceProcessed <<endl;
      exitTranslationLoop = !workerBeeAtLeastOneSentenceUnprocessed;
    }
  }while(!exitTranslationLoop);
  cout<<"All loops completed\n";

  if (useNormalizing){
    cout<<"Closing pstream\n";
    //ClosePBUF();
    if (pbuf.good()){
      pbuf.flush();
      pbuf.close();
    }
    cout<<"pstream closed\n";
  }
  cout<<"Exitting now\n";
  for (int i=0; i< CLanguageModelSAMT::m_languageModels.size(); i++) delete CLanguageModelSAMT::m_languageModels[i];
  for (int i=0; i< CDatabaseSAMTLM::m_samtlmDatabases.size(); i++) delete CDatabaseSAMTLM::m_samtlmDatabases[i];
  exit(0);
  return 0;
}


void DeleteNBestLists(vector<CChartParserHypComplete*>& fullSentenceNBestList,  vector<pair< vector<CChartParserHypComplete*>, double> >& subSentenceNBestListAndAvg){
  for (vector<CChartParserHypComplete*>::iterator iter = fullSentenceNBestList.begin(); iter!= fullSentenceNBestList.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    delete pHyp;
    pHyp= NULL;
  }
  fullSentenceNBestList.clear();

  for (vector< pair<vector<CChartParserHypComplete*>, double> >::iterator subIter = subSentenceNBestListAndAvg.begin(); subIter!= subSentenceNBestListAndAvg.end(); ++subIter){
    //pair<CHypBeam*, double> beamAndAvg = (*subIter);
    //if the supplemenary n best list is the same size as the element you need to add, then create a new n-best list
    //on the next pass around you dont need to do this
    pair< vector<CChartParserHypComplete*>, double>& beamAndAvg = (*subIter);
    for (vector<CChartParserHypComplete*>::iterator iter = beamAndAvg.first.begin(); iter!=  beamAndAvg.first.end(); ++iter){
      CChartParserHypComplete* pHyp = (*iter);
      delete pHyp;
      pHyp= NULL;
    }
    beamAndAvg.first.clear();
  }
  subSentenceNBestListAndAvg.clear();


}

//! Calls retrieve k-best on sub sentence cells to get a more complete perspective during MER training
//vector< pair<CHypBeam*, double> > GenerateSubsentenceNBest(CChartParser& parser, MetricSearch* pMetric, int sentenceNumber, int sourceLength, const int& NBest, int NBestSubSpans){
vector< pair< vector<CChartParserHypComplete*>, double> > GenerateSubsentenceNBest(CChartParser& parser, MetricSearch* pMetric, int sentenceNumber, int sourceLength, const int& NBest, int NBestSubSpans){
  //int jpos = MIN(2, CChartParserInfo::m_sourceLength);
  //int ipos = MAX(1, CChartParserInfo::m_maxCombinationCount-1);

  int ipos = CChartParserInfo::m_maxCombinationCount;
  int jmax = (sourceLength-ipos)+1;
  //<2 means (0) dont do any of this (1) do it only for cell m_maxCombinationCount, 1
  if (NBestSubSpans<2){
    jmax=NBestSubSpans;
  }

  vector<pair<vector<CChartParserHypComplete*>, double> > ret;
  cout<<"ExamineSubsentenceNBest: "<<jmax<<" cells to expand at depth: "<<ipos<<endl;

  for (int jpos=1; jpos<=jmax; jpos++){

    //simplest case, look at the top element
    //int jpos = 1;
    CCompleteHypChartCell& chartCell = parser.m_completeHypsChart[ipos][jpos];

    double avgLength = pMetric->GetAverageSentenceLength(sentenceNumber);
    double artificalLength = avgLength/((double)sourceLength/(double)ipos);
    cout<<"Log: Sub cell: "<<ipos<<","<<jpos<<" sourceLength: "<<sourceLength<<" ipos: "<<ipos<<" avgLength: "<<avgLength<<" artificalLength: "<<artificalLength<<" ";

    if (chartCell.m_productionNodes.count(CChartParserInfo::m_FullSentenceNT)!=0){
      CProductionNode* pTopNode = chartCell.m_productionNodes[CChartParserInfo::m_FullSentenceNT];
      cout<<" num fine-grained: "<<pTopNode->m_fineGrainedNodes.size()<<endl;

      //this needs to be handled in a generic fashion
      //CHypBeam* pBeamTopNode = parser.RetrieveKBestParsesSpan(NBest, pTopNode);
      //pBeamTopNode->HandleUnknownAndTags(); //ARV 09/23 moved into RetrieveKBestParsesSpan
      //ret.push_back(pair<CHypBeam*, double>(pBeamTopNode, artificalLength));

      ret.push_back(pair< vector<CChartParserHypComplete*>, double>(parser.RetrieveKBestParsesSpan(NBest, pTopNode), artificalLength));

    }else{
      cout<<" WARNING: we have no m_FullSentenceNT in cell: "<<ipos<<","<<jpos<<endl;
    }
  }
  return ret;

}

void BuildNBestSubsentence(redi::pstream& pbuf, vector< vector<NBestList*> > suppNBestLists, MetricSearch* p_bSearch, vector<pair< vector<CChartParserHypComplete*>, double> > subSentenceNBestListAndAvg, int subSentenceWorkingNBest, int sentenceNumber, bool useNormalizing, int& AdditionalTranslations ){
  //now we can go thru all the sub sentence n-best lists that have been added as well
  int suppCount = 0;
  vector<int> numUniqueAdded;
  for (vector< pair<vector<CChartParserHypComplete*>, double> >::iterator subIter = subSentenceNBestListAndAvg.begin(); subIter!= subSentenceNBestListAndAvg.end(); ++subIter){
    //pair<CHypBeam*, double> beamAndAvg = (*subIter);
    //if the supplemenary n best list is the same size as the element you need to add, then create a new n-best list
    //on the next pass around you dont need to do this

    pair< vector<CChartParserHypComplete*>, double>& beamAndAvg = (*subIter);

    //if this is the first time we are building supplemental n-best lists for this sentence, then create new NBestList (s)
    if (suppNBestLists[sentenceNumber].size()==suppCount){
      NBestList* pNB = new NBestList;
      //dont keep any translations
      pNB->Init(false, true);
      suppNBestLists[sentenceNumber].push_back(pNB);
    }

    //ARV do all this work before getting the list
    //to be consistent if we apply UniqueBeam
    //if (UniqueBeam) {
    //cout<<"Uniquing sub beam..."<<endl;
    //beamAndAvg.first->Unique();
    //}

    //vector<CChartParserHypComplete*> beamNBestSub = beamAndAvg.first->AsVector();
    int translationsAdded = 0;
    //BuildNBest(pbuf, suppNBestLists[sentenceNumber][suppCount], p_bSearch, beamAndAvg.first, workingNBest, sentenceNumber, useNormalizing, AdditionalTranslations, beamAndAvg.second);
    BuildNBest(pbuf, suppNBestLists[sentenceNumber][suppCount], p_bSearch, beamAndAvg.first, subSentenceWorkingNBest, sentenceNumber, useNormalizing, translationsAdded, beamAndAvg.second);

    cout <<"Additional total translations: "<<translationsAdded <<" effective ref length:" <<beamAndAvg.second<<endl;

    //if (DisplayNBestData>0 && CChartParserInfo::m_verbosity>1){
    //cout<<"Display sub top 1: "<<endl;
    //parser.OutputNBest(beamNBestSub.begin(), beamNBestSub.end(),  1);
    //}

    //make sure you are only doing Unique before deleting
    //int uniqueTranslationsCountTopNode = beamAndAvg.first->Unique();
    //numUniqueAdded.push_back(uniqueTranslationsCountTopNode);
    //cout <<"Additional unique translations: "<<uniqueTranslationsCountTopNode <<" effective ref length:" <<beamAndAvg.second<<endl;

    //just delete the beam now
    //delete beamAndAvg.first;
    //beamAndAvg.first = NULL;
    suppCount++;
  }
  //ARV move this code to after the request is made
  //cout<<"Log: Number of Supp NBestLists: "<<suppCount<<" num translations each ["; copy(numUniqueAdded.begin(), numUniqueAdded.end(), ostream_iterator<int>(cout, " ")); cout<<"]"<<endl;

}

//RawCounts BuildNBest(redi::pstream& pbuf, NBestList* oneNBestList,  MetricSearch* pMetric, CHypBeam* pBeam, int translationCount, int sentenceNumber, bool normalizeTranslations, int& TotalNewTranslations, double artificialRefLength) {

RawCounts BuildNBest(redi::pstream& pbuf, NBestList* oneNBestList,  MetricSearch* pMetric, vector<CChartParserHypComplete*>& nBestVector, int translationCount, int sentenceNumber, bool normalizeTranslations, int& TotalNewTranslations, double artificialRefLength) {

  int numFeatures;
  RawCounts topTranslationRaw;
  //for (CSortedHypContainer::iterator iter = pBeam->m_store.begin(); translationCount-- != 0 && iter!= pBeam->m_store.end(); ++iter){
  //CChartParserHypComplete* pHyp = (*iter).second;
  for (vector<CChartParserHypComplete*>::iterator iter = nBestVector.begin(); translationCount-- != 0 && iter!= nBestVector.end(); ++iter){
    CChartParserHypComplete* pHyp = (*iter);
    numFeatures = pHyp->m_negLogProbs.size();

    stringstream ss(ios::in | ios::out);
    CChartParserInfo::PrintVocabSequence(pHyp->m_translationOutput, ss);
    samtstring translation = ss.str();
    CVocabulary::string_trim(translation);

    if (normalizeTranslations){
      stringstream ss(stringstream::in | stringstream::out);
      //cout<<"Printing to ss"<<endl;
      //cout<<"translation before norm: "<<translation<<endl;
      pbuf<<translation<<"\n"; //az: used to be: <<"\n\r"
      pbuf.flush();
      samtstring str;
      std::getline(pbuf.out(), str);
      translation = str;

      //string returnedTranslation = NormalizeStringPBUF(translation);
      //cout<<"after norm: "<<returnedTranslation<<endl;
      pHyp->m_translationOutput = CChartParserInfo::makeSequence< CWordSequence >(translation);
      //cout<<"after norm list: "<<endl;
      //PrintVocabSequence(pHyp->m_translationOutput);
      //cout<<endl;
      //translation = returnedTranslation;
    }

    RawCounts raw;
    pMetric->ScoreSingleHyp(sentenceNumber, translation, raw);

    //make sure we are using the BLEU / NIST scores
    if (artificialRefLength!=-1){
      raw.referenceLengths = artificialRefLength;
    }

    if (iter==nBestVector.begin()){
      topTranslationRaw = raw;
    }
    //file it

    TotalNewTranslations+= oneNBestList->InsertCandidate(pHyp->m_negLogProbs, CChartParserInfo::makeSequence<CWordSequenceVector>(translation), raw);

  }
  return topTranslationRaw;

}


void LoadText(const char* fileName, vector< CWordSequence >& corpus, CVocabulary* pVoc, int AddBOSEOSInternally){
  ifstream fp(fileName);
  if (!fp.good()){
    cerr<<"Unable to open:"<<fileName<<endl;
  }

  char buf[MAXBUF];
  buf[0] = '\0';
  for (; !fp.eof(); ){
    fp.getline(buf, MAXBUF);
    samtstring sourceString(buf);
    // remove leading or trailing whitespace
    CVocabulary::string_trim(sourceString);
    if (sourceString.size()>0){
      if (AddBOSEOSInternally) {
	sourceString = "<s> " + sourceString + " </s>";
      }
      corpus.push_back(CChartParserInfo::makeSequence<CWordSequence>(sourceString));
    } else {
      fp.getline(buf, MAXBUF);
      if (!fp.eof()) {
	cerr << "Error loading test corpus: empty (or whitespace-only) line in the corpus!" << endl;
	exit(1);
      }
    }
  }
}

void ParseDoubleSequence(samtstring floatSequenceStr, vector<double>& doubleVector) {
  ParseDoubleSequence(floatSequenceStr.c_str(), doubleVector);
}

void ParseDoubleSequence(const char* floatSequenceStr, vector<double>& doubleVector) {
  doubleVector.clear();
  vector<samtstring> tmpscf;
  CRule::Tokenize(floatSequenceStr,tmpscf,",_");
  for (int i = 0; i<tmpscf.size(); i++) {

    double f = atof(tmpscf[i].c_str());
    cout<<f<<" ";
    doubleVector.push_back(f);

  }
}

//! ScalingFactors is the return, not the init value; the last entry
//of trackScalingFactorScores contains the initial ScalingFactors used
void OptimizeParametersInternal( vector<double>& ScalingFactors, vector<NBestList*>& nBestLists, vector< vector<NBestList*> >& suppNBestLists, MetricSearch* pMetric,  vector< vector<double> >& rangeAndInit, ScoreScalingFactorMap& trackScalingFactorScores, samtstring timeAdded){


  //harvest the feats data

  vector<LogProbList*> allFeats;
  vector<RawCounts*> allScores;
  vector<int> allCands;

  //from the base n-best list
  for (vector<NBestList*>::iterator iter = nBestLists.begin(); iter!= nBestLists.end(); ++iter){
    NBestList* oneSent = (*iter);
    oneSent->LoadOptimizationData(allFeats, allScores, allCands );
  }

  int fullSentCands = allCands.size();
  cout<<"Log: OptimizeParametersInternal initial num sent: "<<fullSentCands;
  //from the supplemental n-best lists where each sentence has a vector of n-best lists
  for (int i=0; i<suppNBestLists.size(); i++){
    for (int j=0; j<suppNBestLists[i].size(); j++){
      suppNBestLists[i][j]->LoadOptimizationData(allFeats, allScores, allCands );
    }
  }
  cout<<" after supp: "<<allCands.size()<<endl;

  pair<double, vector<double> > pr = (*(trackScalingFactorScores.rbegin()));
  if (rangeAndInit.size()==2){
    rangeAndInit.push_back(pr.second);
  }

  //we are actively working with parameters
  if (rangeAndInit.size()==3){
    assert(rangeAndInit.size()>0);
    assert(pr.second.size()==rangeAndInit[0].size());
    rangeAndInit[2]= pr.second;
  }

  assert(rangeAndInit.size()==3);

  cout<<"Optimization beginning [rangeAndInit="<<rangeAndInit.size()<<" allFeats="<<allFeats.size()<<" allCands="<<allCands.size()<<"]"<<endl;
  cout<<"Left: "; copy(rangeAndInit[0].begin(), rangeAndInit[0].end(), std::ostream_iterator<double>(cout, "\t")); cout<<endl;
  cout<<"Right: "; copy(rangeAndInit[1].begin(), rangeAndInit[1].end(), std::ostream_iterator<double>(cout, "\t")); cout<<endl;
  cout<<"Init: "; copy(rangeAndInit[2].begin(), rangeAndInit[2].end(), std::ostream_iterator<double>(cout, "\t")); cout<<endl;


  bool doWriteBackupFeats = false; //true;
#ifdef NDEBUG
  doWriteBackupFeats = true;
#endif

  if (true){
    samtstring featsFileName = samtstring("feats.opt.")+timeAdded;
    samtstring candsFileName = samtstring("cands.opt.")+timeAdded;
    samtstring initFileName = samtstring("init.opt.tmp.")+timeAdded;

    ofstream outF(featsFileName.c_str());
    ofstream outC(candsFileName.c_str());
    ofstream outI(initFileName.c_str());

    int sentenceNumber = 0;
    for (vector<NBestList*>::iterator iter = nBestLists.begin(); iter!= nBestLists.end(); ++iter){
      NBestList* oneSent = (*iter);
      oneSent->PrintOptimizationData(sentenceNumber, NSZ, outF, outC);
      sentenceNumber++;
    }

    copy(rangeAndInit[0].begin(), rangeAndInit[0].end(), std::ostream_iterator<double>(outI, " "));
    outI<<endl;
    copy(rangeAndInit[1].begin(), rangeAndInit[1].end(), std::ostream_iterator<double>(outI, " "));
    outI<<endl;
    copy(rangeAndInit[2].begin(), rangeAndInit[2].end(), std::ostream_iterator<double>(outI, " "));
    outI<<endl;

    outC.close();
    outF.close();
    outI.close();
  }


  COptimizeMER test(pMetric);
  test.Init(rangeAndInit[0], rangeAndInit[1], OptiNumPermutes, OptiNumModify, OptiNumStarts, OptiTwoStage, OptiDoNorm, OptiEpsilon, OptiEpsilonScale);

  cout<<"Running opti"<<endl;
  vector<double> returnLambdas = test.minModelCostMaxMetricScore(allFeats, allScores, allCands, rangeAndInit[2], fullSentCands);
  cout<<"Got Params: "; copy(returnLambdas.begin(), returnLambdas.end(), std::ostream_iterator<double>(cout, "_"));   cout<<endl;
  ScalingFactors = returnLambdas;
}
