/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#include "HypFiller.h"

// This function will NOT make any requests for the real LM costs,
// only DummyLookup(s) will be made here
void HypFiller::CreateInitialTuplesAndQueueLookups(
    RulesByProductionResult::iterator begin,
    RulesByProductionResult::iterator end,
    vector<CCombinationTuple*>* initialTuples,
    int& allPossibleExpansions, bool addAll){

  bool debug = false;
  //double fakeExternalCost = 0.0;
  CWordIndex productionResultProcessed;
  //add all rules' best-instantiations
  for (RulesByProductionResult::iterator match_iter=begin;
       match_iter!=end; ++match_iter){

    if (match_iter==begin){
      productionResultProcessed = match_iter->first;
      if (debug){
	cout<<"Spawning for: "<<
            CChartParserInfo::m_pVocabulary->GetWord(
                productionResultProcessed
                )<<endl;
      }
    }else{
      assert(productionResultProcessed == match_iter->first);
    }
    CChartParserHypIncomplete* pNewIncompleteHypPrototype =
        match_iter->second.first;
    CRule* pOneRule = &(match_iter->second.second); // the rule found

    // lets put this hyp's best instantiation onto the beam
    // Using the coarse-grained-EC back pointers provided from pIncomplete,
    // instantiate the <e, 1> element,
    // where the e refers to the rule being used
    // and the 1 refers to the vector of 1st best derivations from each of
    // the back pointer nodes
    vector< std::vector<CChartParserNode*>::iterator > bestDerivationVector;
    //add all the backpointers - using the underlying BP nodes' best cost which
    //is m_bestTotalCost AND LM costs AND boundary LM cost
    // go thru each one and pick out the first best pChartParserNode
    for (vector<CProductionNode*>::const_iterator iter =
             pNewIncompleteHypPrototype->m_productionBackPointers.begin();
         iter!= pNewIncompleteHypPrototype->m_productionBackPointers.end();
         ++iter ){

      CProductionNode* pProdNode = (*iter);

      assert(pProdNode->m_fineGrainedNodes.size()==
             pProdNode->m_lowestToHighestCostFineGrained.size());
      assert(pProdNode->m_lowestToHighestCostFineGrained.size()>0);

      // initialize the derivation vector with (1,...,1) means
      // using for each element an iterator pointing to the
      // best backpointed-node for each NT substitution site
      std::vector<CChartParserNode*>::iterator firstBest =
          pProdNode->m_lowestToHighestCostFineGrained.begin();
      bestDerivationVector.push_back(firstBest);
    }

    // at this point we have all the information we need to queue
    // up all the remaining LM calls. We just need to convert
    // any tags present in the translation output
    CWordSequence translationOutput = pOneRule->m_translationOutput;
    if (pNewIncompleteHypPrototype->m_productionBackPointers.size()==0 &&
        CChartParserInfo::m_tagsForLM.count(
            pOneRule->m_productionResult)!=0){
      //cout<<"Tagging: "; pItem->Display(); cout<<endl;
      // replaces the translation output
      translationOutput = CWordSequence(1, pOneRule->m_productionResult);
    }

    CUtilities::HarvestAllLMStringsAllLMs(
        translationOutput,
        pNewIncompleteHypPrototype->m_productionBackPointers);

    // create tuple
    CCombinationTuple* c = new CCombinationTuple(
        pOneRule, pNewIncompleteHypPrototype, bestDerivationVector);
#ifndef NDEBUG
    if (debug){
      if (pNewIncompleteHypPrototype->m_productionBackPointers.size()==0){
	cout<<"Add phrasal rule to beam: "; c->Display();
      } else {
	cout<<"Add to beam: "; c->Display();
      }
    }
#endif
    initialTuples->push_back(c);
    //
    // add it to the beam and before that instantiate its
    // corresponding complete hyp, calculating LM costs etc.
    //bool wasInserted = pBeam->InsertAndCompleteBaseCombination(c);
    // the InsertBaseCombination never does any pruning and no
    // duplicates should have occurred yet
    //assert(wasInserted==true);
    //===========================================
    // for diagnotic purposes, lets see what the total number
    // expanded would have been
    //===========================================
    int possibleExpansions = 1;
    for (vector<CProductionNode*>::iterator iter =
             pNewIncompleteHypPrototype->m_productionBackPointers.begin();
         iter!= pNewIncompleteHypPrototype->m_productionBackPointers.end();
         ++iter){
      CProductionNode* pPNode = (*iter);
      possibleExpansions *= pPNode->m_fineGrainedNodes.size();
    }
    //cout<<endl;
    allPossibleExpansions += possibleExpansions;

  }//all rules processed
}


// Add a hyp to the m_unfiledHyps vector of the production node
// that is indicated in the hyp
// Returns true iff the hyp was compliant with the beam operating
// in the production node. If a new CProductionNode has to be
// created to deal with this hyp, then we have to indicate that
// in the newlyCreatedNodes vector
bool HypFiller::AddCompletedHyp(
    CChartParserHypComplete* pHyp,
    bool addAll,
    CCompleteHypChartCell* chartCell,
    CCoarseGrainedEq* newlyCreatedNodes
    ) {
  bool s_debugFileHyp = false || (CChartParserInfo::m_verbosity>1);
  if (s_debugFileHyp){
    cout<<"AddCompletedHyp: ";
    pHyp->DisplayRecursive();
  }
  //CCompleteHypChartCell& chartCell =
  //      (*completeHypsChart)[pHyp->I()][pHyp->J()];
  CProductionNode* node; // hypergraph node for pHyp
  bool newlyCreated = false;

  // If there are no elements of this pHyp 's equivalence class
  // (i.e., owning hypergraph node), start a new one and register
  // it in the newlyCreatedNodes vector
  if (chartCell->m_productionNodes.count(pHyp->GetResult())==0) {

    //create a new node right here
    node = new CProductionNode(pHyp->I(), pHyp->J(), pHyp->GetResult());
    chartCell->m_productionNodes.insert(
        std::pair<CWordIndex, CProductionNode*>(pHyp->GetResult(), node));

    if (s_debugFileHyp){
      CChartParserInfo::TabStack();
      cout<<"New node created in this cell - marking it to apply the self rule in the next wave: ";
      node->Display();
      cout<<endl;
    }

    newlyCreatedNodes->insert(
        std::pair<CWordIndex, CProductionNode*>(pHyp->GetResult(), node));
    newlyCreated = true;
  }else{
    //EC exists already and is not the complete-translation node
    //let's add the hyp to the corresponding production node, which will filter down the fine grained nodes
    CCoarseGrainedEq::iterator curr =
        chartCell->m_productionNodes.find(pHyp->GetResult());
    node = (*curr).second;
    if (s_debugFileHyp){
      CChartParserInfo::TabStack(); cout<<"Existing. Node:";
      node->Display();
      cout<<endl;
    }
  }
  bool wasAdded = node->CheckBeamInsertHyp(pHyp, addAll);
  //if you just created the hyp you MUST have kept the hyp
  if (newlyCreated){
    assert(wasAdded==true);
  }
  return wasAdded;
}


pair<int, int> HypFiller::SpawnForProductionResult(
    RulesByProductionResult::iterator begin,
    RulesByProductionResult::iterator end,
    bool addAll,
    CCompleteHypChartCell* chartCell,
    CCoarseGrainedEq* newlyCreatedNodes
    ){

  // now the interesting move. we need to add all the top level
  // rules, then unwind to identify the top k hyps
  CCombinationTupleBeam beam;
  //initialize the beam with the pairs of incomplete
  //hyps and the completing rules
  int allPossibleExpansions = 0;
  //  double fakeExternalCost = 0.0;
  bool debug = false || (CChartParserInfo::m_verbosity>1);


  if (debug){
    cout<<"***BEGIN: HypFiller::SpawnForProductionResult "<<endl;
  }

  //Identify the initialTuples for this rule applications and
  //queue up all the possible LM DummyLookup calls
  vector<CCombinationTuple*> initialTuples;
  CreateInitialTuplesAndQueueLookups(begin,
                      end,
                      &initialTuples,
                      allPossibleExpansions,
                      addAll);
  // all the dummy lookups have been scheduled. Now wait for them
  // to return from all LMs. We have to do this before adding
  // any of the tuples to the final beam since that operation
  // requires real LM calls
  CLanguageModel::WaitForLookupsAllLMs();

  //now add all the initialTuples to the beam!
  for (vector<CCombinationTuple*>::iterator citer = initialTuples.begin();
       citer != initialTuples.end();
       ++citer) {
    CCombinationTuple* c = (*citer);
    bool wasInserted = beam.InsertAndCompleteBaseCombination(c);
    assert(wasInserted);
  }
  // The beam is initialized, start generating the K-best alternative
  // configurations for this rule application

  //accumulate these values over the whole sentence
  clock_t spawnTime = clock();
  int initialNumTuples  = beam.NumTuplesProposed();

  //  double firstFailedCost = FLT_MAX;
  // keep going retrieving tuples and expand these tuples'
  // neighbors until there are no more elements on the beam or
  // you hit a break out condtion:
  //a) hit the limit m_comboPruningBeamSize of elements that
  //   you are allowed to expand
  //b) more than m_comboPruningMinBeamSize tuples have been
  //   considered and the tuple is worse than
  //   bestHyp+prodCategoryPruningMapCostDifference+
  //                  m_comboPruningFuzzCostDifference
  // If a) occurs, we still retrieve what we have on the beam
  //       (since that doesn't cost anything as the LM costs have already
  //       been calculated for these tuples), but we don't expand the
  //       retrieved tuples' neighbors anymore
  // If b) occurs, it would be stupid to still retrieve the rest from
  //       the beam since these tuples must all be worse
  //       than the tuple causing the cutoff based break-out,
  //       therefore none of these would be accepted by AddCompletedHyp

  // whether to expand retrieved tuples' neighbors and put these on the beam
  bool expandNeighbors = true;
  int count = 0; // n.o. tuples retrieved from beam
  // n.o. tuples retrieved that were actually added
  // (not pruned away immediately) by AddCompletedHyp
  int numAdded = 0;
  // n.o. good (i.e., added) tuples that were saved because of an avoided
  // m_comboPruningFuzzCostDifference break-out attempt that was averted
  // due to m_comboPruningMinBeamSize not having been exceeded
  int numAddedPostFailure = 0;
  // after this gets true, added tuples are counted towards numAddedPostFailure
  bool breakOutAttemptAverted = false;
  //  int numTopCalls = 0;
  //see call to beam.NextTuple

#ifndef NDEBUG
  float lastCostDebug = -(FLT_MAX)/10.0;
#endif


  for (CCombinationTuple* pTuple = beam.NextTuple(expandNeighbors);
       pTuple != NULL;
       count++){

#ifndef NDEBUG
    // in cmp=0 we can be sure that each
    // successive cost is higher than the last one
    //same holds when ParserInfo::m_comboPruningUseLM==false
    if (!CChartParserInfo::m_compareTargetWords ||
        !CChartParserInfo::m_comboPruningUseLM){
      assert(pTuple->Cost()>=lastCostDebug);
      lastCostDebug = pTuple->Cost();
    }
    if (debug){
      cout<<"NextTuple: "; pTuple->Display(); cout<<endl;
    }
#endif

    // hyp to be added
    CChartParserHypComplete* pNewCompleteHyp = pTuple->m_pComplete;

    // check whether the hyps cost satisfy m_comboPruningFuzzCostDifference
    float cutoffFromBestCost =
      chartCell->m_productionNodes.count(pNewCompleteHyp->GetResult())==0 ?
        FLT_MAX/10
        :
        chartCell->
        m_productionNodes[pNewCompleteHyp->GetResult()]->m_cutoffLogProb;
    if (!addAll && pTuple->Cost() >
        CChartParserInfo::GetPruningCutoff(
            cutoffFromBestCost,
            CChartParserInfo::m_comboPruningFuzzCostDifference)) {
      // the only thing that can save this tuple and
      // subequent ones now is m_comboPruningMinBeamSize
      if (count >= CChartParserInfo::m_comboPruningMinBeamSize) {
	break; // stop retrieving hyps
      } else {
	// got lucky
	breakOutAttemptAverted = true;
      }
    }

#ifndef NDEBUG
    if (debug){
      cout<<"Was compliant -> ";
    }
    if (debug){
      cout<<"Created: "; pNewCompleteHyp->Display();
    }
    if (debug){
      cout<<" HypCost: "<<pNewCompleteHyp->m_totalCost +
          pNewCompleteHyp->m_scaledFutureCostEstimate<<endl;
    }
#endif

    // now try to add the complete hyp corresponding to this tuple
    // AddCompletedHyp is responsible for deleting the hyp ultimately,
    // no matter whether it accepts it or prunes it
    // that means we have to set its pointer in pTuple to NULL so that
    // it won't get deleted when destructing pTuple
    pTuple->m_pComplete = NULL;
    assert((*(beam.m_uniqeTuples.find(pTuple)))->m_pComplete == NULL);

    if (!AddCompletedHyp(pNewCompleteHyp,
                         addAll,
                         chartCell,
                         newlyCreatedNodes)){
      // failure - the hyp was pruned away

#ifndef NDEBUG
      assert(!addAll);
      if (debug){
        cout<<"Failed beaming -> ";
      }
#endif

    }else{
      // success of adding the hyp
      if (debug){cout<<"Added "<<endl;}
      //track how many hyps you are actually getting thru AddCompletedHyp
      numAdded++;
      if (breakOutAttemptAverted) {
        numAddedPostFailure++;
      }
    }
    // should we still expand retrieved tuples' neighbors or only
    // retrieve the remaining tuples from the beam?
    expandNeighbors = (count < CChartParserInfo::m_comboPruningBeamSize);
    pTuple = beam.NextTuple(expandNeighbors); // retrieve next tuple
  }

  if ( //CChartParserInfo::m_comboPruningBeamSize>=FLT_MAX/10 &&
      CChartParserInfo::m_comboPruningFuzzCostDifference>=FLT_MAX/10){
    //cout<<"Checking to make sure beam is empty"<<endl;
    if (beam.Size()!=0){
      cout<<"BeamSize: "<<beam.Size()<<" count: "<<count<<endl;
    }
    assert(beam.Size()==0);
  }

  // some statistics
  pair<int, int> toReturn(numAdded, allPossibleExpansions);

  CChartParserInfo::m_hypsAdded += numAdded;
  CChartParserInfo::m_numTopCalls += count;
  //this variable is a good measure of how much work was
  // done during combo pruning
  CChartParserInfo::m_numTuplesProposed += beam.NumTuplesProposed();
  CChartParserInfo::m_allPossibleExpansions += allPossibleExpansions;
  CChartParserInfo::m_numAddedPostFailure += numAddedPostFailure;

  //accumulate these values over the whole data set
  clock_t spawnTimeEnd = clock();
  double spawncpu = ((double)  (-spawnTime + spawnTimeEnd)) / CLOCKS_PER_SEC;
  CChartParserInfo::m_spawntimecpu += spawncpu;

  if (debug){
    cout<<"***END: HypFiller::SpawnForProductionResult (numAdded, allPossibleExpansions)="<<toReturn.first<<","<<allPossibleExpansions<<endl;
  }

  return toReturn;
}


pair<int, int> HypFiller::SpawnKBestCompleteHyps(
    RulesByProductionResult& hypRuleCombinations,
    bool addAll,
    CCompleteHypChartCell* chartCell,
    CCoarseGrainedEq* newlyCreatedNodes
    ){

  bool debug = false;
  // do pruning and spawning for each production result that this k
  // split is producing this pruning happens before the BeamPruning call is
  // made on all production results
  if (debug){
    cout<<"Start Spawning === "<<endl;
  }
  pair<int, int> expandedAndPossible;

  for (RulesByProductionResult::iterator iter= hypRuleCombinations.begin();
       iter!=hypRuleCombinations.end(); ){
    CWordIndex productionResult = (*iter).first;
    RulesByProductionResult::iterator end =
        hypRuleCombinations.upper_bound(productionResult);
    //write out the rules for this cell filing
    if (CChartParserInfo::m_verbosity>0){
      for (RulesByProductionResult::iterator tmp = iter; tmp!=end; ++tmp){
	CChartParserHypIncomplete* pNewIncompleteHypPrototype
            = tmp->second.first;
	CRule* pOneItem = &(tmp->second.second); // the rule found
	cout<<"RULE (HypFiller::SpawnKBestCompleteHyps): ";
        DisplayInstantiatedRule(
            pOneItem,
            pNewIncompleteHypPrototype);
        cout<<endl;
      }
    }
    //iter < end defines the range
    pair<int, int> ep = SpawnForProductionResult(iter,
                                                 end,
                                                 addAll,
                                                 chartCell,
                                                 newlyCreatedNodes);
    expandedAndPossible.first+=ep.first;
    expandedAndPossible.second+=ep.second;
    //at the end, reset to the next production result
    iter=end;
  }

  if (debug){
    cout<<"End Spawning === "<<endl;
  }
  return  expandedAndPossible;

}

void HypFiller::DisplayInstantiatedRule(
    CRule* pRule,
    CChartParserHypIncomplete* pIncompleteHyp) {

  cout<<CChartParserInfo::m_pVocabulary->GetWord(pRule->m_productionResult)
      <<" -> ";
  for (int i=0; i<pRule->m_translationOutput.size(); i++){
    if (pRule->m_translationOutput[i]<=0){
      //cout<<" BP="<<m_translationOutput[i]<<" ";
      assert(pIncompleteHyp->m_productionBackPointers.size()
             >-1*pRule->m_translationOutput[i]);
      CProductionNode* pNode =
          pIncompleteHyp->m_productionBackPointers
          [-1*pRule->m_translationOutput[i]];
      cout<<" "<<CChartParserInfo::m_pVocabulary->GetWord(pNode->GetResult())
          <<"["<<pNode->I()<<","<<pNode->J()<<"] ";
    }else{
      cout<<CChartParserInfo::m_pVocabulary->GetWord(
          pRule->m_translationOutput[i])<<" ";
    }
  }
}
