/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#ifndef __LANGUAGEMODEL__
#define __LANGUAGEMODEL__
#include <map>
#include <vector>
#include <cfloat>
#include <climits>
#include "ParserInfo.hh"
#include "Vocabulary.hh"

//! whether our probs are of type float or double
typedef float CProb;
//! whether our log probs are of type float or double
typedef float CLogProb;
//! whether our neg-log probs are of type float or double
typedef float CNegLogProb;
//! type of an ngram frequency: make sure this is consistent with the type trainlm.pl uses
typedef double CFrequency;
//! an index that represents a potentially collapsed LM history
typedef uint64 SLMHistory;

//! type of LM
typedef int CLMType;
#define FORWARDLM 0 //!< left-to-right
#define BACKWARDLM 1 //!< right-to-left
//! context is considered at left and right of the current word
#define CENTERLM 2

//! used by hash-map based LM caches
struct LMCWordSequenceHash{
  size_t operator()(const vector<CWordIndex>& c)const{
    size_t hashCode = 1;
    for (int i=0; i<c.size(); i++){
      hashCode = 31*hashCode + c[i];
    }
    return hashCode;
  }
};

//! used by hash-map based LM caches
struct LMCWordSequenceEquals{
  bool operator()(const vector<CWordIndex>& c1,
                  const vector<CWordIndex>& c2 )const{
    return (c1==c2);
  }
};


// CLanguageModel is the generic interface for all language models
// used in the SAMT system. It defines calls that are relevant
// for both state based and word based LM operations
class CLanguageModel{

public:

  // stores all the LMs we are using - possibly multiple models per file
  static vector<CLanguageModel*> m_languageModels;

  //turn this on or off from OUTSIDE the language model to see what context+scores its operating with
  static bool s_debugLM;

  // type of this LM
  CLMType m_lmType;

  CLanguageModel(CLMType lmType) : m_lmType(lmType)
  {}

  //! Gets the LM prob of the word lmString[wordPos],
  // where the context is given by the other words in lmString.
  /*!
    If context is longer than we need, use only the X words before
    (FORWARDLM) / after (BACKWARDLM) wordPos.
    In case context only has less elements than we need, back off.
    Note: the context may contain m_lmBoundaryWI ("<*>").
    Only the words in lmString between the right-most <*> left of wordPos
    and the left-most <*> right of wordPos are valid context words.
  */
  virtual CNegLogProb GetSingleWordLMCost(
      const CWordSequence& lmString, int wordPos) = 0;

  //! might want to call this after each sentence to
  // let the cache not grow too big
  virtual void ClearCache(){}

  // ==========================================================
  // State and batch based LM call functions
  // ==========================================================

  // Function below match  definition in google prodlm

  virtual SLMHistory getNullState() = 0;

  virtual void DummyLookup(const SLMHistory& lm_history, CWordIndex next_word,
                           SLMHistory* new_history) = 0;

  // Do DummyLookups for a whole sentence. If ebos=true, then
  // the start state is BOS state and we also compute EOS prob
  virtual void DummyLookupSequence(const CWordSequence& sentence) = 0;

  virtual CNegLogProb
  getCostAndNewHistoryAndNWordsFound(const SLMHistory& old_history,
                                     CWordIndex next_word,
                                     SLMHistory* new_history,
                                     int* nwords_found) = 0;

  virtual string HistoryToDebugString(uint32 history) const = 0;

  // Wait for all bactched LM calls to return
  virtual void WaitForLookups() = 0;

  static void WaitForLookupsAllLMs() {
    for (int i=0; i<CLanguageModel::m_languageModels.size(); i++){
      CLanguageModel::m_languageModels[i]->WaitForLookups();
    }
  }

  virtual void PrintContextSequence(const CWordSequence& lmContext) {
    cout<<"[";
    for (CWordSequence::const_iterator iter = lmContext.begin();
         iter != lmContext.end();
         ++iter){
      SLMHistory w = (*iter);
      cout<<HistoryToDebugString(w);
      if (iter+1 != lmContext.end()){
        cout<<",";
      }
    }
    cout<<"]";
  }

  static void PrintContextSequenceAllLMs(const CWordSequence& lmContext) {
    for (int i=0; i<CLanguageModel::m_languageModels.size(); i++){
      cout<<"LM:"<<i<<" ";
      CLanguageModel::m_languageModels[i]->PrintContextSequence(lmContext);
    }
  }

  static void DummyLookupSequenceAllLMs(const CWordSequence& sentence) {
    for (int i=0; i<CLanguageModel::m_languageModels.size(); i++){
      CLanguageModel::m_languageModels[i]->DummyLookupSequence(sentence);
    }
  }
};

#endif // __LANGUAGEMODEL__
