/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#ifndef __MetricSearch__
#define __MetricSearch__
//base interface to scoring metrics that can be
//used to evaluate an n-best list
#include <vector>
#include "MetricRaw.hh"

#include "CompactTSequence.hh"
#include "Vocabulary.hh"
#include "Types.hh"
#include <iostream>
#include <fstream>
#include <sstream>
#include "pstream.h"




//! \var typedefs for single dimensional scores
typedef double Score;

//! \var typedef for a series of returned scores, useful for things like length penalty
typedef vector<Score> AbstractScore;

//! \var maps a score to an index into the n-best list
typedef multimap<Score, int > ScoreSort;

// \var Used as a lookup table for scores
typedef multimap<int, Score > IndexSort;

//! \var A sequence of word indices used to represent translations
typedef std::vector<CWordIndex> VocabVector;

//! \var RawCountsVector defined for use in this class
typedef vector< vector< RawCounts* >* > AllSentRawCountsVector;

/*! \var A configuration is a vector of int that refer to a choice within the nbest list
 *for each source translation. For example configuration [0, 5, 3] would represent
 *picking candidate translation 0, 5, 3 for the first three source sentences.
 */
typedef int* Configuration;


/*! Interface class for scoring object, defines required functionality and provides utility methods
 * Provides utility functions that provide the foundation for any scoring object, and defines and interface that scoring objects must follow. The most important general functionality that MetricSearch provides is the Oracle score calculation that it performs greedily on a vector of RawCount objects.

The MetricSearch interface works with a RawCounts object (defined in MetricRaw.hh). The RawCounts object is set up to handle n-gram precisions, ie number of n-grams correct / n-grams suggested + a reference length field. This format is specifically usefull for bleu/nist scoring, but can
be easily overloaded to support most metrics. This MetricSearch object specified the interface for specific scorer's that can take in RawCounts
object and output a Score
*/
class MetricSearch{

protected:

  //================================================================
  // Utility functions that are used by methods in MetricSearch
  //================================================================
  //void getCounterDifference(Counter v1, Counter v2, Counter& diff, int N);

  //! Returns a new vector representing a [0..0] configuration
  Configuration getInitialConfiguration();

  //Returns the raw counts for a particular configuration
  RawCounts getCountsForConfiguration(const AllSentRawCountsVector& nBestLists, Configuration config);

  //!From an existing configuration, find the top score for this sentence is we could pick
  //any possible translation for this sentence
  std::pair<double, RawCounts> singleSentenceOracle(RawCounts raw, Configuration config, const AllSentRawCountsVector& nBestLists, int sentenceNumber);

  //pair<double, RawCounts> CalculateMetricNew(RawCounts raw, Counter correctDelta, Counter suggestDelta, double closestLenDelta);


  //!Total number of sentences - set after constructor
  int m_numSentences;
  //! Total number of references - set after constructor
  int m_numReferences;

  //!Set of references - build during object construction
  std::vector< vector<VocabVector> > m_referenceBank;

  //!Vocabulary to convert word indices to CWordIndex
  CVocabulary* m_pVoc;


  //!Ngram length that is considered when scoring the RawCounts object. The contant NSZ
  //defined in the MetricRaw.hh file determines the max number of n-gram pairs the the
  //RawCounts object contains.
  int m_nGramConsidered;

public:


  MetricSearch();

  //process and n-best list and stores the count data to calculate scores
  //within the NBest list object
  //virtual void ProcessNBestList(CNBest* oneSourceList, int sentenceNumber)=0;

  //================================================================
  // An extending class must be able to calculate score given counts
  //================================================================

  //!Takes a single translation output and fills its scores into the RawCounts, can only be done by a real instance
  virtual void ScoreSingleHyp(const int sentenceNumber, samtstring candidate, RawCounts& countObject)=0;

  //! Returns the single dimensional score given a RawCounts object
  virtual Score getMetricScore(RawCounts& raw)=0;

  //! Returns the potentially multi dimensional score given a RawCounts object
  virtual AbstractScore getMetricScoreDetailed(RawCounts& raw)=0;

  //!Return the max possible for this scoring metric. This is used to convert the metric score into an error for optimization routines
  virtual Score maxScore()=0;

  //!Minimal differences that is considered a difference between two Scores for the specific error metric
  virtual Score minEpsilon()=0;


//the base implementation doesnt actually do anything here
  //virtual void ScorePairwise(CTranslationItem& hyp, CTranslationItem& ref, RawCounts& raw);

  //match against a set of references, returning the one with the best pairwise score
  //virtual void ScorePairwise(CTranslationItem& hyp, std::vector<CTranslationItem&> refList, RawCounts& raw);

  //virtual AbstractScore ScoreTranslationRun(vector<samtstring>& translationSet, vector<CNBest*>& nBestLists);

  //!Takes the whole set of NBest lists and iteratively selects the best possible score starting from
  //and initial configuration
  virtual Configuration GetOracleConfiguration(const AllSentRawCountsVector& nBestLists);
  //!Takes a configuration and counts up the appropriate RawCounts objects and calculates the metric score
  virtual AbstractScore ScoreConfiguration(const AllSentRawCountsVector& nBestLists, Configuration conf);


  //local scoring between candidates for each sentence, based on some set of raw counts
  //virtual ScoreSort ScoreLocalAlternatives(vector<CTranslationItem>& allTranslations, RawCounts initCounts);
  //local scoring between candidatates for one sentence, with 0 initialization
  //virtual ScoreSort ScoreLocalAlternatives(vector<CTranslationItem>& allTranslations );
  //virtual ScoreSort ScoreLocalAlternatives(CNBest* nBest);

  //global scoring initialized at the oracle configuration
  //virtual ScoreSort ScoreLocalAlternatives(vector<CNBest*>& nBestLists, Configuration conf, int sentenceNumber);

  //================================================================
  // Basic functionality provided by MetricSearch itself
  //================================================================


  // Initialize a raw counts object with 0's in all fields - deprecated in favor of constructor of RawCounts
  //RawCounts initRawCounts(int N);

  //! Incerment a RawCounts object, now deprecated in favor of the + operator on RawCounts
  //void IncrementRawCounts(RawCounts& rawCounts, int N, Counter* correct, Counter* suggested, double* refLengths);

  //This function does the basic work of
  virtual void BuildReferences(const char* fileName, redi::pstream& pbuf, bool normalizeReference, const int numReferences, CVocabulary* pVoc);
  //!Returns the corresponding metric (cause there can be multiple references)
  std::vector<CWordIndex>& GetReference(int sentenceNumber, int referenceNumber=0);

  //!Returns the set of references that are stored in this object
  std::vector< std::vector<CWordIndex> >& GetAllReferences(int sentenceNumber);

  //! Return a scaled artificial reference length for the given sentence
  double GetAverageSentenceLength(int sentenceNumber){
    std::vector< std::vector<CWordIndex> >& references = GetAllReferences(sentenceNumber);
    double sum = 0.0;
    for (std::vector< std::vector<CWordIndex> >::iterator iter = references.begin(); iter!= references.end(); ++iter){
      sum += (*iter).size();
    }
    double avgLength = sum/(double)references.size();
    return avgLength;
  }



  //returns a compact representation for the ngrams contained in the source index sequence upto the size maxNSize
  CCompactTSequence<int>* CollectNGramsBase(std::vector<int> source, int maxNSize);
  void SetVocabulary(CVocabulary* pVoc){m_pVoc = pVoc;}
  void SetNGramConsidered(int n){m_nGramConsidered = n;}
  int GetNGramConsidered(){return m_nGramConsidered;}

  //!Utility method to convert a samtstring to a vector of word indices, now deprecated in favor of MakeSequence in the Vocabulary class
  vector<CWordIndex> stringToVocabVector(samtstring str);


  //!Counts the total number of references stored here
  int TotalNumReferences(){
    int total = 0;
    for (int i=0; i<m_referenceBank.size(); i++){
      vector<VocabVector>& v = m_referenceBank[i];
      total+=v.size();
    }
    return total;
  }
};

#endif
