/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#include "NBestList.hh"
#include <sstream>
#include <iostream>



NBestList::NBestList(){
  m_pTranslations = NULL;
  m_pRawCounts = NULL;
}

NBestList::~NBestList(){
  if (m_pTranslations!=NULL){
    m_pTranslations->clear();
    delete m_pTranslations;
  }
  if (m_pRawCounts){
    m_pRawCounts->clear();
    delete m_pRawCounts;
  }

  m_featureMapDirect.clear();
}


int NBestList::InsertCandidate(LogProbList features, CWordSequenceVector translation, RawCounts raw){

  //get a coder object to hash the translation sequence
  GetTCode coder;
  //get its hash code
  TCode code = coder(translation);

  //set up a lookup object of the LogProbList and the pair
  pair<LogProbList*, TCode> lookup(&features, code);

  //use direct hashing
  //cout<<"Hashing: ";
  //copy(features.begin(), features.end(), ostream_iterator<float>(cout, " "));
  //cout<<" CODE: "<<code;
  
  //see if this translation already exists
  if (m_featureMapDirect.count(lookup)==0){

    //do the busy work of adding the data to the appropriate lists
    LogProbList* added = InsertCandidateGuaranteed(features, translation, raw);
    pair<LogProbList*, TCode> toAdd(added, code);
    m_featureMapDirect.insert(toAdd);
    //cout<<" ADDED"<<endl;
    return 1;
  }else{
    //cout<<" COLLIDED"<<endl;
    return 0;
  }

}


//guaranteed addition to the nbest list
LogProbList* NBestList::InsertCandidateGuaranteed(LogProbList features, CWordSequenceVector translation, RawCounts raw){
      //cout<<"Newing: "<<s<<endl;
  int latestIndex = -1;
  if (m_pTranslations!=NULL){
    m_pTranslations->push_back(translation);
    latestIndex = m_pTranslations->size()-1;
  }
  if (m_pRawCounts!=NULL){
    
    m_pRawCounts->push_back(new RawCounts(raw));
    latestIndex = m_pRawCounts->size()-1;
  }
  if (m_pRawCounts!=NULL && m_pTranslations!=NULL){
    assert(m_pRawCounts->size() == m_pTranslations->size());
  }
  
  LogProbList* fCopy = new LogProbList(features);    
  m_pLogProbs->push_back(fCopy);
  return fCopy;
}

  //load up the details from the nbest list
void NBestList::LoadOptimizationData(vector<LogProbList*>& allFeats, vector<RawCounts*>& allScores, vector<int>& allCands){

  /*
  allCands.push_back(m_featureMap.size());
  for (FeatureMap::iterator iter = m_featureMap.begin(); iter!=m_featureMap.end(); ++iter){
    LogProbList* featurePtr = (*iter).first;
    int dataIndex = (*iter).second;
    RawCounts* raw = (*m_pRawCounts)[dataIndex];
    allFeats.push_back(featurePtr);
    allScores.push_back(raw);
 }
  */
  allFeats.insert(allFeats.end(), m_pLogProbs->begin(), m_pLogProbs->end());
  allScores.insert(allScores.end(), m_pRawCounts->begin(), m_pRawCounts->end());
  assert(m_pRawCounts->size()==m_pLogProbs->size());
  allCands.push_back(m_pLogProbs->size());
}

void NBestList::PrintOptimizationData(const int sentenceNumber, const int NGramsConsidered, ostream& osFeats, ostream& osCands){
  //print number of candidates
  /*
  osCands<<sentenceNumber<<" "<<m_featureMap.size()<<endl;
  //print number of features
  assert(m_pRawCounts!=NULL);
  for (FeatureMap::iterator iter = m_featureMap.begin(); iter!=m_featureMap.end(); ++iter){
    LogProbList* features = (*iter).first;
    int dataIndex = (*iter).second;
    //print the log prob list
    for (int i=0; i<features->size(); i++){
      osFeats<<(*features)[i]<<" ";
    }
    //now for the scores
    RawCounts* raw = (*m_pRawCounts)[dataIndex];
    for (int n=0; n<NGramsConsidered; n++){
      osFeats<<raw->numCorrect[n]<<" "<<raw->numSuggested[n]<<" ";
    }
    osFeats<<raw->referenceLengths<<endl;
  }
  */

  assert(m_pLogProbs->size()==m_pRawCounts->size());
  osCands<<sentenceNumber<<" "<<m_pLogProbs->size()<<endl;
  for (int i=0; i<m_pLogProbs->size(); i++){
    LogProbList* features = (*m_pLogProbs)[i];
    RawCounts* raw = (*m_pRawCounts)[i];
    for (int f=0; f<features->size(); f++){
      osFeats<<(*features)[f]<<" ";
    }
    for (int n=0; n<NGramsConsidered; n++){
      osFeats<<raw->numCorrect[n]<<" "<<raw->numSuggested[n]<<" ";
    }
    osFeats<<raw->referenceLengths<<endl;
  }
}


void NBestList::ClearAll(){
  if (m_pTranslations!=NULL){
    m_pTranslations->clear();
  }
  if (m_pRawCounts){
    m_pRawCounts->clear();
  }

  for (int i=0; i<m_pRawCounts->size(); i++){
    RawCounts* ptr = (*m_pRawCounts)[i];
    delete ptr;
  }
  
  /*
  for (FeatureMap::iterator iter = m_featureMap.begin(); iter!=m_featureMap.end(); ++iter){
    LogProbList* features = (*iter).first;
    features->clear();
    delete features;
  }
  */

  m_featureMapDirect.clear();

  for (int i=0; i<m_pLogProbs->size(); i++){
    LogProbList* ptr = (*m_pLogProbs)[i];
    delete ptr;
  }
  m_pLogProbs->clear();

}

//!Returns the number of elements in the set
int NBestList::Size(){
  return m_featureMapDirect.size();
}
