/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#ifndef __NBESTLIST__
#define __NBESTLIST__

#include "MetricRaw.hh"
#include "Vocabulary.hh"



#include <list>
#include <vector>
#include <map>

#include <hash_map>
#include <hash_set>
#include  <string>
#include <inttypes.h>


typedef vector<CWordIndex> CWordSequenceVector;

//!A list of translations
typedef vector<CWordSequenceVector> TranslationVector;

//!A vector of log probs, need a vector for [x] operations
typedef vector<float> LogProbList; 

//!A vector of RawCounts
typedef vector< RawCounts* > RawCountsVector;

//!A vector of LogProbsList
typedef vector< LogProbList* > LogProbsVector;

//!An index into a the RawCountsVector
typedef int ExternalIndex; 

struct eqcrope
{
  bool operator()(const samtstring& s1, const samtstring& s2) const
  {
    
    if (s1==s2){
      return true;
    }else{
      return false;
    }
  }
};

typedef int TCode; //!Precomputed hash code for a translation
typedef pair<LogProbList*, TCode> LogTransPair;

//! Returns a hashed value for the translation sequence
struct GetTCode{
  size_t operator()(const CWordSequenceVector& t)const{
    int hashCode = 1;
    for (CWordSequenceVector::const_iterator iter = t.begin(); iter!=t.end(); ++iter){
      hashCode = 31*hashCode + (*iter);
    }
    return hashCode;
  }
};

/*!Returns a hash code for the pair LogProbList* and TCode (where TCode is the hash code for the translation sequence)
 * Note, we use the TCode externally here because we dont want to store the whole translation sequence
 */
struct TransVFloatHashCode{
  size_t operator()(const LogTransPair& p) const{
    LogProbList* f = p.first;

    int hashCode = 1;
    for (LogProbList::const_iterator iter = f->begin(); iter!=f->end(); ++iter){
      float value = (*iter);
      unsigned int v = *((unsigned int*)&value);
      hashCode = 31*hashCode + v;
    }
    //add the hashcode from the translation
    hashCode += p.second;

    return hashCode;
  }
};

/*! Equals operator for the pair LogProbList and TCode (see HashCode version for more details)
 */
struct TransVFloatHashEquals{
  bool operator()(const LogTransPair& p1 , const LogTransPair& p2) const{

    const LogProbList* f1 = p1.first;
    const LogProbList* f2 = p2.first;

    for (int i=0; i<f1->size(); i++){
      if ((*f1)[i]!=(*f2)[i]){
	return false;
      }
    }
    
    //if their trans hash code is the same too
    if (p1.second != p2.second){
      return false;
    }

    return true;
  }
};

//a direct hash for the features and TCodes
typedef hash_set<LogTransPair,  TransVFloatHashCode,  TransVFloatHashEquals> FeatureMapDirect;



//!A basic n-best list for 1 source sentence
/**!
 * Keeps features separate (across iterations) - dont worry about the case where 2 uniq trans have exactly the same features (vrare)
 * 
 */
class NBestList{

public:
  //!Just allocate memory for feature map
  NBestList();

  //!delete any non null newed objects
  ~NBestList();

  //!Decide what we want to keep track of in this nbest list
  void Init(bool keepTranslations, bool keepCounts){
    if (keepTranslations){
      m_pTranslations = new TranslationVector;
    }
    if (keepCounts){
      m_pRawCounts = new RawCountsVector;
    }

    //always get this
    m_pLogProbs = new LogProbsVector;
  }

  //!Reserve memory for the n-best lists
  void AllocateNBest(int N);
  
  //! Trim the n-best list down to use only what is required
  void TrimDown();

  //! The basic n-best list can be used in 3 ways, depending on what information we want to associate with the uniqe feature set
  // You pick your poison, either you store the translation or you store the raw counts
  int InsertCandidate(LogProbList features, CWordSequenceVector translation, RawCounts raw);

  //! This version will ALWAYS add new data to the nbest list
  LogProbList* InsertCandidateGuaranteed(LogProbList features, CWordSequenceVector translation, RawCounts raw);
  //void InsertCandidate(LogProbList features, VocabSequence translation);
  //void InsertCandidate(LogProbList features, RawCounts raw);

  //!Write the relevant data for the MER optimizer to the output streams
  void PrintOptimizationData(const int sentenceNumber, const int NGramsConsidered, ostream& osFeats, ostream& osCands);

  //load up the details from the nbest list
  void LoadOptimizationData(vector<LogProbList*>& allFeats, vector<RawCounts*>& allScores, vector<int>& allCands);

  //! clear used when we arent doing optimization
  void ClearAll();

  //!Returns the number of elements in the set ie number of candidates to select from
  int Size();

  //!Returns the RawCountsVector for direct use (be careful with this data)
  RawCountsVector* GetRawCounts(){
    return m_pRawCounts;
  }

protected:

  //Keep track of whether an entry has been seen before (based on a vector hash)
  FeatureMapDirect m_featureMapDirect; 

  //!Initialize this IF you want to keep track of actual translations
  TranslationVector* m_pTranslations;
  //!Initialize this IF you want to track of raw count data
  RawCountsVector* m_pRawCounts;
  //!This come initialized automatically
  LogProbsVector* m_pLogProbs;
  
};

#endif
