#ifndef __ParserInfo__
#define __ParserInfo__

#include "Bool.hh"
//#include <Array.hh>
#include "Vocabulary.hh"
//#include <Text.hh>
#include "Types.hh"

#include "Rule.hh"
#include "ParsingHeader.hh"

#include <cfloat>
#include <climits>
#include <vector>
#include <iostream>
#include <sstream>
#include <iterator>
#include <set>
#include <stdint.h>



typedef vector<int> CWordSequence;
//!<Pruning options for each NT
typedef hash_map<CWordIndex, pair<double, double> > NTPruningMap;

struct CNTOccurrenceStatistics {
  double resultCountPosteriorBased;
  double resultCountSimpleFrequency;
  double logSourceSpanLengthMean;
  double logSourceSpanLengthVariance;
  CNTOccurrenceStatistics() {}
  CNTOccurrenceStatistics(double a, double b, double c, double d) :
      resultCountPosteriorBased(a),
      resultCountSimpleFrequency(b),
      logSourceSpanLengthMean(c/a),
      logSourceSpanLengthVariance(MAX(d/a,0.0001))
  {}
};

//! Maps each NT to its occurrence statistics
typedef hash_map<CWordIndex, CNTOccurrenceStatistics > NTOccurrenceStatisticsMap;


/*! if we see a number tag AND we have no back pointers, then we take the following steps
 * - replace the m_translationOutput with the production result (since in UnwindHyp we directly calculate LM costs based on m_translationOutput)
 * - fill the m_taggedTargets with the pItem->m_translationOutput (these are the words that will finally be subbed in)
 * support for: src_5 -> tgt_5 -> @NUM (this rule will get matched directly on the source side and will enter this phrase)
 *          : @NUM foreign_pt @NUM -> @NUM.@NUM -> @NUM (translation output contains @NUMs, so we dont have to do anything)
 */
typedef hash_set<CWordIndex> TagSet;

/*!
 * Class that allows us to use static variables for Parser setting in the Hyp classes
 * Goal is to separate the hyp and the parser class so that components can use the hyps
 */
class CChartParserInfo{

public:


  //! For CChartParserHyp
  static vector<double> m_scalingFactors;

  //! For CChartParser
  static bool m_GlueAndEOSAssumptions;
  static double FutureCostImpact;
  static int BeamSizeFactor;
  static int m_sourceLength;
  static CVocabulary* m_pVocabulary;
  //! history length assumed for the language models: this will determine the length of the first-words and last-words of a hyp's m_lmContext and thus the granularity of the equivalence classes. This value needn't necessarily correspond to the true Markov-order of the LMs.
  static int C_HistoryLength;

  //! actual history length of the SRI LMs employed (e.g.: 2 for trigram LM)
  static int C_SRI_HistoryLength;

  //! True iff hyps are separated by first-C_HistoryLength and last-C_HistoryLength target words in the CChartParserNode equivalence classes
  static bool m_compareTargetWords;

  //! Not used currently - has to be set to 1
  static bool m_lmInCoarse;

  static int SAMTLM_UseBERFormat;

  //! Recompute LM costs from scratch during N-best extraction whenever a new hypothesis is formed by slurping in backpointed sub-hypotheses - see also help on --RecomputeLMCostsDuringKBest parameter
  static int m_recomputeLMCostsDuringKBest;
  //! Recompute LM costs after K-best retrieval - see also help on --RescoreLM
  static int m_rescoreLM;

  //static CSentence* m_pSentence;
  static CWordSequence m_pSentence;

  static int MAX_HYP_COUNT;
  static int m_maxCombinationCount;
  //! Word index for the whole sentence non-terminal
  static CWordIndex m_FullSentenceNT;
  static CWordIndex m_GlueNT; //!< not used anymore
  static CWordIndex m_unknownWI;
  static CWordIndex m_lmBoundaryWI;
  static CWordIndex m_bosWI;
  static CWordIndex m_eosWI;

  //combo pruning
  static float m_comboPruningBeamSize; //!< used in the spawning process
  static float m_comboPruningMinBeamSize; //!< used in the spawning process
  static float m_comboPruningFuzzCostDifference; //!< used in the spawning process
  static bool m_comboPruningGlobal; //!< does combo pruning across all incomplete hyps
  static bool m_comboPruningUseLM; //!< use the LM during combo pruning
  static int m_updateBeamSize; //!< update the beam size to allow BeamSizeFactor active hyps
  //  static bool m_comboPruningUseLexical; //!< use the Lexical hyps in the same beam during combo pruning

  //rule app count operations to simulate how it used to work before
  static int m_updateHypRuleAppCount; //update the hyp's rule app count with the fine-grained eq class counts
  static int m_lowestRuleAppCount; //use the lowest rule app count instead of the count of the lowest cost hyp
  static int m_multiplicativeRA; //use the rule app count as a multiplicative factor

  //operation conditions
  static int m_verbosity; //!< how verbose is the output

  static int m_stackLevel;
  static bool m_noTranslationOutputOfUnknownWords;

  //NBest extraction options
  static bool m_extractUnique; //!< Extract unique elements only while searching the n-best list
  static int m_extractRecombine; //!< Re-combine items as you do the n-best extraction, only applicable in cmp=0
  static int m_nbestStyle; //!< NBest style 1==A* 2==HC05 3==A* recombined + HC05
  static double m_aStarBeam; //! factor for inclusion of elements in A* search
  static double m_astarFuzz; //! fuzz factor to terminate search

  static int m_globalLMCache; //! if true, then never clear the LM cache
  static int m_dontReloadBeam; //! if true, then dont re-load the beam with a hyp that improves in astar - let hc05 deal with it
  static int m_saveAllAlternatives; //! if true, then we save all alternatives in the eq class (uses much more memory)
  static int m_breadthAStar;

  static TagSet m_tagsForLM; //!< List of tags for the LM
  static int m_normalizeReferences; //! Normalize references with the normalizing script

  //decoding statistics -- see SpawnKBestCompleteHyps for details
  static int m_hypsAdded, m_hypsDeleted, m_hypsDeletedThreshold, m_hypsDeletedHistogram, m_numTopCalls, m_numTuplesProposed, m_allPossibleExpansions, m_numAddedPostFailure;
  static double m_spawntimecpu; // never reset this, its for the whole run

  //n-best extraction statistics -- see RetrieveKBestParses for details
  static int m_noCollisions, m_uniqueCollisions, m_betteredCollisions, m_worseCollisions, m_proposedAddedKBest, m_topCallsKBest, m_deletedKBest, m_numUnique, m_childBetter;

  //lm call statistics
  static uint64_t m_lmCall, m_lmCacheHit;

  // Set directly as parameters
  static string UnknownWord;
  static string FullSentNonTerminal;
  static string NormalizingScript;
  static bool RemoveUnt;
  static int HypsPerNode;
  static int MaxCombinationCount;
  static int MaxRuleAppCountDifference;
  static int CompareTargetWords;

  static string TagSetList;
  static string ComboPruningBeamSize;
  static string ComboPruningMinBeamSize;
  static string ComboPruningFuzzCostDifference;
  static bool ComboPruningGlobal;
  static bool ComboPruningUseLM;
  static string MaxCostDifferencePerCell;
  static string AStarBeam;
  static string PruningMap;
  static bool PruningMapUseBoundary;
  static int NGramsConsidered;
  static int NBest;
  static int DisplayNBestData;
  static bool GenOracleScores;
  static int IterationLimit;
  static bool ExtractUnique;
  static int Verbosity;
  static int NormalizeReferences;
  static bool ExtractRecombine;
  static int NBestStyle;
  static int RescoreLM;
  static int RecomputeLMCostsDuringKBest;
  static bool RA_UpdateHyp;
  static bool RA_LowestCount;
  static bool RA_Mult;
  static bool GlobalLMCache;
  static bool UseSpanFeatures;
  static int BreadthAStar;
  static bool DontReloadBeam;
  static bool SaveAllAlternatives;
  static bool UpdateBeamSize;
  static int HistoryLength;

  static void ResetNBestStatistics(){

    //all of these collision options add up to the m_proposedAddedKBest
    CChartParserInfo::m_noCollisions = 0;
    CChartParserInfo::m_uniqueCollisions = 0;
    CChartParserInfo::m_betteredCollisions = 0;
    CChartParserInfo::m_worseCollisions = 0;

    CChartParserInfo::m_proposedAddedKBest = 0;
    CChartParserInfo::m_topCallsKBest = 0;
    CChartParserInfo::m_deletedKBest = 0;
    CChartParserInfo::m_childBetter = 0; //if you child is better than you in A* unwind

    CChartParserInfo::m_numUnique = 0;
  }

  static void PrintNBestStatistics(){
    cout<<"m_numUnique: "<<m_numUnique<<" m_topCallsKBest: "<<m_topCallsKBest<<" m_proposedAddedKBest: "<<m_proposedAddedKBest<<" (m_noCollisions: "<<m_noCollisions<<" m_uniqueCollisions: "<<m_uniqueCollisions<<" [m_betteredCollisions: "<<m_betteredCollisions<<" m_worseCollisions: "<<m_worseCollisions<<" ] m_deletedKBest: "<<m_deletedKBest<<" m_childBetter: "<<m_childBetter<<endl;
  }

  static void ResetCellStatistics(){
    CChartParserInfo::m_hypsAdded = 0;

    CChartParserInfo::m_hypsDeleted = 0; //total number of hyps deleted when finalizing the node
    CChartParserInfo::m_hypsDeletedThreshold = 0; //total number of hyps deleted when finalizing the node due to threshold
    CChartParserInfo::m_hypsDeletedHistogram = 0; //total number of hyps deleted when finalizing the node due to threshold

    CChartParserInfo::m_numTopCalls = 0;
    CChartParserInfo::m_numTuplesProposed = 0;
    CChartParserInfo::m_allPossibleExpansions = 0;
    CChartParserInfo::m_numAddedPostFailure = 0;
  }

  //DECODER STATE
  //static bool s_maxCombinationCountExceeded; //monotone decoding has begun //arv - dont need this

  //! For Pruning with CompareTargetWords==1
  static NTPruningMap m_pruningMap;
  static double m_maxCostDifferencePerCell; //beam perc for production nodes

  //! Whether to use the backed off first words to help pruning equivalence classes
  static bool m_pruningMapUseBoundary;


  static int m_useSpanFeatures;

  //! Info about all result NTs occurring in the rule database - set up by CBTree constructor
  static NTOccurrenceStatisticsMap m_nonterminalOccurrenceStatsMap;






  //!Loads the static pruning map from the samtstring whose format is "NT-b-beta", and -1 is the generic NT
  static void LoadPruningMap(const char* options);

  //! Processes a float value from a samtstring, taking into account 'inf' for infinity
  static float StringToFloat(const char* s);
  static float StringToFloat(samtstring s);

  //! General conversion routing from samtstring to another type
  template <class T>
  static T from_string(const samtstring& s) {
    std::istringstream iss(s);
    T t;
    if ((iss >> std::dec >> t).fail()) {
      cerr << "Error: Could not convert samtstring '" << s << "' to number." << endl;
      exit(2);
    } else {
      return t;
    }
  }


  //!Load the tags that will be considered directly for the language model
  static void LoadTagsForLM(const char* options);

  //!Defines a pruning strategy across the parsing process given a top hyp
  static float GetPruningCutoff(float lowestCost, double beamValue ) {
    //if (lowestCost>0.0){
    ///double multFactor = 1.0/(1.0-beamValue);
    //return (multFactor)*lowestCost;
    //}else{
    //double beam =  -log(1.0-beamValue);
    //cout<<"Beam is: "<<beam<<" beamValue: "<<beamValue<<endl;

    //  if (beamValue==1.0){
    //    return INT_MAX;
    //  }
    //ARV 10/02/2006 - make the effect of pruning a function of the scaling factors used to make it more independent of these

    double multFactor = 0;
    for (int i=0; i<CRule::C_NumLMs; i++) multFactor += CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i];
    //  return -log(1.0-beamValue)*multFactor + lowestCost;
    //    cout << "GetPruningCutoff: C_NumLMs=" << CRule::C_NumLMs << ",C_LMLocation=" << CRule::C_LMLocation << ",lowestCost=" << lowestCost << ",beamValue=" << beamValue << ",cutOff=" << beamValue*multFactor+lowestCost << endl;
    return beamValue*multFactor + lowestCost;
  }

  static float GetPruningCutoffPerWord(float lowestCost, double beamValue, int numWords ) {
    //  if (beamValue==1.0){
    //    return INT_MAX;
    //  }
    double multFactor = 0;
    for (int i=0; i<CRule::C_NumLMs; i++) multFactor += CChartParserInfo::m_scalingFactors[CRule::C_LMLocation+i];
    //  return -log(1.0-beamValue)*multFactor*(double)(numWords) + lowestCost;
    return beamValue*multFactor*(double)(numWords) + lowestCost;
  }


   //! Prints out the parsing output sequence
  /*!
   * Differs from PrintOutputSequence in CChartParserInfo in the way it annotates the NT with the src span
   * ie #( @NT 1 2 # le chat #) # the cat -> ( @NT-1-2_le_chat the cat )
   * allowing parse viewers to tag the span
   */
  static void PrintParsingOutput(const CWordSequence& v, ostream& os=cout);
  static void PrintParsingOutput_SafePrint(const int i, ostream& os);
  //! Prints a sequence of vocab words with spaces inbetween
  static void PrintVocabSequence(const CWordSequence& v, ostream& os=cout);

  //uses the vocab to convert a samtstring to a vector of word indices
  template <typename Sequence>
  static Sequence makeSequence(samtstring s){
    stringstream ss(stringstream::in | stringstream::out);
    ss<<s;
    ss.flush();
    Sequence sq;
    while (!ss.eof()){
      samtstring oneWord;
      ss>>oneWord;
      //cout<<"oneWord:"<<oneWord<<endl;
      //CString oneWordCString = oneWord.c_str();
      //CWordIndex wordIndex = m_pVocabulary->GetIndex(oneWordCString);
      CWordIndex wordIndex = m_pVocabulary->GetIndex(oneWord);
      sq.insert(sq.end(), wordIndex);
    }
    return sq;
  }

  //!move the output to align nicely
  static void TabStack(){
    for (int i=0; i<m_stackLevel; i++) cout<<"\t";
  }

  static void BERPack(unsigned int unsignedInteger, vector<uint8_t>& byteSeq) {
    assert(unsignedInteger <= INT_MAX);
    if (unsignedInteger==0) {
      byteSeq.push_back(0);
    } else {
      vector<uint8_t> reverseByteSeq;
      //      int shiftAmount = sizeof(unsignedInteger) - 8;
      //      for (int i=0; unsignedInteger != 0; i++) {
      while (unsignedInteger != 0) {
	reverseByteSeq.push_back((unsignedInteger & 127) | 128);
	unsignedInteger >>= 7;
      }
      // add created BER sequence backwards to the byte vector
      byteSeq.insert(byteSeq.end(), reverseByteSeq.rbegin(), reverseByteSeq.rend());
      // mark final BER byte with hi-bit 0
      byteSeq.back() &= 127;
    }
  }

  template <typename UnsignedIntegerType>
  static UnsignedIntegerType BERUnpack(uint8_t*& byteSeqPointer) {
    if (*byteSeqPointer == 0) {
      //      cout << "zero" << endl;
      byteSeqPointer++;
      return 0;
    } else {
      //            cout << "nonzero:" << (int) *byteSeqPointer << endl;
      UnsignedIntegerType unsignedInteger = 0;
      while (true) {
	uint8_t nextByte = *(byteSeqPointer++);
	//	cout << (int) byteSeqPointer << "(" <<(int) nextByte << ")" << endl;
	unsignedInteger = unsignedInteger * 128 + (nextByte & 127);
	// break loop if hi-bit is not set
	//	cout << ((int) nextByte) << ((int) nextByte & 128) << endl;
	if ((nextByte & 128) == 0) break;
      }
      return unsignedInteger;
    }
  }
};

#endif
