/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/
#include "RunDecoder.hh"
#include "BleuSearch.hh"

void RunDecoder::InitializeDecoder(RuleSet* rules,
                                   const vector<CLanguageModel*>& clms,
                                   const vector<double>& Weights) {

  CChartParserInfo::m_verbosity = CChartParserInfo::Verbosity;
  CChartParserInfo::m_extractRecombine = CChartParserInfo::ExtractRecombine;
  CChartParserInfo::m_recomputeLMCostsDuringKBest =
      CChartParserInfo::RecomputeLMCostsDuringKBest;
  CChartParserInfo::m_rescoreLM = CChartParserInfo::RescoreLM;
  CChartParserInfo::m_updateHypRuleAppCount = CChartParserInfo::RA_UpdateHyp;
  CChartParserInfo::m_lowestRuleAppCount = CChartParserInfo::RA_LowestCount;
  CChartParserInfo::m_multiplicativeRA = CChartParserInfo::RA_Mult;
  CChartParserInfo::m_globalLMCache = CChartParserInfo::GlobalLMCache;
  CChartParserInfo::m_normalizeReferences =
      CChartParserInfo::NormalizeReferences;
  CChartParserInfo::m_useSpanFeatures = CChartParserInfo::UseSpanFeatures;
  CChartParserInfo::m_breadthAStar = CChartParserInfo::BreadthAStar;
  CChartParserInfo::m_dontReloadBeam = CChartParserInfo::DontReloadBeam;
  CChartParserInfo::m_saveAllAlternatives =
      CChartParserInfo::SaveAllAlternatives;
  CChartParserInfo::m_updateBeamSize = CChartParserInfo::UpdateBeamSize;
  CChartParserInfo::C_HistoryLength = CChartParserInfo::HistoryLength;

  CChartParserInfo::m_pVocabulary = new CVocabulary;
  CWordIndex unkWI = CChartParserInfo::m_pVocabulary->AddWord(
      samtstring(CChartParserInfo::UnknownWord));
  CWordIndex fullWI =  CChartParserInfo::m_pVocabulary->AddWord(
      samtstring(CChartParserInfo::FullSentNonTerminal));

  cerr<<"FullSentNonTerminalString: "<<
      CChartParserInfo::FullSentNonTerminal<<endl;


  CRule::pVoc = CChartParserInfo::m_pVocabulary;

  bool useNormalizing = false;


  // Set a normalizing script if we plan to use one
  if (CChartParserInfo::NormalizingScript != "") {
    cerr<<" *** SAMT : About to load normalizing script:"<<endl;
    samtstring perlNormalizer = //"perl " +
        samtstring(CChartParserInfo::NormalizingScript);
    pbuf.open(perlNormalizer, std::ios_base::out | std::ios_base::in);
    cerr<<" *** Loaded : "<<perlNormalizer<<endl;
    useNormalizing = true;
  }

  CChartParserInfo::m_pruningMapUseBoundary = (CChartParserInfo::PruningMapUseBoundary);

  CChartParserInfo::m_comboPruningGlobal = ( CChartParserInfo::ComboPruningGlobal );
  CChartParserInfo::m_comboPruningUseLM = ( CChartParserInfo::ComboPruningUseLM );

  if (CChartParserInfo::TagSetList!=""){
    CChartParserInfo::LoadTagsForLM(CChartParserInfo::TagSetList.c_str());
  }
  CChartParserInfo::m_comboPruningBeamSize = CChartParserInfo::StringToFloat(
      CChartParserInfo::ComboPruningBeamSize);
  CChartParserInfo::m_comboPruningMinBeamSize =
      CChartParserInfo::StringToFloat(CChartParserInfo::ComboPruningMinBeamSize);
  CChartParserInfo::m_comboPruningFuzzCostDifference =
      CChartParserInfo::StringToFloat(CChartParserInfo::ComboPruningFuzzCostDifference);
  CChartParserInfo::m_maxCostDifferencePerCell =
      CChartParserInfo::StringToFloat(CChartParserInfo::MaxCostDifferencePerCell);
  CChartParserInfo::m_aStarBeam = CChartParserInfo::StringToFloat(CChartParserInfo::AStarBeam);
  CChartParserInfo::LoadPruningMap(CChartParserInfo::PruningMap.c_str());

  ScalingFactors = Weights;
  CRule::C_NumModelsTotal= ScalingFactors.size();
  cerr<<" *** SAMT Setting CRule::C_NumModelsTotal: "<<CRule::C_NumModelsTotal<<endl;

  // Fill additional fields in the CChartParserInfo that deal with LMs
  CLanguageModel::m_languageModels = clms;
  // set n.o. LMs
  CRule::C_NumLMs = CLanguageModel::m_languageModels.size();
  // also shift the index of Glue and RuleAppCount models
  CRule::C_GlueLocation = CRule::C_NumLMs;
  CRule::C_RuleApplicationCountLocation = CRule::C_NumLMs+1;
  // set n.o. translation models (i.e., cost values per rule in the rule database)
  CRule::C_NumModelsTM = CRule::C_NumModelsTotal - CRule::C_NumLMs - (CChartParserInfo::m_useSpanFeatures? 2:0);
  // also set up the location of the source span penalty feature (this location is only valid if m_useSpanFeatures is true)
  CRule::C_SourceSpanPenaltyLocation = CRule::C_NumModelsTotal - 2;

  // Initialize the parser object
  cerr<<" *** SAMT : About to initialize parser"<<endl;
  parser = new CChartParser(CChartParserInfo::m_pVocabulary,
                            unkWI,
                            CChartParserInfo::RemoveUnt,
                            CChartParserInfo::MaxRuleAppCountDifference,
                            rules,
                            CChartParserInfo::HypsPerNode,
                            CChartParserInfo::MaxCombinationCount,
                            CChartParserInfo::FutureCostImpact,
                            CChartParserInfo::BeamSizeFactor,
                            CChartParserInfo::ExtractUnique);
  parser->SetFullSentenceNonTerminal(fullWI);

  //cout<<"MaxRuleAppCountDifference: "<<CChartParserInfo::MaxRuleAppCountDifference<<endl;
  //cout<<"MaxCombinationCount: "<<CChartParserInfo::MaxCombinationCount<<endl;
  //cout << "CompareTargetWords " << CChartParserInfo::CompareTargetWords << endl;
  parser->SetCompareTargetWords( (CChartParserInfo::CompareTargetWords==1) );
  parser->SetUseLMInCoarse( true );
}

void RunDecoder::RunMERonDev( samtstring ScoringMetric,
                              samtstring ReferenceList,
                              samtstring GainFile,
                              int NumReferences,
                              COptimize* pOptimizer,
                              vector<CWordSequence>& sourceCorpus,
                              vector< vector<double> >& featureRange ){

   // Initialize the scoring metric
   if ((ScoringMetric =="NIST")
       || (ScoringMetric == "NISTBLEU")
       || (ScoringMetric == "IBMBLEU")){

     CChartParserInfo::NGramsConsidered = 4;
     if (NSZ<4){
       CChartParserInfo::NGramsConsidered=NSZ;
       cerr<<"WARNING: Dropping NGramsConsidered for"
           "ScoringMetric to: "
           <<CChartParserInfo::NGramsConsidered<<" since NSZ = "<<NSZ<<endl;
     }
   }
   bool normalizeReference =
       (CChartParserInfo::m_normalizeReferences>0) ? 1 : 0;

   cerr<<"Initializing modelBestCounts to handle upto: "<<
       CChartParserInfo::NGramsConsidered<<"-grams"<<endl;

   MetricSearch* pMetric = new BleuSearch(
       ReferenceList.c_str(),
       pbuf,
       normalizeReference,
       GainFile.c_str(),
       ScoringMetric,
       NumReferences,
       CChartParserInfo::NGramsConsidered,
       CChartParserInfo::m_pVocabulary);

   pOptimizer->SetMetric(pMetric);

   //this object is properly initialized in the constuctor
   RawCounts modelBestCounts;
   //tracking for the top level n-best lists
    vector<NBestList*> nBestLists; //track them all
    vector< vector<NBestList*> > suppNBestLists; // won't be filling this
    vector<pair< vector<CChartParserHypComplete*>, double> > subSentenceNBestListAndAvg ;

    for (int i=0; i<sourceCorpus.size(); i++){
      NBestList* pNB = new NBestList;
      //dont keep any translations
      pNB->Init(false, true);
      nBestLists.push_back(pNB);
    }

    //keep track of scaling factors
    ScoreScalingFactorMap trackScalingFactorScores;
    bool MEROptimize = (pOptimizer!=NULL);
    bool MERComplete = false;
    //start the main MER loop
    bool exitTranslationLoop = false;
    int iteration = 0;
    //save sentence IDs for which no translation was found
    vector<int> noTranslationSentences;
    int sentenceNumber = 0;
    int TotalNewTranslations = 0;

    // =========================================================================
    // The main MER training loop
    // =========================================================================
    do {
      for (vector<CWordSequence>::const_iterator corpusIter =
               sourceCorpus.begin(); corpusIter != sourceCorpus.end(); ++corpusIter){

        cout<<endl<<endl<<endl<<"==============================================";
        cout<<"Sentence " << sentenceNumber << " (starting with 0): ";
	const CWordSequence& pSourceSentence = (*corpusIter);
	CChartParserInfo::PrintVocabSequence(pSourceSentence);

        parser->SetScalingFactors(ScalingFactors);
        // The key parsing step
	parser->InitSearch(pSourceSentence);
	parser->ParseChart();
        vector<CChartParserHypComplete*> fullSentenceNBestList =
            parser->RetrieveKBestParses(CChartParserInfo::NBest);
        if (fullSentenceNBestList.empty()){
	  cerr << "ERROR: No translation found for sentence "<<sentenceNumber<<endl;
	  cout << "ERROR: No translation found for sentence "<<sentenceNumber<<endl;
	  exit(-1);
	}

        // Update the scores for the first best sentence
        if (pMetric!=NULL) {
          RawCounts topTranslationRaw =
              CChartParser::BuildNBest(
                  pbuf,
                  nBestLists[sentenceNumber],
                  pMetric,
                  fullSentenceNBestList,
                  CChartParserInfo::NBest,
                  sentenceNumber,
                  m_useNormalizing,
                  TotalNewTranslations);

	  modelBestCounts = modelBestCounts + topTranslationRaw;
        }
        // If we were asked to display
        if (CChartParserInfo::DisplayNBestData>0){
	  parser->OutputNBest(
              fullSentenceNBestList.begin(),
              fullSentenceNBestList.end(),
              CChartParserInfo::DisplayNBestData);
          cout<<endl;
	}

        // if there is any need for global keeping of nbest
	// lists then we cant delete them immediately
	if ((CChartParserInfo::GenOracleScores || MEROptimize)){
	  //leave the nbest list alone!
	}else{
	  //BUT we dont clear the sub-sentence n-best lists...
	  nBestLists[sentenceNumber]->ClearAll();
	}
        CChartParser::DeleteNBestLists(fullSentenceNBestList,
                                       subSentenceNBestListAndAvg);
        parser->ClearSearch();
        sentenceNumber++;
      }
      // Now one whole dev set worth of sentences have been translated
      if (pMetric!=NULL){
        cout<<"Model best: ";
	for (int i=0; i<CChartParserInfo::NGramsConsidered; i++){
	  cout<<modelBestCounts.numCorrect[i]
              <<" / "<<modelBestCounts.numSuggested[i]<<" ";
	}
	cout<<" ref: "<<modelBestCounts.referenceLengths<<endl;
	cout<<endl;
        AbstractScore detailedScore = pMetric->getMetricScoreDetailed(
            modelBestCounts);
	assert(detailedScore.size()>0);
	double tscore = detailedScore[0];
	trackScalingFactorScores.insert(
            pair<double, vector<double> >(tscore, ScalingFactors));
	cout<<"TrainTranslationScore: ";
	copy(detailedScore.begin(), detailedScore.end(), std::ostream_iterator<double>(cout, " "));
	cout<<" for n.o. ScalingFactors: "<<ScalingFactors.size()<<endl;
	cout<<"TotalNewTranslations: "<<TotalNewTranslations<<endl;

	// this criteria is probably a bit too strict on its own
        // so we also check to make sure the metric score hasnt stagnated. we can
	// check this by looking at the score tracker
	if (CChartParserInfo::GenOracleScores){
	  AllSentRawCountsVector accumRawCounts;
	  for( vector< NBestList* >::iterator iter=nBestLists.begin(); iter!=nBestLists.end(); ++iter ) {
	    NBestList* pNB = (*iter);
	    accumRawCounts.push_back(pNB->GetRawCounts());
	  }

	  Configuration config = pMetric->GetOracleConfiguration(accumRawCounts);
	  cout<<"OracleTranslationScore: "<<
              pMetric->ScoreConfiguration(accumRawCounts, config)[0]<<endl;
	  delete [] config;
	}

        if (TotalNewTranslations==0){
          MERComplete = true;
	  cout<<"Potential early termination triggered since TotalNewTranslations==0"<<endl;
	}
      } else{
	cout<<"No metric search object was not initialized\n, were no references were provided?"<<endl;
      }
      cout<<"Done with my first pass!!!"<<endl;
      iteration++;
      // if there is optimization going on
      if (pOptimizer!=NULL){
	CChartParser::OptimizeParametersInternal(
            ScalingFactors,
            nBestLists,
            suppNBestLists,
            pOptimizer,
            featureRange,
            trackScalingFactorScores);
      }
      exitTranslationLoop = iteration >= CChartParserInfo::IterationLimit || !MEROptimize || MERComplete;

    }while(!exitTranslationLoop);

    delete pMetric;
  }

void RunDecoder::ShutdownDecoder() {
    // ===========================================================================
    // Clean up after translation is complete
  // ===========================================================================
  cout<<"All loops completed\n";
  if (m_useNormalizing){
    cout<<"Closing pstream\n";
    //ClosePBUF();
    if (pbuf.good()){
      pbuf.flush();
      pbuf.close();
    }
    cout<<"pstream closed\n";
  }
  cout<<"Exitting now\n";
  for (int i=0; i< CLanguageModel::m_languageModels.size(); i++) {
    delete CLanguageModel::m_languageModels[i];
  }
}
