/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#include "UniqueCompleteHyps.hh"
#include "ChartParserHyp.hh"


//============================================================================
// CChartParserHypComplete unique target samtstring class
//============================================================================

// add hashing terms with respect to the terminals and backpointers to the hashCode
void HashUniqueTargetsHyp::TerminalsAndBackpointers(const CChartParserHypComplete* pHyp, size_t& hashCode)const{
  for (CWordSequence::const_iterator iter=pHyp->m_translationOutput.begin(); iter!=pHyp->m_translationOutput.end(); ++iter){
    CWordIndex nextWord = *iter; 
    //if its a word
    if (nextWord>0){
      hashCode = 31*hashCode + nextWord;
    }else{
      hashCode = 31*31*hashCode + (long)(pHyp->m_backPointers[-1*nextWord]);      
    } 
  }
}

//returns a HashCode for the pHyp based on the addresses of the back pointers and the words translated so far
size_t HashUniqueTargetsHyp::operator()(const CChartParserHypComplete* pHyp)const{

  size_t hashCode = pHyp->m_completelyUnwound? 2 : 1; 
  //either you are cmp=1, or you are cmp=0 but are not doing eq classing
  //if (CChartParserInfo::m_compareTargetWords==1 || CChartParserInfo::m_extractRecombine==0){
  TerminalsAndBackpointers(pHyp, hashCode);    
  //}else{
  //assert(CChartParserInfo::m_compareTargetWords==0 && CChartParserInfo::m_extractRecombine==1);
  //LMContextAndBackpointers(pHyp, hashCode);    
  //}
  return hashCode;

}


CChartParserHypComplete* CUniqueCompleteHyps::CheckUnique(CChartParserHypComplete* pHyp){

#ifndef NDEBUG
   bool debug = false;
   if (debug){
     cout<<"Checking: "; cout<<" "<<CChartParserInfo::m_pVocabulary->GetWord(pHyp->GetResult())<<" -> "; CChartParserInfo::PrintVocabSequence(pHyp->m_translationOutput); cout<<" PARSE "; 
     CChartParserInfo::PrintParsingOutput(pHyp->m_parsingOutput); cout<<" ";
   }
#endif

   HashUniqueTargetsHyp hasher;
   size_t mycode = hasher(pHyp);
   if (m_uniqueHyps.count(mycode)==0){

#ifndef NDEBUG
     if (debug){       cout<<"Passed: " << endl;      }
#endif

     //ARV 09/23/06 just looking at pHyp->TotalCost() is fine when using the beam style approach, but not if we have left the LM as m_scaledFutureCostEstimate
     //az 06/11/05: changed to AllComponentCost() to include the LM boundary cost estimate
     //while this is conceptually more sensible, it doesn't have any practical effect since potential
     //hyps to be compared have the same translationOutput and therefore the same boundary LM estimate
     //m_uniqueHyps[mycode] = pHyp->TotalCost();
     //     m_uniqueHyps[mycode] = pHyp->ScaledTMLMCost();
     m_uniqueHyps[mycode] = pHyp->AllComponentCost();
     
     //no collision at all - immediate pass thru for now, a hyps might come in the future and kick another one out
     CChartParserInfo::m_noCollisions++;
   }else{
     
#ifndef NDEBUG
     if (debug){       cout<<"Collided: ";      }
#endif
     //did have a collision
     //CChartParserInfo::m_uniqueCollisions++;

     //this might be a junk hyp; if so, throw it away
    TUniqueHyps::iterator iter = m_uniqueHyps.find(mycode);
    double bestCost = (*iter).second;
    //pair<double, CChartParserHypComplete*>& existing = (*iter).second;
    
    //if (pHyp->TotalCost()<bestCost){
    if (pHyp->AllComponentCost()<bestCost){

#ifndef NDEBUG
      if (debug){	cout<<"Better: " << endl;       }
#endif

      //just save the new costs as the best costs now
      //m_uniqueHyps[mycode] = pHyp->TotalCost();
      m_uniqueHyps[mycode] = pHyp->AllComponentCost();

      //you collided and unseated another hyp
      //CChartParserInfo::m_betteredCollisions++;

    }else{

#ifndef NDEBUG
      if (debug){	cout<<"Worse: " << endl;       }
#endif

      //you collided and got unseated
      //CChartParserInfo::m_worseCollisions++;

      //DONT DELETE, if you do it becomes hard to memory manage 
      //delete pHyp;
      //pHyp = NULL;
      return NULL;
    }
  }
  return pHyp;
}
