/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// CVocabulary
// Base class for vocabulary handling
//

#include <stdlib.h>
#include <stdio.h>
#include "samtstring.hh"
#include <fstream>
using std::ifstream;
using std::ofstream;

//#include <utils.hh>
#include "Vocabulary.hh"

int
CVocabulary::Size() const
{
  return m_VocItems.size();
}



const samtstring& CVocabulary::GetWord ( CWordIndex WordIndex ) const
{
  if (!IsInRange(WordIndex)){
    cerr<<"Asked for vocab index: "<<WordIndex<<" not in range of voc"<<endl;
    exit(-1);
  }
  return m_VocItems[ WordIndex ];
} // GetWord

const samtstring* CVocabulary::GetWordPtr ( CWordIndex WordIndex ) const
{
  if (IsInRange(WordIndex)){
    return &(m_VocItems[ WordIndex ]);
  }else{
    return NULL;
  }
} // GetWord


CVocabulary::CVocabulary()
{

  InitBlank();
} // CVocabulary



CVocabulary::~CVocabulary()
{
  // delete the VocItems
  Clear();
} // ~CVocabulary



bool
CVocabulary::Allocate( int MaxNumOfVocItems )
{
  //m_VocItems.SetIncrement( MaxNumOfVocItems );
  m_VocItems.reserve(MaxNumOfVocItems);
  return true;
}



void
CVocabulary::Clear()
{

  m_VocItems.clear();
  m_VocIndices.clear();

}



bool
CVocabulary::Load( const char *FileName )
{
  ifstream InStream( FileName );

  if ( !InStream.good() ) {
      cerr <<  "CVocabulary::Load(): Vocabulary file " << FileName << " was not opened for read" << endl << flush;
      exit(1);
      //return false;
  } // endif: error opening file

  bool Result = Load( InStream );
  InStream.close();

  return Result;
} // Load

//initialize the vocabulary for use in translation
bool CVocabulary::InitBlank(){
  // add a dummy to make sure that word index 0 is not used
  int unkWI = AddWord("<DUMMY_061101bla>");
  assert(unkWI==0);
  return true;
}

bool
CVocabulary::Load( istream &InStream )
{

  // make sure no word has been added yet
  if (Size() != 1) {
    cerr << "Error in CVocabulary::Load: vocab isn't a virgin anymore. Make sure you call Load before any calls of AddWord / GetIndex." << endl;
    exit(1);
  }
  if ( !InStream.good() ) {
      cerr <<  "CVocabulary::Load(): Could not read from ifstream" << endl << flush;
      //      return false;
      exit(1);
  } // endif: error reading stream
  // skip comment lines
  //  InStream.getline( str, MAXBUF );
  //  while ( (str[ 0 ] == '#') && !InStream.eof() ) {
  //    InStream.getline( str, MAXBUF );
  //  }

  int wordsAddedCount=0;
  char buf[MAXBUF];
  buf[0] = '\0';
  for (; !InStream.eof(); ){
    InStream.getline(buf, MAXBUF);
    samtstring sourceString(buf);
    //    CVocabulary::string_trim(sourceString); // remove leading or trailing whitespace
    if (sourceString.size()>0){
      AddWord(sourceString);
      wordsAddedCount++;
    } else {
      InStream.getline(buf, MAXBUF);
      if (!InStream.eof()) {
	cerr << "CVocabulary::Load(): Error loading vocab file: empty (or whitespace-only) line in the file in line " << (wordsAddedCount+1) << endl;
	exit(1);
      }
    }
  }
  if (wordsAddedCount != Size()-1) {
    cerr << "CVocabulary::Load(): Error loading vocab file: at least one of the words in the file occurred twice!" << endl;
    exit(1);
  }
  return true;
} // Load



bool
CVocabulary::Store( const char *FileName ) const
{
  ofstream OutStream( FileName );

  if ( !OutStream.good() ) {
    cerr <<  "File was not opened for write\n";
    exit (-1);
  }

  //bool Result = Store( OutStream, WithSGML );
  bool Result = Store( OutStream);
  OutStream.close();

  return Result;
} // Store



bool
CVocabulary::Store( ostream &OutStream ) const
{
  if ( !OutStream.good() ) {
    cerr << "File could not be opened for write\n";
    exit (-1);
  }
  
  for (int i=0; i<m_VocItems.size(); i++){
    cout<<m_VocItems[i]<<endl;
  }

  /*

  CWordIndex index = 0;
  
  if ( WithSGML ) {
    OutStream << m_VocItems[ index ]->Word() << endl;
  }

  for ( index = 1; index <= Size(); index++ ) {
    OutStream << m_VocItems[ index ]->Word() << endl;
  }

  if ( WithSGML ) {
    OutStream << m_VocItems[ index ]->Word() << endl;
  }
  */

  return true;
} // Store



// ********************************************************************
// Search-Methods

const bool CVocabulary::IsInRange( CWordIndex WordIndex ) const{
  //if ((WordIndex>=m_VocItems.MinIndex()) && (WordIndex<=m_VocItems.MaxIndex())){
  if (WordIndex>=0 && WordIndex<m_VocItems.size()){
    return true;
  }else{
    return false;
  }
}

CWordIndex CVocabulary::GetIndex( const samtstring& Word ) {
  return AddWord(Word);
}

CWordIndex CVocabulary::AddWord(const samtstring & word) {
  //verify that the word does not exist
  if (m_VocIndices.count(word)==0){
    // doesn't exist - add it
    if (word.length() == 0) {
      cerr << "Warning: Adding empty word to the vocab - probably due to leading or trailing spaces in the source-side test data. Remove these to avoid bad translation results!" << endl;
      //cout << "Warning: Adding empty word to the vocab - probably due to leading or trailing spaces in the source-side test data. Remove these to avoid bad translation results!" << endl;
    }
    int currentSize = m_VocItems.size();
    m_VocItems.push_back(word);
    m_VocIndices[word] = currentSize;
    return currentSize;
  }else{
    //sgi::hash_map< const char *, CWordIndex, sgi::hash<const char*>, eqstr >::const_iterator FindIter;
    //FindIter = m_VocIndices.find( (const char *)word );
    //return (*FindIter).second;
    return (*(m_VocIndices.find(word))).second;
  }

  //m_VocIndices[pVocItem->PointerToWord()]= currentLastIndex;
  //return currentLastIndex;
}

// ********************************************************************
// Display-Methods

void
CVocabulary::Display() const
{
  cout << "Vocabulary:\n";
  cout << " Index Word\n";
  for (int i=0; i<m_VocItems.size(); i++){
    cout<<i<<"="<<m_VocItems[i]<<endl;
  }
} // DisplayVocabulary
