/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#ifndef Vocabulary_hh
#define Vocabulary_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
// CVocabulary
// Base class for vocabulary handling
//
#include <iostream>
#include <hash_map>
#include <hash_set>
#include "samtstring.hh"

#include "Types.hh"



struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};

struct __StringHash{
  size_t operator()(const samtstring &s1) const{
    return hash<const char*>()(s1.c_str());
  }
};

//! Note: it's important that CWordIndex be a 32-bit int to ensure consistency when loading word indices from e.g. Berkeley DBs. Also note that we need signed ints as we use negative values or zero to denote backpointer sites
typedef int CWordIndex;

class CVocabulary
{
public:
  CVocabulary();
  ~CVocabulary();

  //! Initialze some constants so that we can use a blank vocab
  bool InitBlank();

  // allocate memory for MaxNumOfVocItems vocabulary items
  // does not affect the size of the vocabulary, but allows faster loading
  bool Allocate( int MaxNumOfVocItems );

  // remove all entries in the vocabulary
  void Clear();

  // load vocabulary with/without comment lines/<UNK>/</s>
  bool Load( const char *FileName );
  bool Load( std::istream &InStream ); 

  // store vocabulary with/without <UNK>/</s>
  // no comment line is written
  bool Store( const char *FileName ) const;
  bool Store( std::ostream &OutStream) const;

  // number of stored vocabulary items
  int Size() const;

  //const CVocabularyItem & GetItem( CWordIndex WordIndex ) const;
  //const CString &         GetWord( CWordIndex WordIndex ) const;
  //const CString *         GetWordPtr( CWordIndex WordIndex ) const;
  const samtstring &         GetWord( CWordIndex WordIndex ) const;
  const samtstring *         GetWordPtr( CWordIndex WordIndex ) const;
  
  const bool         IsInRange( CWordIndex WordIndex ) const;

  //inline bool IsContentWord( CWordIndex WordIndex ) const;
  //inline bool IsMultiWord( CWordIndex WordIndex ) const;

  // for multi-word items, get index of first and last word
  //inline CWordIndex FirstWord( CWordIndex WordIndex ) const;
  //inline CWordIndex LastWord( CWordIndex WordIndex ) const;

  // return index of Word if in vocabulary
  // if Word is not in the vocab, add it on-the-fly 
  // Behavior is thus identical to AddWord
  //CWordIndex GetIndex( const CString & Word ) const;
  CWordIndex GetIndex( const samtstring& Word );

  // search for the nearest vocabulary item <= Word
  //CWordIndex SearchWord( const char * Word ) const;

  // add a word to the vocabulary if not existing yet
  // return the word's word index
  CWordIndex AddWord(const samtstring& word);
  //CWordIndex AddWord(const CString & word);


  // convert samtstring into CWordIndices
  //BOOL ConvertToIndices( char * const str, CArray<CWordIndex> *pSent, int *lenn);

  // display methods
  void Display() const;
  //! trims an samtstring
  static void string_trim(samtstring& str)
  {
    samtstring::size_type pos = str.find_last_not_of(" \n\r");
    if(pos != samtstring::npos) {
      str.erase(pos + 1);
      pos = str.find_first_not_of(" \n\r");
      if(pos != samtstring::npos) str.erase(0, pos);
    }
    else str.erase(str.begin(), str.end());
  }




protected:
  //void UpdateMultiWords();
  //CArray< CVocabularyItem * > m_VocItems;
  vector< samtstring > m_VocItems;
  hash_map< samtstring, CWordIndex, __StringHash > m_VocIndices;
};

#endif
