/*
  Syntax Based Statistical Machine Translation
  Chart Parsing Tool for Generalized, Annotated Synchronous Grammars
  Copyright (C) 2006  Andreas Zollmann, Ashish Venugopal
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

*/

#include "Rule.hh"
#include <iostream>
#include <sstream>



//! initialized in FastTranslateChart main
//C_NumModelsTotal = 0;
//C_NumModelsTM = 0;

// these values will be set in FastTranslateChart.cc
CVocabulary* CRule::pVoc=NULL; //! statically shared vocab for all rules
int CRule::C_NumModelsTM=-1; //number of tm models represented in each hyp:
//glue feature, word count, phrase count (1),
//RF|Result, RF|Target, RF|Source,
//OccurrenceCount, ...
//this should be C_NumModelsTotal-C_NumLMs
int CRule::C_NumModelsTotal=-1; //!number of models represented in each hyp (LM + TMs)
int CRule::C_NumLMs=0; //!< number of LMs
int CRule::C_LMLocation=0; //!< Location of the LMs in the feature vector (always zero)
int CRule::C_GlueLocation=0; //!< Will be incremented for each LM added
int CRule::C_RuleApplicationCountLocation=1; //!< Will be incremented for each LM added
int CRule::C_SourceSpanPenaltyLocation;

CRule::CRule(const samtstring& strkey, const samtstring& strdata) : 
  //start up m_negLogProbs with zeros for the language model costs
  m_negLogProbs(C_NumLMs, 0.0)
{
  bool debug = false;
  bool formaterror = false;

  // save the actual rule as a string
  m_rule = strkey + "#" + strdata;

  //cout<<"Loading rule: src="<<strkey<<" data="<<strdata<<endl;
  
  //get rule parts
  vector<samtstring> ruleparts;
  Tokenize(strdata,ruleparts,"#");
  if  (ruleparts.size() != 3) {
     cerr << "Rule body has less or more than 3 parts! (namely " << ruleparts.size() << ") ";
     formaterror = true;
  } else {
    //first part is the translation output
    //vector<samtstring> translationwords;
     //Tokenize(ruleparts[0],translationwords," ");
    
    stringstream ss(ruleparts[0], ios::in | ios::out);
    //cout<<"Translation part="<<ruleparts[0]<<endl;
     while (!ss.eof()){
       samtstring oneWord;
       CWordIndex wi;
       ss>>oneWord;
       
       //cout<<"oneWord="<<oneWord<<endl;
       //are we talking about an abstraction marker?
       if (oneWord.at(0) == '@'){
	 //remove the @
	  oneWord.erase(oneWord.begin());
	  //cout<<"leftword: "<<oneWord<<endl;
	  
	  //read off the abstraction number and subtract one because internally they will 
	  //correspond to backpointer indices starting with 0
	  int repPoint = int(atof(oneWord.c_str()))-1;
	  //!!az: to avoid potential exploit, check that this number
	  //is between 0 and 'number of nonterminals (i.e. n.o. @ symbols) in strkey'-1
	  //for now, just check that it's >= 0:
	  if (repPoint < 0) {
	    cerr << "Argument markers @N need to have N > 0. ";
	    formaterror = true;
	  }

	  //negation to distinguish backpointer indices from word indices
	  wi = -1*repPoint;
	  //cout<<"found arg:"<<oneWord<<" wi: "<<wi<<endl;
	}else{
	  //its a regular word
	 //cout<<"regular="<<oneWord<<endl;
	 //CString oneWordCString = (const char*)oneWord.c_str();
	  //lookup this word in the vocab, if it's not there, GetIndex
	  //will add it and assign it a unique ID
	 wi = pVoc->GetIndex(oneWord);
	 //	  if (wi==CVocabulary::EOT || wi==CVocabulary::UNK){
	 //	    //wi = pVoc->AddWord(oneWordCString);
	 //	    wi = pVoc->AddWord(oneWord);
	 //	  }
	}
	//now push the word index into the translationOutput vector
	m_translationOutput.push_back(wi);
     }

     /*
     cout<<"Final translation output=";
     for (vector<CWordIndex>::iterator iter=m_translationOutput.begin(); iter!=m_translationOutput.end(); ++iter){
       cout<<(*iter)<<" ("<<pVoc->GetWord(*iter)<<") ";
     }
     cout<<endl;
     */
     //second part is the result nonterminal
     stringstream ss2(ruleparts[1], ios::in | ios::out);
     if (ss2.eof()) {
       cerr << "Result part of rule is empty! ";
       formaterror = true;
    } else {
       CWordIndex targetItem;
       samtstring nonTerminalWord;
       ss2>>nonTerminalWord;

       //make sure this is really a nonterminal symbol
       if (nonTerminalWord.at(0) != '@'){
	 cerr << "Result category must be a nonterminal symbol (i.e., starting with '@')! ";
	 formaterror = true;
       } else {
	 //CString oneWordCString = (const char*)nonTerminalWord.c_str();
	 //lookup this word in the vocab
	 //if it's not there, GetIndex will add it and assign it a
	 //unique ID
	 CWordIndex wi = pVoc->GetIndex(nonTerminalWord);
	 //	 if (wi==CVocabulary::EOT || wi==CVocabulary::UNK){
	 //	   //wi = pVoc->AddWord(oneWordCString);
	 //	   wi = pVoc->AddWord(nonTerminalWord);
	 //	 }
	 m_productionResult = wi;
	 if (!ss2.eof()) {
	   cerr << "Result part of rule has more than one words! ";
	   formaterror = true;
	 } else {
	   //now process the cost vector
	   
	   stringstream ss3(ruleparts[2], ios::in | ios::out);
	   while (!ss3.eof()){
	     double neglogprob;
	     ss3>>neglogprob;
	     m_negLogProbs.push_back(neglogprob);
	   }

	 }
       }
    }
   }
  	  
   if (formaterror) {
      cerr << "Error in database format: wrong format of record's data '" <<strdata << "' (key is: '" << strkey << "')" << endl;
      exit(-1);
   }

   //cout<<"END of rule building;"<<endl;
   
}

CRule::CRule(const CRule& r){
  m_productionResult = r.m_productionResult;
  m_translationOutput = r.m_translationOutput;
  m_negLogProbs = r.m_negLogProbs;
  m_rule = r.m_rule;
}

  //!Assignment operator
CRule& CRule::operator = (const CRule& r){

  if (this == &r){
    return *this;
  }
  m_productionResult = r.m_productionResult;
  m_translationOutput = r.m_translationOutput;
  m_negLogProbs = r.m_negLogProbs;
  m_rule = r.m_rule;
  return *this;
}


//!Utilty display method for debugging
void CRule::Display() {
  cout<<"m_productionResult="<<m_productionResult;
  if (pVoc!=NULL){
    cout<<" ("<<pVoc->GetWord(m_productionResult)<<") ";
  }
  cout<<" m_translationOutput=";
  if (pVoc!=NULL){
    
    for (int i=0; i<m_translationOutput.size(); i++){
      if (m_translationOutput[i]<=0){
	cout<<" BP="<<m_translationOutput[i]<<" ";
      }else{
	cout<<pVoc->GetWord(m_translationOutput[i])<<" ";
      }
    }
  }else{
    cout<<" NULL-voc ";
  }

  cout<<" m_negLogProbs=";
  for (int i=0; i<m_negLogProbs.size(); i++){
    cout<<m_negLogProbs[i]<<" ";
  }
  
}
